%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: decl_support.ql,v 1.6 2000/11/15 06:29:44 itp Exp $
%%
%%  This file contains support procedures for 'declarations'
%%  (see lib/decl_expansions).  It must not contain any
%%  declarations itself!

#include "../lib/ergo.h"

%% private update_theory(Fact, Theory).
%%
%%  Update Fact in Theory.
%%  If there is a similar Fact in Theory, supersede it with a warning.
%%  A Fact is similar if it has the same functor and first argument.
%%  The warning (and the retract/assert!) are avoided if the new
%%  and old facts are ``identical''.  At present, ``identical'' is implemented
%%  by ==/2 -- but this is too strong!
%% 
%%  mode update_theory(@compound, @theory).
%%
update_theory(Fact, Theory) :-
    functor(Fact, Fun, Arity),
    functor(OldFact, Fun, Arity),
    arg(1, Fact, Arg),
    arg(1, OldFact, Arg),
    query_theory(OldFact, Theory),
    !,
    update_theory1(Fact, OldFact, Fun, Arg, Theory).
update_theory(Fact, Theory) :-
    assert_theory(Fact, Theory).

update_theory1(Fact, OldFact, _, _, _) :-
    Fact == OldFact,
%% TO DO: Pick happy medium between = and == in update_theory1. MB 10/11/1997
%% Would be nice to test for variants here.  Requiring == is too strong,
%% but = is too weak.
%% When this is fixed, update description of update_theory.
    !.
update_theory1(Fact, OldFact, FactType, Name, Theory) :-
    writelist(['WARNING: superseding ', w(FactType), ' declaration of ',
               w(Name), ' in theory ', w(Theory), nl]),
    retract_theory(OldFact, Theory),
    assert_theory(Fact, Theory).



%% command add_doc(Thy, Name, Header, Doc, CmdList).
%%
%%  Adds a doc node (called Name) to the Thy part of the
%%  Ergo database.  Thy must be 'global' or a theory name.
%%
%%  Header is the header for the node.
%% 
%%  Doc is all the text within the doc node.  It may also
%%  contain cross-references to other doc nodes (see writelist/3).
%%
%%  CmdList is a list of cmd(Thy,Table,CmdName,CmdArity)
%%  terms which each point to a Thy:Table(Call,_,_) fact
%%  where the outermost functor of Call is CmdName/CmdArity.
%%
%%  Eg. add_doc(global,
%%               'show/1',
%%               ['show(Topic)'],
%%               ['Shows information about Topic.',nl,...],
%%               [cmd(global,elves,show,1)]).
%%
%%  mode add_doc(@atom, @atom, @writelist, @writelist, @list(compound)).
%%
add_doc(Thy, Name, Header, Doc, Cmds) :-
    check_is_legal_writelist(Header),
    check_is_legal_writelist(Doc),
    check_atom(Name),
    list(Cmds),
    atom(Thy),
    update_theory(doc_node(Name,Header,Doc,Cmds), Thy).

%% command add_doc_fragment(Thy, Name, Doc).
%%
%%  Adds a doc fragment called Name (an atom or a compound term
%%  whose parameters are distinct metavariables) to Thy.
%%  Thy must be 'global' or a theory name.
%%
%%  Doc is all the text within the doc fragment; it should refer to the
%%  parameter vales e.g. with w(Param).  It may also contain cross-references
%%  to other doc nodes (see writelist/3).
%%
%%  mode add_doc_fragment(@atom, @nonvar, @writelist).
%%
add_doc_fragment(Thy, Name, Doc) :-
    check_is_legal_writelist(Doc),
    check_or_error(
        ( Name =.. [F|Args],
          atom(F),
          check_distinct_metavars(Args)
        ),
        ['illegal doc fragment name: ', q(Name)]),
    atom(Thy),
    update_theory(doc_fragment(Name, Doc), Thy).


%% command add_doc_extension(ThisThy, ExtNode, ExtDoc).
%%
%%  Add a doc extension in ThisThy, extending node ExtNode with ExtDoc.
%%
%%  mode add_doc_extension(@thyname, @atom, @writelist).
%%
add_doc_extension(Thy, Name, Doc) :-
    check_atom(Thy),
    check_atom(Name),
    check_is_legal_writelist(Doc),
    %% NOT update_theory, as we want to extend it!
    assert_theory(doc_extension(Name, Doc), Thy).


%% command add_command(Thy, Table, Call, Modes, Impl, DocNode).
%%
%%  Adds a command to the table of commands called Table.
%%  Thy must be 'global' or the name of the theory that this
%%  command is associated with.  
%%
%%  Call is the template for the command.  Its outermost functor
%%  must be an atom that does not include a theory prefix.
%%
%%  Modes is a list of mode declarations for the command.
%%
%%  Impl is Prolog code that implements the command.
%%
%%  DocNode is the name of the most relevant doc node for this command.
%%  It should not include the theory name.  If the command has no
%%  relevant doc node, DocNode should be 'none'.
%%
%%  Eg. add_command(global, elves, show(Topic),
%%                  [show(@list(atom)),show(@atom)],
%%                  show(Topic),
%%                  'show/1').
%%
%%  Eg. add_command(jprop, tactic, cases(E),
%%                  cases(?thy(term)),
%%                  'jprop:cases'(E),
%%                  'cases/1').
%%
%%  mode add_command(@theory, @atom, @cmd, @mode, @code, @atom).
%%
add_command(Thy, Table, Call, Modes, Impl, DocNode) :-
    %writelist(['Adding ', w(Call), ' to ', w(Table), nl]),
    check_atom(Thy),
    check_atom(Table),
    check_atom(DocNode),
    check_is_legal_cmd(Call),
    functor(Call,F,N),
    check_or_error(
        nonvar(Impl),
        ['implementation of ',q(Table),' command ',q(F/N),' is missing.']),
    check_or_error(
        is_legal_mode_list(Modes,F,N),
        ['bad mode declaration ',q(Modes),' for command ',q(F/N)]),
    %%SEE expansions.ql:
    %%declare_meta_predicate_modes(Modes),
    ( Table = rule_constraint ->
        rule_constraint_check_impl(Call, Impl)
    ;
        true
    ),
    Fact =.. [Table,Call,Modes,Impl,DocNode],
    update_theory(Fact, Thy).


%% command delete_command(Thy, Table, Call).
%%
%%  Deletes a command from the table of commands called Table.
%%  Thy must be 'global' or the name of the theory that this
%%  command is associated with.  
%%
%%  Call is the template for the command.  Its outermost functor
%%  must be an atom that does not include a theory prefix.

delete_command(Thy, Table, Call) :-
    check_atom(Thy),
    check_atom(Table),
    check_is_legal_cmd(Call),
    Fact =.. [Table,Call,_,_,_],
    retract_theory(Fact, Thy).


%% find_command(Cmd, Table, Thy, Modes, Impl, DocNode).
%% find_command(Cmd, Table, Modes, Impl, DocNode).
%%
%%  Look up Cmd from Table in Thy.
%%  Modes, Impl and DocNode specify the command; Impl is suitable to pass
%%    to call_tac/1 (and must be a --->/2 term if the returned command is a
%%    proof command with subproofs).
%%  Can enumerate multiple matching commands on backtracking.
%%  If Thy is omitted, or is a variable, look for Cmd from Table in all
%%  ancestor theories and in the global area.
%% 
%%  mode find_command(?cmd, ?cmdtab, @theory, -mode, -code, -atom).
%%  mode find_command(?cmd, ?cmdtab, -mode, -code, -atom).
%%
find_command(Cmd, Tab, Modes, Impl, DocNode) :-
    find_command(Cmd, Tab, _, Modes, Impl, DocNode).

find_command(Cmd, Tab, Thy, Modes, Impl, DocNode) :-
    command_table(Tab,_),
    ( nonvar(Cmd) ->
        check_is_legal_cmd(Cmd)
    ;
        true
    ),
    Fact =.. [Tab,Cmd,Modes,Impl,DocNode],
    query_global_or_ancestors(Fact, Thy).


%% command add_parameter(Thy, Name, DocNode, SetProc, GetProc).
%%
%%  Add a parameter to the parameter table.
%% 
%%  Thy must be 'global' or the name of the theory that this
%%  parameter is associated with.  The SetProc and GetProc may store
%%  the value any way they choose, not necessarily as a fact in Thy.
%%
%%  mode add_parameter(@theory, @atom, @atom, @atom, @atom).
%% 
add_parameter(Thy, Name, DocNode, Set, Get) :-
    check_atom(Thy),
    check_atom(Name),
    check_atom(DocNode),
    check_atom(Set),
    check_atom(Get),
    update_theory(parameter(Name, DocNode, Set, Get), Thy).


