%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: fact_types.ql,v 1.76 2002/04/28 05:03:54 itp Exp $

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The procedures in this file determine what clauses can be stored
%% in the Ergo database, which part of the database they are stored
%% in, whether or not they should be saved between sessions etc.
%%
%% The logical organisation of the Ergo database is that it is comprised
%% of a global area (named 'global') and a separate area for each
%% theory that has been defined (named by the theory name).
%%
%% The current physical organisation is that clauses in the global area
%% are saved into the 'global.erg' file and for each theory called T,
%% the clauses in theory T that need saving are saved into a file
%% called 'T.erg'.  This physical organisation could easily be
%% changed, whereas changing the logical organisation would be harder.
%%
%% NB. All clauses that are essential to the soundness of the prover
%%     must be declared in this file, rather than being added using
%%     the new_*_fact/2 procedures below.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% global_fact(FactName, Arity).
%% global_fact(FactName, Arity, Props).
%% global_fact(FactName, Arity, Core, Saved, Assert, Retract, Props).
%%
%% For backwards compatibility only:
%%
%% global_saved_fact(FactName, Arity).
%% global_local_fact(FactName, Arity).
%% global_tactic_fact(FactName, Arity).
%%
%%  global_fact/2 succeeds iff FactName/Arity has been declared 
%%  to be a clause that can be stored in the global area of the
%%  (runtime) Ergo database.
%%
%%  global_fact/3 returns additional properties of the FactName/Arity
%%  facts.  So does global_fact/5, but it should not be used outside
%%  the core prover, since its interface is more likely to change.
%%
%%  The remaining clauses succeed on various subsets of the clauses
%%  identified by global_fact/2.
%%
%%  The global_saved_fact/2 and global_local_fact/2 procedures 
%%  partition all global facts according to whether or not they 
%%  are saved between sessions.
%%
%%  global_tactic_fact/2 succeeds on the subset of the global facts
%%  that are not essential to the soundness of the prover, so can be
%%  freely manipulated by tactics.
%%
%%  Each Name/Arity fact can have one or more of the following
%%  properties associated with it:
%%
%%    assert(Ord), where Ord is a or z.  (default is a).
%%
%%    key(K), where K is a number in 1..Arity.  (default is no keys).
%%
%%       Means that all Name/Arity facts in the same theory
%%       should have different values in field K.
%%
%%       WARNING: Even if there are no keys, you should not rely on 
%%       storing duplicate facts within a single theory, since the 
%%       technique used for retracting loaded facts doesn't work with
%%       repeated identical (or unifiable!) facts.
%%
%%    retract(Flag), where Flag is true or false.  (default is true).
%%
%%       Flag=false means that individual Name/Arity facts cannot be 
%%       retracted (or updated) once they have been added (except by
%%       deleting the whole theory).  Such facts are slightly
%%       faster to query than retractable facts.
%%
%%    save(Flag), where Flag is true or false.  (default is false).
%%
%%       Name/Arity facts will be saved between sessions iff Flag=true.
%%
%%    core(Flag), where Flag is true or false.  (default is false).
%%
%%       Means that Name/Arity facts are essential to the soundness
%%       of the core prover, so tactics cannot update them, except
%%       via calling prover procedures.
%%
%%  mode global_fact(?atom, ?integer).
%%  mode global_fact(?atom, ?integer, ?list(compound)).
%%  mode global_fact(?atom, ?integer, ?bool,?bool,?atom,?bool,?list(compound)).
%%
%%  mode global_saved_fact(?atom, ?integer).
%%  mode global_local_fact(?atom, ?integer).
%%  mode global_tactic_fact(?atom, ?integer).
%%

global_fact(F,N) :-
    global_fact(F,N,_).

global_fact(F,N,P) :-
    global_fact(F,N,Core,Save,Assert,Retract,Props),
    P = [core(Core),save(Save),assert(Assert),retract(Retract)|Props].

global_saved_fact(F,N) :-
    global_fact(F,N,_,true,_,_,_).

global_local_fact(F,N) :-
    global_fact(F,N,_,false,_,_,_).

global_tactic_fact(F,N) :-
    global_fact(F,N,false,_,_,_,_).

%% global_fact(FactName, Arity, Core,  Save,  AZ, Retract, Props).
global_fact(global_fact,7,	true,  true,  z,  false, []).
global_fact(theory_fact,7,	true,  true,  z,  false, []).
global_fact(global_fact_local,7,true,  false, z,  false, []).
global_fact(theory_fact_local,7,true,  false, z,  false, []).
global_fact(ancestor,2,		true,  true,  a,  true,  [key(1)]).
global_fact(parent,2,		true,  true,  a,  true,  []).
global_fact(subthy_owner,2,	true,  true,  a,  true,  [key(1)]).
global_fact(tactic_file,1,	true,  true,  z,  true,  [key(1)]).
global_fact(dbms_interface,1,	true,  true,  a,  true,  [key(1)]).
global_fact(load_path,1,	true,  true,  z,  true,  [key(1)]).
global_fact(path,1,		true,  true,  z,  true,  [key(1)]). % OBSOLETE
global_fact(request,2,		true,  true,  z,  true,  [key(1)]).
global_fact(announcement,2,	true,  true,  z,  true,  [key(1)]).
global_fact(handler,2,		false, false, a,  true,  []).
global_fact(listener,2,		false, false, a,  true,  []).
global_fact(doc_node,4,		false, false, a,  true,  []).
global_fact(doc_fragment,2,	false, false, a,  true,  []).
global_fact(doc_extension,2,	false, false, a,  true,  []).
%% type(Name, Type, DocNode, TypeCheckPred).
global_fact(type,4,             false, false, z,  true,  []).
global_fact(tactic_wrapper,3,	true,  false, a,  true,  []).
global_fact(parameter,4,	false, false, a,  true,  []).
global_fact(command_table,2,	false, false, a,  true,  []).
global_fact(needs_saving,1,	true,  false, z,  true,  [key(1)]).
global_fact(F,1,		true,  true,  z,  true,  [key(1)]) :-
    query_global(ancestor(Thy, _)),
    Thy \= global,
    atom_concat2(ancestor_of_, Thy, F).
global_fact(F,4,		false, false, a,  true,  [key(1)]) :-
    command_table(F,_).
global_fact(F,N,Core,Save,Assert,Retract,Props) :-
    query_global(global_fact(F,N,Core,Save,Assert,Retract,Props)).
global_fact(F,N,Core,Save,Assert,Retract,Props) :-
    query_global(global_fact_local(F,N,Core,Save,Assert,Retract,Props)).


%% theory_fact(FactName, Arity).
%% theory_fact(FactName, Arity, Props).
%% theory_fact(FactName, Arity, Core, Saved, Assert, Retract, Props).
%%
%% For backwards compatibility only:
%% theory_saved_fact(FactName, Arity).
%% theory_local_fact(FactName, Arity).
%% theory_tactic_fact(FactName, Arity).
%%
%%  theory_fact/2 succeeds iff FactName/Arity has been declared 
%%  to be a theory-specific clause that can be stored in the 
%%  (runtime) Ergo database.
%%
%%  theory_fact/3 returns additional properties of the FactName/Arity
%%  facts.  So does theory_fact/5, but it should not be used outside
%%  the core prover, since its interface is more likely to change.
%%
%%  The remaining clauses succeed on various subsets of the clauses
%%  identified by theory_fact/2.
%%
%%  The theory_saved_fact/2 and theory_local_fact/2 procedures partition
%%  all theory specific facts according to whether or not they are 
%%  saved in theory files.  Those that are saved will persist 
%%  between prover sessions, whereas those that are not saved
%%  will persist for the current session only.  
%%
%%  The theory_tactic_fact/2 procedure identifies those clauses that
%%  are not essential to the soundness of the prover, so can be
%%  freely manipulated by tactics.
%%
%%  Note: for fastest access, the theory_saved_fact/2 clauses should
%%  be ordered by decreasing access frequency.
%%
%%  mode theory_fact(?atom, ?integer).
%%  mode theory_fact(?atom, ?integer, ?list(compound)).
%%  mode theory_fact(?atom, ?integer, ?bool,?bool,?atom,?bool,?list(compound)).
%%
%%  mode theory_saved_fact(?atom, ?integer).
%%  mode theory_local_fact(?atom, ?integer).
%%  mode theory_tactic_fact(?atom, ?integer).
%%
theory_fact(F,N) :-
    theory_fact(F,N,_).

theory_fact(F,N,P) :-
    theory_fact(F,N,Core,Save,Assert,Retract,Props),
    P = [core(Core),save(Save),assert(Assert),retract(Retract)|Props].

theory_saved_fact(F,N) :-
    theory_fact(F,N,_,true,_,_,_).

theory_local_fact(F,N) :-
    theory_fact(F,N,_,false,_,_,_).

theory_tactic_fact(F,N) :-
    theory_fact(F,N,false,_,_,_,_).

%% Ergo 5 facts.
%% theory_fact(FactName, Arity, Core, Saved, Assert, Retract, Props).
theory_fact(context_name,1, true, true, a, false, [key(1)]).
    %% The available kinds of context for each node of a proof in
    %% theory T are all names N such that context_name(N) has been
    %% declared in T or one of its ancestors.
    %% Note that these facts cannot be retracted, since to do so would
    %% invalidate the rules in the theory that use that context.

theory_fact(infrule,7,	true, true, a, false, [key(1)]).
    %% Used for Ergo 5 inference rules (and axioms and theorems).
    %% Eg. infrule(Name,FullName,Concl,Id,Seqs,Before,After).
    %% FullName must be a std_nonvar term and Name is the functor
    %% of FullName.  The arguments of FullName are typically used
    %% to constrain the way that the rule is instantiated.
    %% All args have associated types.
    %% All infrule facts should be asserted via rule_add,
    %% rather than using assert_theory directly.
    %% Similarly, they should be queried using rule_query.
theory_fact(infrule_apply,5, true, true, a, false, [key(1)]).
    %% infrule_apply(Name, FullName, Concl, Id, SeqIds) :- Code.
    %% Equivalent to infrule/7 facts, but in a more efficient form
    %% for applying rules.  These facts are generated automatically
    %% by rule_add whenever an infrule/7 fact is added.
    %% These facts cannot be queried (especially when compiled),
    %% because of the Code part.

theory_fact(infrule_status,2, true, true, a, false, [key(1)]).
    %% Every Ergo 5 inference rule that is not a postulate has
    %% one of these facts.
    %%
    %% For axioms (or fundamental inference rules) it is
    %% infrule_status(Name,fundamental).
    %% For proved theorems (or inference rules that have been proved)
    %% it is infrule_status(Name,DepList) where DepList is a list of
    %% all the rules (in Thy:ShortName format) that were used to prove Name. 


%% Ergo 4 facts (many used by Ergo 5 as well)
theory_fact(rule,5,	true, true, a, true, [key(1)]).
    %% Used for axioms and theorems.
    %% Eg. rule(Name,Kind,Shape,Stmt,Body), where
    %%  - Kind is axiom, axiom(FullName), defn, postulate or theorem(Deps).
    %% All rule facts should be asserted using assert_axthm/6,
    %% rather than using assert_theory directly.
theory_fact(openrule,8,	true, true, a, true, []).
theory_fact(symbol,3,	true, true, a, true, []).
    %% symbol(Name,Arity,Kind), or
    %% symbol(param_const(Name,Value),when(Code),Kind), where
    %%
    %% - Name is (fully qualified) atom.
    %% - Arity is a list for functions/quantifiers (see check_arity/1)
    %% - Code is Prolog code that checks the legality of Value.
    %% - Kind is 'declared' or 'defined'.
    theory_fact(equivalence,1,	true, true, a, true, [key(1)]).
    %% Records the names of all binary functions that have been
    %% declared to be equivalence relations in the theory.
theory_fact(preorder,1,	true, true, a, true, [key(1)]).
    %% Records the names of all binary functions that have been
    %% declared to be preorder relations in the theory.
theory_fact(reverse_of,2,	true, true, a, true, []).
    %% Records converse relationships between preorder relations.
    %% Eg. reverse_of(r1,r2) means that r1(A,B) <=> r2(B,A).
theory_fact(stronger_than,2,	true, true, a, true, []).
    %% Records partial order relationships between preorder relations.
    %% Eg. stronger_than(r1,r2) means that r1(A,B) => r2(A,B).
theory_fact(tactic_file,1,	true, true, z, true, [key(1)]).
    %% Records the names (without directories or extensions) of
    %% the tactic files that are associated with the theory.
theory_fact(compiled_rule,8,	true, true, a, true, []).
theory_fact(save_clauses,3,	true, true, z, true, []).
theory_fact(interpretation,2,	true, true, z, true, []).
    %% Records translations in an interpretation.
theory_fact(interpreted_theory,2, true, true, z, true, []).
    %% interpreted_theory(Abs, Int) records which abstract theory was
    %% interpreted to generate this subtheory, and where the interpretation
    %% facts can be found.
theory_fact(interpreted_subtheory,2, true, true, z, true, []).
    %% interpreted_subtheory(Abs, Sub) records that this theory contains facts
    %% that pertain to a subinterpretation Abs->Sub.
theory_fact(object_vars,1,	true, true, z, true, [key(1)]).
    %% Defines the legal object variable prefixes.
theory_fact(status,1,	true, true, a, true, []).
    %% Eg. status(S), where S=open or S=closed.

theory_fact(tactic_wrapper,3,	true,  false, a, true,  [key(1)]).
    %% tactic_wrapper(Call, SubProofList, WrappedCall).
    %%   This is only used by wrap_tac, not call_tac.
    %%   The outermost functor of Call must be fully qualified:
    %%   e.g., 'thy:tacname'(...) for theory-specific tactics
    %%   (global_tactics are just tacname(...)).
theory_fact(doc_node,4,		false, false, a, true,  [key(1)]).
    %% doc_node(Name, Header, Doc, OptCmdRef).
    %%   where OptCmdRef is a list of cmd(Thy,Table,CmdName,CmdArity)
    %%   terms which each point to a Thy:Table(Call,_,_) fact
    %%   where the outermost functor of Call is CmdName/CmdArity.
theory_fact(doc_fragment,2,	false, false, a, true,  [key(1)]).
    %% doc_fragment(Head, Doc).
    %%   where Doc can refer to any arguments of Head.
theory_fact(doc_extension,2,	false, false, a, true,  []).
    %% doc_extension(Name, Doc).
    %%   Typically, all the doc_extension/2 clauses are sorted according
    %%   to the Doc writelist, then appended onto the writelist of the
    %%   doc node call Name.  Each doc_extension/2 fact usually contains
    %%   one cross reference in Doc.
theory_fact(parameter,4,	false, false, a, true,  [key(1)]).
    %% parameter(Name, DocNode, SetProc, GetProc).
    %%   SetProc and GetProc will be called with one argument to set/get the
    %%   value of the parameter.  They are entirely responsible for storing and
    %%   retrieving the value.  The parameter/4 facts just declare the
    %%   existence of the parameter and its documentation.
%% TO DO: Consider adding `Type' and `DefaultValue' to the information in
%% theory_fact(parameter,4, ...).
%% Ray's last comment: Probably not worth the bother.
theory_fact(F,4,		false, false, a,  true,  [key(1)]) :-
    command_table(F,_).
theory_fact(F,N,Core,Save,Assert,Retract,Props) :-
    query_global(theory_fact(F,N,Core,Save,Assert,Retract,Props)).
theory_fact(F,N,Core,Save,Assert,Retract,Props) :-
    query_global(theory_fact_local(F,N,Core,Save,Assert,Retract,Props)).


%% command_table(Functor, Doc).
%%
%%  Functor/4 is a global_local and theory_local fact type for command
%%  declarations, and Doc describes its purpose.
%%  Functor is the name of the table, and the arguments are
%%  Call, Mode, RealCode, DocNode.
%%   NB. If a proof command may have subproofs, then RealCode must be of
%%   the form (Call2 ---> SubProofs).  If it definitely has no
%%   subproofs (and therefore its last argument is not a list of
%%   subproofs), then RealCode must be just Call2, which does
%%   not have '--->'/2 as the outermost functor.  In both cases,
%%   Call2 must match the first argument of the tactic_wrapper fact
%%   for the corresponding tactic.
%%   
%%  mode command_table(?atom, ?writelist).
%% 
builtin_command_table(general,
                      ['general commands']).
builtin_command_table(database,
                      ['database commands']).
builtin_command_table(proof,
                      ['proof commands']).
builtin_command_table(rule_constraint,
                      ['Constraints on Rules']).
builtin_command_table(library,
                      ['tactic support `commands'' that should only be',nl,
	               'called from other tactics, not interactively.']).

command_table(Table, Doc) :-
    query_global(command_table(Table, Doc)).


%% new_command_table(Table, Doc).
%%
%%  Declare a new command table.
%%
%%  mode new_command_table(@atom, @writelist).
%%
new_command_table(Table, Doc) :-
    \+ query_global(command_table(Table, _)),
    assert_global(command_table(Table, Doc)),
    (
        %% declare the associated global predicate.
        dbms_asserted_name(global,Table,RealF),
        dynamic(RealF/4),
        fail
    ;
        %% declare the new fact as dynamic in all loaded theories.
        loaded_theory(Thy,_),
        dbms_asserted_name(Thy, Table, RealF),
        dynamic(RealF/4),
        fail
    ;
        true
    ).




%% new_global_fact(FactName, Arity, Properties).
%% new_theory_fact(FactName, Arity, Properties).
%%
%% For backwards compatibility only these used to exist.
%% Unless something unusual happens they should never need to be reinserted.
%% new_global_local_fact(FactName, Arity).
%% new_theory_local_fact(FactName, Arity).
%%
%% For backwards compatibility only.
%% Left in to allow old global.erg files to work.
%% new_global_saved_fact(FactName, Arity).
%% new_theory_saved_fact(FactName, Arity).
%%
%%  These procedures declare new fact types in the Ergo database.
%%  Each FactName/Arity pair may be declared as being both global
%%  and theory-specific.  Properties is a list of properties as
%%  described in the global_fact/3 documentation above.  Note that
%%  Properties may contain several key/1 properties, but must contain
%%  at most one of the other properties.
%%
%%  mode new_global_fact(@atom, @integer, @list(gcomp)).
%%  mode new_theory_fact(@atom, @integer, @list(gcomp)).
%%

% new_global_local_fact(F,N) :- new_global_fact(F,N,[save(false)]).

% new_theory_local_fact(F,N) :- new_theory_fact(F,N,[save(false)]).

%% TO DO: Eventually the backwards compatibility for the calls to
%% new_global_fact should be removed.
%% This can be done once no global.erg files or PRT and potentially something
%% else in the Cogito refinement stuff, command line interfaces, ...
%% rely upon the old formats of new_global_saved_fact and
%% new_theory_saved_fact.
%% MB 29/7/1997

new_global_saved_fact(F,N) :- new_global_fact(F,N,[save(true)]).

new_theory_saved_fact(F,N) :- new_theory_fact(F,N,[save(true)]).

new_global_fact(F,N,Props0) :-
    check_new_fact(global_fact,F,N,Props0,Decl),
    assert_global(Decl),
    dbms_asserted_name(global,F,RealF),
    dynamic(RealF/N).

new_theory_fact(F,N,Props0) :-
    check_new_fact(theory_fact,F,N,Props0,Decl),
    assert_global(Decl),
    (
        %% declare the new fact as dynamic in all loaded theories.
        loaded_theory(Thy,_),
        dbms_asserted_name(Thy, F, RealF),
        dynamic(RealF/N),
        fail
    ;
        true
    ).


%% check_new_fact(Kind,F,N,Props,Decl).
%%
%%  Kind must be either 'global_fact' or 'theory_fact'.
%%  This procedure succeeds if F/N is not already declared as a
%%  global/theory-specific fact.  Generates an error message otherwise.
%%
%%  mode check_new_fact(@atom, @atom, @integer, @list(gcomp), @gcomp).
%%
check_new_fact(global_fact,F,N,_,_) :-
    global_fact(F,N),
    !,
    handle_error(['global fact type ',q(F/N),' is already in use.']).
check_new_fact(theory_fact,F,N,_,_) :-
    theory_fact(F,N),
    !,
    handle_error(['theory fact type ',q(F/N),' is already in use.']).
check_new_fact(Where,F,N,Props0,Decl) :-
    atom(F),
    integer(N),
    0 =< N, N < 256,
    !,
    extract_property(core(Core), core(false),   Props0, Props1),
    extract_property(save(Save), save(false),   Props1, Props2),
    extract_property(assert(AZ), assert(a),     Props2, Props3),
    extract_property(retract(R), retract(true), Props3, Props4),
    check_extra_fact_properties(Props4,N),
    ( Save = true ->
        Where2 = Where
    ;
        atom_concat2(Where, '_local', Where2)
    ),
    Decl = Where2(F,N,Core,Save,AZ,R,Props4).
check_new_fact(Where,F,N,_,_) :-
    handle_error(['illegal ',w(Where),' name or arity: ',q(F/N)]).


extract_property(P, _, Props0, Props) :-
    delete(P, Props0, Props1),
    ground(P),
    !,
    Props = Props1.
extract_property(Default, Default, Props0, Props0).


check_extra_fact_properties([],N).
check_extra_fact_properties([P|PS],N) :-
    check_extra_fact_property(P,N),
    !,
    check_extra_fact_properties(PS,N).
check_extra_fact_properties([P|_],_) :-
    handle_error(['illegal fact property: ',q(P)]).

check_extra_fact_property(key(N),Arity) :-
    integer(N),
    between(1,Arity,N).


%% create_theory_database(Thy).
%%
%%  Creates an empty theory, called Thy.
%%  It does not create a `searcher' predicate for Thy,
%%  or assert the loaded_theory/2 fact.
%%
%%  mode create_theory_database(@atom).
%%
create_theory_database(Thy) :-
    theory_fact(F,N),
    dbms_asserted_name(Thy, F, RealF),
    dynamic(RealF/N),
    fail.
create_theory_database(_).


%% delete_theory_database(Thy).
%%
%%  Deletes all facts in theory Thy, including the
%%  loaded_theory/2 fact.
%%
%%  mode delete_theory_database(@theory).
%%
delete_theory_database(Thy) :-
    retract(loaded_theory(Thy,_)),
    %%
    %% We don't abolish any loaded facts, since they will not be
    %% accessed again, even if Thy is rebuilt, because searcher
    %% predicates for new theories only look at asserted facts.
    %% 
    %% Abolish all asserted facts.
    theory_fact(F,A),
    dbms_asserted_name(Thy, F, AF),
    abolish(AF/A),
    fail.
delete_theory_database(_).


%% create_global_database.
%%
%%  Declare all the dynamic predicates that the prover uses
%%  and do anything that is necessary to set up the global.erg
%%  database area.
%%
%%  mode create_global_database.
%%
create_global_database :-
    write('    creating:   Global Database...\n'),
    dynamic(loaded_theory/2),
    global_fact(F,N),
    dbms_asserted_name(global, F, RealF),
    dynamic(RealF/N),
    fail.
create_global_database :-
    dynamic(batch_mode/0),
    dynamic(elves/0),
    dynamic(dbg/1),
    dynamic(foreign_dir/1),
    dynamic(global_file_location/1),
    dynamic(incomplete_retry_depth/1),
    dynamic(library_location/1),
    dynamic(loaded/3),
    dynamic(macro_stack/1),
    assert(msg_verbosity(1)),
    assert(loaded_theory(global, query_global)).

