%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: query.ql,v 1.84 2001/06/13 03:21:19 itp Exp $


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file contains procedures for querying the Ergo database.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../lib/ergo.h"


%% query_global(Fact).
%% query_theory(Fact).
%% query_theory(Fact, Thy).
%% query_ancestors(Fact).
%% query_ancestors(Fact, Thy).
%% query_global_or_ancestors(Fact).
%% query_global_or_ancestors(Fact, Thy).
%%
%%  These procedures are the lowest level way of extracting information
%%  from the Ergo database.  In all cases, the functor and arity of
%%  Fact must be instantiated before the procedure is called.
%%  These procedures iterate through all matching facts (and ancestors,
%%  if applicable) upon backtracking.
%%
%%  query_global/1 succeeds iff Fact is in the global area of the database.
%%
%%  query_theory/1 succeeds iff Fact is in the current theory.
%%
%%  query_theory/2 succeeds iff Fact is in the Thy part of the database.
%%                   (Thy may also be `global').
%%
%%  query_ancestors/1 succeeds iff Fact is in the current theory or
%%    any one of its ancestors.
%%
%%  query_ancestors/2 is the same as query_ancestors/1, but also returns
%%    the name of the theory in which the fact was found.  
%%
%%  query_global_or_ancestors/1,2 is like query_ancestors/1,2 but searches the
%%    global database first.
%% 
%%  mode query_global(?compound).
%%  mode query_theory(?compound).
%%  mode query_theory(?compound, @atom).
%%  mode query_ancestors(?compound).
%%  mode query_ancestors(?compound, ?atom).
%%  mode query_global_or_ancestors(?compound).
%%  mode query_global_or_ancestors(?compound, ?atom).
%%

%% This is the hand-coded `searcher' predicate for the `global' theory.
%% For speed, we expand out dbms_asserted_form for the known global facts.
%%
query_global(global_fact(A,B,C,D,E,F,G)) :- !,'/A/global_fact'(A,B,C,D,E,F,G).
query_global(theory_fact(A,B,C,D,E,F,G)) :- !,'/A/theory_fact'(A,B,C,D,E,F,G).
query_global(global_fact_local(A,B,C,D,E,F,G)) :- !,'/A/global_fact_local'(A,B,C,D,E,F,G).
query_global(theory_fact_local(A,B,C,D,E,F,G)) :- !,'/A/theory_fact_local'(A,B,C,D,E,F,G).
query_global(ancestor(A,B)) :-          !,'/A/ancestor'(A,B).
query_global(parent(A,B)) :-            !,'/A/parent'(A,B).
query_global(subthy_owner(A,B)) :-      !,'/A/subthy_owner'(A,B).
query_global(tactic_file(A)) :-         !,'/A/tactic_file'(A).
query_global(dbms_interface(A)) :-      !,'/A/dbms_interface'(A).
query_global(load_path(A)) :-           !,'/A/load_path'(A).
query_global(path(A)) :-                !,'/A/path'(A).
query_global(request(A,B)) :-           builtin_request(A,B).
query_global(request(A,B)) :-           !,'/A/request'(A,B).
query_global(announcement(A,B)) :-      builtin_announcement(A,B).
query_global(announcement(A,B)) :-      !,'/A/announcement'(A,B).
query_global(handler(A,B)) :-           !,'/A/handler'(A,B).
query_global(listener(A,B)) :-          !,'/A/listener'(A,B).
query_global(doc_node(A,B,C,D)) :-      !,'/A/doc_node'(A,B,C,D).
query_global(doc_fragment(A,B)) :-      !,'/A/doc_fragment'(A,B).
query_global(doc_extension(A,B)) :-     !,'/A/doc_extension'(A,B).
query_global(type(A,B,C,D)) :-          builtin_type(A,B,C,D).
query_global(type(A,B,C,D)) :-          !,'/A/type'(A,B,C,D).
query_global(tactic_wrapper(A,B,C)) :-  !,'/A/tactic_wrapper'(A,B,C).
query_global(tactic(A,B,C)) :-          !,'/A/tactic'(A,B,C).
query_global(parameter(A,B,C,D)) :-     !,'/A/parameter'(A,B,C,D).
query_global(command_table(A,B)) :-     builtin_command_table(A,B).
query_global(command_table(A,B)) :-     !,'/A/command_table'(A,B).
query_global(needs_saving(A)) :-        !,'/A/needs_saving'(A).
query_global(ancestor_of(A,B)) :-        
    !,atom_concat2('/A/ancestor_of_', A, ANC),
    call_predicate(ANC, B).
query_global(Fact) :-
    dbms_asserted_form(global, Fact, RealFact),
    call_predicate(RealFact).

query_theory(Fact) :-
    global_state_lookup(current_dir,Thy),
    query_theory(Fact, Thy).

query_theory(Fact, Thy) :-
    atom(Thy),
    loaded_theory(Thy, Searcher),
    !,
    call_predicate(Searcher,Fact).
query_theory(_, Thy) :-
    handle_error(['Theory ',q(Thy),' is not a legal/loaded theory.']).


query_ancestors(Fact) :-
    query_ancestors(Fact, _).

query_ancestors(Fact, Thy) :-
    ancestor(Thy),
    loaded_theory(Thy, Searcher),
    call_predicate(Searcher,Fact).

query_global_or_ancestors(Fact) :-
    query_global_or_ancestors(Fact, _).

query_global_or_ancestors(Fact, global) :-
    query_global(Fact).
query_global_or_ancestors(Fact, Thy) :-
    ancestor(Thy),
    query_theory(Fact, Thy).


%% private builtin_request(Event,Msg).
%% private builtin_announcement(Event,Msg).
%%
%%  These are predefined entries in the global part of the
%%  database.  Such entries can never be deleted.
%%
%%  mode builtin_request(?compound, ?writelist).
%%  mode builtin_announcement(?compound, ?writelist).
%%
builtin_request(
    msg(_,_,_),
    ['msg(Caller,Importance,Msg).',nl,
     'The tactic or package called Caller wishes to output',nl,
     'an informational message, Msg.  Higher Importance numbers',nl,
     'mark more important/top-level messages.']).
builtin_request(
    error(_,_),
    ['error(Class,Msg).',nl,
     'Class is: data, database, or program.']).
builtin_request(
    write(_,_,_),
    ['write(Stream,Depth,Data) -- write Data to Stream with limit Depth.',nl,
     'Data is: thy(Thy, Term) or code(Thy, Term).']).
builtin_request(
    question(_,_,_),
    ['question(Prompt,Default,Answer) -- ask a question.',nl,
     'Prompt is a writelist, Default is ''yes'' or ''no'', and Answer',nl,
     'should be returned either ''yes'' or ''no''']).
builtin_request(
    reader_special_execute_hook(_),
    ['reader_special_execute_hook(Term) -- execute Term without reprompting.']).
builtin_request(
    signal(_,_),
    ['signal(Signal,Severity) -- handle an asynchronous signal.']).
builtin_request(
    reannotate(_,_),
    ['reannotate(Node,Annotation) -- reannotate a proof node.',nl,
     'A proof is being rerun: Node is the number of the node in the new',nl,
     'proof, and the corresponding node of the old proof had Annoation.',nl,
     'Handler should add an annotation to Node, if that is appropriate.']).
builtin_announcement(
    prover(_),
    ['prover(Action).',nl,
     'Action is: startup, exit, abandon or resume.']).
builtin_announcement(
    theory(_,_),
    ['theory(Name,Action).',nl,
     'Action is: created, deleted, inherit(T), loaded, selected or saved.']).
builtin_announcement(
    assert(_,_),
    ['assert(Thy,Fact) -- Fact has been added to the database.',nl,
     'Thy is ''global'' (for global facts) or a theory name.']).
builtin_announcement(
    proof(_,_,_),
    ['proof(Thy,Name,Action).',nl,
     'This announces the beginning or end of a proof.',nl,
     'Action is one of: started, failed, suspended, resumed,',nl,
     'finished(Saved), where Saved is true/false according to whether',nl,
     'the theorem has been added to the theory or not.',nl,
     'Note that in_proof is true at the time of this announcement,',nl,
     'so procedures like is_focus can be used to get information',nl,
     'about the proof.']).
builtin_announcement(
    wrap_tac(_,_,_,_),
    ['wrap_tac(CallId,Msg,Kind,Action).',nl,
     'This announces the progress of a tactic or one of its subproofs.',nl,
     nl,
     'CallId is an integer that uniquely distinguishes this particular',nl,
     'call to Tac from other tactic calls within the same proof.',nl,
     nl,
     'Kind may be tactic(Tac), where Tac is the call to the tactic',nl,
     '(excluding the list of subproofs), or subproof(N) for the N''th',nl,
     'subproof of the currently executing tactic.',nl,
     nl,
     'Msg is a writelist message about the tactic/subproof.',nl,
     nl,
     'Action will be one of: call, exit or fail.']).
builtin_announcement(
    hyp(Action,SubProof,Number,Term),
    ['hyp(Action, SubProof, Number, Term).',nl,
     'Announces a change to the hypotheses of SubProof.',nl,
     'Action is one of: add(Kind), delete or discharge, where',nl,
     'Kind is inherited, translated, new, proved or conjectured.']).
builtin_announcement(
    read_input(_, _),
    ['read_input(Caller, Input).',nl,
     'Input has been read by Caller.',nl,
     'Caller may be:',nl,
     tab, 'get_command(Prompt, OpTable, Input)',nl,
     tab,tab, '(Input is a term)',nl,
     tab, 'get_ergo_line(Stream, Input)',nl,
     tab,tab, '(Input is a list of character codes).'
    ]).
builtin_announcement(
    param_changed(Param, NewVal),
    ['param_changed(Param, NewVal).', nl,
     'The Ergo parameter ''Param'' changed its value to NewVal.'
    ]).
builtin_announcement(
    need_operators(Thy),
    ['need_operators(Thy).',nl,
     'The prover is about to use the operator table for Thy.'
    ]).
     

%% is_theorem(?Name, +Form)
%% is_theorem(?Name, +Form, +Assumptions)
%%
%%   Perform search for an axiom or theorem in the inherited theories.
%%   Executes the body rather than returning it as an argument.  Consequently
%%   it is unadvisable to pass uninstantiated variables as Form or Assumptions.
%%
is_theorem(Name, Form) :- 
    find_rule(Name, Form, [],  Body), 
    call_predicate(Body).

is_theorem(Name, Form, Ass) :- 
    find_rule(Name, Form, Ass, Body), 
    call_predicate(Body).


%% find_rule(?Name, ?Form, ?Body)
%% find_rule(?Name, ?Form, ?Assumptions, ?Body)
%%
%%   Perform search for an axiom or theorem in the inherited theories.
%%   Returns the guard as an argument (Body), which must be executed later.
%%
find_rule(Name, Form, Body) :- 
    find_rule(Name, Form, [],  Body).

find_rule(Name, Form, Ass, Body) :-
    decode_rule(Name, Th, RuleName), 
    find_rule2(Th, RuleName, Form, Ass, Body).




%% find_rule2('proof', ?Name, ?Form, '[]', 'true')
%% find_rule2(?Theory, ?Name, ?Form, ?Ass, ?Body)
%%
%%   Perform search for an axiom or theorem in the inherited theories.
%%
%%   Internal predicate for rule searching.  Doesn't recognise the Theory:Rule
%%   conventions.
%%
find_rule2(proof, Name, Form, [], true) :-
    is_ass(Name, Form).
find_rule2(Th, Name, Form, Ass, Body) :-
    look_for_theorem(Th:Name, Form),
    rule_at(Th, Name, Form, Ass, Body).


%% rule_at(+Theory, ?Name, ?Form, ?Assumptions, ?Body)
%%
%%   Perform search for an axiom or theorem in a specified theory
%%
rule_at(Th,Name,Form,[],Body) :-
    query_theory(rule(Name,_,_,Form,Body),Th).


%% look_for_theorem(Name, Stmt).
%%
%% Restrict search theories for a partially instantiated theorem by
%% examining the outermost function symbols of the head and (eventually) 
%% assumption clauses.

%% TO DO: Re-enable look_for_theorem.
%% REMOVE this one all references to look_for_theorem have been removed.
%% It has been subsumed by the rule browser.
%% Ray's last comment: Not 100% certain that it has been subsumed.

look_for_theorem(PT:_Name, _Form) :-
    %% (nonvar(PT) ; var(_Form)), !,
    ancestor(PT).
/*
look_for_theorem(PT:Name, Form) :-
    functor(Form, Op, Arity),
    !,
    ( atom(Op) ->
        inherited_symbol(Op, Arity, Children),
        member(PT, Children)
    ;
        ancestor(PT)
    ).

look_for_theorem(PT:Name, Form) :-
    quantify(Form, Q, _, _),
    !,
    ( atom(Q) ->
        inherited_symbol(Q, quant, Children),
        member(PT, Children)
    ;
        ancestor(PT)
    ).

look_for_theorem(PT:Name, Form) :-
    ancestor(PT).
*/


%% known_symbol(Name, Arity).
%%
%%  Succeeds iff Name is declared/defined in the current theory or
%%  one of its ancestors and Arity is compatible with the arity of
%%  that declaration/definition.
%%
%%  mode known_symbol(@atom, @list(ground)).
%%  mode known_symbol(@compound, @compound).
%%
known_symbol(Name, Arity) :-
    query_ancestors(symbol(Name, Arity2, _)),
    (Arity2 = when(Body) ->
        call_predicate(Body)
    ;
        compatible_arity(Arity, Arity2)
    ).
%% TO DO: Do we need to decode Name in known_symbol to get straight to
%% the right theory?
%%
%% The new code would be something like...
%%
%% known_symbol(Name, Arity) :-
%%    split_atom ??? (Name, Theory, UnprefixedName ),
%%    query_theory(symbol(Name, Arity2, _), Theory),
%%    (Arity2 = when(Body) ->
%%        call_predicate(Body)
%%    ;
%%        compatible_arity(Arity, Arity2)
%%    ).
%%
%% Perhaps time it? RN reckone it's almost definitely worth doing.




%% compatible_arity(Arity1, Arity2).
%%
%%  Succeeds iff Arity1 is compatible (an instance of) Arity2.
%%  For the moment (with no unbounded quantifier lists), arity
%%  lists only contain ground terms, so we use unification to
%%  check compatibility.
%%
%%  mode compatible_arity(@list(ground), @list(ground)).
%%
compatible_arity(A, A).


%% decode_rule(RuleName, Thy, Name).
%%
%%  Decodes the Thy:Name convention for rules.
%%
%%  mode decode_rule(?term, ?term, ?term).
%%
decode_rule(Thy:Name, T, N)  :- !, Thy = T, Name = N.
decode_rule(Name, _Th, Name) :- \+ integer(Name).


%% decode_comp_rule(RuleName, Name, Trace).
%%
%%  RuleName is the internal name of a particular application of the
%%  compiled rule called Name (without any theory prefix) and Trace
%%  describes what how the compiled rule executed.
%%
%%  mode decode_comp_rule(+nonvar, ?atom, ?nonvar).
%%
decode_comp_rule(cr(Name, Tr), Name, Trace) :- !, Tr = Trace.
decode_comp_rule(Name, Name, _Trace).


%% decode_subthy_rule(Rule, Thy, Name).
%%
%%  Either Rule is the unqualified Name and Thy is the current theory,
%%  or Rule is Thy:Name and Thy is a subtheory of the current theory.
%%
%% mode decode_subthy_rule(@compound, -atom, -atom).
%%
decode_subthy_rule(Rule, Thy, Name) :-
    check_current_theory(CD),
    decode_rule(Rule, Thy, Name),
    ( Thy = CD ->
        true
    ;
        check_or_error(
            query_global(subthy_owner(Thy, CD)),
            [q(Thy), ' is not a subtheory of the current theory.'])
    ).


%% is_preorder(Reln).
%%
%%  Succeeds iff Reln is a known preorder (includes equivalence relations)
%%  possibly with one or more rev(_) wrappers to reverse its direction.
%%
%%  mode is_preorder(@ground).
%%
is_preorder(0) :-
    %% catch the var case.
    !,
    handle_illegal_arg(is_preorder, 1, 'preorder or rev(...)').
is_preorder(ERGO_EQUALS) :-
    !.
is_preorder(ERGO_IMPLIES) :-
    !.
is_preorder(ERGO_EQUIV) :-
    !.
is_preorder(rev(Rel)) :-
    !,
    is_preorder(Rel).
is_preorder(Rel) :-
    query_ancestors(preorder(Rel)),
    !.


%% is_equivalence(Reln).
%%
%%  Succeeds iff Reln is a known equivalence relation.
%%  One or more rev(_) wrappers are allowed, although they
%%  do not change the meaning of equivalence relations.
%%
%%  mode is_equivalence(@ground).
%%
is_equivalence(0) :-
    %% catch the var case.
    !,
    handle_illegal_arg(is_equivalence, 1, 'equivalence or rev(...)').
is_equivalence(ERGO_EQUALS) :-
    !.
is_equivalence(ERGO_EQUIV) :-
    !.
is_equivalence(rev(Rel)) :-
    !,
    is_equivalence(Rel).
is_equivalence(Rel) :-
    query_ancestors(equivalence(Rel)),
    !.


%% is_stronger_than(Reln1, Reln2).
%%
%%  Succeeds iff Reln1 is stronger than (or equal to) Reln2
%%  and both relations are known to be preorders or rev/1 applied
%%  once or twice to a known preorder.
%%
%%  NB. The implementation of this is tricky and has been through
%%      at least ten revisions!  Soundness is easy, but completeness
%%      can be quite difficult to achieve.  The current implementation
%%      relies on the following invariants of the Ergo database:
%%
%%        1) for all equivalences E, it contains: reverse_of(E,E).
%%           (see add_equivalence).
%%        2) for all equivalences E and non-equivalence preorders P,
%%           it contains stronger_than(E, rev(P)) whenever it
%%           contains stronger_than(E,P).  (see add_stronger_than).
%%        3) whenever it contains reverse_of(P,RP), it also contains
%%           reverse_of(RP,P).  (see add_reverse_of).
%%
%%  mode is_stronger_than(@ground, @ground).
%%
is_stronger_than(0,_) :-
    %% catch the var case first.
    !,
    handle_illegal_arg(is_stronger_than, 1, 'preorder or rev(...)').
% now the common cases from the ergo theory (avoids calling query_ancestors).
is_stronger_than(ERGO_EQUALS,Rel) :-
    equals_is_stronger_than(Rel),
    !.
is_stronger_than(ERGO_EQUIV,Rel) :-
    equiv_is_stronger_than(Rel, Flag),
    !,
    Flag = true.
is_stronger_than(ERGO_IMPLIES,Rel) :-
    implies_is_stronger_than(Rel, Flag),
    !,
    Flag = true.
is_stronger_than(rev(Rel1),Rel2) :-
    rev_is_stronger_than(Rel1, Rel2, Flag),
    !,
    Flag = true.
% the general case.
is_stronger_than(A,B) :-
    simplify_relation(A,A2),
    simplify_relation(B,B2),
    (A2 = B2 ->
        is_preorder(A2)
    ; A2 = rev(A3) ->
        query_ancestors(stronger_than(A3, B3)),
        simplify_rev_relation(B3,B2)
    ;
        query_ancestors(stronger_than(A2, B3)),
        simplify_rev_relation(B3,B2)
    ),
    !.

equals_is_stronger_than(0) :- !,fail.  %% catch var case.
equals_is_stronger_than(ERGO_EQUALS) :- !.
equals_is_stronger_than(ERGO_EQUIV) :- !.
equals_is_stronger_than(ERGO_IMPLIES) :- !.
equals_is_stronger_than(rev(Rel)) :- equals_is_stronger_than(Rel).

equiv_is_stronger_than(0, _) :- !,fail.  %% catch var case.
equiv_is_stronger_than(ERGO_EQUALS, false) :- !.
equiv_is_stronger_than(ERGO_EQUIV, true) :- !.
equiv_is_stronger_than(ERGO_IMPLIES, true) :- !.
equiv_is_stronger_than(rev(Rel), Flag) :- equiv_is_stronger_than(Rel, Flag).

implies_is_stronger_than(0, _) :- !,fail.  %% catch var case.
implies_is_stronger_than(ERGO_EQUALS, false) :- !.
implies_is_stronger_than(ERGO_EQUIV, false) :- !.
implies_is_stronger_than(ERGO_IMPLIES, true) :- !.
implies_is_stronger_than(rev(Rel), Flag) :-
    implies_is_stronger_than_rev(Rel, Flag).

implies_is_stronger_than_rev(0, _) :- !,fail.  %% catch var case.
implies_is_stronger_than_rev(ERGO_EQUALS, false) :- !.
implies_is_stronger_than_rev(ERGO_EQUIV, false) :- !.
implies_is_stronger_than_rev(ERGO_IMPLIES, false) :- !.

rev_implies_is_stronger_than(0, _) :- !,fail.  %% catch var case.
rev_implies_is_stronger_than(ERGO_EQUALS, false) :- !.
rev_implies_is_stronger_than(ERGO_EQUIV, false) :- !.
rev_implies_is_stronger_than(ERGO_IMPLIES, false) :- !.
rev_implies_is_stronger_than(rev(Rel), Flag) :-
    implies_is_stronger_than(Rel, Flag).

rev_is_stronger_than(0, _, _) :-
    %% catch var case.
    !,
    fail.
rev_is_stronger_than(ERGO_EQUALS, Rel2, _) :-
    !,
    equals_is_stronger_than(Rel2).
rev_is_stronger_than(ERGO_EQUIV,  Rel2, Flag) :-
    !,
    equiv_is_stronger_than(Rel2, Flag).
rev_is_stronger_than(ERGO_IMPLIES, Rel2, Flag) :-
    rev_implies_is_stronger_than(Rel2, Flag).


%% simplify_relation(Reln0, Reln).
%% simplify_rev_relation(RevReln0, Reln).
%%
%%  simplify_relation simplifies Reln0 down to an equivalent relation
%%  Reln which is in canonical form, by removing rev/1 wrappers where
%%  possible.  Calling simplify_rev_relation(P, R) is equivalent to
%%  calling simplify_relation(rev(P),R).
%%  
%%  They do not require the first argument to be a known preorder.
%%
%%  NB. the implementation of these relies on the fact that all
%%     equivalence relations E have reverse_of(rev(E),E) recorded.
%%     (add_equivalence/1 ensures that this is the case).
%%
%%  mode simplify_relation(@ground, ?ground).
%%  mode simplify_rev_relation(@ground, ?ground).
%%
simplify_relation(0, _) :-
    %% catch the var case.
    !,
    handle_illegal_arg(simplify_relation, 1, 'preorder or rev(...)').
simplify_relation(rev(A), B) :-
    !,
    simplify_rev_relation(A, B).
simplify_relation(A, A) :-
    atom(A).

simplify_rev_relation(0, _) :-
    %% catch the var case.
    !,
    handle_illegal_arg(simplify_rev_relation, 1, 'preorder or rev(...)').
simplify_rev_relation(ERGO_EQUALS, Rel) :-
    !,
    Rel = ERGO_EQUALS.
simplify_rev_relation(ERGO_EQUIV, Rel) :-
    !,
    Rel = ERGO_EQUIV.
simplify_rev_relation(rev(A), B) :-
    !,
    simplify_relation(A,B).
    %% TO DO: If the above clause of simplify_relation works,
    %% find out why it does.
simplify_rev_relation(A, B) :-
    query_ancestors(reverse_of(A,B)),
    !.
simplify_rev_relation(A, rev(A)) :-
    atom(A).
