%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: save.ql,v 1.67 2002/06/27 05:10:19 itp Exp $

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The procedures in this file handle the saving and loading
%% of theory-specific and global database facts to and from files.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% The current version number of the *.erg file format.
theory_file_version('2.0').


database command
  touch_theory(Thy)
extends
  save_exit_cmds
doc
  'Marks Thy so that it will be saved by the next save command.',nl,
  'Thy must be the name of a loaded theory, or ''global''.'
mode
  touch_theory(@atom).

touch_theory(Thy) :-
    loaded_theory(Thy, _),
    touch_database(Thy).
touch_theory(Thy) :-
    handle_error(['Theory ', q(Thy), ' is not loaded.',nl]).


%% This is not a general command, because that might lead to confusion.
%% (If does not save incomplete proofs, for instance).

database command
  save
extends
  save_exit_cmds
mode
  save
doc
  'Save all modified theories and global data.'.

save :-
    op_table(proof),
    obvar_prefix_table(proof),
    save_global_data,
    save_theory_data.

%% private save_theory_data.
%%
%%  Save the facts in all updated theories.
%%
%%  mode save_theory_data.
%%
save_theory_data :-
    query_global(needs_saving(Thy)),
    atom_append(Thy, '.erg.qge', FileName),
    save_to(FileName, Stream),
    dbms_loaded_searcher_name(Thy, Searcher),
    concat_atom([Thy, '/Version/'], VerPred),
    theory_file_version(FileFormat),
    save_thy_line(Stream, VerPred(FileFormat)),
    save_thy_line(Stream, (?- asserta(loaded_theory(Thy,Searcher)))),
    save_theory_searcher(Stream, Thy, Searcher),
    save_thy_lines(Stream, Thy),
    saved(Stream),
    announce_event(theory(Thy, saved)),
    fail.
save_theory_data :-
    retractall_global(needs_saving(_)).


%% private save_theory_searcher(Stream, Theory, Searcher).
%%
%%  Writes out the `searcher' predicate (called Searcher/1) for Theory.
%%  The searcher predicate is an index of all the fact types in
%%  the saved file.  For example, a searcher called `s'
%%  (their real names are more complex) for theory tt might
%%  look like, where ff/1 facts have the assert(z) and retract(false)
%%  properties and gg/2 facts have the assert(a), retract(true) and
%%  key(2) properties:
%%
%%     s(ff(F))   :- 'tt/L/ff'(F).  (loaded ff facts)
%%     s(ff(F))   :- !,'tt/A/ff'(F).  (ff facts asserted during this session)
%%     s(gg(F,G)) :- ,'s/A/gg'(F,G).
%%     s(gg(F,G)) :- !,'tt/L/gg'(F,G),
%%                   \+ 'tt/R/gg'(G).
%%     ...
%%
%%  mode save_theory_searcher(@stream, @theory).
%%
save_theory_searcher(Stream, Thy, _) :-
    %% Output a dynamic declaration for each `retraction' fact.
    theory_fact(F,N,_,true,_,true,_),
    %% Filter out facts that do not appear in the theory being saved.
    is_present_in_theory(F, N, Thy),
    functor(Fact, F, N),
    dbms_retracted_form(Thy, Fact, RFact),
    functor(RFact, RF, RN),
    save_thy_line(Stream, (?- dynamic(RF/RN))),
    fail.
save_theory_searcher(Stream, Thy, Searcher) :-
    %% Do the clauses for each kind of saved fact that contains data.
    theory_fact(F,N,_,true,AZ,Retract,_),
    functor(Fact, F, N),
    save_searcher_clause(AZ,Retract,Stream,Thy,Searcher,Fact),
    fail.
save_theory_searcher(Stream, Thy, Searcher) :-
    %% Finish with the generic case to handle future kinds of facts.
    save_thy_line(Stream,
                  Searcher(Fact),
                  ( dbms_asserted_form(Thy,Fact,AFact),
                    call_predicate(AFact))).


save_searcher_clause(_, _, Stream, Thy, Searcher, Fact) :-
    functor(Fact, F, N),
    \+ is_present_in_theory(F, N, Thy),
    !,
    %% No matching saved facts, so we only need to output the asserted case.
    dbms_asserted_form(Thy, Fact, AFact),
    save_thy_line(Stream, Searcher(Fact), (!,AFact)).
save_searcher_clause(a, Retract, Stream, Thy, Searcher, Fact) :-
    dbms_loaded_form(Thy, Fact, LFact),
    dbms_asserted_form(Thy, Fact, AFact),
    save_thy_line(Stream, Searcher(Fact), AFact),
    ( Retract = true ->
        dbms_retracted_form(Thy, Fact, RFact),
        save_thy_line(Stream, Searcher(Fact), (!,LFact,\+ RFact))
    ;
        save_thy_line(Stream, Searcher(Fact), (!,LFact))
    ).
save_searcher_clause(z, Retract, Stream, Thy, Searcher, Fact) :-
    dbms_loaded_form(Thy, Fact, LFact),
    dbms_asserted_form(Thy, Fact, AFact),
    ( Retract = true ->
        dbms_retracted_form(Thy, Fact, RFact),
        save_thy_line(Stream, Searcher(Fact), (LFact,\+ RFact))
    ;
        save_thy_line(Stream, Searcher(Fact), LFact)
    ),
    save_thy_line(Stream, Searcher(Fact), (!,AFact)).


%% is_present_in_theory(F, N, Thy).
%%
%%  Succeeds if Thy contains any F/N facts.
%%
is_present_in_theory(infrule_apply, _, Thy) :-
    !,
    is_present_in_theory(infrule, 7, Thy).
is_present_in_theory(F, N, Thy) :-
    functor(Fact, F, N),
    query_theory(Fact, Thy),
    !.


%% private save_thy_lines(Stream, Theory).
%%
%%  Save every instance of every fact in the given theory.
%%
%%  mode save_thy_lines(@stream, @atom).
%%
save_thy_lines(Stream, Theory) :-
    theory_saved_fact(F,N),
    functor(Fact, F, N),
    debug(save_theory, ['saving ',q(F/N)]),
    %% construct the internal equivalent of Fact.
    dbms_loaded_form(Theory, Fact, LFact),
    ( Fact = infrule_apply(SName,Name,Concl,Id,SeqIds) ->
        %% We cannot query these facts, so we generate them from
        %% the corresponding infrule/7 facts instead.
        query_theory(infrule(SName,TName,Concl,Id,Seqs,Before,After), Theory),
        mode_from_rule(TName, Name, _),
        debug(save_theory, ['saving ',q((LFact :- Code))]),
        rule_compile(Id, (rule_unify(Concl, BConcl), Before), 
                     After, Seqs, SeqIds, Code),
        functor(LFact, LFun, _),
        BLFact = LFun(SName,Name,BConcl,Id,SeqIds),
        debug(save_theory, ['saving ',q((BLFact :- Code))]),
        save_thy_line(Stream, BLFact, Code),
        fail
    ;
        query_theory(Fact, Theory),
        save_thy_line(Stream, LFact),
        fail
    ).
save_thy_lines(Stream, Theory) :-
    query_theory(save_clauses(Name, Arity, ThyPos), Theory),
    functor(Head, Name, Arity),
    (ThyPos > 0 ->
        arg(ThyPos, Head, Theory)
    ;
        true
    ),
    clause(Head, Body),
    save_thy_line(Stream, Head, Body),
    fail.
save_thy_lines(Stream, Theory).


%% private save_thy_line(Stream, Head).
%% private save_thy_line(Stream, Head, Body).
%%
%%  Write the clause  (Head :- Body)  to the stream Stream.
%%
%%  mode save_thy_line(@stream, @nonvar).
%%  mode save_thy_line(@stream, @nonvar, @nonvar).
%%
save_thy_line(Stream, Head) :-
    encoded_write_term(Stream, Head, []).

save_thy_line(Stream, Head, Body) :-
    encoded_write_term(Stream, (Head :- Body), []).


%% private encoded_save_thy_line(Stream, Head).
%% private encoded_save_thy_line(Stream, Head, Body).
%%
%%  Write the clause  (Head :- Body)  to the stream Stream in
%%  encoded form.
%%
%%  mode encoded_save_thy_line(@stream, @nonvar).
%%  mode encoded_save_thy_line(@stream, @nonvar, @nonvar).
%%
encoded_save_thy_line(Stream, Head) :-
    encoded_write_term(Stream, Head, []).
encoded_save_thy_line(Stream, Head, Body) :-
    encoded_write_term(Stream, ':-'(Head, Body), []).


%% private save_clause_body(Stream, Code).
%%
%%  auxillary procedure for save_thy_line/2.
%%
save_clause_body(Stream, (A,B)) :-
    !,
    tab(Stream, 8),
    writeq(Stream, A),
    write(Stream, ','),
    nl(Stream),
    save_clause_body(Stream, B).
save_clause_body(Stream, B) :-
    tab(Stream, 8),
    writeq(Stream, B),
    write(Stream, '.'),
    nl(Stream).



%% private save_global_data.
%%
%%  Save all global facts if the global database has been updated.
%%
%%  mode save_global_data.
%%
save_global_data :-
    query_global(needs_saving(global)),
    !,
    save_to('global.erg', Stream),
    unwind_protect(save_global(Stream),
                   saved(Stream)),
    retractall_global(needs_saving(global)).
save_global_data.


%% private save_global.
%%
%%  Saves all global information to the current output stream.
%%
%%  mode save_global.
%%
save_global(Stream) :-
    %% Declare all extra dynamic facts first.
    query_global(global_fact(Name,Arity,_,_,_,_,_)),
    dbms_asserted_name(global, Name, RealName),
    encoded_write_term(Stream, (?- dynamic(RealName/Arity)), []),
    fail.
save_global(Stream) :-
    global_saved_fact(F,Arity),
    %% NB. we cannot call query_global here, because it would
    %%     return the inbuilt facts, which we do not need to save.
    dbms_asserted_name(global, F, RealF),
    functor(RealFact, RealF, Arity),
    call_predicate(RealFact),
    encoded_save_thy_line(Stream, RealFact),
    fail.
save_global(_).



proof_file_name(Theory:Name, FileName) :-
    atom_append(Name, '.pf', PfName),
    atomlist_to_atom([Theory, PfName], FileName).

save_to(File, Stream) :-
    write('Saving '),
    write(File),
    write(' ... '),
    flush_output,
    open(File, write, Stream).

saved(Stream) :-
    close(Stream),
    write('saved.'), nl.

save_to(File) :-
    write('Saving '),
    write(File),
    write(' ... '),
    flush_output,
    tell(File).

saved :-
    told,
    write('saved.'), nl.


%% consult_thy(+TheoryName, -Ver, -Path)
%%
%%  Reads and asserts the contents of theory TheoryName.
%%
%%  Returns the path in which the file was found, and the version of 
%%  'save' which was used to create the file.
%%
%%  mode consult_thy(@theory, -atom, -nonvar).
%%
consult_thy(Theory, Ver, Path) :-
    get_load_path(PathList),
    member(Path, PathList),
    full_file_name(Path, Theory, Dir, PSFile),
    ( complete_filename(PSFile, '.erg.qo', File) ->
        msg(loading, ['Theory ',q(Theory),' (qo)  from  ',w(Dir)]),
        load(File)
    ; complete_filename(PSFile, '.erg.qge', File) ->
        msg(loading, ['Theory ',q(Theory),' (qge)  from  ',w(Dir)]),
        op_table(user),
        obvar_prefix_table(user),
        consult(File)
    ;
        %% For backwards compatibility, we read old theory files.
        complete_filename(PSFile, '.erg', File) ->
        msg(loading, ['Theory ',q(Theory),'  from  ',w(Dir)]),
        op_table(user),
        obvar_prefix_table(user),
        consult(File)
    ),
    !,
    %% NB. loading/consulting the file will have asserted
    %%     a loaded_theory(Theory,...) fact and declared various
    %%     predicates to be dynamic.
    concat_atom([Theory, '/Version/'], VerPred),
    VerPred(Ver).
consult_thy(Theory, _V, _P) :-
    handle_error(['unable to load theory: ',q(Theory)]).


database command
  save_snapshot(File)
extends
  save_exit_cmds
mode
  save_snapshot(@filename)
doc
  'Save a snapshot of the current session into File.',nl,
  'A saved session can be resumed later using the shell',nl,
  'command, "restore File".',
  para,
  'Expect some warnings about new definitions being ignored',nl,
  'when you execute this command and when you restore the',nl,
  'resulting snapshot.'.
%%
%%  Unfortunately, there is no good way of distinguishing the current
%%  session from the restored one, so we *always* print banners etc.
%%  after the save.
%%
save_snapshot(File) :-
    atom(File),
    save_all_ctrs,
    save(File),
    %% os(argv(Args)),
    %% command_args(Args),
    banner,
    writelist(['Preloaded version.',nl,
               '(Please ignore the load foreign warnings)',nl]), 
    do_load_foreign,
    restore_all_ctrs,
    reconsult_global,
    announce_event(prover(resume)).
