%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% ERGO COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1993 by The University of Queensland, Queensland 4072 Australia
%% 
%% 
%% Permission to use, copy and distribute this software for any
%% non-commercial purpose and without fee is hereby granted, provided
%% that the above copyright notice and this permission notice and
%% warranty disclaimer appear in all copies and in supporting
%% documentation, and that the name of the University of Queensland not
%% be used in advertising or publicity pertaining to distribution of
%% the software without specific, written prior permission.
%% 
%% Source code modifications are prohibited except where written 
%% agreement has been given in advance by the University of Queensland.
%% 
%% The University of Queensland disclaims all warranties with regard to
%% this software, including all implied warranties of merchantability
%% and fitness.  In no event shall the University of Queensland be
%% liable for any special, indirect or consequential damages or any
%% damages whatsoever resulting from loss of use, data or profits,
%% whether in an action of contract, negligence or other tortious
%% action, arising out of or in connection with the use or performance
%% of this software.
%% 
%% CVS $Id: syntax2.ql,v 1.12 1999/10/25 18:48:20 itp Exp $

#include "../lib/ergo.h"


/*
* Check a function definition
* Check 
* (a) the only FREE object variables in Term are members of ObVarList;
* (b) Term contains only legal syntax
*/


legal_func_term(Term, ObVarList) :-  
    substitute(Term, Sub, SubTerm),
    legal_func_sub(Sub, ObVarList, NewObVarList),
    !,
    legal_func_subterm(SubTerm, NewObVarList).
legal_func_subterm(Term, ObVarList) :- 
    obvar(Term),
    !,
    check_or_error(
        member(Term, ObVarList),
        ['illegal object variable ',q(Term),' in function definition']).
legal_func_subterm(Term, ObVarList) :- 
    var(Term),
    !,
    handle_error(['illegal meta-variable ',q(Term),' in function definition']).
legal_func_subterm(Term, ObVarList) :- 
    quantify(Term, Q, V, Body),
    !,
    check_or_error(
        inherited_symbol(Q, quant, _),
        ['quantifier ', q(Q), ' not defined in this theory']),
    legal_func_term(Body, [O|ObVarList]).
legal_func_subterm(Term, ObVarList) :- 
    Term =.. [Func|Args],
    length(Args, N),
    !,
    check_or_error(
        inherited_symbol(Func, N, _),
        ['function ', q(Func), '/', q(N), ' not defined in this theory']),
    legal_func_terms(Args, ObVarList).
legal_func_terms(AL, ObVars) :-
    \+ (member(A, AL),
        \+ legal_func_term(A, ObVarList)).

legal_func_sub([], ObVarList, ObVarList). 
legal_func_sub([H|Sub], ObVarList, NewObVarList) :-
    legal_func_pllsub(H, ObVarList, Vars),
    append(Vars, ObVarList, ObVarList1),
    legal_func_sub(Sub, ObVarList1, NewObVarList).

legal_func_pllsub([], _, []).
legal_func_pllsub([(R/D)|T], ObVarList, [D|Vars]) :-
    legal_func_term(R, ObVarList), 
    legal_func_pllsub(T, ObVarList, Vars).


/*
* Check a quantifier definition
* Check 
* (a) all occurrences of Body are in the binding of Bound
* (b) the only schematic variable in Term is Body
* (c) there are no free object variables
* (d) Term contains only legal syntax
*/

legal_quant_term(Term, Bound, Body) :-
    legal_var_term(Term, Bound, [], Body, [], Provided_bound_vars),
    gen_nfi(Body, Provided_bound_vars, Bound),
    !.

legal_var_term(Term, X, ObVarList, Body, PBVars0, PBVars) :-  
    substitute(Term, Sub, SubTerm),
    legal_var_sub(Sub, X, ObVarList, NewObVarList, Body, PBVars0, PBVars1),
    !,
    legal_var_subterm(SubTerm, X, NewObVarList, Body, PBVars1, PBVars).

legal_var_subterm(Term, X, ObVarList, Body, PBVars0, PBVars0) :- 
    obvar(Term),
    !,
    (member(Term, ObVarList) ->
        true
    ;
        write('illegal object variable in definition'), nl,
        fail
    ).
legal_var_subterm(Term, X, ObVarList, Body, PBVars0, PBVars0) :- 
    var(Term),
    !,
    (Term = Body -> 
        (member(X, ObVarList) ->
            true
        ; 
            write('the bound variable is not correctly bound in definition'),
            nl,
            fail
        )
    ;
        write('illegal schematic variables in definition'),
        nl,
        !,
        fail
    ).
legal_var_subterm(Term, X, ObVarList, Body, PBVars0, PBVars) :- 
    quantify(Term, TQ, TOV, TB),
    !,
    check_or_error(
        inherited_symbol(Q, quant, _),
        ['quantifier ', q(Q), ' not defined in this theory.']),
    legal_var_term(TB, X, [TOV|ObVarList], Body, [TOV|PBVars0], PBVars).
legal_var_subterm(Term, X, ObVarList, Body, PBVars0, PBVars) :- 
    Term =.. [Func|Args],
    !,
    length(Args, N),
    check_or_error(
        inherited_symbol(Func, N, _),
        ['function ',q(Func),'/',q(N),' not defined in this theory.']),
    legal_var_terms(Args, X, ObVarList, Body, PBVars0, PBVars).

legal_var_terms([], _, _, _, _, []) :-
    !.
legal_var_terms([A|AL], X, ObVarList, Body, PBVars0, PBVars) :-
    !, 
    legal_var_term(A, X, ObVarList, Body, PBVars0, PBVars1),
    legal_var_terms(AL, X, ObVarList, Body, PBVars0, PBVars2),
    append(PBVars1, PBVars2, PBVars).

legal_var_sub([], _, ObVarList, ObVarList, _, PBVars, PBVars). 
legal_var_sub([H|Sub], X, ObVarList, SubtedVars, Body, PBVars0, PBVars) :-
    legal_var_pllsub(H, X, ObVarList, Vars, Body, PBVars0, PBVars1),
    append(Vars, ObVarList, ObVarList1),
    legal_var_sub(Sub, X, ObVarList1, SubtedVars, Body, PBVars1, PBVars).

legal_var_pllsub([], _, _, [], _, PBVars, PBVars).
legal_var_pllsub([(R/D)|T], X, ObVarList, [D|Vars], Body, PBVars0, PBVars) :-
    legal_var_term(R, X, ObVarList, Body, PBVars0, PBVars1), 
    legal_var_pllsub(T, X, ObVarList, Vars, Body, [D|PBVars1], PBVars2).

gen_nfi(_, [], _) :-
    !.
gen_nfi(Body, [Bound|A], Bound) :-
    !,
    gen_nfi(Body, A, Bound).
gen_nfi(Body, [Z|A], Bound) :-
    member(Z, A),
    !,
    gen_nfi(Body, A, Bound).
gen_nfi(Body, [Z|A], Bound) :-
    Z not_free_in Body,
    gen_nfi(Body, A, Bound).




/*
** replace object vars of the term (given in the obvarlist) with metavars
** done by generating a parallel sub that replaces the obvars in list with
** new metavars
** e.g. rep_obs_with_vars(f(x,y), [x,y], [X/x,Y/y] * f(x,y))
*/

rep_obs_with_vars(Term, [], Term) :-
    !.
rep_obs_with_vars(Term, ObVarList, NewTerm) :-
    gen_sub_list(ObVarList, SubList),
    substitute(NewTerm, [SubList], Term).


/*
** Generate a list representing a parallel sub from a list of
** objectvars e.q. gen_sub_list([x1,...,xn], [X1/x1,...,Xn/xn]).
*/

gen_sub_list([], []).
gen_sub_list(OList, SList) :-
    OList = [Y|OL],
    SList = [(X/Y)|SL],
    gen_sub_list(OL, SL).


/*
** Check if list contains only entries of form x not_free_in A
** and execute them
*/

check_nfi([]).
check_nfi(true).
check_nfi([(X not_free_in A)|Rest]) :-
    var(A),
    X not_free_in A,
    check_nfi(Rest).

