%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: unix.ql,v 1.104 2006/02/05 22:15:23 itp Exp $

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file contains procedures that may be UNIX-specific, such
%% as handling command line arguments, invoking sub-processes
%% and manipulating directory names and file extensions.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% cmdline_args(ArgList).
%%
%%  Processes the command line arguments given when Ergo is started.
%%
%%  mode cmdline_args(@list(ground)).
%%
cmdline_args([]).
cmdline_args([ergo4 | Tail]) :-
    !,
    cmdline_args(Tail).
cmdline_args(['-V' | Tail]) :-
    !,
    set_verbosity(2),
    cmdline_args(Tail).
cmdline_args(['-G', Global | Tail]) :-
    !,
    assertz(global_file_location(Global)),
    cmdline_args(Tail).
cmdline_args(['-dbg', Name | Tail]) :-
    atom(Name),
    !,
    assertz(dbg(Name)),
    cmdline_args(Tail).
cmdline_args(['-listen', Name | Tail]) :-
    query_global(announcement(Ann, _)),
    functor(Ann,Name,_),
    !,
    add_listener(Ann, writelist([nl,'ANNOUNCEMENT: ',q(Ann),nl])),
    cmdline_args(Tail).
cmdline_args(['-spy', Name | Tail]) :-
    atom(Name),
    !,
    spy(Name),
    cmdline_args(Tail).
cmdline_args(['-lib', L | Tail]) :-
    !,
    assertz(library_location(L)),
    cmdline_args(Tail).
cmdline_args(['-foreign', ForeignDir | Tail]) :-
    !,
    retractall(foreign_dir(_)),
    assertz(foreign_dir(ForeignDir)),
    cmdline_args(Tail).
cmdline_args(['-buildforeign', ForeignDir | _]) :-
    !,
    retractall(foreign_dir(_)),
    assertz(foreign_dir(ForeignDir)),
    writelist(['Building foreign files ...', nl]),
    do_create_foreign,
    writelist(['Building foreign files ... done', nl]),
    halt(0).
cmdline_args([Arg | _]) :-
    handle_error(['unknown command line argument: ',q(Arg)]).


%% full_file_name(Path, BaseName, RealPath, FullName).
%%
%%  Path may be an atom or a lib(P) term, where P is an atom.
%%
%%  RealPath is constructed from Path by expanding lib(P)
%%  into ErgoLibraryPath/P, where ErgoLibraryPath is defined
%%  using the -lib switch to Ergo.
%%
%%  FullName is an atom corresponding to RealPath/BaseName.
%%
%%  mode full_file_name(@ground, @atom, -atom, -atom).
%%
full_file_name('.', File, '.', File) :-
    !.
full_file_name(lib(P), File, Dir, FullFile) :-
    !,
    library_location(L),
    atomlist_to_atom([L, P], Dir),
    atomlist_to_atom([L, P, File], FullFile).
full_file_name(P,   File, P, FullFile) :-
    atomlist_to_atom([P, File], FullFile).


%% do_consult(File, Expansions).
%%
%%  Use to load tactic files only.  Use consult_thy and consult_global
%%  for theory and global files respectively.
%%  Expansions can include: decl(Thy), macro, prefix(Thy), thy(Thy).
%%
%%  mode do_consult(@atom, @list(nonvar)).
%%
do_consult(File, _) :-
    loaded(File,_,_),
    !,
    handle_error(['File already loaded: ',q(File)]).
do_consult(File, Exps) :-
    get_load_path(PathList),
    member(P, PathList),
    full_file_name(P, File, Dir, PSFile),
    msg(loading, 2, ['  looking for ',w(File),' in ',w(Dir)]),
    complete_filename(PSFile, Extn, NewFileName),
    !,
    msg(loading, ['  ',w(File),w(Extn),'  from  ',w(Dir)]),
    ( query_global(handler(fileoperation(Extn,consult,PSFile),Handler)) ->
        call_predicate(Handler)
    ; Extn = '.qo' ->
        load(NewFileName)
    ; Extn = '.ql' ->
        pathname(PSFile,_,CutFileName),
        concat_atom(['/tmp/',CutFileName,'.qi'], QIFile),
        concat_atom(['$QPPATH/qppp -C -P ', NewFileName, ' > ', QIFile], Cmd),
        os(system(Cmd, 0)),
        op_table(user),
        obvar_prefix_table(user),
        consult_with_expansions(Exps, QIFile),
        concat_atom(['/bin/rm ', QIFile], RmCmd),
        os(system(RmCmd))
    ; Extn = '.qi' ->
        op_table(user),
        obvar_prefix_table(user),
        consult_with_expansions(Exps, NewFileName)
    ; Extn = '.qg' ->
        op_table(user),
        obvar_prefix_table(user),
        consult_with_expansions([], NewFileName)
    ;        
        fail
    ),
    assertz(loaded(File,Extn,NewFileName)).
do_consult(File, _) :-
    handle_error(['Unable to find file: ',q(File)]).


%% do_reconsult(File, Expansions).
%%
%%  Reconsult a given tactic file.
%%  Only applicable to non-compiled tactic files.
%%
%%  Warning: the consulted file should not contain multi-file predicates. 
%%
do_reconsult(FileName, _, _, Exps) :-
    loaded(FileName, Type, FullFileName),
    !,
    msg(reconsulting, ['  ',w(FullFileName)]),
    ( query_global(handler(fileoperation(Type,reconsult,FullFileName),
                   Handler)) ->
        call_predicate(Handler)
    ; Type = '.ql' ->
        pathname(FileName,_,CutFileName),
        concat_atom(['/tmp/',CutFileName,'.qi'], QIFile),
        concat_atom(['$QPPATH/qppp -C -P ', FullFileName, ' > ', QIFile], Cmd),
        os(system(Cmd,0)),
        op_table(user),
        obvar_prefix_table(user),
        reconsult_with_expansions(Exps, QIFile),
        concat_atom(['/bin/rm ',QIFile], RmCmd),
        os(system(RmCmd))
    ; Type = '.qi' ->
        op_table(user),
        obvar_prefix_table(user),
        reconsult_with_expansions(Exps, FullFileName)
    ; Type = '.qg' ->
        op_table(user),
        obvar_prefix_table(user),
        reconsult_with_expansions([], FullFileName)
    ; Type = '.qo' ->
        handle_error(['cannot reconsult compiled file: ',q(FileName)])
    ;
        fail
    ).
do_reconsult(FileName, Thy, File, Exps) :-
    % \+ loaded(FileName, Type, FullFileName)
    %% FileName is not loaded yet, but it should have been.
    %% This typically happens when errors were detected in the
    %% tactic file as the theory was loaded.
    query_theory(tactic_file(File), Thy),
    !,
    do_consult(FileName, Exps).
do_reconsult(FileName, _, _, _) :-
    handle_error(['file ',q(FileName),' is not loaded.']).


%% do_compile(File, Expansions, Always).
%%
%%  Compile/recompile File with Expansions.
%%  Expansions can include: decl(Thy), macro, thy(Thy), prefix(Thy).
%%  Compiles only if necessary, unless Always=yes.
%%
%%  mode do_compile(@filename, @list(nonvar), @atom).
%%
do_compile(File, Exps, Always) :-
    get_load_path(PathList),
    member(P, PathList),
    full_file_name(P, File, Dir, PSFile),
    msg(compile, 2, ['looking for ',w(File),' in ',w(Dir)]),
    complete_filename(PSFile, Extn, Source),
    ( query_global(handler(fileoperation(Extn,compile(Always),PSFile),Handler)) ->
        !,
        call_predicate(Handler)
      ;
        Extn = '.ql',
        !,
        ( Always \== yes,
          complete_filename(PSFile, '.qo', Object),
          file_newer_than(Object, Source) ->
              msg(compile, [w(Source),' does not need recompiling.'])
        ;
            check_compile_okay(Dir,Source),
            msg(compile, ['compiling ',w(Source),'.']),
            op_table(user),
            obvar_prefix_table(user),
            compile_with_expansions(Exps, Source)
        )
    ).
do_compile(File, _, _) :-
    handle_error(['Unable to find file: ',q(File)]).

check_compile_okay('.', _) :-
    !.
check_compile_okay(_, Source) :-
    question(['WARNING: compiling ',w(Source),nl,
              ' will overwrite any existing .qo file in that directory.',nl,
              'Do you still wish to compile it?'],
             no).


%% do_expand(File, Expansions).
%%
%%  Process File with Expansions, leaving a *.qge file.
%%  Expansions can include: decl(Thy), macro, thy(Thy), prefix(Thy).
%%
%%  mode do_expand(@filename, @list(nonvar)).
%%
do_expand(File, Exps) :-
    get_load_path(PathList),
    member(P, PathList),
    full_file_name(P, File, Dir, PSFile),
    msg(compile, 2, ['looking for ',w(File),' in ',w(Dir)]),
    complete_filename(PSFile, '.ql', Source),
    !,
    msg(compile, ['expanding ',w(Source),'.']),
    op_table(user),
    obvar_prefix_table(user),
    expand_with_expansions(Exps, Source).
do_expand(File, _) :-
    handle_error(['Unable to find file: ',q(File)]).


%% consult_global.
%%
%%  Load the global file.  Uses the global path specified in the
%%  command line arguments.  If unable to find 'global.erg' in
%%  these directories, it fails, and the proof session 
%%  should be aborted.
%%
%%  mode consult_global.
%%
consult_global :-
    global_file_location(G), % set by -G arguments.
    msg(loading, 2, ['looking for ',w(G)]),
    access(G, 4, 0),
    msg(loading, ['Global information from ',w(G)]),
    op_table(user),
    obvar_prefix_table(user),
    encoded_consult(G),
    !.
consult_global :-
    handle_error(['Unable to find ''global.erg'' file.',nl,
               'Use -G <filename> to specify a global file, or create',nl,
               'an empty ''global.erg'' file in the current directory.']).

%% encoded_consult(FileName).
%%
%% Consult an encoded file.
%%
%% mode encoded_consult(@atom).

encoded_consult(FileName) :-
    open(FileName, read, InStream),
    unwind_protect(
      encoded_read_assert_clauses(InStream, []),
      close(InStream)
    ).

%% encoded_read_assert_clauses(Stream, ReadPred).
%%
%% Read an encoded clause and assert it.
%%
%% mode read_assert_clauses(@integer, @list(gcomp)).
%%

encoded_read_assert_clauses(Stream, ReadPred) :-
    encoded_read_term(Stream, Clause0, [variable_names(VarNames)]),
    expand_term(Clause0, Clause1, VarNames),
    ('$consult_clauses'(Clause1, ReadPred, NewReadPred, [])
      ->
        encoded_read_assert_clauses(Stream, NewReadPred)
    ;
	true
    ).
encoded_read_assert_clauses(_, _).


%% reconsult_global.
%%
%%  Rereads the global file and unions the facts therein with the
%%  facts that are currently in the global area of the Ergo database.
%%
%%  mode reconsult_global.
%%

reconsult_global :-
    global_file_location(G), % set by -G arguments.
    msg(loading, 2, ['looking for ',w(G)]),
    access(G, 4, 0),
    msg(loading, ['Extra global information from ',w(G)]),
    open(G, read, GS),
    !,
    unwind_protect(
        reconsult_global_facts(GS),
        close(GS)).


reconsult_global_facts(GS) :-
    repeat,
      encoded_read_term(GS, Term, []),
      union_global_fact(Term),
      stream_property(GS, end_of_stream(at)),
    !.

union_global_fact(Fact) :-
    msg(reconsult_global, 3, ['Read fact: ',q(Fact)]),
    fail.
union_global_fact('?-'(_)) :-
    !.
union_global_fact(end_of_file) :-
    !.
union_global_fact(RealFact) :-
    compound(RealFact),
    call_predicate(RealFact),
    !.
union_global_fact(RealFact) :-
    compound(RealFact),
    functor(RealFact, RealName, Arity),
    %% Simulate dbms_asserted_form backwards!
    dbms_asserted_name(global, Name, RealName),
    functor(Fact, Name, Arity),
    same_args(RealFact, Fact, 1, Arity),
    global_fact(Name, Arity),
    !,
    msg(reconsult_global, 2, ['Adding fact: ',q(Fact)]),
    assert_global(Fact).
union_global_fact(RealFact) :-
    handle_error(['unknown global.erg fact: ',q(RealFact)]).


load_path parameter
  [set_load_path, get_load_path]
extends
  parameters
doc
  'The `load_path'' parameter determines which directories will be',nl,
  'searched when theories and tactic files are being loaded.',
  para,
  'It is a list of directory specifiers.',nl,
  'Each directory specifier must be one of the following:',nl,
  begin(list),
  item(*), 'a directory name (eg. ''/homes/prover/theories'').',nl,
  '  Note that directories containing non-alphanumeric characters',nl,
  '  must be quoted.  Eg. ''.'' means the current directory.',nl,
  item(*), 'a term of the form `lib(RelPath)'', where RelPath is an atom.',nl,
  '  This specifies the directory ERGOHOME/RelPath, where ERGOHOME is the',nl,
  '  `library directory'' of Ergo, which is set up during the installation',nl,
  '  process, and passed to the Ergo executable using the -lib argument.',nl,
  '  Eg. `lib(''thy/obj'')''.',nl,
  end(list),
  'When the load_path is being set, the special atom ''...'' appearing',nl,
  'in the new load_path list is an abbreviation for all the directories',nl,
  'that are in the current load_path list and do not appear explicitly in',nl,
  'the new load_path list.',
  para,
  'An Example:  If load_path is [''.'',''/homes/prt/thys'',lib(thy)],',nl,
  '  the current directory is /homes/marku/ergo, and the Ergo installation',nl,
  '  directory is /usr/ergo, then the following directories will be',nl,
  '  searched for theory files and global tactic files:',
  para,
  '    ''/homes/marku/ergo'', ''/homes/prt/thys'', ''/usr/ergo/thy''.',nl,
  para,
  'Theory-specific tactic files (*.ql) are expected to be within',nl,
  'a theory subdirectory, so when trying to load a tactic file called',nl,
  '''tac2'' in a theory called ''arith'', with the above load_path,',nl,
  'the following paths to the tactic file would be tried:',nl,
  begin(list),
  item([]),'    ''/homes/marku/ergo/arith/tac2.ql'',',
  item([]),'    ''/homes/prt/thys/arith/tac2.ql'',',
  item([]),'    ''/usr/ergo/thy/arith/tac2.ql''.',
  end(list), nl,
  '(various alternative extensions, such as *.qo, would also be tried',nl,
  'in each of these directories).'.


set_load_path(L) :-
    closed_list(L),
    append(L1, ['...'|L2], L),
    !,
    map(check_legal_path, [L1]),
    map(check_legal_path, [L2]),
    get_load_path(Curr),
    diff_list(Curr, L, Curr2),
    append(Curr2, L2, Final1),
    append(L1, Final1, Final),
    retractall_global(load_path(_)),
    assert_global(load_path(Final)).
set_load_path(L) :-
    closed_list(L),
    map(check_legal_path, [L]),
    retractall_global(load_path(_)),
    assert_global(load_path(L)).

%% If the cleanup of old global.erg files has not been completed
%% (including those of off-site users) the the line
%%    retractall_global(path(_))
%% should be reinserted as the final line in both definitions of set_load_path.
%% Ray's last comment: Should be OK by now.


get_load_path(L) :-
    query_global(load_path(L0)),
    !,
    L = L0.
get_load_path(L) :-
    %% for backwards compatibility.
    findall(P, query_global(path(P)), L).


is_legal_path(A) :- atom(A).
is_legal_path(lib(A)) :- atom(A).

check_legal_path(P) :-
    check_or_error(
        is_legal_path(P),
        ['illegal path: ',q(P),'.']).


%% complete_filename(RootName, Extension, FullName).
%%
%%  Find a valid Extension (.qo, .qg, .qi, .ql, .erg or .erg.qo) for RootName.
%%  Try to find a newer object file first.
%%
%%  The filename returned in FullName is suitable for use in load/1
%%  when Extension='.qo' or '.erg.qo', and suitable for use in consult/1 
%%  when Extension='.ql', '.qi', '.qg' or '.erg'.
%%
%%  mode(@atom, ?atom, -atom).
%%
complete_filename(RootName, Type, FullName) :-
    query_global(handler(filevariant(RootName, Type, FullName) ,Handler)),
    call_predicate(Handler),
    !. 
complete_filename(RootName, '.qo', FullName) :-
    atom_append(RootName, '.qo', FullName),
    access(FullName, 4, 0),   % Read OK?
    atom_append(RootName, '.ql', FullName1),
    (
        \+ access(FullName1, 4, 0)  %% .ql does not exist
    ;
        file_newer_than(FullName, FullName1) %% .qo newer than .ql
    ),
    !.
complete_filename(RootName, '.qg', FullName) :-
    atom_append(RootName, '.qg', FullName),
    access(FullName, 4, 0),   % Read OK?
    atom_append(RootName, '.ql', FullName1),
    (
        \+ access(FullName1, 4, 0)  %% .ql does not exist
    ;
        file_newer_than(FullName, FullName1) %% .qg newer than .ql
    ),
    !.
complete_filename(RootName, '.qi', FullName) :-
    atom_append(RootName, '.qi', FullName),
    access(FullName, 4, 0),   % Read OK?
    atom_append(RootName, '.ql', FullName1),
    (
        \+ access(FullName1, 4, 0)  %% .ql does not exist
    ;
        file_newer_than(FullName, FullName1) %% .qi newer than .ql
    ),
    !.
complete_filename(RootName, '.ql', FullName) :-
    atom_append(RootName, '.ql', FullName),
    access(FullName, 4, 0),   % Read OK?
    !.
complete_filename(RootName, '.erg.qo', FullName1) :-
    atom_append(RootName, '.erg.qo', FullName1),
    access(FullName1, 4, 0),   % Read OK?
    atom_append(RootName, '.erg.qge', FullName),
    (
        \+ access(FullName, 4, 0)  %% .erg does not exist
    ;
        file_newer_than(FullName1, FullName) %% .erg.qo newer than .erg
    ),
    !.
complete_filename(RootName, '.erg.qge', FullName) :-
    atom_append(RootName, '.erg.qge', FullName),
    access(FullName, 4, 0),   % Read OK?
    !.
complete_filename(RootName, '.erg', FullName) :-
    atom_append(RootName, '.erg', FullName),
    access(FullName, 4, 0),   % Read OK?
    !.

%% pathname(PathName, Directory, FileName).
%% pathname(PathName, Directory, FileName, FileRoot, Extension).
%% existing_pathname(PathName, Directory, FileName).
%% existing_pathname(PathName, Directory, FileName, FileRoot, Extension).
%%
%%  Pathname is Directory/FileName, where FileName has no `/'.
%%  FileName is FileRootExtension, where Extension starts with `.' and has no
%%  other `.', or Extension is null and FileRoot has no `.'.
%%  For existing_pathname/3,5, the file must exist, and we can backtrack
%%  through matching files.
%%
%%  mode pathname(@atom, ?atom, ?atom).
%%  mode pathname(@atom, ?atom, ?atom, ?atom, ?atom).
%%  mode pathname(-atom, @atom, @atom).
%%  mode pathname(-atom, @atom, @atom, ?atom, ?atom).
%%  mode existing_pathname(@atom, ?atom, ?atom).
%%  mode existing_pathname(@atom, ?atom, ?atom, ?atom, ?atom).
%%  mode existing_pathname(?atom, @atom, ?atom).
%%  mode existing_pathname(?atom, @atom, ?atom, ?atom, ?atom).
%%
pathname(Path, Dir, File) :-
    atom(Path),
    !,
    atom_codes(Path, PathC),
    pathname1(PathC, [], _, [], DirR, [], FileR),
    reverse(DirR, DirC),
    reverse(FileR, FileC),
    atom_codes(Dir, DirC),
    atom_codes(File, FileC).
pathname(Path, Dir, File) :-
    atom(Dir),
    atom(File),
    !,
    concat_atom([Dir, /, File], Path).
pathname(Path, Dir, File) :-
    handle_error(program, [w(pathname(Path, Dir, File)), ': bad arguments.']).

pathname(Path, Dir, File, Root, Ext) :-
    pathname(Path, Dir, File),
    atom_codes(File, FileC),
    filename1(FileC, [], _, [], RootR, [], ExtR),
    reverse(RootR, RootC),
    reverse(ExtR, ExtC),
    atom_codes(Root, RootC),
    atom_codes(Ext, ExtC).

existing_pathname(Path, Dir, File) :-
    atom(Path),
    var(Dir),
    !,
    existing_pathname(Path, Dir, File).
existing_pathname(Path, Dir, File) :-
    atom(Dir),
    !,
    directory_files(Dir, Files),
    member(File, Files),
    concat_atom([Dir, /, File], Path).
existing_pathname(Path, Dir, File) :-
    handle_error(program, [w(existing_pathname(Path, Dir, File)),
                           ': bad arguments.']).

existing_pathname(Path, Dir, File, Root, Ext) :-
    existing_pathname(Path, Dir, File),
    atom_codes(File, FileC),
    filename1(FileC, [], _, [], RootR, [], ExtR),
    reverse(RootR, RootC),
    reverse(ExtR, ExtC),
    atom_codes(Root, RootC),
    atom_codes(Ext, ExtC).

pathname1([], CD, CD, D, D, F, F).
    %% writelist(['P  is ', qchars([]), tab,
    %%            'CD is ', qchars(CD), tab,
    %%            'D  is ', qchars(D), tab,
    %%            'F  is ', qchars(F), nl]).
pathname1([0'/|P], CD0, CD, _D0, D, _F0, F) :-
    %% writelist(['P  is ', qchars([0'/|P]), tab,
    %%            'CD is ', qchars(CD0), tab,
    %%            'D  is ', qchars(_D0), tab,
    %%            'F  is ', qchars(_F0), nl]),
    !,
    pathname1(P, [0'/|CD0], CD, [0'/|CD0], D, [], F).
pathname1([C|P], CD0, CD, D0, D, F0, F) :-
    %% writelist(['P  is ', qchars([C|P]), tab,
    %%            'CD is ', qchars(CD0), tab,
    %%            'D  is ', qchars(D0), tab,
    %%            'F  is ', qchars(F0), nl]),
    pathname1(P, [C|CD0], CD, D0, D, [C|F0], F).

filename1([], CR, CR, R0, R, X0, X) :-
    ( R0 = [] ->
        R = CR, X = []
    ; R0 = [0'.|R] ->
        X = X0
    ).
filename1([0'.|P], CR0, CR, _R0, R, _X0, X) :-
    !,
    filename1(P, [0'.|CR0], CR, [0'.|CR0], R, [0'.], X).
filename1([C|P], CR0, CR, R0, R, X0, X) :-
    filename1(P, [C|CR0], CR, R0, R, [C|X0], X).

%% directory_files(Directory, Files).
%%
%%  Files is a list of files in Directory.
%%
%%  mode directory_files(@atom, -list(atom)).
%%
directory_files(Dir, Files) :-
    opendir(Dir, Dirp, Stat1),
    ( Stat1 =:= 0 ->
        true
    ;
        unix_error(Stat1, Msg),
        handle_error(['Could not open directory ', q(Dir), ': ', w(Msg)])
    ),
    readdir(Dirp, Ino, Name, Stat2),
    ( Stat2 =:= 0 ->
        directory_files(Ino, Name, Dir, Dirp, Files),
        closedir(Dirp)
    ; Stat2 < 0 ->
        directory_files(end_of_stream, _, Dir, Dirp, Files),
        closedir(Dirp)
    ).

directory_files(end_of_stream, _, _, _, []) :-
    !.
directory_files(_, File, Dir, Dirp, [File|Files]) :-
    readdir(Dirp, Ino, Name, Stat),
    ( Stat =:= 0 ->
        directory_files(Ino, Name, Dir, Dirp, Files)
    ; 
        directory_files(end_of_stream, _, Dir, Dirp, Files)
    ).

%% file_newer_than(File1, File2).
%%
%%  True iff File1 is newer than File2.  Both files must already exist.
%%
%%  mode file_newer_than(@filename, @filename).
%%
file_newer_than(File1, File2) :-
    stat(File1, _, _, _, _, _, _, _, _, _, MTime1, _, _, _, Stat1),
    stat(File2, _, _, _, _, _, _, _, _, _, MTime2, _, _, _, Stat2),
    ( Stat1 =\= 0 ->
        unix_error(Stat1, Msg),
        handle_error(['Could not check age of file ', q(File1), ': ', w(Msg)])
    ; Stat2 =\= 0 ->
        unix_error(Stat2, Msg),
        handle_error(['Could not check age of file ', q(File2), ': ', w(Msg)])
    ;
        MTime1 > MTime2
    ).

%% file_type(File, Type).
%%
%%  File exists and has type Type.
%%  Type can be:
%%     fifo   fifo special
%%     chr    character special
%%     dir    directory
%%     blk    block special
%%     reg    ordinary file
%%  Fails silently if File does not exist or if its type will not unify with
%%    Type.  Thus, this can be used to test the existence of File.
%%
%%  mode file_type(@filename, ?atom).
%%
file_type(File, Type) :-
    stat(File, Mode, _,_,_,_,_,_,_,_,_,_,_,_, Stat),
    /* see /usr/include/sys/stat.h */
    S_IFMT  = 16'F000,  /* type of file */
    S_IFIFO = 16'1000,  /* fifo */
    S_IFCHR = 16'2000,  /* character special */
    S_IFDIR = 16'4000,  /* directory */
    S_IFBLK = 16'6000,  /* block special */
    S_IFREG = 16'8000,  /* regular */
    ( Stat =:= 2 /* ENOENT */ ->
        %% File does not exist: fail silently
        fail
    ; Stat =\= 0 ->
        %% Error
        unix_error(Stat, Msg),
        handle_error(['file_type/2: ', q(File), ': ', w(Msg)])
    ; (Mode /\ S_IFMT) =:= S_IFIFO ->
        Type = fifo
    ; (Mode /\ S_IFMT) =:= S_IFCHR ->
        Type = chr
    ; (Mode /\ S_IFMT) =:= S_IFDIR ->
        Type = dir
    ; (Mode /\ S_IFMT) =:= S_IFBLK ->
        Type = blk
    ; (Mode /\ S_IFMT) =:= S_IFREG ->
        Type = reg
    ;
        handle_error(program, ['file_type: stat botch'])
    ).

%% delete_file(File).
%%
%%  Delete (unlink) File.
%%
%%  mode delete_file(@filename).
%%
delete_file(File) :-
    unlink(File, Stat),
    ( Stat =:= 0 ->
        true
    ;
        unix_error(Stat, Msg),
        handle_error(['Could not delete file ', q(File), ': ', w(Msg)])
    ).

%% make_directory(Dir).
%%
%%  Create directory Dir.
%%
%%  mode make_directory(@filename).
%%
make_directory(Dir) :-
    mkdir(Dir, 8'777, Stat),
    ( Stat =:= 0 ->
        true
    ;
        unix_error(Stat, Msg),
        handle_error(['Could not create directory ', q(Dir), ': ', w(Msg)])
    ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%% FOREIGN FUNCTIONS START HERE %%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

?- retractall(foreign(_,_,_)),
   retractall(foreign_file(_,_)).

%% private stat(File, Mode, Ino, Dev, Rdev, Nlink, Uid, Gid, Size,
%%              Atime, Mtime, Ctime, Blksize, Blocks, Return).
%%
%%  Interface to the stat(2) Unix system call.
%%  First arg is the file name, last arg is the return value (0 or errno).
%%  Other args are as in struct stat.
%%
?- assert(foreign(p_stat, c,
                  stat(+atom, -integer, -integer, -integer, -integer,
                       -integer, -integer, -integer, -integer, -integer,
                       -integer, -integer, -integer, -integer, [-integer]))).

%% private unlink(File, Return).
%%
%%  Interface to the unlink(2) Unix system call.
%%  File is the file name to unlink; Return is the return value (0 or errno).
%%
?- assert(foreign(p_unlink, c, unlink(+atom, [-integer]))).

%% private unix_error(Errno, Message).
%%
%%  Interface to the strerror(3) C library routine.
%%  Message is the message for Errno.
%%
?- assert(foreign(p_error, c, unix_error(+integer, [-atom]))).

%% private mkdir(Dir, Mode, Return).
%%
%%  Interface to the mkdir(2) system call.
%%  Dir is the pathname to create; Mode is the mode to create it with; Return
%%  is the return value (0 or errno).
%%
?- assert(foreign(p_mkdir, c, mkdir(+atom, +integer, [-integer]))).

%% private ctr_alloc(InitVal, Ind).
%% private ctr_get1(Ind, Val).
%% private ctr_set1(Ind, Val).
%%
%%  Global persistent counters.
%%  Counters are identified by (small) integer indexes.
%%  Their values are integers.
?- assert(foreign(ctr_alloc, c, ctr_alloc(+integer, [-integer]))),
   assert(foreign(ctr_get, c, ctr_get1(+integer, [-integer]))),
   assert(foreign(ctr_set, c, ctr_set1(+integer, +integer))).

%% private getpid(Return).
%%
%%  Interface to the getpid(2) system call.
%%  Return is the return value (the current process's process ID).
%%
?- assert(foreign(p_getpid, c, getpid([-integer]))).

%% private opendir(Dir, Dirp, Stat).
%% private readdir(Dirp, Inode, Name, Stat).
%% private closedir(Dirp).
%%
%%  Interface to the opendir(3), readdir(3) and closedir(3) system calls.
%%  Dir is a directory name (as an atom).
%%  Dirp is a directory pointer.
%%  Inode is an inode number.
%%  Name is a file name (as an atom).
%%  Stat is a return status (0 or errno, or 1 for readdir at end of stream).
%%
opendir(Dir, dirp(H,L), Stat) :-
    opendir(Dir, H, L, Stat).

readdir(dirp(H,L), I, N, Stat) :-
    readdir(H, L, I, N, Stat).

closedir(dirp(H,L)) :-
    closedir(H, L).

?- assert(foreign(p_opendir, c,
                  opendir(+atom, -integer, -integer, [-integer]))),
   assert(foreign(p_readdir, c,
                  readdir(+integer, +integer, -integer, -atom, [-integer]))),
   assert(foreign(p_closedir, c,
                  closedir(+integer, +integer))).

%% do_create_foreign.
%%
%%  Call this (from inside a development version of Ergo) to generate a new
%%  unix_interface.cc when the foreign functions have changed.
%% 
do_create_foreign :-
    foreign_dir(FDir),
    !,
    atom_concat2(FDir, '/unix.o', FFile),
    atom_concat2(FDir, '/unix_interface', IFile),
    assert(foreign_file(FFile, [p_stat, p_unlink, p_mkdir, p_error,
                                ctr_alloc, ctr_get, ctr_set, p_getpid,
                                p_opendir, p_readdir, p_closedir])),
    '$convert_foreign'([FFile], [], IFile).
do_create_foreign :-
    handle_error(['Specify foreign directory!']).

%% TO DO: Get the makefile to carry out the generation of unix_interface.cc
%% instead of leaving it to do_create_foreign.

%% do_load_foreign.
%%
%%  Called when initializing the prover to load the foreign files.
%%  Assumes that unix_interface.cc has been built and compiled.
%% 
do_load_foreign :-
    foreign_dir(FDir),
    !,
    atom_concat2(FDir, '/unix.o', FFile),
    atom_concat2(FDir, '/unix_interface.o', IFile),
    msg(loading, ['C++ files from ',w(FDir)]),
    load_foreign([FFile,IFile],
                 [stat/15=p_stat_interface,
                  unlink/2=p_unlink_interface,
                  unix_error/2=p_error_interface,
                  mkdir/3=p_mkdir_interface,
                  ctr_alloc/2=ctr_alloc_interface,
                  ctr_get1/2=ctr_get_interface,
                  ctr_set1/2=ctr_set_interface,
                  p_getpid/1=p_getpid_interface,
                  opendir/4=p_opendir_interface,
                  readdir/5=p_readdir_interface,
                  closedir/2=p_closedir_interface
                 ]).
do_load_foreign :-
    handle_error(['Specify foreign directory!']).
