%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: win_rule.ql,v 1.26 2006/02/01 23:36:24 itp Exp $

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Procedures for adding window inference rules (opening rules)
%% to the current theory.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%% openrule(Type, Focus, Position, NewHyps, HypTrans, RelnTrans, GoalTrans).
%%
%%  Adds an opening rule to the current theory.
%%  Type must be 'primitive' (no other types are supported yet).
%%
%%  mode openrule(@atom, @term, @list(nonvar), @list(term),
%%                @list(compound), @list(compound), @list(compound)).
%%
openrule(Type, Focus, Pos, NH, ShiftHyp, ShiftEqu, ShiftGoal) :-
    check_open(Thy),
    fix_path(Focus, Pos, CPos, OW, SF, CW),
    var(OW),
    check_term(Focus),
    check_terms(NH),
    check_ruleequ(ShiftEqu),
    check_id(ShiftHyp, OtNH, ID),
    check_id(ShiftGoal, OtNG, _),
    (Type = primitive ->
        true
    ;
        %% These alternatives need work/justification.
        \+ \+ (ShiftEqu = ['--->'(E1,E2)],
               CEq =..[E2, OW, SF],
               PEq =..[E1, Focus, CW],
               list_to_and(NH, AndNH),
               freeze(PEq, CEq),
               (
                   %% Co 5.2
                   Type = defined(Co),
                   ID == true,
                   check_Co52(Co, AndNH, Focus, CEq, PEq)
               ; 
                   %% SC (b) .. (d)
                   Type = defined(B, CList, DList),
                   split_fn(ShiftHyp, Dom, Ran),
                   check_SC_b(B, AndNH, Dom, Ran, CEq, PEq),
                   check_SC_c(CList, Dom, PEq),
                   check_SC_d(DList, AndNH, Dom, Ran, CEq, PEq)
              ))),
    assert_theory(openrule(Focus,CPos,OW,NH,
                           OtNH,ShiftEqu,OtNG,'--->'(SF,CW)),Thy).


check_id([],              [],            false).
check_id(id,              ['--->'(A,A)], true).
check_id([id|_],          ['--->'(A,A)], true).
check_id(['--->'(A,B)|_], ['--->'(A,A)], true) :- 
    A == B,
    var(A),
    !.
check_id(['--->'(A,B)|C], ['--->'(A,B)|D], E) :- 
    check_id(C, D, E).


split_fn([], [], []).
split_fn(['--->'(A,B)|C], [A|D], [B|E]) :-
    split_fn(C, D, E).


%% TO DO: Check whether the window rule proposed in check_Co52 and
%% check_SC_[bcd] satisfies Corollary 5.2 or
%% Soundness Conditions B through D respectively from PJR's paper
%% 'Soundness Condtions for Window Opening Rules' 
%% These have not been fully checked out for Ergo 4.1 yet and are
%% commented out at the moment.
/*
%% private check_Co52(...).
%% private check_SC_b(...).
%% private check_SC_c(...).
%% private check_SC_d(...).
%%
check_Co52(NH, F, CEq, PEq) :-
    check_rule(_, ERGO_IMPLIES(H,PEq), 
               [ERGO_IMPLIES(ERGO_AND(H,NH),CEq)]).

check_SC_b(ANH, Dom, Ran, F, CEq, PEq) :-
    list_to_and(Ran, ANH, CHyps),
    list_to_and(Dom, PHyps),
    check_rule(_, ERGO_IMPLIES(PHyps,PEq), [ERGO_IMPLIES(CHyps,CEq)]).

check_SC_c(Dom, PEq) :-
    \+ (member(X, Dom),
        \+ check_is_theorem(_, or(X,PEq)).
%% TO DO: Fix the above line in check_SC_c.

check_SC_d(NH, Shift, CEq, PEq) :-
    \+ (member(X, Shift),
        copy(X, '--->'(K1,SK1)),
        copy(X, '--->'(K2,SK2)),
        freeze_vars((K1, K2, SK1, SK2)),
        \+ (is_theorem(_, ANH => ((SK1 and SK2 => SK) or CEq)),
%% TO DO: Fix these two calls to is_theorem in check_SC_d.
            is_theorem(_,         (K1 and K2 => K) or PEq),
            member('--->'(K,SK), Shift))).
*/



%% private fix_path(OriginalTerm, Path, CPath, Subterm, NewSubterm, NewTerm).
%%
%%  Finds Subterm at position Path in OriginalTerm and constructs a copy
%%  (NewTerm) of OriginalTerm with NewSubterm substituted at this position.
%%  The corrected form of Path is returned as CPath.
%%
%%  The following table shows how various paths index into terms:
%%
%%      Term                 Subterm        Paths (last one is preferred)
%%    ===================================================================
%%      X                    X              []
%%      f(X,Y)               X              [1]
%%      f(X,Y)               Y              [2]
%%      q [] B               B              [2],     [b([])]
%%      q [x] B              B              [2],     [b([x])]
%%      q [x:T,y] B          B              [2],     [b([x,y])]
%%      q [x:T] B            T              [1,1,2], [t(1)]
%%      q [x:T,y:U] B        U              [1,2,2], [t(2)]
%%      q [x,y,z:U] B        U              [1,3,2], [t(3)]
%%
%%  mode fix_path(+nonvar,+list(nonvar),-list(nonvar),?term,?term,-nonvar).
%%
fix_path(Term, _, _, _, _, _) :-
    sub(Term),
    !,
    handle_error(['substitutions not allowed in: ',Term,'.']).
fix_path(Term, [], [], Term, A, A) :-
    !.
fix_path(Term, Pos0, [t(N)|NewPos], Subterm, A, NewTerm) :-
    quant(Term),
    (
        Pos0 = [1,N,2|Pos]
    ;
        Pos0 = [t(N)|Pos]
    ),
    integer(N),
    !,
    quantify(Term, Quant, BVars, Body),
    check_or_error(
        closed_list(BVars),
        [q(BVars),' is not a closed list.']),
    check_or_error(
        select(N, X:Type, BVars, X:NewType, NewBVars),
        ['element ',w(N),' of ',q(BVars),' is not typed.']),
    quantify(NewTerm, Quant, NewBVars, Body),
    fix_path(Type, Pos, NewPos, Subterm, A, NewType).
fix_path(Term, [P|Pos], [b(ObVars)|NewPos], Subterm, A, NewT) :-
    quant(Term),
    ( 
        P == 2
    ;
        P = b(ObVars)
    ),
    !,
    quantify(Term, Quant, BVars, Body),
    check_bndvar_list(BVars, _, ObVars2, _),
    ObVars = ObVars2,
    quantify(NewT, Quant, BVars, NewBody),
    !,
    fix_path(Body, Pos, NewPos, Subterm, A, NewBody).
fix_path(Term, [P|Pos], [P|NewPos], Subterm, A, NewT) :-
    check_or_error(
        ( compound(Term),
          functor(Term, F, N),
          0 < P,
          P =< N),
        ['illegal position, [...,',q(P),'|',q(Pos),'] in openrule']),
    Term =.. [F|TList],
    select(P, Arg, TList, NewArg, NewTList),
    NewT =.. [F|NewTList],
    fix_path(Arg, Pos, NewPos, Subterm, A, NewArg).
