%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: decl_expansions.ql,v 1.8 2001/03/22 02:03:13 itp Exp $
%%
%% See expansions.ql for how to use this file.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% WARNING!
%% This file should be able to run OUTSIDE Ergo.
%% Beware of using Ergo-defined procedures in this file!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% DECLARATION EXPANSIONS (can be outside Ergo, unless wrappers are used).
%%     Declarations look like:
%%      Table command
%%         Cmd, ...
%%      header
%%         Docstring, ...
%%      doc
%%         DocString, ...
%%      extends
%%         Node, ...
%%      mode
%%         ModeSpec, ...
%%      wrapper
%%         TacticWrapper ---> SubProofWrappers
%%      implementation
%%         Impl.
%%     All clauses but the first command are optional; not all clauses are
%%      meaningful for all command Tables.
%%
%%     All doc clauses are combined (in order) with the descriptions of the
%%      command clauses to define a single doc node.  Wrapper and
%%      implementation clauses pertain to the latest command clause.
%%      A command clause MUST appear first.
%%
%%     Example 1:
%%     dbms command
%%       check_thy
%%     mode
%%       check_thy
%%     implementation
%%       current_dir(Thy),
%%       check_thy(Thy)
%%     command
%%       check_thy(Thy)
%%     extends
%%       misc_cmds
%%     mode
%%       check_thy(@theory)
%%     doc
%%       'Check Thy (default: current theory) for its quality',nl,
%%       '(e.g. duplicated rules).'.
%%     
%%     This generates two commands check_thy/0 and check_thy/1 in the dbms
%%     command table.  Only check_thy/1 needs to be implemented directly in
%%     Prolog, since the `implementation' clause furnishes an implementation
%%     for check_thy/0.  A doc node (cross-linked to both dbms commands) is
%%     created from the `command' and `doc' clauses, and linked into the
%%     misc_cmds menu node by the `extends' clause.  No explicit `header' is
%%     provided, so the generated doc node will be identified by the command
%%     names themselves.  The mode specifies that the argument of check_thy/1
%%     is a theory name; certain modes trigger type checks or automatic
%%     deabbreviation when the command is used.
%%
%%   Example 2:
%%   proof command
%%       fwd_cases(Cases, Target)
%%   extends
%%       hyp_proofcmds
%%   doc
%%       'This tactic uses case analysis on Cases to prove Target.  Cases', nl,
%%       'Should be a hypothesis or easily deducible (i.e. the call', nl,
%%       'use([t], Cases) should succeed).'
%%   mode
%%       fwd_cases(?thy(term), ?thy(term))
%%   wrapper
%%     ['Case analysis on: ',thy(Cases)] --->
%%     [['Case 1:'],['Case 2:']].
%%
%%   This generates a command fwd_cases/2 in the proof command table.  The
%%   tactic takes two theory terms as asrguments; they will be automatically
%%   deabbreviated when the tactic is called.  Also, the 'wrapper' fact
%%   provides messages that are displayed and incorporated in proof wrappers
%%   when fwd_cases/2 is used, and indicates that the tactic has two subproofs.
%%   The wrapper includes the message 'Case analysis on: ...' when the tactic
%%   is called, and 'Case 1:' and 'Case 2:' for the subproofs.
%%

?- op(1150, xfy, command),
   op(1150, xfy, header),
   op(1150, xfy, doc),
   op(1150, xfy, doc_fragment),
   op(1150,  fy, elf),
   op(1150, xfy, parameter),
   op(1150, xfy, extends),
   op(1150, xfy, mode),
   op(1140, xfy, arity),
   op(1150, xfy, wrapper),
   op(990,  xfy, '--->'),
   op(1150, xfy, implementation).

%% command enable_declaration_expansions(Thy).
%% command disable_declaration_expansions.
%%
%%  Install/remove expansions for declarations in Thy.
%%  Thy can be 'global' or a theory name.
%%
%%  mode enable_declaration_expansions(@theory).
%%  mode disable_declaration_expansions.
%%
enable_declaration_expansions(Thy) :-
    add_ergo_expansion(expand_decls(Thy)).

disable_declaration_expansions :-
    remove_ergo_expansion(expand_decls(_)).

%% expand_decls(ThyOrGlobal, Decls, Queries, Vars).
%%
%%  Expand the declarations that preceed Ergo procedure definitions into
%%  queries that build command entries, doc nodes etc.
%%
%%  mode expand_decls(@theory, @compound, -list(term), @list(compound)).
%%
expand_decls(Thy, command(Table, Rest), Qs, Vars) :-
    exp_debug(['trying decl expansion on ', w(command(Table, Rest))]),
    ( expand_decls(Rest, command, [], Qs0, Vars, Table, Thy,
                   [], Doc, _, LC, [], Hdr),
      exp_debug(['Parsed declarations; now completing queries.']),
      expand_complete_queries(Qs0, Qs1, Thy, DocName, Qs0),
      exp_debug(['Completed queries ; now building doc.']),
      expand_create_cmd_doc_query(Doc, Hdr, Thy, DocName, Table, LC, Qs1, Qs),
      exp_debug(['Built doc; expansion succeeded.']) ->
          true
    ;
        exp_data_error(warning,
                       declaration,
                       ['Unacceptable declaration ignored:', nl,
                        tab(4), w(command(Table, Rest))]),
        Qs = []
    ),
    exp_debug(['done decl expansion on ', w(command(Table,Rest)), nl,
               '  resulting queries:', nl,tab(4),tab(4), w(Qs)]).

%%% Caution: this clause is called from the dbms `doc' command as well as from
%%% expansions.
%%% TO DO: Abstract out the appropriate functionality for the
%%% doc node creation in expand_decls.
expand_decls(Thy, doc(Name, Rest), Qs, _Vars) :-
    ( expand_decls(Rest, doc, [], Qs0, [], concept, Thy,
                   [], Doc, _, _, [Name-[]], Hdr) ->
        expand_create_simple_doc_query(Doc, Hdr, Thy, Name, Qs0, Qs1),
        expand_complete_queries(Qs1, Qs, Thy, Name, Qs1)
    ;
        exp_data_error(warning,
                       declaration,
                       ['Unacceptable declaration ignored:', nl,
                        tab(4), w(doc(Name, Rest))]),
        Qs = []
    ),
    exp_debug(['done decl expansion on ', w(doc(Name,Rest))]).

expand_decls(Thy, header(Name, Rest), Qs, _Vars) :-
    ( expand_decls(Rest, header, [], Qs0, [], concept, Thy,
                   [], Doc, _, _, [Name-[]], Hdr) ->
        expand_create_simple_doc_query(Doc, Hdr, Thy, Name, Qs0, Qs1),
        expand_complete_queries(Qs1, Qs, Thy, Name, Qs1)
    ;
        exp_data_error(warning,
                       declaration,
                       ['Unacceptable declaration ignored:', nl,
                        tab(4), w(header(Name, Rest))]),
        Qs = []
    ),
    exp_debug(['done decl expansion on ', w(header(Name,Rest))]).

expand_decls(Thy, elf(Rest), Qs, _Vars) :-
    ( expand_decls(Rest, elf, [], Qs0, [], elf, Thy,
                   [], Doc, _, _, [], Hdr) ->
        %% Find out the DocName
        make_query(elf(Thy, _ElfName, DocName), Q),
        member(Q, Qs0),
        expand_create_simple_doc_query(Doc, Hdr, Thy, DocName, Qs0, Qs1),
        expand_complete_queries(Qs1, Qs, Thy, DocName, Qs1)
    ;
        exp_data_error(warning,
                       declaration,
                       ['Unacceptable declaration ignored:', nl,
                        tab(4), w(elf(Rest))]),
        Qs = []
    ),
    exp_debug(['done decl expansion on ', w(elf(Rest))]).

expand_decls(Thy, doc_fragment(Head, Rest), Qs, _Vars) :-
    ( expand_decls(Rest, doc, [], Qs0, [], fragment, Thy,
                   [], Doc1, _, _, [], _) ->
        reverse(Doc1, Doc),
        make_query(frag(Thy, Head, Doc), Q),
        Qs1 = [Q | Qs0],
        expand_complete_queries(Qs1, Qs, Thy, Head, Qs1)
    ;
        exp_data_error(warning,
                       declaration,
                       ['Unacceptable declaration ignored:', nl,
                        tab(4), w(doc_fragment(Head, Rest))]),
        Qs = []
    ),
    exp_debug(['done decl expansion on ', w(doc_fragment(Head,Rest))]).

expand_decls(Thy, parameter(Name, Rest), Qs, _Vars) :-
    ( expand_decls(Rest, parameter, [], Qs1, [], Name, Thy,
                   [], Doc, _, _, [], Hdr),
      expand_complete_queries(Qs1, Qs2, Thy, DocName, Qs1),
      expand_create_simple_doc_query(Doc, Hdr, Thy, DocName, Qs2, Qs) ->
          true
    ;
        exp_data_error(warning,
                       declaration,
                       ['Unacceptable declaration ignored:', nl,
                        tab(4), w(parameter(Name, Rest))]),
        Qs = []
    ),
    exp_debug(['done decl expansion on ', w(parameter(Name,Rest))]).


%% private expand_decls(Decls, Command, Qs0, Qs, Vars, Table, Thy,
%%                      Doc0, Doc, LastCmd0, LastCmd, Header0, Header).
%%
%%  A Decls has a Command, a Clause and (maybe) some more Decls.
%%
%%  mode expand_decls(@compound, @atom, +list(compound), -list(compound),
%%            @list(compound), @cmdtab, @thyname,
%%            +writelist, -writelist, +term, -term, +atom, -atom).
%%
expand_decls(command(Clause, Rest), Cmd, Qs0, Qs, Vs, T, Thy,
             H0, H, LC0, LC, Hdr0, Hdr) :-
    !,
    expand_decls_clause(Clause, Cmd, Qs0, Qs1, Vs, T, Thy,
                        H0, H1, LC0, LC1, Hdr0, Hdr1),
    expand_decls(Rest, command, Qs1, Qs, Vs, T, Thy,
                 H1, H, LC1, LC, Hdr1, Hdr).
expand_decls(doc(Clause, Rest), Cmd, Qs0, Qs, Vs, T, Thy,
             H0, H, LC0, LC, Hdr0, Hdr) :-
    !,
    expand_decls_clause(Clause, Cmd, Qs0, Qs1, Vs, T, Thy,
                        H0, H1, LC0, LC1, Hdr0, Hdr1),
    expand_decls(Rest, doc, Qs1, Qs, Vs, T, Thy,
                 H1, H, LC1, LC, Hdr1, Hdr).
expand_decls(header(Clause, Rest), Cmd, Qs0, Qs, Vs, T, Thy,
             H0, H, LC0, LC, Hdr0, Hdr) :-
    !,
    expand_decls_clause(Clause, Cmd, Qs0, Qs1, Vs, T, Thy,
                        H0, H1, LC0, LC1, Hdr0, Hdr1),
    expand_decls(Rest, header, Qs1, Qs, Vs, T, Thy,
                 H1, H, LC1, LC, Hdr1, Hdr).
expand_decls(extends(Clause, Rest), Cmd, Qs0, Qs, Vs, T, Thy,
             H0, H, LC0, LC, Hdr0, Hdr) :-
    !,
    expand_decls_clause(Clause, Cmd, Qs0, Qs1, Vs, T, Thy,
                        H0, H1, LC0, LC1, Hdr0, Hdr1),
    expand_decls(Rest, extends, Qs1, Qs, Vs, T, Thy,
                 H1, H, LC1, LC, Hdr1, Hdr).
expand_decls(mode(Clause, Rest), Cmd, Qs0, Qs, Vs, T, Thy,
             H0, H, LC0, LC, Hdr0, Hdr) :-
    !,
    expand_decls_clause(Clause, Cmd, Qs0, Qs1, Vs, T, Thy,
                        H0, H1, LC0, LC1, Hdr0, Hdr1),
    expand_decls(Rest, mode, Qs1, Qs, Vs, T, Thy,
                 H1, H, LC1, LC, Hdr1, Hdr).
expand_decls(wrapper(Clause, Rest), Cmd, Qs0, Qs, Vs, T, Thy,
             H0, H, LC0, LC, Hdr0, Hdr) :-
    !,
    expand_decls_clause(Clause, Cmd, Qs0, Qs1, Vs, T, Thy,
                        H0, H1, LC0, LC1, Hdr0, Hdr1),
    expand_decls(Rest, wrapper, Qs1, Qs, Vs, T, Thy,
                 H1, H, LC1, LC, Hdr1, Hdr).
%expand_decls(subwrappers(Clause, Rest), Cmd, Qs0, Qs, Vs, T, Thy,
%             H0, H, LC0, LC, Hdr0, Hdr) :-
%    !,
%    expand_decls_clause(Clause, Cmd, Qs0, Qs1, Vs, T, Thy,
%                        H0, H1, LC0, LC1, Hdr0, Hdr1),
%    expand_decls(Rest, subwrappers, Qs1, Qs, Vs, T, Thy,
%                 H1, H, LC1, LC, Hdr1, Hdr).
expand_decls(implementation(Clause, Rest), Cmd, Qs0, Qs, Vs, T, Thy,
             H0, H, LC0, LC, Hdr0, Hdr) :-
    !,
    expand_decls_clause(Clause, Cmd, Qs0, Qs1, Vs, T, Thy,
                        H0, H1, LC0, LC1, Hdr0, Hdr1),
    expand_decls(Rest, implementation, Qs1, Qs, Vs, T, Thy,
                 H1, H, LC1, LC, Hdr1, Hdr).
expand_decls(Clause, Cmd, Qs0, Qs, Vs, T, Thy,
             H0, H, LC0, LC, Hdr0, Hdr) :-
    %% now we've processed all the keywords out of Clause, so it should
    %% be a comma-separated list of `lines', or a single `line'.
    expand_decls_clause(Clause, Cmd, Qs0, Qs, Vs, T, Thy,
                        H0, H, LC0, LC, Hdr0, Hdr).

%% private expand_decls_clause(Clause, Cmd, Qs0, Qs, Vars, Table, Thy,
%%                             Doc0, Doc, LastCmd0, LastCmd, Header0, Header).
%%
%%  A Clause is a comma-separated list of Lines.
%%
%%  mode expand_decls_clause(@compound, @atom,
%%                           +list(compound), -list(compound),
%%                           @list(compound), @cmdtab, @thyname,
%%                           +writelist, -writelist, +term, -term,
%%                           +atom, -atom).
expand_decls_clause((L1,Ls), Cmd, Qs0, Qs, Vs, T, Thy,
                    H0, H, LC0, LC, Hdr0, Hdr) :-
    !,
    expand_decls_line(Cmd, L1, Qs0, Qs1, Vs, T, Thy,
                      H0, H1, LC0, LC1, Hdr0, Hdr1),
    expand_decls_clause(Ls, Cmd, Qs1, Qs, Vs, T, Thy,
                        H1, H, LC1, LC, Hdr1, Hdr).
expand_decls_clause(L, Cmd, Qs0, Qs, Vs, T, Thy,
                    H0, H, LC0, LC, Hdr0, Hdr) :-
    expand_decls_line(Cmd, L, Qs0, Qs, Vs, T, Thy,
                      H0, H, LC0, LC, Hdr0, Hdr).

%% private expand_decls_line(Cmd, Line, Qs0, Qs, Vars, Table, Thy,
%%                           Doc0, Doc, LastCmd0, LastCmd, Header0, Header).
%% 
%%  A Line is processed depending on the Cmd it pertains to.
%%  Cmd is the keyword immediately preceeding the group of
%%  lines in which this Line appears.
%%
%%  mode expand_decls_line(@atom, @compound,
%%                    +list(compound), -list(compound),
%%                    @list(compound), @cmdtab, @thyname,
%%                    +writelist, -writelist, +term, -term,
%%                    +atom, -atom).
%%
%%% command: Create the query, leaving CallModes and Impl to be filled in.
%%% Also add an entry to the header -- may be modified by later header clauses.
expand_decls_line(command, _, _, _, _, concept, _, _, _, _, _, _, _) :-
    !,
    exp_data_error(unrecoverable,
                   declaration,
                   ['Command clause not permitted in doc declaration.']).
expand_decls_line(Type, Line, _Qs0, _Qs, _Vs, _Tab, Thy,
                  _H0, _H, _LC0, _LC, _Hdr0, _Hdr) :-
    exp_debug(['Expanding ', w(Type), ' line: ', q(Line), ' in ', w(Thy)]),
    fail.
expand_decls_line(command, Call, Qs, [Q | Qs], Vs, Tab, Thy,
                  H, H, _, Call, Hdr0, Hdr) :-
    make_query(command(Thy, Tab, Call, _Modes, _Impl, _HN), Q),
    copy_term((Call,Vs),  (CallC,VsC)),
    exp_bind_vars_names(VsC),
    open_string(write, Strm),
    write_term(Strm, CallC, [op_table(Thy)]),
    stream_to_atom(Strm, CallA),
%    close(Strm),
    Hdr = [CallA-[] | Hdr0],
    exp_debug(['Created command description ', q(CallA)]).

%%% parameter: create the query.
%%% Also add an entry to the header -- may be modified by later header clauses.
expand_decls_line(parameter, [Set,Get], Qs, [Q|Qs], _, Name, Thy,
                  H, H, _, _, Hdr, [Name-[]|Hdr]) :-
    make_query(parm(Thy,Name,DocNode,Set,Get), Q),
    make_doc_name(parameter, Name, DocNode).

%%% elf: create the query.
%%% Also add an entry to the header -- may be modified by later header clauses.
expand_decls_line(elf, ElfName, Qs, [Q|Qs], _, _, Thy,
                  H, H, _, _, Hdr, [ElfName-[]|Hdr]) :-
    atom(ElfName),
    make_doc_name(elf, ElfName, DocName),
    make_query(elf(Thy, ElfName, DocName), Q).

%%% header: add a line to the first header; doc query will be created later.
expand_decls_line(header, Line, Qs, Qs, _, _, _, H, H, LC, LC,
                  [LCD-Hd0|Hdrs], [LCD-Hd|Hdrs]) :-
    Hd = [Line|Hd0].

%%% doc: Add a line to the doc; the query will be created at the end.
expand_decls_line(doc, Line, Qs, Qs, _, _, _, H, [Line|H], LC, LC, Hdr, Hdr) :-
    ( true /* prefer: is_legal_writelist_elem(Line) */ ->
        true
    ;
        exp_data_error(unrecoverable,
                       declaration,
                       ['Bad doc entry: ', q(Line)])
    ).

%%% extends: Add a doc extension.
expand_decls_line(extends, ExtNode, Qs, [Q|Qs], _Vs, _Tab, Thy,
                  Doc, Doc, LC, LC, Hdr, Hdr) :-
    make_query(ext(Thy, ExtNode, XRefTag, _ThisNode), Q),
    ( Hdr = [Def-[]|_] ->
        XRefTag = [Def]
    ; Hdr = [Def-User|_],
      strip_leading_nls(User, [Tag|_]) ->
        XRefTag = [Tag]
    ;
        exp_data_error(unrecoverable,
                       declaration,
                       ['Cannot create xref with empty header.'])
    ).

%%% mode: Add an entry to the CallModes in the appropriate query.

expand_decls_line(mode, arity(Mode, Arity), Qs, Qs, _, Tab, Thy,
                  Doc, Doc, LC, LC, Hdr, Hdr) :-
    functor(Mode, F, N),
    functor(Call, F, N),
    make_query(command(Thy, Tab, Call, Modes, _Impl, _Doc), Q),
    ( member(Q, Qs) ->
        open_append(Modes, [arity(Mode, Arity)])
    ;
        exp_data_error(unrecoverable,
                       declaration,
                       ['Mode declaration ', w(Mode), nl,
                        tab(4), 'has no corresponding command declaration.'])
    ).
expand_decls_line(mode, Mode, Qs, Qs, _, Tab, Thy,
                  Doc, Doc, LC, LC, Hdr, Hdr) :-
    functor(Mode, F, N),
    functor(Call, F, N),
    make_query(command(Thy, Tab, Call, Modes, _Impl, _Doc), Q),
    ( member(Q, Qs) ->
        open_append(Modes, [Mode])
    ;
        exp_data_error(unrecoverable,
                       declaration,
                       ['Mode declaration ', w(Mode), nl,
                        tab(4), 'has no corresponding command declaration.'])
    ).

%%% wrapper: create a wrapper query for the last command
expand_decls_line(wrapper, Line, Qs, [Q|Qs], _Vs, Tab, Thy,
                  Doc, Doc, Call, Call, Hdr, Hdr) :-
    !,
    expand_wrapper(Line, Call, Wrapper, SubWrappers),
    make_query(wrap(Thy, Call, Wrapper, SubWrappers), Q),
    ( \+ member_eq(Tab, [proof,tactic]) ->
        exp_data_error(unrecoverable,
                       declaration,
                       [w(Call), ': only tactics can have wrappers!'])
    ; make_query(wrap(_, Call, _, _), Q1),
      member(Q1, Qs) ->
        exp_data_error(unrecoverable,
                       declaration,
                       [w(Call), ': multiple wrappers!'])
    ;
        true
    ).

%%% implementation: add to the Impl for the last command.
expand_decls_line(implementation, Impl, Qs, Qs, _Vs, Tab, Thy,
                  H, H, LC, LC, Hdr, Hdr) :-
    make_query(command(Thy, Tab, LC, _, RC, _Doc), Q),
    ( member(Q, Qs) ->
        open_append_commas(RC, (Impl,_))
    ;
        exp_data_error(unrecoverable,
                       declaration,
                       ['Implementation declaration ', w(Impl), nl,
                        tab(4), 'has no corresponding command declaration.'])
    ).


%% private expand_wrapper(+Line, +Cmd, -Wrapper, -SubWrappers).
%%
expand_wrapper('--->'(Wrapper,SubWrappers), _, Wrapper, SubWrappers) :-
    list(Wrapper),  %% simple version of check_is_legal_writelist(Wrapper)
    list(SubWrappers),  %% may be open or closed.
    !.
expand_wrapper(Wrapper, _, Wrapper, []) :-
    %% The tactic has no subproofs.
    list(Wrapper),  %% simple version of check_is_legal_writelist(Wrapper),
    !.
expand_wrapper(_, Call, _, _) :-
    exp_data_error(unrecoverable,
                   declaration,
                   [w(Call), ': bad wrapper!']).


%% private expand_complete_queries(Queries0,Queries, Thy, DocName, AllQueries).
%%
%%  Complete a list of queries: bind Thy and DocName, close each modes list,
%%  and bind Impl if not already provided.  Note that AllQueries must
%%  be the complete Queries0 list.
%%
%%  mode expand_complete_queries(+list(compound), ?list(compound),
%%                               @theory, @atom, +list(compound)).
%%
expand_complete_queries([], [], _, _, _).
expand_complete_queries([Q|_], _, _, _, _) :-
    exp_debug(['About to complete query: ', w(Q)]),
    fail.
expand_complete_queries([Q|Qs0], [Q2|Qs], Thy, DocName, AllQs) :-
    make_query(command(Thy, Tab, Call, Modes, Impl0, DocName), Q),
    !,
    expand_complete_queries_modes(Modes, Call),
    expand_complete_queries_impl(Call, AllQs, Impl0, Impl),
    make_query(command(Thy, Tab, Call, Modes, Impl, DocName), Q2),
    exp_debug(['Completed command query: ',
               'Call: ', q(Call),nl,
               'Modes: ', q(Modes),nl,
               'Impl: ', q(Impl),nl,
               'DocName: ', q(DocName)]),
    expand_complete_queries(Qs0, Qs, Thy, DocName, AllQs).
expand_complete_queries([Q|Qs0], [Q|Qs], Thy, DocName, AllQs) :-
    make_query(ext(Thy, ExtNode, XRefTag, DocName), Q),
    !,
    exp_debug(['Completed query: ', q(extqy(Thy,ExtNode,XRefTag,DocName))]),
    expand_complete_queries(Qs0, Qs, Thy, DocName, AllQs).
expand_complete_queries([Q|Qs0], [Q|Qs], Thy, DocName, AllQs) :-
    make_query(parm(Thy, Name, DocName, Set, Get), Q),
    !,
    exp_debug(['Completed query: ', q(parm(Thy, Name, DocName, Set, Get))]),
    expand_complete_queries(Qs0, Qs, Thy, DocName, AllQs).
expand_complete_queries([Q|Qs0], [Q|Qs], Thy, DocName, AllQs) :-
    expand_complete_queries(Qs0, Qs, Thy, DocName, AllQs).

%% private expand_complete_queries_modes(Modes, Call).
%%
%%  Complete the modes by closing the list and issuing a warning if no modes
%%  were given.
%%
%%  mode expand_complete_queries_modes(?mode, @cmd).
%% 
expand_complete_queries_modes(Modes, Call) :-
    open_to_closed(Modes),
    ( Modes = [] ->
        exp_data_error(warning, declaration, ['no modes for ', w(Call)])
    ;
        true
    ).

%% private expand_complete_queries_impl(Call, Queries, Impl0, Impl).
%%
%%  Complete the implementation: add the subproofs (if any).
%%
%%  mode expand_complete_queries_impl(@cmd, @list(compound), ?code, -code).
%% 
expand_complete_queries_impl(Call, Queries, Impl0, Impl) :-
    make_query(wrap(_, Call, _, SubMs), Q),
    member(Q, Queries),
    !,
    %% The tactic has a wrapper, so we must build the correct
    %% implementation clause so that call_tac/wrap_tac will handle
    %% any subproofs correctly.
    ( SubMs == [] ->
        Impl = Call
    ;
        Impl = '--->'(Call,_)
    ),
    ( var(Impl0) ->
        true
    ;
        exp_data_error(warning, var(Impl0),
                       ['Tactic ',q(Call),' implementation was ignored.'])
    ).
expand_complete_queries_impl(_, _, Impl0, Impl) :-
    nonvar(Impl0),
    !,
    open_append_commas(Impl0, true),
    strip_trailing_true(Impl0, Impl).
expand_complete_queries_impl(Call, _, _, Call).


strip_trailing_true((A,true), A) :- !.
strip_trailing_true((A,B0), (A,B)) :- !, strip_trailing_true(B0,B).
strip_trailing_true(A, A).


%% private expand_create_cmd_doc_query(Doc, Hdr, Thy, Name,
%%                                     Table, LastCmd,
%%                                     Queries0, Queries).
%% private expand_create_simple_doc_query(Doc, Hdr, Thy, Name,
%%                                        Queries0, Queries).
%%
%%  Build the doc node and add a query to install it.
%%
%%  The *_cmd_* form does a few extra things (eg. builds up the
%%  commands list for the doc_node) and uses Table and LastCmd
%%  to create Name (the name of the doc_node).
%%
%%  mode expand_create_cmd_doc_query(
%%                       +writelist, +list(compound), @theory, ?atom,
%%                       @cmdtab, +compound,
%%                       +list(compound), -list(compound)).
%%  mode expand_create_simple_doc_query(
%%                       +writelist, +list(compound), @theory, @atom,
%%                       +list(compound), -list(compound)).
%%
expand_create_cmd_doc_query(Doc, Hdr, Thy, Name, Tab, LC, Qs, [Q|Qs]) :-
    make_doc_header(Hdr, Header),
    reverse(Header, HList),
    functor(LC, F, N),
    make_doc_name(Tab, F, N, Name),
    reverse(Doc, DList),
    make_query(doc(Thy, Name, HList, DList, Cmds), Q),
    expand_decls_get_commands(Qs, Cmds).

expand_create_simple_doc_query(Doc, Hdr, Thy, Name, Qs, [Q|Qs]) :-
    make_doc_header(Hdr, Header),
    reverse(Header, HList),
    reverse(Doc, DList),
    make_query(doc(Thy, Name, HList, DList, []), Q).

make_doc_header([], []).
make_doc_header([LCD-[]|Hdr], [LCD|Header]) :-
    !,
    make_doc_header(Hdr, Header2),
    add_separator(Header2, nl, Header).
make_doc_header([_-[H|HH]|Hdr], Header) :-
    make_doc_header(Hdr, Header1),
    add_separator(Header1, nl, Header2),
    append([H|HH], Header2, Header).
    
%% private make_query(Struct, Query).
%%
%%  Query is a query that will create a fact corresponding to Struct.
%%  Eventually, this procedure should generate *clauses* (of multifile
%%  procedures) rather than queries.
%%  
%%  mode make_query(@compund, -compound).
%%
make_query(command(Thy, Tab, Call, Modes, Impl, Doc),
   (?- add_command(Thy, Tab, Call, Modes, Impl, Doc))).
make_query(doc(Thy, Name, Header, Doc0, Cmds),
   (?- add_doc(Thy, Name, Header, Doc, Cmds))) :-
    dump_documentation(Name, Doc0, Doc).
make_query(ext(ThisThy, ExtNode, XRefTag, ThisNode),
   (?- add_doc_extension(ThisThy, ExtNode,
                         [xref(XRefTag, ThisThy, ThisNode)]))).
make_query(frag(Thy, Head, Doc),
   (?- add_doc_fragment(Thy, Head, Doc))).
make_query(wrap(Thy, Call, Msg, Subs),
   (?- add_tactic_wrap_info(Thy, Call, Msg, Subs))).
make_query(parm(Thy, Name, DocNode, Set, Get),
   (?- add_parameter(Thy, Name, DocNode, Set, Get))).
make_query(elf(Thy, Name, DocNode),
   (?- add_elf_name(Thy, Name, DocNode))).

%% private dump_documentation(Name, Doc0, Doc).
%%
dump_documentation(Name, Doc0, Doc) :-
    clause(dumping_documentation,true),
    !,
    map(wl_atom_to_codes, [Doc0,Doc1]),
    atom_codes(Name, NameC),
    make_doc_file_name(NameC, FileC),
    atom_codes(File, FileC),
    open(File, write, Strm),
    encoded_write_term(Strm, doc(Name,Doc1), []),
    close(Strm),
    Doc = [file(File)].
dump_documentation(_, Doc, Doc).

make_doc_file_name([], ".doc.qge").
make_doc_file_name([C|Cs], [D|Ds]) :-
    ( C == 0'/ ->
        D = 0'@
    ;
        D = C
    ),
    make_doc_file_name(Cs, Ds).

wl_atom_to_codes(A, chars(C)) :-
    atom(A),
    \+ member(A, [nl,tab]),
    !,
    atom_codes(A, C).
wl_atom_to_codes(A, A).

%% private make_doc_name(Table, Functor, Arity, Name).
%% private make_doc_name(Table, Functor, Name).
%%
%%  Name is the name of the doc node for Functor/Arity in Table.
%%  Arity may be omitted for parameters.
%%
%%  mode make_doc_name(@cmdtab, @atom, @integer, ?atom).
%%  mode make_doc_name(@cmdtab, @atom, ?atom).
%% 
make_doc_name(Tab, F, N, Name) :-
    concat_atom([Tab, :, F, /, N], Name).
make_doc_name(Tab, F, Name) :-
    concat_atom([Tab, :, F], Name).

%% private expand_decls_get_commands(Queries, Commands).
%%
%%  Create the command cross-references for the doc node.
%%
%%  mode expand_decls_get_commands(+list(compound), -list(compound)).
%%
expand_decls_get_commands([], []).
expand_decls_get_commands([Q|Qs], [cmd(Thy, Table, Fun, Arity)|Cs]) :-
    make_query(command(Thy, Table, Call, _, _, _), Q),
    !,
    functor(Call, Fun, Arity),
    expand_decls_get_commands(Qs, Cs).
expand_decls_get_commands([_Q|Qs], Cs) :-
    expand_decls_get_commands(Qs, Cs).

%% private exp_bind_vars_names(VarsAndNames).
%%
%%  Takes a list of Variable=Name pairs and binds the Variables to their Names.
%%
%%  mode exp_bind_vars_names(+list(compound)).
%%
exp_bind_vars_names([]).
exp_bind_vars_names([N=N|Ns]) :-
    exp_bind_vars_names(Ns).

%% private open_append_commas(Commas, Tail).
%%
%%  Binds the tail of Commas (a comma-separated sequence with an uninstantiated
%%  tail) to Tail.
%%
%%  mode open_append_commas(+term, +term).
%%
open_append_commas(Var, Tail) :-
    var(Var),
    !,
    Var = Tail.
open_append_commas((_,Rest), Tail) :-
    open_append_commas(Rest, Tail).

%% strip_leading_nls(WriteList0, WriteList).
%%
%%  WriteList is WriteList0 with leading newlines removed.
%% 
%%  mode strip_leading_nls(@writelist, ?writelist).
%%
strip_leading_nls([], []).
strip_leading_nls([nl|W0], W) :-
    !,
    strip_leading_nls(W0, W).
strip_leading_nls(W, W).

%% add_separator(RestOfList0, Separator, RestOfList).
%%
%%  If RestOfList0 is empty, return it.
%%  Otherwise, add a Separator.
%%
%%  mode add_separator(@list(term), @term, ?list(term)).
%% 
add_separator([], nl, []).
add_separator([X|Xs], Sep, [Sep,X|Xs]).

