%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: io.ql,v 1.93 2003/05/05 05:37:55 itp Exp $

#include "../lib/ergo.h"

%% get_command(Prompt, OpTable, Cmd).
%% get_command(Prompt, OpTable, Cmd, String).
%%
%%  Get next command from user or macro file.
%%  It reads the next term from the current input stream
%%  (after closing any empty streams) and returns it in Cmd.
%%
%%  The four argument version also returns the string of characters
%%  that were parsed to produce Cmd.  This is useful if you want
%%  to preserve whitespace, comments or anonymous variable names.
%%  The current implementation of the four argument version handles
%%  terms that straddle several lines, but does not handle multiple
%%  terms on a single line -- it produces an error then retries.
%%  This could be fixed by storing any unused portion of each input
%%  line in some persistent area between calls to get_command/4.
%%
%%  The three argument version is likely to be faster because it
%%  is implemented using the Qu-Prolog C-level tokeniser, whereas
%%  the four argument version handles String at the Prolog level.
%%
%%  The reader treats forms starting with 'READER SPECIAL' specially:
%%   'READER SPECIAL''ECHO'(Chars) -- echo Chars and parse it to get Cmd.
%%   'READER SPECIAL''EXEC'(Cmd)   -- pass Cmd to reader_special_execute_hook
%%                                    and read again.
%% 
%%  mode get_command(@list(term), @atom, -term).
%%
get_command(Prompt, OpTable, C) :-
    repeat,
    (\+ in_macro ->
        writelist(Prompt), flush_output
    ;
        true
    ),
    get_command2(Prompt, OpTable, C),
    !.

get_command(Prompt, OpTable, C, String) :-
    repeat,
    (\+ in_macro ->
        writelist(Prompt), flush_output
    ;
        true
    ),
    get_term_lines(String),
    debug(get_command, ['read string: ', chars(String)]),
    ( String = "" ->
        %% Qu-Prolog bug?  Cannot parse empty string stream.
        C = end_of_file
    ;
        open_string(read(String), Strm),
        unwind_protect(
            read_term(Strm,C,
                      [remember_name(true),
                       op_table(OpTable),
                       obvar_prefix_table(OpTable)]),
            close(Strm))
    ),
    !,
    debug(get_command, ['parsed command: ', w(C)]),
    announce_event(read_input(get_command(Prompt, OpTable, C), C)).


%% private get_command2(Prompt, OpTable, Cmd).
%% private get_command3(Prompt, OpTable, Cmd0, Cmd).
%%
%%  mode get_command2(@list(term), @atom, -term).
%%  mode get_command3(@list(term), @atom, +term, -term).
%%
get_command2(Prompt, OpTable, C) :-
    current_ergo_input(S),
    read_term(S,C1,
              [remember_name(true),
               op_table(OpTable), obvar_prefix_table(OpTable)]),
    !,
    debug(get_command, ['read command: ', w(C1)]),
    get_command3(Prompt, OpTable, C1, C),
    debug(get_command, ['returning command: ', w(C)]),
    announce_event(read_input(get_command(Prompt, OpTable, C), C)).

get_command3(_, _, A, _) :-
    A == end_of_file,
    in_macro,
    !,
    exit_macro,
    fail.
get_command3(_, _, A, Cmd) :-
    A == end_of_file,
    %% \+ in_macro
    !,
    nl,
    Cmd = end_of_file.
get_command3(_, OpTable, A, Cmd) :-
    nonvar(A),
    A = 'READER SPECIAL''ECHO'(CmdChars),
    !,
    %% Special: echo, parse and return cmd
    writelist([chars(CmdChars)]),
    open_string(read(CmdChars), Strm),
    unwind_protect(
        read_term(Strm,Cmd,
                  [remember_name(true),
                   op_table(OpTable),
                   obvar_prefix_table(OpTable)]),
        close(Strm)).
get_command3(Prompt, OpTable, A, Cmd) :-
    nonvar(A),
    A = 'READER SPECIAL''EXEC'(C),
    !,
    %% Special: execute and read again (without prompt)
    ( less_verbose(99, handle_event(reader_special_execute_hook(C))) ->
        get_command2(Prompt, OpTable, Cmd)
    ;
        writelist(['Failed to handle reader special command ', w(C)]),
        get_command(Prompt, OpTable, Cmd)
    ).
get_command3(_, _, C, C).
    %% C \== end_of_file or special: silently return command.


%% get_term_lines(String).
%%
%%  Reads input lines until a complete term has been read.
%%  Currently, it raises an error if more than one term per line.
%%
%%  mode get_term_lines(-chars).
%%
get_term_lines(Lines) :-
    get_term_lines2("", Lines).

get_term_lines2(Lines0, Lines) :-
    current_ergo_input(S),
    debug(get_command, ['Getting lines for command term']),
    get_ergo_line(S, Line, Eof),
    append(Lines0, Line, Lines2),
    ( io_term(Lines2, Tail) ->
        %% Lines2..Tail contains one complete term.
        ( io_whitespace(Tail, "") ->
            Lines = Lines2
        ;
            handle_error(['multiple commands per line not allowed here.'])
        )
    ; Eof == no ->
        get_term_lines2(Lines2, Lines)
    ;
        Lines2 = Lines
    ).



%% get_ergo_line(Stream, Line).
%%
%%  Reads the remainder of the current line of input and
%%  returns it in Line.  Line includes the trailing nl.
%%
%%  The reader treats forms starting with 'READER SPECIAL' specially:
%%   'READER SPECIAL''ECHO'(Chars) -- echo Chars and return as Line.
%%   'READER SPECIAL''EXEC'(Cmd)   -- pass Cmd to reader_special_execute_hook
%%                                    and read again.
%% 
%%  mode get_ergo_line(@stream, ?list(integer)).
%%
get_ergo_line(S, Line) :-
    get_ergo_line(S, Line, _Eof).

get_ergo_line(S, Line, Eof) :-
    get_code(S,Char),
    debug(get_ergo_line, ['Got code: ', w(Char)]),
    get_ergo_line2(Char, S, [], Line, Eof),
    debug(get_ergo_line, ['Got line: ', chars(Line)]),
    announce_event(read_input(get_ergo_line(S, Line), Line)).

get_ergo_line2(10, _, Line, LineR, no) :-
    !,
    reverse([10|Line], LineR).
get_ergo_line2(-1, S, Line, LineR, yes) :-
    !,
    %% end of file, so we close the file if it is the current
    %% macro file that we are reading from.  This is a bit unclean.
    ( current_ergo_input(S),in_macro ->
        exit_macro,
        fail
    ;
        true
    ),
    reverse(Line, LineR).
get_ergo_line2(39, S, "LAICEPS REDAER'", Line, Eof) :-
    !,
    read(S, Term),
    debug(get_ergo_line, ['Read special command: ', q(Term)]),
    ( Term = 'ECHO'(Line) ->
        %% Special: echo and return.
        writelist([chars(Line)]),
        Eof = no
    ; Term = 'EXEC'(Cmd) ->
        %% Special: execute and read again
        ( less_verbose(99, handle_event(reader_special_execute_hook(Cmd))) ->
            get_ergo_line(S, Line, Eof)
        ;
            writelist(['Failed to handle reader special command ', w(Cmd)]),
            get_ergo_line(S, Line, Eof)
        )
    ;
        handle_error(['Cannot parse reader special command ', w(Term)]),
        get_ergo_line(S, Line, Eof)
    ).
get_ergo_line2(C, S, Line0, Line, Eof) :-
    get_code(S,C2),
    debug(get_ergo_line, ['Got code: ', w(C2)]),
    get_ergo_line2(C2, S, [C|Line0], Line, Eof).



%% command set_batch_mode(Flag).
%%
%%  This determines whether interactive questions asked via
%%  question/2 should be answered automatically or not.
%%  If Flag=true then all questions will automatically be answered
%%  with their default answer (which should be the 'safe' option, in
%%  some sense).  If Flag=false, then the usual interactive behaviour
%%  will result.
%%
%%  mode set_batch_mode(?truefalse).
%%


batch_mode parameter
  [set_batch_mode, get_batch_mode]
doc
  'This determines whether interactive questions asked via',nl,
  'question/2 should be answered automatically or not.',nl,
  'If batch_mode===true, all questions will automatically be answered',nl,
  'with their default answer (which should be the ''safe'' option, in',nl,
  'some sense).  If batch_mode===false, the usual interactive behaviour',nl,
  'will result.'
extends
  parameters.

set_batch_mode(true) :-
    ( batch_mode ->
        true
    ;
        asserta(batch_mode)
    ).
set_batch_mode(false) :-
    retractall(batch_mode).


%% private get_batch_mode(Flag).
%%
%%  mode get_batch_mode(?truefalse).
%%
get_batch_mode(Flag) :-
    ( batch_mode ->
        Flag = true
    ;
        Flag = false
    ).


%% question(Prompt, Default).
%%
%%  Ask the user a question, with the given Default (which must be
%%  'yes' or 'no').
%%  Note that if batch_mode has been set to true, (see set_batch_mode),
%%  the default answer may be used automatically. 
%%
%%  mode question(@writelist, @atom).
%%
question(Question, Default) :-
    batch_mode,
    !,
    writelist(Question),
    writelist([' : ', q(Default), '  (batch_mode).',nl]),
    Default = yes.
question(Question, Default) :-
    check_or_error(
        question_default(Default, YesNum, NoNum),
        ['default answer in question/2 must be yes or no, not: ',q(Default)]),
    handle_question(Question, Default, YesNum, NoNum, Answer),
    !,
    Answer = yes.

handle_question(Question, Default, _, _, Answer) :-
    handle_event(question(Question, Default, Answer)).
handle_question(Question, Default, YesNum, NoNum, Answer) :-
    writelist(Question),
    writelist([' (', q(Default), ') : ']),
    current_ergo_input(S),
    get_ergo_line(S,Line),
    ( answer_matches(Line, "yes", YesNum) ->
        Answer = yes
    ; answer_matches(Line, "no", NoNum) ->
        Answer = no
    ;
        writelist(['Please reply ''yes.'' or ''no.''.',nl]),
        handle_question(Question, Default, YesNum, NoNum, Answer)
    ).

question_default(yes, 0, 1).
question_default(no,  1, 0).


%% answer_matches(Line, ExpectedAnswer, NumChars).
%%
%%  Succeeds iff Line is a prefix of ExpectedAnswer (ignoring whitespace
%%  at the beginning and end of Line and ignoring any trailing fullstop on
%%  the end of Line) and at least NumChars of ExpectedAnswer have been
%%  matched.  The matching is case sensitive and the usual convention
%%  is for ExpectedAnswer to be all lower case.
%%
%%  mode answer_matches(@list(integer), @list(integer), integer).
%%
answer_matches([C|Line], Answer, N) :-
    whitespace(C),
    !,
    answer_matches(Line, Answer, N).
answer_matches(Line, Answer, N) :-
    answer_matches2(Line, Answer, N).

answer_matches2([], _, N) :-
    PRE('answer_matches2: [], N =< 0', integer(N)),
    N =< 0.
answer_matches2([A|Line], [A|Answer], N) :-
    !,
    PRE('answer_matches2: N - 1', integer(N)),
    N1 is N - 1,
    answer_matches2(Line, Answer, N1).
answer_matches2(Line, _, N) :-
    PRE('answer_matches2: Line, N =< 0', integer(N)),
    N =< 0,
    filter(nonwhitespace, Line, Rem),
    (
        Rem = ""
    ;
        Rem = "."
    ),
    !.


%% whitespace(Char).
%%
%%  Succeeds if Char is whitespace (^I, ^J or space).
%%
whitespace(32).
whitespace(10).
whitespace(9).


%% nonwhitespace(Char).
%%
%%  Succeeds if Char is not whitespace.
%%
nonwhitespace(C) :-
    \+ whitespace(C).


%% A simplified tokeniser for get_command_and_string.
%% Its main purpose is to accept a complete term (finishing with
%% full stop followed by whitespace), without being tricked by
%% full stops inside quoted atoms or strings etc.
%% It does not handle 0'... notation for characters yet.
%%
%% Each io_<kind>(String, Rest) procedure succeeds iff String
%% begins with a <kind> and Rest is the remaining portion of String.
%%

io_term        --> io_token, io_tokens, io_endterm.

io_tokens      --> io_token, !, io_tokens.
io_tokens      --> "".

io_token       --> [Q], {member(Q,"'\"")}, !, io_qchars(Q), [Q].
io_token       --> io_comment, !.
io_token       --> io_char.

io_graphics_char --> [Q], {member(Q, "-/+*<=>#@$\\^&~`:.?|")}, 
                     !, io_graphics_char.
io_graphics_char --> "".

io_qchars(Q)   --> io_qchar(Q), !, io_qchars(Q).
io_qchars(_)   --> "".

io_qchar(Q)    --> [Q,Q].
io_qchar(Q)    --> "\\", [_].
io_qchar(Q)    --> [C], {C =\= Q}.

io_char        --> io_endterm,!,{fail}.
io_char        --> [Q], {member(Q, "-/+*<=>#@$\\^&~`:.?|")}, 
                   !, io_graphics_char.
io_char        --> [_].

io_endterm     --> ".", io_whitechar.

io_comment     --> "%", !, io_restofline.

io_restofline  --> "\n", !.
io_restofline  --> [_], io_restofline.

io_whitespace  --> io_whitechar, !, io_whitespace.
io_whitespace  --> io_comment, !.
io_whitespace  --> "".

io_whitechar   --> " ".
io_whitechar   --> "\t".
io_whitechar   --> "\n".




%% is_legal_writelist(List).
%% check_is_legal_writelist(List).
%%
%%  Succeeds iff List is a legal argument to writelist/1.
%%
%%  mode is_legal_writelist(@writelist).
%%  mode check_is_legal_writelist(@writelist).
%%
is_legal_writelist([]).
is_legal_writelist([E|ES]) :-
    is_legal_writelist_elem(E),
    is_legal_writelist(ES).

check_is_legal_writelist([]).
check_is_legal_writelist([E|ES]) :-
    check_or_error(
        is_legal_writelist_elem(E),
        ['illegal writelist element: ',q(E)]),
    check_is_legal_writelist(ES).


is_legal_writelist_elem(0) :- !, fail.  %% catch variables.
is_legal_writelist_elem(tab(_)).
is_legal_writelist_elem(w(_)).
is_legal_writelist_elem(chars(_)).
is_legal_writelist_elem(q(_)).
is_legal_writelist_elem(canon(_)).
is_legal_writelist_elem(qchars(_)).
is_legal_writelist_elem(thy(_)).
is_legal_writelist_elem(thy(_,_)).
is_legal_writelist_elem(code(_)).
is_legal_writelist_elem(code(_,_)).
is_legal_writelist_elem(xref(_,_,_)).
is_legal_writelist_elem(xinc(_,_)).
is_legal_writelist_elem(begin(_)).
is_legal_writelist_elem(end(_)).
is_legal_writelist_elem(item(_)).
is_legal_writelist_elem(max_depth(_,_)).
is_legal_writelist_elem(wl(_)).
is_legal_writelist_elem(emph(_)).
is_legal_writelist_elem(left(_,_)).
is_legal_writelist_elem(right(_,_)).
is_legal_writelist_elem(centre(_,_)).
is_legal_writelist_elem(file(_)).
is_legal_writelist_elem(para).
is_legal_writelist_elem(br).
is_legal_writelist_elem(A) :- atom(A).


%% writelist_first_line(WriteList, FirstLine).
%%
%%  FirstLine is the first line of WriteList, excluding the 'nl'.
%%  That is, List is the smallest prefix of WriteList that is
%%  immediately followed by an 'nl' atom, or the whole of
%%  WriteList if it does not contain any 'nl'.
%%
%%  mode writelist_first_line(+writelist, ?writelist).
%%
writelist_first_line(E, E) :-
    atom(E),
    !.
writelist_first_line([nl|_], Line) :-
    !,
    Line = [].
writelist_first_line([E|List], [E|Line]) :-
    writelist_first_line(List,Line).

writelist header
  'Writelists'
extends
  concepts
doc
  'Most Ergo output is produced by the writelist/1,2 or msg/2,3 commands.',nl,
  'These commands take a ''writelist'' argument.',
  para,
  'A writelist is a list of atoms and terms, to be written in turn,',nl,
  'with the following interpretation of each element:',
  para,
  begin(list),
  item('''nl'':'),
  '  begin a new line, then write current indentation.',nl,
  item('''tab'':'),
  '  write a tab character.',nl,
  item('tab(N):'),
  '  write N spaces.',nl,
  item('w(T):'),
  '  write T using the Qu-Prolog write command and default operators.',nl,
  item('q(T):'),
  '  write T using the Qu-Prolog writeq command and default operators.',nl,
  item('canon(T):'),
  '  write T using the Qu-Prolog write_canonical command.',nl,
  item('thy(T):'),
  '  write T, viewed as a term in the current theory.',nl,
  item('thy(Thy,T):'),
  '  write T, viewed as a term in theory Thy.',nl,
  item('code(T):'),
  '  write T, viewed as Prolog code in the current theory.',nl,
  item('code(Thy,T):'),
  '  write T, viewed as Prolog code in theory Thy.',nl,
  item('chars(T):'),
  '  write the chars list T (eg. "abc") as an unquoted string.',nl,
  item('qchars(T):'),
  '  write the chars list T (eg. "abc") as a quoted string.',nl,
  item('begin(Environment):'),
  '  begins a new Environment, within which certain commands',nl,
  '  behave specially.',nl,
  '  At present, the only defined values for Environment are `list'' and',nl,
  '  `verb''.',nl,
  '  Within a ''list'' environment the command `item(Tag)'' produces',nl,
  '  a line break and text for Tag.  Tag may be a number, an atom, or',nl,
  '  a writelist.',
  item('end(Environment):'),
  '  ends the most recently started Environment.',nl,
  item('xref(WL,Thy,Name):'),
  '  this writes out the atom/writelist WL and marks it as a',nl,
  '  cross reference to doc node Name in theory Thy (Thy may be ''global'').',nl,
  '  Name may also be cmd(CmdName/Arity) or ''parameter:ParamName''.',nl,
  '  WL must not contain any newlines and should not begin or end',nl,
  '  with whitespace (if whitespace is desired, it should be put into',nl,
  '  the adjacent elements of the surrounding writelist).',
  para,
  '  Typical examples are:',
  para,
  '    [''See '',xref(''fred/2'',global,cmd(fred/2)),''.'']',
  para,
  '    [''See '',xref(''fred/2'',global,''proof:fred/2''),''.'']',
  para,
  '  These may print out just ''See fred/2.'', or may',nl,
  '  underline ''fred/2'', or may (in a printed manual) produce',nl,
  '  something like ''See fred/2 (page 89).''',nl,
  item('xinc(Thy, Frag(Args)):'),
  '  this writes the instantiated doc fragment from Thy.',nl,
  item('wl(L):'),
  '  writes the writelist L.',nl,
  item('emph(L):'),
  '  writes the writelist L, ', emph([emphasizing]), ' it in some way.',nl,
  item('verb(L):'),
  '  writes the writelist L in verabtim mode in some way.',nl,
  item('left(W,L):'),
  '  left justify writelist(L) in a column of width W.',nl,
  item('right(W,L):'),
  '  right justify writelist(L) in a column of width W.',nl,
  item('centre(W,L):'),
  '  centre writelist(L) in a column of width W.',nl,
  item('para:'),
  '  Write a paragraph break.',nl,
  item('br:'),
  '  Write a hard line break.',nl,
  end(list),
  nl,
  'Note: the q(T), canon(T), qchars(T), thy(T), thy(Thy,T),',nl,
  'code(T) and code(Thy,T) items all write T in a way which can be',nl,
  'parsed back in, whereas the other ways of writing terms may not.',nl.


%% writelist_to_chars(List, Chars).
%% writelist_to_chars(List, Indent, Chars).
%% writelist_to_chars(List, Indent, Depth, Chars).
%%
%%  Converts a writelist into a string (list of characters).
%%
%%  mode writelist_to_chars(@writelist, ?chars).
%%  mode writelist_to_chars(@writelist, @writelist, ?chars).
%%  mode writelist_to_chars(@writelist, @writelist, @integer, ?chars).
%%
writelist_to_chars(WL, Chars) :-
    writelist_to_chars(WL, [], 0, Chars).

writelist_to_chars(WL, Indent, Chars) :-
    writelist_to_chars(WL, Indent, 0, Chars).

%writelist_to_chars(WL, Indent, Depth, Chars) :-
%    open_string(write, S),
%    unwind_protect(
%        ( writelist2(WL, S, Indent, Depth),
%          stream_to_chars(S, Chars)),
%        close(S)).
% fixed close problem

writelist_to_chars(WL, Indent, Depth, Chars) :-
    open_string(write, S),
    (writelist2(WL, S, Indent, Depth) ->  
       stream_to_chars(S, Chars)
    ;  
       close(S)
    ).
 

%% writelist(List).
%% writelist(Stream, List).
%% writelist_indent(List, Indent).
%% writelist_indent(Stream, List, Indent).
%%
%%
%%  Write the elements of List to Stream (or to the current output
%%  stream if Stream is not specified).
%%
%%  Indent must be a list of terms to be written by writelist/1.
%%
%%  mode writelist(@writelist).
%%  mode writelist(@stream, @writelist).
%%  mode writelist_indent(@writelist, @writelist).
%%  mode writelist_indent(@stream, @writelist, @writelist).
%%
writelist(L) :-
    noinline_writelist(L).

noinline_writelist(L) :-
    current_output(S),
    writelist2(L, S, [], 0).

writelist(S, L) :-
    writelist2(L, S, [], 0).

writelist_indent(L, I) :-
    current_output(S),
    writelist2(L, S, I, 0).

writelist(S, L, I) :-
    writelist2(L, S, I, 0).


writelist2([],_,_,_) :-
    !.
writelist2([T|TS], S, Indent, Depth) :-
    !,
    writelist_elem(T, S, Indent, Depth),
    writelist2(TS, S, Indent, Depth).
writelist2(A, S, _, _) :-
    atom(A),
    !,
    write_atom(S, A).
writelist2(T, _, _, _) :-
    handle_illegal_arg(writelist,1,list,T).


%% private writelist_elem(Elem).
%% private writelist_elem(Elem, Stream, Indent, MaxDepth).
%%
%%  mode writelist_elem(@term).
%%  mode writelist_elem(@term, @stream, @writelist, @integer).
%%
writelist_elem(T) :-
    noinline_writelist_elem(T).

noinline_writelist_elem(T) :-
    current_output(S),
    writelist_elem(T, S, [], 0).

writelist_elem(T, S, _, _) :-
    var(T),
    !,
    error('Warning: writelist sees a variable: '),
    errornl(T),
    writeR(S, T).
writelist_elem(nl, S, Indent, _) :-
    !,
    nl(S),
    writelist2(Indent, S, [], 0).
writelist_elem(tab, S, _, _) :-
    !,
    put_code(S, 9).
writelist_elem(para, S, _, D) :-
    !,
    handle_write(para, D, S).
writelist_elem(br, S, _, D) :-
    !,
    handle_write(br, D, S).
writelist_elem(A, S, _, _) :-
    atom(A),
    !,
    write_atom(S, A).
writelist_elem(tab(N), S, _, _) :-
    !,
    tab(S, N).
writelist_elem(w(T), S, _, D) :-
    !,
    write_term(S, T, [remember_name(true), max_depth(D),
                      op_table(user), obvar_prefix_table(user)]).
writelist_elem(q(T), S, _, D) :-
    !,
    write_term(S, T, [quoted(true), remember_name(true), max_depth(D),
                      op_table(user), obvar_prefix_table(user)]).
writelist_elem(canon(T), S, _, D) :-
    !,
    write_term(S, T, [quoted(true), ignore_ops(true),
                      remember_name(true), max_depth(D)]).
writelist_elem(max_depth(D,L), S, Indent, _) :-
    !,
    writelist2(L, S, Indent, D).
writelist_elem(left(W,L), S, _, D) :-
    !,
    writelist_to_chars(L, [], D, Chars),
    length(Chars, Len),
    ( Len < W ->
        Pad is W - Len,
        writelist(S, [chars(Chars), tab(Pad)])
    ;
        length(Front, W),
        append(Front, _Back, Chars),
        writelist(S, [chars(Front)])
    ).
writelist_elem(right(W,L), S, _, D) :-
    !,
    writelist_to_chars(L, [], D, Chars),
    length(Chars, Len),
    ( Len < W ->
        Pad is W - Len,
        writelist(S, [tab(Pad), chars(Chars)])
    ;
        length(Front, W),
        append(Front, _Back, Chars),
        writelist(S, [chars(Front)])
    ).
writelist_elem(centre(W,L), S, _, D) :-
    !,
    writelist_to_chars(L, [], D, Chars),
    length(Chars, Len),
    ( Len < W ->
        Pad is W - Len,
        PadL is Pad//2,
        PadR is Pad - PadL,
        writelist(S, [tab(PadL), chars(Chars), tab(PadR)])
    ;
        length(Front, W),
        append(Front, _Back, Chars),
        writelist(S, [chars(Front)])
    ).
writelist_elem(chars(T), S, _, _) :-
    !,
    write_chars(T, S, false).
writelist_elem(qchars(T), S, _, _) :-
    !,
    put_code(S, 34),
    write_chars(T, S, 34),
    put_code(S, 34).
writelist_elem(thy(T), S, _, D) :-
    !,
    global_state_lookup(current_dir,Thy),
    handle_write(thy(Thy,T), D, S).
writelist_elem(thy(Thy,T), S, _, D) :-
    !,
    handle_write(thy(Thy,T), D, S).
writelist_elem(code(T), S, _, D) :-
    !,
    (
        global_state_lookup(current_dir,Thy)
    ;
        Thy = global
    ),
    !,
    handle_write(code(Thy,T), D, S).
writelist_elem(code(Thy,T), S, _, D) :-
    !,
    handle_write(code(Thy,T), D, S).
writelist_elem(xref(WL,Thy,Name), S, _, D) :-
    !,
    ( atom(Name) ->
        RealName = Name
    ; Name = cmd(Cmd/Arity),
      query_theory(doc_node(RealName, _, _, Cmds), Thy),
      member(cmd(_, _, Cmd, Arity), Cmds) ->
          true
    ;
        handle_error(['Bad xref: ', w(Name)])
    ),
    handle_write(xref(WL,Thy,RealName), D, S).
writelist_elem(begin(E), S, _, D) :-
    !,
    handle_write(begin(E), D, S).
writelist_elem(end(E), S, _, D) :-
    !,
    handle_write(end(E), D, S).
writelist_elem(item(T), S, _, D) :-
    !,
    handle_write(item(T), D, S).
writelist_elem(wl(L), S, _, D) :-
    !,
    handle_write(wl(L), D, S).
writelist_elem(emph(L), S, _, D) :-
    !,
    handle_write(emph(L), D, S).
writelist_elem(xinc(Thy, Frag), S, _, D) :-
    !,
    handle_write(xinc(Thy, Frag), D, S).
writelist_elem(file(F), S, _, D) :-
    !,
    handle_write(file(F), D, S).
writelist_elem(Term, S, _, D) :-
    handle_write(Term, D, S),
    !.
writelist_elem(Term, S, _, _) :-
    error('Warning: bad writelist term: '),
    errornl(Term),
    write_term(S, Term, [remember_name(true),
                         op_table(user), obvar_prefix_table(user)]).


%% private write_chars(Chars, Stream, QuoteChar).
%%
%%  Writes Chars as a string (without quotes) to Stream.
%%  It prints two of any character whose numeric value is QuoteChar.
%%  (Set QuoteChar to a non-integer if you do not want this to happen).
%%
%%  mode write_chars(+list(integer), +stream, +integer).
%%
write_chars([], _, _).
write_chars([C|Cs], S, Q) :-
    put_code(S, C),
    (C = Q ->
        put_code(S, C)
    ;
        true
    ),
    write_chars(Cs, S, Q).


%% private handle_write(Item, Depth, Stream).
%%
handle_write(Item, Depth, Stream) :-
    handle_event(write(Stream, Depth, Item)),
    !.
handle_write(code(_,Code), Depth, Stream) :-
    write_term(Stream, Code,
               [quoted(true), remember_name(true),
                op_table(user), obvar_prefix_table(user),
                max_depth(Depth)
               ]).
handle_write(thy(_,Term), Depth, Stream) :-
    write_term(Stream, Term,
               [quoted(true), remember_name(true),
                op_table(global), obvar_prefix_table(global),
                max_depth(Depth)
               ]).
handle_write(xref(WL,_Thy,Name), _Depth, Stream) :-
    writelist2(WL,Stream,[],0),
    write_atom(Stream, '<<'),
    write_atom(Stream, Name),
    write_atom(Stream, '>>').
handle_write(begin(list), _, _).
handle_write(end(list), _, Stream) :-
    nl(Stream).
handle_write(begin(verb), _, _).
handle_write(end(verb), _, Stream) :-
    nl(Stream).
handle_write(item(Tag), _, Stream) :-
    nl(Stream),
    write_list_item_tag(Tag, Stream),
    write_atom(Stream, ' ').
handle_write(xinc(Thy, Frag), _Depth, Stream) :-
    query_theory(doc_fragment(Frag, Doc), Thy),
    writelist(Stream, Doc).
handle_write(file(F), D, S) :-
    get_load_path(LPath),
    member(Dir, [lib(lib)|LPath]),
    full_file_name(Dir, F, _, File),
    access(File, 4, 0),
    !,
    open(File, read, Strm),
    encoded_read_term(Strm, doc(_,WL), []),
    close(Strm),
    writelist2(WL, S,  [], D).
handle_write(wl(L), D, S) :-
    writelist2(L, S, [], D).
handle_write(emph(L), D, S) :-
    write(S, '**'),
    writelist2(L, S, [], D),
    write(S, '**').    
handle_write(verb(L), D, S) :-
    writelist2(L, S, [], D).
handle_write(para, _D, S) :-
    nl(S),nl(S).
handle_write(br, _D, S) :-
    nl(S).

write_list_item_tag(WL, Stream) :-
    list(WL),
    !,
    writelist(Stream, WL).
write_list_item_tag(A, Stream) :-
    atomic(A),
    !,
    write(Stream, A).



general command
  msg(Caller, Msg),
  msg(Caller, Level, Msg)
extends
  general_cmds
doc
  'This is the recommended printing mechanism for all informational',nl,
  'messages from tactics and from other parts of the prover.',
  para,
  'Caller:  the name of the tactic or package that generated the message.',br,
  'Level:   an integer specifying the importance of the message.',br,
  'Msg:     a valid ', xref([writelist], global, writelist), '.',
  para,
  'msg/2 always succeeds, but only prints Msg if the current ',
  xref([verbosity],global,'parameter:verbosity'),nl,
  'number is greater than or equal to Level',nl,
  '(if there is no Level argument, then Level defaults to 1).',
  para,
  'Msg is automatically indented and followed by a newline, except',nl,
  'when it is the empty list, in which case nothing is printed. '
mode
  msg(@nonvar, @writelist),
  msg(@nonvar, @integer, @writelist).

msg(Caller, Msg) :-
    msg(Caller, 1, Msg).

msg(Caller, Level, Msg) :-
    msg_verbosity(V),
    ip_lookup(less_verbose, Delta),
    V - Delta >= Level,
    !,
    Importance is V - Delta - Level,
    handle_msg(Msg, Caller, Importance).
msg(_, _, _).

handle_msg([], _, _) :-
    !.
handle_msg(Msg, Caller, Importance) :-
    handle_event(msg(Caller, Importance, Msg)),
    !.
handle_msg(Msg, Caller, _) :-
    writelist_indent(['    ',q(Caller),':',tab|Msg],[tab,tab]),
    nl.


verbosity parameter
  [set_verbosity, get_verbosity]
doc
  'Determines how many trace messages will be displayed by tactics.',nl,
  'Zero means no messages.',nl,
  'Higher numbers mean more messages.',
  para,
  'This is typically set by the user to determine how many',nl,
  'tracing messages he/she wants to see.'
extends
  parameters.

set_verbosity(New) :-
    New >= 0,
    retractall(msg_verbosity(_)),
    asserta(msg_verbosity(New)).

get_verbosity(V) :-
    msg_verbosity(V0),
    ip_lookup(less_verbose,Delta),
    V is V0 - Delta.


%% less_verbose(+Tactic).
%% less_verbose(+N, +Tactic).
%%
%%  This subtracts N (which defaults to 1 in the first form) 
%%  from the current verbosity level during the execution of Tactic.
%%  It resets it to the original level when Tactic succeeds or fails.
%%
%%  Negative values of N *raise* the verbosity level for Tactic,
%%  which is something that tactics should probably not do.
%%
%%  This is typically used when a tactic calls an auxillary tactic
%%  and does not want too much output from the auxillary tactic.
%%
%%  mode less_verbose(+tactic).
%%  mode less_verbose(+integer, +tactic).
%%
less_verbose(Tac) :-
    less_verbose(1, Tac).

less_verbose(N, Tac) :-
    ip_lookup(less_verbose,Delta0),
    Delta is Delta0 + N,
    ip_set(less_verbose,Delta),
    ( ergo_version(ergo4) ->
        call_tac(Tac)
    ;
        call_predicate(Tac)
    ),
    ip_set(less_verbose, Delta0).


