%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: prefix_expansions.ql,v 1.7 2000/11/15 06:29:46 itp Exp $
%%
%% See expansions.ql for how to use this file.

#include "../lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% WARNING!
%% This file should be able to run OUTSIDE Ergo.
%% Beware of using Ergo-defined procedures in this file!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% TACTIC PREFIXING (can be outside Ergo, but normally will not be).
%%     Every definition and call on a procedure defined in the file is prefixed
%%     with a `module' name: the current theory.

%% command enable_prefix_expansions(Thy, File).
%% command disable_prefix_expansions.
%%
%%  Enable/disable module prefixing for Thy in File.
%%
%%  mode enable_prefix_expansions(@theory, @filename).
%%  mode disable_prefix_expansions.
%%
enable_prefix_expansions(Thy, File) :-
    build_tac_prefix_procs(Thy, File),
    add_ergo_expansion(tac_add_prefixes).

disable_prefix_expansions :-  %  errornl(dpeb),
    retractall(tac_prefix_procs(_, _)),
    remove_ergo_expansion(tac_add_prefixes).
%    errornl(dpef).

?- dynamic(tac_prefix_procs/2).

%% tac_add_prefixes(Term0, Term, Vars).
%%
%%  Add a module prefix to all instances of procedures defined in the file.
%%  We DCG-expand here too, for convenience.
%%
%%  mode tac_add_prefixes(@nonvar, -nonvar, @ignored).
%%  mode tac_add_prefixes(@nonvar, -nonvar).
%%
tac_add_prefixes(Var, _, _) :-
    var(Var),
    !,
    fail.
tac_add_prefixes([], [], _) :-
    !.
tac_add_prefixes([T0|Ts0], [T|Ts], _) :-
    !,
    tac_add_prefixes(T0, T),
    tac_add_prefixes(Ts0, Ts, _).
tac_add_prefixes(T0, T, _) :-
    tac_add_prefixes(T0, T).

tac_add_prefixes(Term0, Term) :-
    exp_debug(['trying prefix expansion on ', w(Term0)]),
    tac_add_prefixes_2(Term0, Term, Flag),
    nonvar(Flag),
    !,
    exp_debug(['done prefix expansion on ', w(Term)]).
tac_add_prefixes(Term, Term) :-
    %% we mustn't fail here, as there is iteration down a list above.
    exp_debug(['no prefix expansion needed on ', w(Term)]).

tac_add_prefixes_2(Term, Term, _) :-
    var(Term),
    !.
tac_add_prefixes_2(Higher, Higher, _) :-
    functor(Higher, Var, _),
    var(Var),
    !.
tac_add_prefixes_2((Head0 --> Body0), Clause, Flag) :-
    dcg((Head0 --> Body0), Clause0),
    tac_add_prefixes_2(Clause0, Clause, Flag).
tac_add_prefixes_2(--->(Tac0,CodeList0), --->(Tac,CodeList), Flag) :-
    !,
    exp_debug(['About to prefix ', w(--->(Tac0, CodeList0))]),
    functor(Tac0, Fun, Ar0),
    nonvar(Fun),
    Ar1 is Ar0 + 1,
    functor(RTac, Fun, Ar1),
    ( tac_prefix_procs(RTac, PTac) ->
        Flag = true,
        functor(PTac, PFun, Ar1),
        functor(Tac, PFun, Ar0),
        same_args(Tac0, Tac, 1, Ar0)
    ;
        Tac = Tac0
    ),
    tac_add_prefixes_codelist(CodeList0, CodeList, Flag).
tac_add_prefixes_2(Term0, Term, Flag) :-
    tac_prefix_procs(Term0, Term1),
    !,
    Flag = true,
    exp_debug(['prefixed ', q(Term0), ' giving ', q(Term1), nl]),
    tac_add_prefixes_2(Term1, Term, Flag).
tac_add_prefixes_2(Term0, Term, Flag) :-
    meta_predicate(Term0, GoalArgs),
    !,
    exp_debug(['About to prefix meta-args ', w(GoalArgs), ' of ', w(Term0)]),
    functor(Term0, Fun, Ar),
    functor(Term,  Fun, Ar),
    tac_add_prefixes_2(1, Ar, GoalArgs, Term0, Term, Flag).
tac_add_prefixes_2(Term, Term, _).

tac_add_prefixes_2(N, Ar, _, _, _, _) :-
    N > Ar,
    !.
tac_add_prefixes_2(N, Ar, GA, T0, T, F) :-
    PRE(['tac_add_prefixes_2'], N =< Ar),
    arg(N, T0, A0),
    arg(N, T, A),
    ( GA = [N|RGA] ->
        tac_add_prefixes_2(A0, A, F)
    ;
        A0 = A,
        GA = RGA
    ),
    N1 is N + 1,
    tac_add_prefixes_2(N1, Ar, RGA, T0, T, F).

tac_add_prefixes_codelist(V, V, _) :-
    var(V),
    !.
tac_add_prefixes_codelist([], [], _) :-
    !.
tac_add_prefixes_codelist([Code0 | Codes0], [Code | Codes], Flag) :-
    !,
    tac_add_prefixes_2(Code0, Code, Flag),
    tac_add_prefixes_2(Codes0, Codes, Flag).
tac_add_prefixes_codelist(Code0, [Code], Flag) :-
    tac_add_prefixes_2(Code0, Code, Flag).

%% QP built-ins
meta_predicate((?- _),                   [1]).
meta_predicate((:- _),                   [1]).
meta_predicate((_ :- _),                 [1,2]).
meta_predicate((_ , _),                  [1,2]).
meta_predicate((_ ; _),                  [1,2]).
meta_predicate((_ -> _),                 [1,2]).
meta_predicate((_ ^ _),                  [2]).
meta_predicate((\+ _),                   [1]).
meta_predicate(bagof(_,_,_),             [2]).
meta_predicate(call(_),                  [1]).
meta_predicate(catch(_,_,_),             [1,3]).
meta_predicate(findall(_,_,_),           [2]).
meta_predicate(once(_),                  [1]).
meta_predicate(setof(_,_,_),             [2]).
meta_predicate(unwind_protect(_,_),      [1,2]).
meta_predicate(delay(_,_),               [2]).

%% Ergo declarations
meta_predicate(add_command(_,_,_,_,_,_), [5]).
meta_predicate(check_or_error(_,_),      [1]).    
meta_predicate(add_listener(_,_),        [2]).
meta_predicate(add_handler(_,_),         [2]).
meta_predicate(less_verbose(_,_),        [2]).
meta_predicate(save_hyps(_),             [1]).
meta_predicate(thawing_new_vars(_),      [1]).
meta_predicate(open_pos(_,_),            [2]).
meta_predicate(open_pos(_,_,_),          [3]).
meta_predicate(open_hyp(_,_),            [2]).
meta_predicate(open_hyp(_,_,_),          [2]).
%% TO DO: make tactic arg last in meta_predicate(open_hyp(_,_,_)?

%%meta_predicate(Goal,                     GoalArgs) :-
%%    declared_meta_predicate(Goal, GoalArgs).
%% TO DO: Enable meta_predicate(Goal, GoalArgs) once declared_meta_predicate
%% is enabled.

%% TO DO: Test the use of declared_meta_predicate and enable if successful.
%% Once enabled, meta_predicate(Goal, GoalArgs) can be enabled.
%% Ray's last comment (verbatim): I think this probably a big job.

%%%% declared_meta_predicate(Goal, GoalArgs).
%%?- dynamic(declared_meta_predicate/2).

%%%% declare_meta_predicate_modes(Modes).
%%%%
%%%%  Examine Modes and declare a meta-predicate if appropriate.
%%%%
%%%%  PROBLEM: this uses the declared form of the command, not the implemented
%%%%  form!  So commands from theory-specific tactic files, with explicit
%%%%  implementations or with wrapped subproofs will be wrongly declared.
%%%%
%%%%  mode declare_meta_predicate_modes(@list(modespec)).
%%%%
%%declare_meta_predicate_modes([]).
%%declare_meta_predicate_modes([Mode|Modes]) :-
%%    functor(Mode, Fun, Arity),
%%    functor(Skel, Fun, Arity),
%%    declare_meta_predicate_modes(Modes, Mode, Skel, Arity).

%%declare_meta_predicate_modes([], Mode, Skel, Arity) :-
%%    declare_meta_predicate_mode(Mode, Skel, Arity).
%%declare_meta_predicate_modes([Mode2|Modes], Mode, Skel, Arity) :-
%%    \+ Skel = Mode2,
%%    !,
%%    declare_meta_predicate_mode(Mode, Skel, Arity),
%%    declare_meta_predicate_modes([Mode2|Modes]).
%%declare_meta_predicate_modes([_Mode2|Modes], Mode, Skel, Arity) :-
%%    %% Skel and Mode2 are unifiable, so can discard Mode2,
%%    declare_meta_predicate_modes(Modes, Mode, Skel, Arity).

%%declare_meta_predicate_mode(Mode, Skel, Arity) :-
%%    declare_meta_predicate_args(Mode, 1, Arity, GoalArgs),
%%    ( GoalArgs = [] ->
%%        true
%%    ;
%%        assert(declared_meta_predicate(Skel, GoalArgs))
%%    ).

%%declare_meta_predicate_args(_, Arg, Arity, []) :-
%%    Arg > Arity,
%%    !.
%%declare_meta_predicate_args(Mode, Arg, Arity, GoalArgs) :-
%%    %% Arg =< Arity,
%%    arg(Arg, Mode, ArgMode),
%%    ( arg(1, ArgMode, code) ->
%%        GoalArgs = [Arg|GoalArgs1]
%%    ;
%%        GoalArgs = GoalArgs1
%%    ),
%%    Arg1 is Arg + 1,
%%    declare_meta_predicate_args(Mode, Arg1, Arity, GoalArgs1).
    
%% private build_tac_prefix_procs(Module, File).
%% private build_tac_prefix_proc(Term, Module).
%%
%%  Record the need to prefix the procedures in File with Module.
%%
%%  mode build_tac_prefix_procs(@atom, @atom).
%%  mode build_tac_prefix_proc(@term, @atom).
%%
%% N.B. We explicitly do DCG expansion here, so lookup needs to do the same.
%%
build_tac_prefix_procs(Mod, File) :-
    open(File, read, Strm),
    repeat,
      %% do_consult will have set up op and obvar tables
      read(Strm, Term),
      ( dcg(Term, Term1) ->
          true
      ;
          Term = Term1
      ),
      build_tac_prefix_proc(Term1, Mod),
      Term == end_of_file,
    !,
    close(Strm).

build_tac_prefix_proc([], _) :-
    !.
build_tac_prefix_proc([T|Ts], Mod) :-
    !,
    build_tac_prefix_proc(T, Mod),
    build_tac_prefix_proc(Ts, Mod).
build_tac_prefix_proc(end_of_file, _) :-
    !.
build_tac_prefix_proc((?- _Query), _) :-
    !.
build_tac_prefix_proc((:- _Decl), _) :-
    !.
build_tac_prefix_proc((@ _Decl), _) :-
    !.
build_tac_prefix_proc((command(_,_)), _) :-
    !.
build_tac_prefix_proc((doc(_,_)), _) :-
    !.
build_tac_prefix_proc((Head :- _Body), Mod) :-
    !,
    build_tac_prefix_proc(Head, Mod).
build_tac_prefix_proc(Head, Mod) :-
    (
        atom(Head)
    ;
        compound(Head),
        functor(Head, F, _),
        atom(F)
    ),
    !,
    ( tac_prefix_procs(Head, _) ->
        true
    ;
        functor(Head, Fun, Ar),
        functor(UnPref, Fun, Ar),
        concat_atom([Mod, ':', Fun], PreFun),
        functor(Pref, PreFun, Ar),
        same_args(UnPref, Pref, 1, Ar),
        assert(tac_prefix_procs(UnPref, Pref))
    ).

