%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: uniqvars.ql,v 1.8 2000/11/15 06:29:46 itp Exp $
%%
%% See expansions.ql for how to use this file.
%% 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% WARNING!
%% This file should be able to run OUTSIDE Ergo.
%% Beware of using Ergo-defined procedures in this file!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%  This file generates warnings whenever:
%%
%%  1) a clause contains singleton variables.  That is, non-anonymous 
%%     variables that only occur once in the clause.
%%  2) a repeated variable has a name that starts with an underscore.
%%  3) the clause redefines '.'/2.
%%     (This usually indicates a full stop in the wrong place.)
%%
%%  To use these facilities, do:
%%   ?- enable_uniqvars_expansions.
%%
%%  before consulting/compiling some Qu-Prolog file.

%% command enable_uniqvars_expansions.
%% command disable_uniqvars_expansions.
%%
%%  Enable/disable singleton variable checking.
%%
enable_uniqvars_expansions :-
    add_ergo_expansion(check_uniqvars).

disable_uniqvars_expansions :-
    remove_ergo_expansion(check_uniqvars).    

%% Only the clauses that match this will be checked.
%% (Note that facts are not).
uniqvars_checkable_fact((_ :- _)).
uniqvars_checkable_fact((?- _)).
uniqvars_checkable_fact((:- _)).


%% check_uniqvars(In, Out, Names).
%%
%%  This procedure is designed to be passed to add_expansion_vars/1.
%%
check_uniqvars((L,R), _, Names) :-
    uniqvar_warning(' you are defining '',''/2.  A misplaced fullstop?',
                    '', Names, (L,R)),
    fail.
check_uniqvars(In, _, Names) :-
    uniqvars_checkable_fact(In),
    collect_variables(In, Vars),
    '$merge_sort'(Vars, Vars2, '@<'),  %% this should not have a dollar?
    check_uniqvars2(Vars2, Names, In),
    fail.

check_uniqvars2([], _, _).
check_uniqvars2([V], Names, Id) :-
    !,
    check_uniqvar_name(V, Names, Id).
check_uniqvars2([V1,V2|Vars], Names, Id) :-
    V1 == V2,
    !,
    %% V1 is a repeated variable.
    check_non_uniqvar_name(V1, Names, Id),
    uniqvars_skip(Vars, V1, Vars2),
    check_uniqvars2(Vars2, Names, Id).
check_uniqvars2([V|Vars], Names, Id) :-
    %% V is a singleton variable.
    check_uniqvar_name(V, Names, Id),
    check_uniqvars2(Vars, Names, Id).


%% check_uniqvar_name(Var, Names, Id).
%%
%%  Prints a warning if Var has a name in Names and that
%%  name does not begin with an underscore.
%%
%%  Id is some term that identifies the clause containing
%%  this variable -- it is used as part of the warning message.
%%
check_uniqvar_name(V, Names, Id) :-
    uniqvar_member(Names, V, Name),
    \+ atom_search(Name, 1, '_', 1),
    !,
    uniqvar_warning('singleton variable ', Name, Names, Id).
check_uniqvar_name(_, _, _).


%% check_non_uniqvar_name(Var, Names, Id).
%%
%%  Check that Var has a name in Names and that it does not
%%  begin with an underscore.
%%
%%  Id is some term that identifies the clause containing
%%  this variable -- it is used as part of the warning message.
%%
check_non_uniqvar_name(_, _, _) :- !.   %% UNTIL @THY EXPANSION IS FIXED.
check_non_uniqvar_name(V, Names, Id) :-
    uniqvar_member(Names, V, Name),
    !,
    ( atom_search(Name, 1, '_', 1) ->
        uniqvar_warning('repeated anonymous variable ',Name,Names,Id)
    ;
        true
    ).
check_non_uniqvar_name(V, Names, Id) :-
    uniqvar_warning('no name for repeated variable ',V,Names,Id).


%% uniqvar_warning(Msg, Value, Names, Id).
%%
%%  Prints a warning message containing Msg, Value and Id.
%%
%%  mode uniqvar_warning(@atom, @term, @list(compound), @term).
%%
uniqvar_warning(Msg,Val,Dict,Id) :-
    error('WARNING: '),
    error(Msg),
    error(Val),
    errornl,
    error('         in clause: '),
    %% To make the output more like the input clause, we
    %% instantiate all variables to their names, temporarily.
    uniqvar_instantiate_names(Dict),
    write_term(user_error, Id, [max_depth(3),numbervars(true)]),
    errornl,
    fail.
uniqvar_warning(_,_,_,_).

uniqvar_instantiate_names([]).
uniqvar_instantiate_names([A=A|CS]) :-
    !,
    uniqvar_instantiate_names(CS).
uniqvar_instantiate_names([_|CS]) :-
    %% instantiation has failed, probably because A is an obvar.
    uniqvar_instantiate_names(CS).


%% uniqvars_skip(List0, Var, List).
%%
%%  List is List0 with all initial occurrences of Var removed.
%%  Uses == when comparing Var.
%%
%%  mode uniqvars_skip(@list(term), @term, ?list(term)).
%%
uniqvars_skip([X|XS], V, Out) :-
    V == X,
    !,
    uniqvars_skip(XS, V, Out).
uniqvars_skip(In, _, In).
    

%% uniqvar_member(Pairs, Var, Name).
%%
%%  Var=Name is a member of Pairs.
%%  Uses == when comparing Var.
%%
%%  mode uniqvar_member(@list(compound), @term, ?term).
%%
uniqvar_member([X=N|Pairs], V, Name) :-
    X == V,
    !,
    Name = N.
uniqvar_member([_|Pairs], V, Name) :-
    uniqvar_member(Pairs, V, Name).
