#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

extern "C"
int
p_stat (char *file, int *mode, int *ino, int *dev, int *rdev,
        int *nlink, int *uid, int *gid, int *size,
        int *atime, int *mtime, int *ctime, int *blksize, int *blocks)
{
  struct stat buf;

  if (stat (file, &buf) == -1)
    {
      *mode = *ino = *dev = *rdev = *nlink = *uid = *gid = *size =
        *atime = *mtime = *ctime = *blksize = *blocks = 0;
      return errno;
    }
  else
    {
      *mode = buf.st_mode;
      *ino = buf.st_ino;
      *dev = buf.st_dev;
      *rdev = buf.st_rdev;
      *nlink = buf.st_nlink;
      *uid = buf.st_uid;
      *gid = buf.st_gid;
      *size = buf.st_size;
      *atime = buf.st_atime;
      *mtime = buf.st_mtime;
      *ctime = buf.st_ctime;
      *blksize = buf.st_blksize;
      *blocks = buf.st_blocks;
      return 0;
    }
}

extern "C"
int
p_unlink (char *file)
{
  if (unlink (file) == -1)
    return errno;
  else
    return 0;
}

extern "C"
int
p_mkdir (char *path, int mode)
{
  if (mkdir (path, mode) == -1)
    return errno;
  else
    return 0;
}

extern "C"
char *
p_error (int errno)
{
  char *msg = strerror (errno);

  if (errno == 0)
    return "No error";
  else if (msg == 0)
    return "Bad error";
  else
    return msg;
}

static int ctr_nextctr=0;  /* Next numbered counter to be allocated */
static int ctr_nctrs=0;    /* Number of counters available */
static int *ctr_ctrs=0;    /* Array of counter values */

extern "C"
int
ctr_alloc (int initval)
{
  if (ctr_nextctr >= ctr_nctrs)
    {
      if (ctr_nctrs == 0)
        ctr_nctrs = 16;
      else
        ctr_nctrs *= 2;
      ctr_ctrs = (int *)realloc (ctr_ctrs, ctr_nctrs*sizeof(int));
    }
  ctr_ctrs[ctr_nextctr] = initval;
  return ctr_nextctr++;
}

extern "C"
int
ctr_get (int ind)
{
  if (ind < ctr_nctrs)
    return ctr_ctrs[ind];
  else
    return -1;
}

extern "C"
void
ctr_set (int ind, int val)
{
  if (ind < ctr_nctrs)
    ctr_ctrs[ind] = val;
}

extern "C"
long
p_getpid (void)
{
  return (long)getpid();
}

extern "C"
long
p_opendir (char *filename, long *p_dirp_h, long *p_dirp_l)
{
  DIR *dirp = opendir(filename);
  
  if (dirp == 0)
    {
      *p_dirp_h = *p_dirp_l = 0;
      return errno;
    }
  else
  {
    *p_dirp_h = (long)dirp >> 16;
    *p_dirp_l = (long)dirp & 0xffff;
    return 0;
  }
}

extern "C"
long
p_readdir (long p_dirp_h, long p_dirp_l, long *inode, char **name)
{
  struct dirent *res = readdir((DIR *)((p_dirp_h<<16)|p_dirp_l));
  
  if (res == 0)
    {
      *name = "";
      return 1;
    }
  else
    {
      *inode = (long)res->d_ino;
      *name  = res->d_name;
      return 0;
    }
}

extern "C"
void
p_closedir (long p_dirp_h, long p_dirp_l)
{
  closedir ((DIR *)((p_dirp_h<<16)|p_dirp_l));
}
