%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: constrain.ql,v 1.20 2000/11/15 06:29:47 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% The procedures in this file keep track of not_free_in and
%% object variable distinctness constraints.
%% They are used by Ergo to avoid over-specialization of proofs.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../lib/ergo.h"


proof command
  accept_constraints
extends
  nfi_proofcmds
doc
  'This initialises the constraint tracking system.',nl,
  'Call it whenever you want to accept all',nl,
  'constraints on fixed proof variables',nl,
  'that are currently in the Qu-Prolog delay list.'
mode
  accept_constraints.

accept_constraints :-
    ip_lookup(fixedvars, Vars),
    retry_delays,
    collect_constraints(Vars, Dist, NFI, Special),
    ip_set(accepted_distinctions, Dist),
    ip_set(accepted_not_free_ins, NFI),
    ip_set(accepted_specials,     Special),
    ( Special = [] ->
        true
    ;
        msg(accept_constraints, 2, ['illegal constraints: ',q(Special)]),
        fail
    ).



%% command show_fixedvar_constraints.
%% command show_proofvar_constraints.
%%
%%  Print current constraints involving fixed/fixed&new proof variables.
%%  It uses writeR, so names will be given to otherwise unnamed variables.
%%  We do not fail out of this (to reclaim heap space) in case the
%%  caller wishes to keep such names.
%%
%%  mode show_fixedvar_constraints.
%%  mode show_proofvar_constraints.
%%
show_fixedvar_constraints :-
    fixedvar_constraints(OldDist,OldNFI,OldSpecial,
                         RecentDist,RecentNFI,RecentSpecial),
    writelist(['    Known constraints on fixed variables:',nl]),
    show_constraint_list(OldNFI),
    show_constraint_list(OldDist),
    show_constraint_list(OldSpecial),
    show_empty_constraints(OldNFI, OldDist, OldSpecial),
    writelist(['    Recent constraints on fixed variables:',nl]),
    show_constraint_list(RecentNFI),
    show_constraint_list(RecentDist),
    show_constraint_list(RecentSpecial),
    show_empty_constraints(RecentNFI, RecentDist, RecentSpecial),
    writelist([nl]).

show_proofvar_constraints :-
    proofvar_constraints(OldDist,OldNFI,OldSpecial,
                         RecentDist,RecentNFI,RecentSpecial,
                         NewDist,NewNFI,NewSpecial),
    writelist(['    Known constraints on fixed variables:',nl]),
    show_constraint_list(OldNFI),
    show_constraint_list(OldDist),
    show_constraint_list(OldSpecial),
    show_empty_constraints(OldNFI, OldDist, OldSpecial),
    writelist(['    Recent constraints on fixed variables:',nl]),
    show_constraint_list(RecentNFI),
    show_constraint_list(RecentDist),
    show_constraint_list(RecentSpecial),
    show_empty_constraints(RecentNFI, RecentDist, RecentSpecial),
    writelist(['    Constraints on new variables:',nl]),
    show_constraint_list(NewNFI),
    show_constraint_list(NewDist),
    show_constraint_list(NewSpecial),
    show_empty_constraints(NewNFI, NewDist, NewSpecial),
    writelist([nl]).


%% private show_constraint_list(List).
%%
%%  Prints out all the constraints in List.  It uses a recursive loop to
%%  do this, rather than a repeat-fail loop, so that the writes
%%  generate distinct names for distinct variables.
%%
%%  mode show_constraint_list(@list(compound)).
%%
show_constraint_list([]).
show_constraint_list([C|CS]) :-
    writelist([tab,code(C),nl]),
    show_constraint_list(CS).


%% private show_empty_constraints(List1,List2,List3).
%%
%%  Prints a '<none>' message iff List1, List2 and List3 are all [].
%%
show_empty_constraints([],[],[]) :-
    !,
    writelist([tab,'<none>',nl]).
show_empty_constraints(_,_,_).


%% fixedvar_constraints(OldDist,    OldNFI,    OldSpecial,
%%                      RecentDist, RecentNFI, RecentSpecial).
%% proofvar_constraints(OldDist,    OldNFI,    OldSpecial,
%%                      RecentDist, RecentNFI, RecentSpecial,
%%                      NewDist,    NewNFI,    NewSpecial).
%%
%%  These procedures return various sets of constraints that
%%  affect the current proof.
%%
%%  The Old* lists are fixedvar constraints that have been accepted.
%%  The Recent* lists are fixedvar constraints that have not been accepted.
%%  The New* lists are newvar constraints.
%%  The *Dist lists contain object variable distinctness information,
%%  in the form Obvar not_free_in ObVarList.
%%  The *NFI lists contain ObVar not_free_in Term constraints.
%%  The *Special lists contain all other kinds of constraints.
%%
%%  WARNING: these may give peculiar results if called when the fixedvars
%%           or newvars lists are not up-to-date (eg. when variables
%%           have been instantiated).
%%
%%  mode fixedvar_constraints(?list(compound),?list(compound),?list(compound),
%%                            ?list(compound),?list(compound),?list(compound)).
%%  mode proofvar_constraints(?list(compound),?list(compound),?list(compound),
%%                            ?list(compound),?list(compound),?list(compound),
%%                            ?list(compound),?list(compound),?list(compound)).
%%

%% TO DO: To fix the peculiar results in fixedvar_constraints and
%% proofvar_constraints when he fixedvars or newvars lists are not up-to-date
%% (eg. when variables have been instantiated),
%% we should perhaps check if it is being called within thawing_newvars,
%% and if so, collect_vars first?

fixedvar_constraints(OldDist,OldNFI,OldSpecial,
                     RecentDist,RecentNFI,RecentSpecial) :-
    retry_delays,
    ip_lookup(fixedvars, Vars),
    collect_constraints(Vars, Dist, NFI, Special),
    ip_lookup(accepted_distinctions, OldDist),
    ip_lookup(accepted_not_free_ins, OldNFI),
    ip_lookup(accepted_specials,     OldSpecial),
    diff_list(Dist,    OldDist,    RecentDist),
    diff_list(NFI,     OldNFI,     RecentNFI),
    diff_list(Special, OldSpecial, RecentSpecial).


proofvar_constraints(OldDist,OldNFI,OldSpecial,
                     RecentDist,RecentNFI,RecentSpecial,
                     NewDist,NewNFI,NewSpecial) :-
    ip_lookup(fixedvars, FixedVars),
    ip_lookup(newvars, NewVars),
    append(FixedVars, NewVars, Vars),
    retry_delays,  %% for speed, we don't want to do this twice!
    collect_constraints(FixedVars, FixedDist, FixedNFI, FixedSpecial),
    ip_lookup(accepted_distinctions, OldDist),
    ip_lookup(accepted_not_free_ins, OldNFI),
    ip_lookup(accepted_specials,     OldSpecial),
    diff_list(FixedDist,    OldDist,    RecentDist),
    diff_list(FixedNFI,     OldNFI,     RecentNFI),
    diff_list(FixedSpecial, OldSpecial, RecentSpecial),
    collect_constraints(Vars, AllDist, AllNFI, AllSpecial),
    diff_list(AllDist,    FixedDist,    NewDist),
    diff_list(AllNFI,     FixedNFI,     NewNFI),
    diff_list(AllSpecial, FixedSpecial, NewSpecial).


proof command
  constraints_accepted
extends
  nfi_proofcmds
doc
  'constraints_accepted/0 fails silently if any new',nl,
  'constraints that affect the fixed proof variables have been',nl,
  'added to the system since the last call to accept_constraints.'
mode
  constraints_accepted.

constraints_accepted :-
    \+ \+ fixedvar_constraints(_,_,_,[],[],[]).


proof command
  check_constraints_accepted
extends
  nfi_proofcmds
doc
  'check_constraints_accepted/0 generates an error if',nl,
  'some recent proof step has constrained the fixed proof variables.'
mode
  check_constraints_accepted.

check_constraints_accepted :-
    constraints_accepted,
    !.
check_constraints_accepted :-
    handle_error(['A recent proof step over-constrained the proof.']).


proof command
  ensure_constraints_accepted,
  ensure_constraints_accepted(Msg)
extends
  nfi_proofcmds
doc
  'ensure_constraints_accepted/0 succeeds iff there are no',nl,
  'recently added (that is, not yet accepted) constraints',nl,
  'on the fixed proof variables or if the user decides to',nl,
  'accept all the recent constraints (it displays Msg, shows',nl,
  'the accepted and recent constraints, then asks the user whether',nl,
  'those constraints should be accepted (the default answer is no).'
mode
  ensure_constraints_accepted,
  ensure_constraints_accepted(@writelist).

ensure_constraints_accepted :-
    ensure_constraints_accepted(
        [nl,
         'WARNING: A constraint has been added to the current theorem.',nl]).

ensure_constraints_accepted(_) :-
    constraints_accepted,
    !.
ensure_constraints_accepted(_) :-
%% TO DO: The following optimisation should be done some day.
%% It would be more efficient to merge this with the above
%% call to constraints_accepted, but this point is reached
%% infrequently, so I haven't bothered yet.
    fixedvar_constraints(_,_,_,_,_,L),
    L = [_|_],
    !,
    handle_error(['delayed unifications were introduced: ',q(L)]).
ensure_constraints_accepted(Msg) :-
    writelist(Msg),
    show_fixedvar_constraints,
    question('Accept all recent constraints?', no),
    accept_constraints.

