%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%%
%% CVS $Id: itrans.ql,v 1.42 2001/06/04 01:24:09 itp Exp $

#include "../lib/ergo.h"


%% itrans -- Incremental 'trans'.
%%
%%  The itrans command provide similar functionality to a trans/4
%%  (or h_trans or d_trans) command, but in a more flexible way,
%%  because the functionality of trans is broken into small
%%  pieces, as follows:
%%
%%  itrans_start                   -- begin a new itrans transformation
%%  in_itrans                      -- succeed iff an itrans is in progress
%%  itrans_input_name(?Name)       -- the input of the transformation
%%  itrans_rule_name(?Thy,?Name)   -- the primary rule of the transformation
%%  itrans_direction(?Dir)         -- the direction of the primary rule
%%  itrans_conditional(?Cond)      -- Cond=c if the rule is conditional,
%%                                    Cond=u otherwise.
%%  itrans_eval_body               -- execute the body of the primary rule
%%  itrans_kind(?Kind)             -- what kind of rule is the primary rule
%%  itrans_relation(?Rel)          -- the relation of the transformation
%%  itrans_condlist(?CondList)     -- the list of conditions
%%  itrans_stmt(?Stmt)             -- the statement of the primary rule.
%%  itrans_body(?Code)             -- the side condition of the primary rule.
%%  itrans_input(?Input)           -- the input term of the transformation
%%  itrans_input_shape(?Shape)     -- the shape of the input term
%%  itrans_output_shape(?Shape)    -- the shape of the output term
%%  itrans_rule_relation(?Rel)     -- the relation of the primary rule
%%  itrans_output(?Output)         -- the output term of the transformation
%%  itrans_finish(?Dest)           -- complete the transformation and apply
%%                                     it to the current proof.
%%  itrans_dischargelist(?DList)   -- the completed list of rule names used
%%                                     to discharge the conditions (can
%%                                     only be called after itrans_satisfy).
%%
%%
%%  The minimal calls that are required in an itrans operation are:
%%
%%      1. itrans_start first, then
%%      2. itrans_rule_name, itrans_direction and itrans_conditional
%%         (in any order, but all three must be called),
%%      3. itrans_satisfy/2, to discharge all conditions.
%%      4. finally, itrans_finish.
%%  plus: 
%%      - itrans_input_name must be called exactly once sometime after
%%        itrans_start and before itrans_finish.
%%
%%  However, if the Dest name passed to itrans_finish is null, then
%%  step three is optional, and itrans_eval_body is not done
%%  automatically within itrans_finish.
%%
%%  To extend this minimal set of calls, any of the above procedures
%%  may be called between itrans_start and itrans_finish, except for:
%%
%%    itrans_start         -- don't call this unless you want to 
%%                            start a new transformation
%%    itrans_eval_body     -- can only be called after itrans_rule_name.
%%    itrans_condlist      -- can only be called after itrans_conditional
%%                          (and after itrans_rule_name if conditional=true).
%%    itrans_dischargelist -- can only be called after itrans_satisfy.
%%    itrans_body          -- can only be called after itrans_finish.
%%    itrans_*_shape       -- can only be called after itrans_rule_name,
%%                          itrans_conditional and itrans_direction.
%%
%%  Most of the procedures can also be called after itrans_finish,
%%  but only until some other itrans transformation begins, or
%%  some other trans-like or open/close command is executed.
%%  
%%  If another itrans operation is started before itrans_finish succeeds,
%%  the first (incomplete) itrans operation is cancelled.  However, its
%%  state will be restored upon backtracking out of the more recent
%%  itrans operation.  Similarly, an itrans operation can be cancelled
%%  at any time by calling the itrans_abort procedure, (but the itrans
%%  state will still be restored upon backtracking over the itrans_abort).
%%  Open_hyp, trans, h_trans, d_trans and similar context changing
%%  procedures automatically invalidate any itrans in progress
%%  (but, again, the itrans state is restored upon backtracking).
%%
%%  Most of the above procedures are highly non-deterministic, so it is
%%  the caller's responsibility to restrict backtracking as desired,
%%  by instantiating the arguments to the above procedures appropriately
%%  and/or inserting cuts after some of the procedures.
%%
%%  It is also the caller's responsibility to ensure that these calls
%%  take place with newvars thawed, if that is desired.  Typically 
%%  this is done by putting all calls to the above procedures within
%%  a single thawing_new_vars/1 wrapper.  If newvars are not thawed,
%%  the transformations will still be sound, but instantiations of
%%  newvars will not be allowed (the newvars will still be frozen),
%%  so some rule applications may be rejected.
%%
%%
%%  Future Enhancements (Keith):
%%  ============================
%%  The most obvious future enhancement would be to provide for
%%  nested transformations.
%% 
%%  With such an arrangement, this trans command would have the 
%%  functional capabilities of all possible compiled rules.  This would be an
%%  obvious device to use for rerunning altered compiled rules, especially
%%  if the rerun information for the two devices shared a common structure.    
%%
%%  In creating such a facility, it would be necessary to also modify the
%%  discharge capacity to avoid disappointing attempts at constructing
%%  circular discharges.


%% Internal data structures.
%% =========================
%%
%%  All communication between the various routines takes place using
%%  the following variables.  These variables are represented using
%%  several implicit parameters and can be accessed using the itr/3
%%  procedure.
%%
%%    vars          --   a list of terms that have come from the
%%                       database, and may contain variables not already
%%                       present in the proof.  This may include
%%                       variables from the input, output and/or some
%%                       conditions of the primary rule.  Basically, if
%%                       the caller wants to look at any part of the
%%                       primary rule before we know how that part is going
%%                       to be discharged, we must record that part, in
%%                       order to do legal-term-checking at the end.
%%
%%    input_name    -- Describes where the input of the transformation
%%                     is to come from.  Possible values are:
%%                         focus       -- the focus of the current window.
%%                         true        -- the implicit hypothesis: true.
%%                         null        -- implies that OutName = null.
%%                         hyp:Num     -- hypothesis Num of current window.
%%                         hyp:Term    -- hypothesis that matches Term
%%                         proof:Ass   -- assumption Ass of current proof.
%%                         Thy:Name    -- database rule.
%%
%%    input_term    -- The term that is the input of the transformation.
%%
%%    input_flag    -- Instantiated to true when we no longer need to
%%                     worry about collecting the input term's variables.
%%
%%    output_name   -- This is not instantiated until itrans_finish.
%%                     It describes where the output of the transformation
%%                     is to go.  Possible values are:
%%                         focus       -- InName must be focus too.
%%                         null        -- output will be discarded.
%%                         browse(Depth, Result)  
%%                                  -- rule discarded and unify restricted to
%%				    -- Depth at which Result is exceeded or not
%%                         hyp:Num     -- a newly generated hypothesis.
%%                         true        -- iff InName = hyp:_ and
%%                                        hyp:_ is a conjecture.
%%
%%    output_term   -- The term that is the output of the transformation.
%%
%%    output_flag   -- Instantiated to true when we no longer need to
%%                     worry about collecting the output term's variables.
%%
%%    relation      -- The relation of the transformation.
%%                     (eg. for a focus transformation, this would be
%%                      the relation of the current window).
%%
%%    kind          -- kind of rule: axiom/theorem/compiled_rule etc.
%%    theory        -- the theory of primary rule (or 'hyp' or 'proof')
%%    name          -- canonical name/number of primary rule.
%%    stmt          -- statement of the primary rule.
%%    shape         -- shape of the primary rule.
%%    body          -- body of the primary rule.
%%    body_flag     -- flag to show whether Body has been executed.
%%                     Initially a variable, then instantiated (to true)
%%                     when Body has been executed.
%%    conditional   -- 'c' if the primary rule is conditional, else 'u'.
%%    condlist      -- list of all conditions to be discharged.  This is
%%                     instantiated by itrans_condlist or itrans_satisfy.
%%    condlist_flag -- Instantiated to true when we no longer need 
%%                     collect the variables in the conditions.
%%    concl         -- conclusion of rule (Stmt minus Conditions)
%%    direction     -- direction of primary rule:  f/r/t/rt.
%%    rule_relation -- relation of primary rule.
%%                     WARNING: this may be in an unsimplified rev(_) form,
%%                              such as rev(=).
%%    dischargelist -- list of the rule names used to discharge conditions.
%%
%%    condbody      -- the collected bodies of the conditions.
%%    cdep          -- collected 'conjecture' dependencies of the
%%                     primary rule and all condition discharge rules.
%%    udep          -- collected 'use' dependencies of the
%%                     primary rule and all condition discharge rules.
%%    frozen        -- a true/false flag that determines whether variables
%%                     in the primary rule should be frozen before being
%%                     unified with any other term.
%%    stronger_flag -- Instantiated to true after we have checked
%%                     is_stronger_than(<rule_relation>, <relation>).


@define(itr_num_fields, 23).

%% itr(Component, Value, IT).
%%
%%  Value is the current value of Component in IT,
%%  where IT is the structure returned by itr_valid/working.
%%
%%  mode itr(+atom, ?term, +compound).
%%
@define(itr(input_name,    V, IT), arg(1,IT,V)).
@define(itr(input_term,    V, IT), arg(2,IT,V)).
@define(itr(input_flag,    V, IT), arg(3,IT,V)).
@define(itr(output_name,   V, IT), arg(4,IT,V)).
@define(itr(output_term,   V, IT), arg(5,IT,V)).
@define(itr(output_flag,   V, IT), arg(6,IT,V)).
@define(itr(relation,      V, IT), arg(7,IT,V)).
@define(itr(theory,        V, IT), arg(8,IT,V)).
@define(itr(name,          V, IT), arg(9,IT,V)).
@define(itr(kind,          V, IT), arg(10,IT,V)).
@define(itr(stmt,          V, IT), arg(11,IT,V)).
@define(itr(body,          V, IT), arg(12,IT,V)).
@define(itr(body_flag,     V, IT), arg(13,IT,V)).
@define(itr(conditional,   V, IT), arg(14,IT,V)).
@define(itr(condlist,      V, IT), arg(15,IT,V)).
@define(itr(condlist_flag, V, IT), arg(16,IT,V)).
@define(itr(concl,         V, IT), arg(17,IT,V)).
@define(itr(direction,     V, IT), arg(18,IT,V)).
@define(itr(rule_relation, V, IT), arg(19,IT,V)).
@define(itr(dischargelist, V, IT), arg(20,IT,V)).
@define(itr(frozen,        V, IT), arg(21,IT,V)).
@define(itr(stronger_flag, V, IT), arg(22,IT,V)).
@define(itr(shape,         V, IT), arg(23,IT,V)).
@define(itr(condbody,      V, _), ip_lookup('itrans/condbody', V)).
@define(itr(vars,          V, _), ip_lookup('itrans/vars', V)).
@define(itr(cdep,          V, _), ip_lookup('itrans/cdep', V)).
@define(itr(udep,          V, _), ip_lookup('itrans/udep', V)).

itr(Name, _, _) :-
    nonvar(Name),
    handle_error(program,['illegal call: ',q(itr(Name,...))]).


%% itr_valid(IT).
%% itr_working(IT).
%%
%%  These both get itrans data from the implicit parameters.
%%
%%  itr_valid checks that an itrans operation is in progress,
%%  or has been recently completed and no other trans-like operation
%%  has been started since the itrans operation was completed.
%%
%%  itr_working checks that an itrans operation is in progress,
%%  and has not yet been completed.
%%
%%  Note that the 'itrans/info' implicit parameter is set to
%%  'false' when there is no valid itrans operation.
%%
%%  mode itr_valid(-compound).
%%  mode itr_working(-compound).
%%
@define(itr_valid(IT),
        ( ip_lookup('itrans/info', IT),
          compound(IT))).

@define(itr_working(IT),
        ( ip_lookup('itrans/info', IT),
          compound(IT),
          @@itr(output_name, OutName, IT),
          var(OutName))).


%% For debugging purposes only.
itr_write :-
    @@itr_valid(IT),
    @@itr(input_name,    V1, IT),writelist(['input_name:     ',q(V1),nl]),
    @@itr(input_term,    V2, IT),writelist(['input_term:     ',q(V2),nl]),
    @@itr(input_flag,    V3, IT),writelist(['input_flag:     ',q(V3),nl]),
    @@itr(output_name,   V4, IT),writelist(['output_name:    ',q(V4),nl]),
    @@itr(output_term,   V5, IT),writelist(['output_term:    ',q(V5),nl]),
    @@itr(output_flag,   V6, IT),writelist(['output_flag:    ',q(V6),nl]),
    @@itr(relation,      V7, IT),writelist(['relation:       ',q(V7),nl]),
    @@itr(theory,        V8, IT),writelist(['theory:         ',q(V8),nl]),
    @@itr(name,          V9, IT),writelist(['name:           ',q(V9),nl]),
    @@itr(kind,          V10, IT),writelist(['kind:           ',q(V10),nl]),
    @@itr(stmt,          V11, IT),writelist(['stmt:           ',q(V11),nl]),
    @@itr(body,          V12, IT),writelist(['body:           ',q(V12),nl]),
    @@itr(body_flag,     V13, IT),writelist(['body_flag:      ',q(V13),nl]),
    @@itr(conditional,   V14, IT),writelist(['conditional:    ',q(V14),nl]),
    @@itr(condlist,      V15, IT),writelist(['condlist:       ',q(V15),nl]),
    @@itr(condlist_flag, V16, IT),writelist(['condlist_flag:  ',q(V16),nl]),
    @@itr(concl,         V17, IT),writelist(['concl:          ',q(V17),nl]),
    @@itr(direction,     V18, IT),writelist(['direction:      ',q(V18),nl]),
    @@itr(rule_relation, V19, IT),writelist(['rule_relation:  ',q(V19),nl]),
    @@itr(dischargelist, V20, IT),writelist(['dischargelist:  ',q(V20),nl]),
    @@itr(frozen,        V21, IT),writelist(['frozen:         ',q(V21),nl]),
    @@itr(stronger_flag, V22, IT),writelist(['stronger_flag:  ',q(V22),nl]),
    @@itr(shape,         V23, IT),writelist(['shape:          ',q(V23),nl]),
    @@itr(condbody,      V30,  _),writelist(['condbody:       ',q(V30),nl]),
    @@itr(vars,          V31,  _),writelist(['vars:           ',q(V31),nl]),
    @@itr(cdep,          V32,  _),writelist(['cdep:           ',q(V32),nl]),
    @@itr(udep,          V33,  _),writelist(['udep:           ',q(V33),nl]).


%% The following procedures are used to access/update the *changing*
%% portion of the itrans state.

itr_add_condbody(true) :-
    !.
itr_add_condbody(Body) :-
    ip_lookup('itrans/condbody', Body0),
    ip_set('itrans/condbody', (Body0,Body)).


itr_add_term(Term, Flag) :-
    (Flag == true ->
        true
    ;
        ip_lookup('itrans/vars', Vars0),
        ip_set('itrans/vars', [Term|Vars0]),
        Flag = true
    ).

itr_add_termlist(TermList, Flag) :-
    (Flag == true ->
        true
    ;
        ip_lookup('itrans/vars', Vars0),
        append(TermList, Vars0, Terms),
        ip_set('itrans/vars', Terms),
        Flag = true
    ).


itr_add_cdep(CDep) :-
    ip_lookup('itrans/cdep', CDep0),
    cdep_union(CDep, CDep0, NewCDep),
    ip_set('itrans/cdep', NewCDep).


itr_add_udep(UDep) :-
    ip_lookup('itrans/udep', UDep0),
    udep_union(UDep, UDep0, NewUDep),
    ip_set('itrans/udep', NewUDep).



%% itrans_start
%%
%%  Starts a new itrans operation.
%%
%%  mode itrans_start.
%%
itrans_start :-
    functor(IT, itinfo, @@itr_num_fields),
    ip_set('itrans/info', IT),  %% Overrides any current itrans data.
    ip_set('itrans/condbody', true),
    ip_set('itrans/vars', []),
    cdep_empty(CDep),
    udep_empty(UDep),
    ip_set('itrans/cdep', CDep),
    ip_set('itrans/udep', UDep).


%% itrans_abort
%%
%%  A predicate to be called by operations such as open_pos, which
%%  would invalidate an in-progress transformation.
%%
%%  mode itrans_abort.
%%
itrans_abort :-
    ip_set('itrans/info', false).



%% in_itrans
%%
%%  Succeeds iff an itrans is currently in progress (and is unfinished).
%%
%%  mode in_itrans.
%%
in_itrans :-
    @@itr_working(_).



%% itrans_input_name(+Name).
%%
%%  This sets 'input_name' and (maybe) 'input_term' and 'relation',
%%  plus sets some flags to save work for certain kinds of inputs.
%%
%%  mode itrans_input_name(+compound).
%%
itrans_input_name(focus) :-
    !,
    @@itr_working(IT),
    @@itr(input_name, focus, IT),
    @@itr(input_term, Input, IT),
    @@itr(input_flag, true,  IT),
    is_focus(Input).
itrans_input_name(null) :-
    !, 
    @@itr_working(IT),
    @@itr(input_name,    null, IT),
    @@itr(input_flag,    true, IT),
    @@itr(output_flag,   true, IT),
    @@itr(condlist_flag, true, IT).
itrans_input_name(true) :-
    !,              
    @@itr_working(IT),
    @@itr(input_name,    ergo:true, IT),
    @@itr(input_term,    ERGO_TRUE, IT),
    @@itr(input_flag,    true,      IT).
itrans_input_name(Hyp) :-
    is_hypname(Hyp),
    !,
    @@itr_working(IT),
    @@itr(input_name,    hyp:Num, IT),
    @@itr(input_term,    Input,   IT),
    @@itr(input_flag,    true,    IT),
    is_hyp(Hyp, Num, Input).
itrans_input_name(conj:Num) :-
    var(Num),
    !,
    @@itr_working(IT),
    @@itr(input_term,    Input,    IT),
    @@itr(input_flag,    IFlag,    IT),
    itr_add_term(Input,IFlag),
    add_conj(Input, Num, CDep, UDep),
    @@itr(input_name, hyp:Num, IT),
    itr_add_cdep(CDep),
    itr_add_udep(UDep),
    !.
%% TO DO: extend itrans_input_name to handle compiled rules.
%%       (currently only handles axiom/theorem/lemma).
%% Ray's last comment: These are obsolete too.
itrans_input_name(Rule) :-
    decode_rule(Rule, Thy, Name),
    @@itr_working(IT),
    @@itr(input_name,    Thy:Name, IT),
    @@itr(input_term,    Input,    IT),
    find_rule2(Thy, Name, Input, [], Body),
    !,  %% we don't allow backtracking over rules -- too slow.
    itr_add_condbody(Body).



%% itrans_kind(?Kind).
%%
%%  mode itrans_kind(?term).
%%
itrans_kind(Kind) :-
    @@itr_valid(IT),
    @@itr(kind, Term, IT),
    Kind = Term.


%% itrans_frozen(Flag).
%%
%%  Flag=true means that all variables in the selected rule are
%%  frozen before they are unified with any user-supplied terms.
%%  By default, Flag=false.
%%
%%  This procedure must be called before itrans_rule_name in order
%%  to set the flag to true, since it is false by default and
%%  it cannot be changed after itrans_rule_name succeeds.
%%
%%  If it is called after the frozen Flag has already been set,
%%  (e.g., after itrans_rule_name), it returns the value of the Flag.
%%
%%  mode itrans_frozen(?atom).
%%
itrans_frozen(Flag) :-
    @@itr_valid(IT),
    @@itr(frozen, Flag, IT),
    (
        Flag = false
    ;
        Flag = true
    ),
    !.


%% itrans_rule_name(Thy,Name).
%%
%%  Selects the primary rule of the transformation, using
%%  Thy and Name to restrict the search if they are instantiated.
%%
%%  If itrans_frozen(true) was called before this procedure, then
%%  all variables in the selected rule are frozen before they are
%%  unified with any user-supplied terms (input, output etc.).
%%
%%  Subsequent calls return the theory and name of the selected rule.
%%
%%  mode itrans_rule_name(?atom, ?gcomp).
%%  mode itrans_rule_name(?atom, ?gcomp, @atom).
%%
itrans_rule_name(Thy,Name) :-
    @@itr_valid(IT),
    @@itr(theory, Thy,   IT),
    @@itr(name,   Name2, IT),
    (var(Name2) ->
        %% select a rule.
        @@itr(kind,        Kind,        IT),
        @@itr(stmt,        Stmt,        IT),
        @@itr(shape,       Shape,       IT),
        @@itr(body,        Body,        IT),
        (@@itr(frozen, false, IT) ->
            itr_find_rule(Kind, Thy, Name, Name1,
                          Stmt, Shape, Body, CDep, UDep)
        ;
           @@itr(frozen, true, IT),
           itr_find_rule(Kind, Thy, Name, Name1,
                         Stmt2, Shape, Body, CDep, UDep),
           freeze_vars(Stmt2),
           Stmt = Stmt2
        ),
        Name2 = Name1,
        itr_add_cdep(CDep),
        itr_add_udep(UDep)
    ;
        %% return the name of the previously selected rule.
        decode_comp_rule(Name2, Name, _Trace)
    ).


%% private itr_find_rule(Kind, Theory, Name0, Name,
%%                       Stmt, Shape, Body, CDep, UDep).
%%
%%  Auxillary procedure for itrans_rule_name.
%%  This is the itrans version of get_th_or_hyp.
%%
%%  mode itr_find_rule(?atom, ?theory, ?nonvar, ?nonvar,
%%                     ?term, ?nonvar, ?nonvar, -cdep_dms, -udep_dms).
%%
itr_find_rule(hypothesis, hyp, Name0, Name, Stmt, Shape, true, CDep, UDep) :-
    get_hyp(hyp:Name0, HRec),
    hrec_numterm(HRec, Name, Stmt),
    rule_shape(Stmt, Shape),
    hrec_cdep(HRec, CDep),
    hrec_udep(HRec, UDep).
%itr_find_rule(assumption, proof, Name0, Name0, Stmt, true, CDep, UDep) :-
%    is_ass(Name0, Stmt, UDep),
%    cdep_empty(CDep).
%itr_find_rule(compiled_rule, Thy, Name0, Name, Stmt, Body, CDep, UDep) :-
%    ancestor(Thy),
%    get_compiled_rule(Thy, Name0, _, Concl),
%    %%
%    %% CDep and UDep will be discharged when the compiled rule 
%    %% is executed, which happens when itrans_eval_body is called.
%    %%
%    cdep_create(CDep),
%    udep_create(UDep),
%    Name = cr(Name0,Trace),
%    Body = eval_compiled_rule(Thy,Name0,_,Concl,Stmt,CDep,UDep,Trace).
itr_find_rule(Kind0, Thy, Name0, Name0, Stmt, Shape, Body, CDep, UDep) :-
    cdep_empty(CDep),
    udep_empty(UDep),
    ancestor(Thy),
    query_theory(rule(Name0, Kind, Shape, Stmt, Body), Thy),
    functor(Kind, Kind0, _).
%% TO DO: Do the check functor(Kind, Kind0, _) earlier in itr_find_rule
%% if atom(Kind0).


%% itrans_direction(Dir).
%%
%%  The first call performs the 'choose_direction' component of the
%%  transformation (non-deterministically, if Dir is not instantiated).
%%  Subsequent calls unify Dir with the chosen direction.
%%
%%  mode itrans_direction(?atom).
%%
itrans_direction(Dir) :-
    @@itr_valid(IT),
    @@itr(direction, Direction, IT),
    (var(Direction) ->
        @@itr(concl,          Concl,     IT),
        @@itr(input_term,     Input,     IT),
        @@itr(rule_relation,  RuleRel,   IT),
        @@itr(output_term,    Output,    IT),
        choose_direction(Dir, Concl, Input, RuleRel, Output)
    ;
        true
    ),
    Dir = Direction.


%% itrans_conditional(Cond).
%%
%%  Performs the 'choose_conditional' component of the transformation
%%  (non-deterministically, if Cond is not instantiated).
%%  Cond=c for conditional rules, u for unconditional rules.
%%
%%  mode itrans_conditional(?atom).
%%
itrans_conditional(Cond) :-
    @@itr_valid(IT),
    @@itr(conditional, Cond,  IT),
    @@itr(stmt,        Stmt,  IT),
    @@itr(concl,       Concl, IT),
    %% a variant of choose_conditional/4 that sets Cond to c or u.
    itr_choose_conditional(Cond, Stmt, Concl),
    debug(itrans,['conditional=',q(Cond)]).

itr_choose_conditional(c, ERGO_IMPLIES(_,Concl), Concl).
itr_choose_conditional(u, Concl, Concl).

%% itrans_eval_body
%%
%%  This evaluates the Prolog side-condition of the primary rule.
%%  It can only be called after itrans_rule_name.
%%
%%  mode itrans_eval_body.
%%
itrans_eval_body :-
    @@itr_working(IT),
    @@itr(body,       Body,  IT),
    @@itr(body_flag,  Flag,  IT),
    PRE(itrans_eval_body, nonvar(Body)), 
    (Flag == true ->
        true
    ;
        itr_eval(Body),
        Flag = true
    ).

itr_eval(true) :-
    !.
itr_eval(Code) :-
    call_predicate(Code).



%% itrans_satisfy(RuleList, ConjList).
%%
%%  Satisfies all conditions (if any) of the primary rule.
%%
%%  The i'th element of RuleList specifies how the i'th condition
%%  should be discharged.  A special notation, conj:N, means that
%%  it should be discharged by generating a conjecture (N is instantiated
%%  to the number of the resulting conjecture.
%%
%%  However, ConjList can be used to place additional constraints
%%  upon the conjectures that can be generated.  ConjList is the
%%  list of the terms that were conjectured.  For instance, if
%%  ConjList=[], no conditions can be discharged by adding conjectures.
%%  If ConjList=[T1,T2], then exactly two conditions must generate
%%  conjectures, and T1 is unified with the statement of the first
%%  conjecture and T2 is unified with the statement of the second
%%  conjecture.  Note that the first conj:N entry in RuleList gives
%%  the number N of the T1 conjecture.
%%
%%  mode itrans_satisfy(?list(rulename), ?list(thy(term))).
%%
itrans_satisfy(Rules, Conjs) :-
    @@itr_working(IT),
    @@itr(conditional, Cond, IT),
    @@itr(body,        Body, IT),
    @@itr(direction,   Dir,  IT),
    nonvar(Cond),
    nonvar(Body),
    nonvar(Dir),
    !,
    itr_check_relations(IT),
    itr_calculate_condlist(Cond, IT, CondList),
    debug(itrans,['satisfy condlist=',q(CondList),nl,
                  'conjs=',q(Conjs),nl,
                  'rules=',q(Rules)]),
    @@itr(dischargelist, DisList, IT),
    itr_satisfy(CondList,Conjs,Rules,DisList),
    debug(itrans,['satisfy condlist=',q(CondList),nl,
                  'conjs=',q(Conjs),nl,
                  'rules=',q(Rules),nl,
                  'dislist=',q(DisList)]).
itrans_satisfy(_, _) :-
    handle_error(['itrans_satisfy called before itrans_rule_name,',nl,
                  'itrans_conditional or itrans_direction.']).


itr_calculate_condlist(u, _, []).
itr_calculate_condlist(c, IT, CondList) :-
    @@itr(stmt, ERGO_IMPLIES(Conds, _), IT),
    @@itr(condlist, CondList, IT),
    CondList = [_|_], % ensure there is at least one condition.
    conjunction_to_list(Conds, CondList),
    !,
    debug(itrans,['Condlist=',q(CondList)]).


%% private itr_satisfy(CondList,ConjectureList,UserRuleList,FinalRuleList).
%%
%%  mode itr_satisfy(+list(thy(term)), ?list(thy(term)),
%%                   ?list(rulename), ?list(fullrulename)).
%%
itr_satisfy([],[],[],[]).
itr_satisfy([Cond|CS],Conjs,[Rule0|RS0],[Rule|RS]) :-
    (
        get_th_or_hyp(Rule0, Rule, Cond, CBody, CDep, UDep),
        ConjTail = Conjs
    ;
        Rule0 = conj:No,   %% means add a new conjecture.
        Conjs = [Cond|ConjTail],
        itr_add_term(Cond, _),
        add_conj(Cond, No, CDep, UDep),
        Rule = hyp:No,
        CBody = true
    ),
    !,
    debug(itrans,['satisfied condition ',q(Cond),' using ',q(Rule0)]),
    itr_add_condbody(CBody),
    itr_add_cdep(CDep),
    itr_add_udep(UDep),
    itr_satisfy(CS,ConjTail,RS0,RS).



%% itrans_relation(Rel).
%%
%%  Unifies Rel with the relation of the current transformation.
%%  It is itrans_finish that checks that Rel is a preorder.
%%
%%  mode itrans_relation(?gcomp).
%%
itrans_relation(Rel) :-
    @@itr_valid(IT),
    @@itr(relation, Rel, IT).


%% itrans_condlist(CondList).
%%
%%  mode itrans_condlist(?list(term)).
%%
itrans_condlist(CondList) :-
    @@itr_valid(IT),
    @@itr(conditional, Cond, IT),
    @@itr(body,        Body, IT),
    @@itr(direction,   Dir,  IT),
    PRE(itrans_condlist, (nonvar(Cond), nonvar(Body), nonvar(Dir))),
    itr_check_relations(IT),
    itr_calculate_condlist(Cond, IT, CL),
    @@itr(condlist_flag, Flag, IT),
    itr_add_termlist(CL, Flag),
    CondList = CL.


%% itrans_stmt(Stmt).
%%
%%  mode itrans_stmt(?thy(term)).
%%
itrans_stmt(Stmt) :-
    @@itr_valid(IT),
    @@itr(stmt, Term, IT),
    @@itr(input_flag, Flag, IT),
    @@itr(output_flag, Flag, IT),
    @@itr(condlist_flag, Flag, IT), 
    itr_add_term(Term, Flag),
%    writeq(2,itrans_stmt(Stmt)),
%    writeq(2,itr_add_term(Term, Flag)),
    Stmt = Term. 


%% itrans_conclusion(Concl).
%%
%%  mode itrans_conclusion(?thy(term)).
%%
itrans_conclusion(Concl) :-
    @@itr_valid(IT),
    @@itr(concl, Concl, IT).



%% itrans_input(Input).
%%
%%  mode itrans_input(?term).
%%
itrans_input(In) :-
    @@itr_valid(IT),
    @@itr(input_term, Term, IT),
    @@itr(input_flag, Flag, IT),
    itr_add_term(Term, Flag),
    In = Term.


%% itrans_input_shape(Shape).
%%
%%  mode itrans_input_shape(?nonvar).
%%
itrans_input_shape(Shape) :-
    @@itr_valid(IT),
    @@itr(body, Body, IT),
    nonvar(Body),  %% check that thy:rule is determined!
    @@itr(shape, StmtShape, IT),
    @@itr(conditional, Cond, IT),
    @@itr(direction, Dirn, IT),
    atom(Cond),
    atom(Dirn),
    get_rule_shape(Cond, Dirn, _, StmtShape, Shape).


%% itrans_rule_relation(Rel).
%%
%%  mode itrans_rule_relation(?term).
%%
itrans_rule_relation(Rel) :-
    @@itr_valid(IT),
    @@itr(rule_relation, Rel, IT).


%% itrans_output(Output).
%%
%%  mode itrans_output(?term).
%%
itrans_output(Out) :-
    @@itr_valid(IT),
    @@itr(output_term, Term, IT),
    @@itr(output_flag, Flag, IT),
    itr_add_term(Term, Flag),
    Out = Term.


%% itrans_output_shape(Shape).
%%
%%  mode itrans_output_shape(?nonvar).
%%
itrans_output_shape(Shape) :-
    @@itr_valid(IT),
    @@itr(body, Body, IT),
    nonvar(Body),  %% check that thy:rule is determined!
    @@itr(shape, StmtShape, IT),
    @@itr(conditional, Cond, IT),
    @@itr(direction, Dirn, IT),
    atom(Cond),
    atom(Dirn),
    get_rule_shape(Cond, _, Dirn, StmtShape, Shape).

%% itrans_finish(OutName).
%%
%%  itrans_finish attempts to apply the transformation to some part
%%  of the current proof, as determined by OutName and/or the way
%%  that itrans_input_name was called.  The possible values of
%%  OutName are:
%%
%%     'focus'  -- the focus has been transformed.
%%     'true'   -- the input conjecture has been discharged.
%%     hyp:Num  -- a new hypothesis has been added.
%%     null     -- the proof is unchanged.
%%     browse(Depth, Result)  
%%              -- rule discarded and unify restricted to
%%	        -- Depth at which Result is exceeded or not
%%
%%  All conditions must have been discharged before calling itrans_finish.
%%
%%  mode itrans_finish(?compound).
%% 
itrans_finish(OName1) :-
    @@itr_working(IT),
    @@itr(input_name,      IName,  IT),
    @@itr(direction,       Dir,    IT),
    nonvar(IName),
    nonvar(Dir),
    !,
    itr_check_complete(OName1, IT),
    itr_check_relations(IT),
    @@itr(condbody, CBody, IT),
    debug(itrans, ['checking condbodies: ', q(itr_eval(CBody))]),
    itr_eval(CBody),
    debug(itrans, ['checking incomplete_retry_delays']),
    (compound(OName1), OName1 = browse(Depth, Res) ->
        incomplete_retry_delays(Depth, Res),
        OName = null
    ;
        OName = OName1,
        incomplete_retry_delays
    ),
    !,
    debug(itrans, ['all checks completed successfully.']),
    %% THE TRANSFORMATION IS NOW COMPLETE, SO WE APPLY IT.
    @@itr(output_term,   Output,  IT),
    @@itr(output_name,   OName,   IT),
    @@itr(relation,      Rel,     IT),
    (OName == null ->
        true
    ;
        %% Add the variables to the proof.
        @@itr(output_flag,   OFlag,   IT),
        itr_add_term(Output, OFlag),
        @@itr(vars,          Vars,    IT),
        thawing_new_vars(ip_set(newvars, Vars))
    ),
    @@itr(theory,    Theory,  IT),
    @@itr(name,      RName,   IT),
    debug(itrans, ['calling ',q(itr_finish(OName,IName,'...'))]),
    itr_finish(OName, IName, Theory:RName, Output, Rel, IT).
itrans_finish(OName) :-
    @@itr_working(IT),
    @@itr(input_name, IName, IT),
    var(IName),
    !,
    handle_error([q(itrans_finish(OName)),
                  ' called before itrans_input_name.']).
itrans_finish(OName) :-
    @@itr_working(IT),
    @@itr(direction, Dir, IT),
    nonvar(Dir),
    !,
    handle_error([q(itrans_finish(OName)),
                  ' called before itrans_direction.']).
itrans_finish(OName) :-
    handle_error([q(itrans_finish(OName)),
                  ' called when no current itrans.']).


%% itr_check_complete(OutName, IT).
%%
%%  Checks that conditions have been discharged and the primary
%%  rule body has been called.  However, if OutName=null, or browse(_,_) these
%%  are optional.
%%
itr_check_complete(null, _) :-
    !.
itr_check_complete(browse(_,_), _) :-
    !.
itr_check_complete(_, IT) :-
    %% check that all conditions are discharged.
    %% (this also implies that itrans_rule_name and 
    %%  itrans_conditional have both been called).
    @@itr(dischargelist,   DList,  IT),
    list(DList),
    !,
    debug(itrans, ['[re]checking rule body: ', q(itrans_eval_body)]),
    itrans_eval_body.
itr_check_complete(_, _) :-
    handle_error(['itrans_finish called with undischarged conditions.']).



%% private itr_check_relations(IT).
%%
%%  This checks that the rule relation is instantiated and is at least
%%  as strong as relation of the whole itrans operation.
%%  It should only be called after itrans_{rule_name,direction,conditional}
%%  have all been called.
%%
itr_check_relations(IT) :-
    @@itr(stronger_flag, Flag, IT),
    atom(Flag),
    !.
itr_check_relations(IT) :-
    @@itr(relation, Rel, IT),
    @@itr(rule_relation, RuleRel, IT),
    debug(itrans, [q(check_relations(RuleRel, Rel))]),
    is_ground_relation(RuleRel),
    ( var(Rel) ->
        simplify_relation(RuleRel, Rel),
        is_preorder(Rel)
    ;
        is_stronger_than(RuleRel, Rel)
    ),
    @@itr(stronger_flag, true, IT),
    debug(itrans, ['check_relations ends with Rel = ',q(Rel)]).


%% private itr_finish(OutputName, InputName, Rule, Output, Reln, IT).
%%
%%  The Input-Output-specific parts of itrans_finish.
%%
%%  mode itr_finish(?compound, +compound, +compound, +term, +gcomp, +compound).
%%
itr_finish(focus, focus, Rule, Output, Rel, IT) :-
    !,
    is_reln(WindowReln),
    is_stronger_than(Rel, WindowReln),
    @@itr(cdep,          CD,    IT),
    @@itr(udep,          UD,    IT),
    @@itr(direction,     Dir,   IT),
    @@itr(dischargelist, DList, IT),
    udep_simplify(UD, UDep),
    add_proof_step(Output, tr(Rule,Dir,Output,DList), CD, UDep).
itr_finish(true, hyp:Num, Rule, ERGO_TRUE, Rel, IT) :-
    !,
    %% discharge conjecture Num (same as d_trans).
    get_hyp(Num, HRec),
    hrec_just(HRec, cj(J)),
    var(J),
    @@itr(cdep,          CD,    IT),
    @@itr(udep,          UD,    IT),
    @@itr(direction,     Dir,   IT),
    @@itr(dischargelist, DList, IT),
    is_stronger_than(Rel, rev(ERGO_IMPLIES)),
    justify_hrec(HRec, tr(Rule, Dir, ERGO_TRUE, DList), CD, UD).
itr_finish(hyp:Num, IName, Rule, Output, Rel, IT) :-
    IName = _ : _,
    !,
    %% add Output as a new hypothesis Num.
    @@itr(cdep,          CD,    IT),
    @@itr(udep,          UD,    IT),
    @@itr(direction,     Dir,   IT),
    @@itr(dischargelist, DList, IT),
    is_stronger_than(Rel, ERGO_IMPLIES),
    add_hyp(Output, Num, ht(IName,Rule,Dir,DList), CD, UD).
itr_finish(null, _, _, _, _, _) :-
    %% do nothing.
    !.
itr_finish(OName, IName, _, _, _, _) :-
    handle_error([q(itrans_finish(OName)), 
                  ' conflicts with input=', q(IName),'.']).



%% itrans_dischargelist(DList).
%%
%%  DList is the list of rule/hyp names used to discharge all the
%%  conditions of the most recent itrans transformation.  This
%%  procedure can only be called *after* itrans_satisfy succeeds.
%%
%%  mode itrans_dischargelist(?list(compound)).
%%
itrans_dischargelist(DList) :-
    @@itr_valid(IT),
    @@itr(dischargelist, Term, IT),
    list(Term),  %% check that itrans_satisfy has finished.
    DList = Term.


%% itrans_body(Body).
%%
%%  Body is the Prolog code that checks side conditions of the
%%  primary rule of the most recent itrans transformation.  This
%%  procedure can only be called *after* itrans_finish succeeds.
%%
%%  mode itrans_body(?code).
%%
itrans_body(Body) :-
    @@itr_valid(IT),
    @@itr(output_name, OutName, IT),
    @@itr(body,        Term,    IT),
    nonvar(OutName),  %% check that itrans is finished.
    Body = Term.


%% private get_rule_shape(Conditional, InDirn, OutDirn, StmtShape, Shape).
%%
%%  mode get_rule_shape(?atom, ?atom, ?atom, @nonvar, ?integer).
%%
get_rule_shape(c, f, r,  b(_,b(Shape,_)), Shape).
get_rule_shape(c, r, f,  b(_,b(_,Shape)), Shape).
get_rule_shape(c, t, rt, b(_,Shape0), Shape) :-
    simplify_shape(Shape0, Shape).
get_rule_shape(c, rt, t, b(_,_), 1).
get_rule_shape(u, f, r,  b(Shape,_), Shape).
get_rule_shape(u, r, f,  b(_,Shape0), Shape) :-
    simplify_shape(Shape0, Shape).
get_rule_shape(u, t, rt, Shape0, Shape) :-
    simplify_shape(Shape0, Shape).
get_rule_shape(u, rt, t, _, 1).

simplify_shape(b(X,Y0), Z) :-
    !,
    simplify_shape(Y0, Y),
    Z is X+Y+1.
simplify_shape(Z, Z) :-
    integer(Z).
