%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%%
%% CVS $Id: var_dms.ql,v 1.18 2000/11/15 06:29:47 itp Exp $

#include "../lib/ergo.h"

%% cdep_create(Node).
%% cdep_create_empty(Node).
%% cdep_create_discharged(Node,CDep).
%% cdep_empty(CDep).
%% cdep_union(CDep1, CDep2, CDep).
%% cdep_simplify(CDep0, CDep).
%% cdep_discharge(Node, CDep).
%% cdep_depends_on(CDep, Node).
%% cdep_depends_on_nothing(CDep).
%%
%%  A CDep (conjecture dependency) system records the dependencies of 
%%  each 'CDep' term on a set of 'nodes'.  
%%  The CDep terms are built up (from nodes) using the cdep_empty, 
%%  cdep_union and cdep_simplify procedures and by 'discharging'
%%  a node (which converts it into a CDep term that depends upon
%%  other nodes).  Note that nodes are a subset of the CDep terms.
%%
%%  Creating Nodes
%%  ==============
%%  cdep_create creates a new node (Node) in the cdep dependency
%%    management system (DMS).
%%
%%  cdep_create_empty(Node) is an abbreviation for
%%    (cdep_create(Node),cdep_empty(Empty),cdep_discharge(Node,Empty)).
%%    This corresponds to adding a new hypothesis that does not 
%%    depend upon any conjectures.
%%
%%  cdep_create_discharged(Node,CDep) is an abbreviation for
%%    (cdep_create(Node),cdep_discharge(Node,CDep)).
%%
%%  Building up CDep terms
%%  ======================
%%  cdep_empty(CDep) creates a CDep term that does not depend upon
%%    any nodes.
%%
%%  cdep_union combines all the dependencies of CDep1 and CDep2 into CDep.
%%
%%  cdep_simplify simplifies CDep0 into another term CDep that has the
%%    same dependencies but may allow future processing to be more
%%    efficient.  Typically, this would be called after a series of
%%    union and/or discharge calls have been used to construct a 
%%    complex CDep term.
%%
%%  cdep_discharge(Node, CDep) replaces Node by the dependencies
%%    CDep.  CDep must have been built via cdep_union, cdep_simplify
%%    and/or cdep_discharge.  Node must have been created by cdep_create
%%    and must not have been discharged before.  This procedure will
%%    fail if CDep depends upon Node, since that would result in a
%%    circular dependency structure.
%%    If Node represents a conjecture, cdep_discharge would be called 
%%    when that conjecture was discharged and CDep would reflect all the
%%    conjectures that were used to discharge Node.
%%
%%  Checking dependencies
%%  =====================
%%  cdep_depends_on(CDep, Node) succeeds iff CDep depends upon Node.
%%    Note that Node must have been created via cdep_create and not
%%    yet discharged by cdep_discharge.
%%
%%  cdep_depends_on_nothing(CDep) succeeds iff CDep does not depend
%%    upon any nodes.
%%
%%  mode cdep_create(-cdep_node).
%%  mode cdep_create_empty(-cdep_node).
%%  mode cdep_empty(-cdep_dms).
%%  mode cdep_union(@cdep_dms, @cdep_dms, -cdep_dms).
%%  mode cdep_simplify(@cdep_dms, -cdep_dms).
%%  mode cdep_discharge(+cdep_node, @cdep_dms).
%%  mode cdep_depends_on(@cdep_dms, @cdep_node).
%%  mode cdep_depends_on_nothing(@cdep_dms).
%%
%% NB. The data structures used by this version are more complicated 
%%     than necessary, because it allows extra precondition checking
%%     to ensure that the cdep_* procedures are being used correctly.
%%
cdep_create(cdep_node(_)).

cdep_create_empty(cdep_node(cdep_empty)).

cdep_create_discharged(Node,CDep) :-
    PRE(cdep_create_discharged(1), var(Node)),
    PRE(cdep_create_discharged(2), cdep_legal_dms(CDep)),
    Node = CDep.

cdep_empty(cdep_empty).

% watchout itrans relies on union as is - do not try this:
%cdep_union(A, cdep_empty, A).  
%    PRE(cdep_union(1), cdep_legal_dms(A)).
%cdep_union(cdep_empty, B, B).
%    PRE(cdep_union(1), cdep_legal_dms(B)).

cdep_union(A,B,cdep_union(A,B)) :-
    PRE(cdep_union(1), cdep_legal_dms(A)),
    PRE(cdep_union(2), cdep_legal_dms(B)).

cdep_simplify(CDep0, Simp) :- 
    PRE(cdep_simplify(1), cdep_legal_dms(CDep0)),
    freeze_term(CDep0, Vars),
    thaw_term(Vars),
    (Vars = [] ->
        Simp = cdep_empty
    ;
        Simp = cdep_simplify(Vars)
    ).

cdep_discharge(Node, CDep) :-
    PRE(cdep_discharge(1), cdep_legal_node(Node)),
    PRE(cdep_discharge(2), cdep_legal_dms(CDep)),
    Node = cdep_node(Var),
    PRE(cdep_discharge, var(Var)),
    %% add this (and below) if you remove the cdep_node wrapper
    %% around Node:   CDep \== Node,
    Var = CDep.  %% assumes occurs_check is enabled.

cdep_depends_on(CDep, Node) :-
    PRE(cdep_depends_on(2), cdep_legal_node(Node)),
    PRE(cdep_depends_on(1), cdep_legal_dms(CDep)),
    Node = cdep_node(Var),
    PRE(cdep_depends_on(2), var(Var)),
    \+ Var = CDep.   %% assumes occurs_check is enabled.

cdep_depends_on_nothing(CDep) :-
    PRE(cdep_depends_on_nothing, cdep_legal_dms(CDep)),
    collect_vars(CDep, []).


cdep_legal_node(N) :-
    compound(N),
    N = cdep_node(_).

cdep_legal_dms(DMS) :- var(DMS),!,fail.
cdep_legal_dms(cdep_empty).
cdep_legal_dms(cdep_node(_)).
cdep_legal_dms(cdep_union(_,_)).
cdep_legal_dms(cdep_simplify(_)).


%% udep_create(Node).
%% udep_create_empty(Node).
%% udep_create_discharged(Node, UDep).
%% udep_node_dependencies(Node, UDep).
%% udep_empty(UDep).
%% udep_union(UDep1, UDep2, UDep).
%% udep_simplify(UDep0, UDep).
%% udep_discharge(Node, UDep).
%% udep_delete(Node).
%% udep_depends_on(UDep, Node).
%%
%%  This is an extended version of the CDep DMS which allows testing
%%  for dependencies on discharged nodes.  That is, a node that has
%%  been discharged can still be used as the second argument of
%%  udep_depends_on/2.  No 'depends_on_nothing' test is provided,
%%  since only the empty CDep structure would satisfy it.
%%
%%  An additional feature is udep_delete, which allows a node to
%%  be deleted (at most once).  A deleted node is removed from the
%%  dependency graph, so that no DMS term will depend upon it.
%%
%%  Another additional feature is udep_node_dependencies which returns
%%  a DMS structure in UDep that captures all the dependencies of Node,
%%  *excluding* Node itself.
%%
%%  Finally, a list of udep DMS structures is acceptable wherever 
%%  a single udep DMS structure would be.
%%
%%  Implementation: each node is represented by a term udep(Id,Deps)
%%      where Id is a variable that acts as the identity of the node
%%      and Deps records the nodes that this node depends upon.
%%      (Deps must have been built from other nodes via udep_empty,
%%      udep_union and udep_simplify calls).
%%
%%  mode udep_create(-udep_node).
%%  mode udep_create_empty(-udep_node).
%%  mode udep_create_discharged(-udep_node, +udep_dms).
%%  mode udep_empty(-atom).
%%  mode udep_union(@udep_dms, @udep_dms, -udep_dms).
%%  mode udep_simplify(@udep_dms, -udep_dms).
%%  mode udep_discharge(+udep_node, @udep_dms).
%%  mode udep_delete(+udep_node).
%%  mode udep_depends_on(@udep_dms, @udep_node).
%%
udep_create(udep_node(_,_)).

udep_create_empty(udep_node(_,udep_empty)).

udep_create_discharged(Node,UDep) :-
    PRE(udep_create_discharged(1), var(Node)),
    PRE(udep_create_discharged(2), udep_legal_dms(UDep)),
    Node = udep_node(_,UDep).

udep_node_dependencies(Node, UDep) :-
    PRE(udep_node_dependencies(1), udep_legal_node(Node)),
    Node = udep_node(_, UDep).

udep_empty(udep_empty).

%udep_union(A, udep_empty, A).
%    PRE(udep_union(1), udep_legal_dms(A)).

%udep_union(udep_empty, B, B).
%    PRE(udep_union(1), udep_legal_dms(B)).

udep_union(A,B,udep_union(A,B)) :-
    PRE(udep_union(1), udep_legal_dms(A)),
    PRE(udep_union(2), udep_legal_dms(B)).

udep_simplify(UDep0, udep_simplify(Vars)) :-
    freeze_term(UDep0, Vars),
    thaw_term(Vars).


udep_discharge(Node, UDep) :-
    PRE(udep_discharge(1), udep_nondeleted_node(Node)),
    PRE(udep_discharge(2), udep_legal_dms(UDep)),
    Node = udep_node(_,UD),
    PRE(udep_discharge, var(UD)),
    UD = UDep.    %% assumes occurs_check is enabled.

udep_delete(Node) :-
    PRE(udep_delete, udep_nondeleted_node(Node)),
    Node = udep_node(true,_).

udep_depends_on(UDep, Node) :-
    PRE(udep_depends_on(1), udep_legal_dms(UDep)),
    PRE(udep_depends_on(2), udep_legal_node(Node)),
    Node = udep_node(Id,_),
    var(Id), % check that it has not been deleted.
    \+ Id = UDep.    %% assumes occurs_check is enabled.


udep_nondeleted_node(Node) :-
    compound(Node),
    Node = udep_node(Id,_),
    var(Id).

udep_legal_node(Node) :-
    compound(Node),
    Node = udep_node(_,_).

udep_legal_dms(UDep) :-
    var(UDep),
    !,
    fail.
udep_legal_dms(udep_empty).
udep_legal_dms(udep_node(Id,_)) :-
    var(Id).
udep_legal_dms(udep_node(Id,_)) :-
    Id == true.    %% kludge: allow this because prover/save.ql needs it.
udep_legal_dms(udep_union(_,_)).
udep_legal_dms(udep_simplify(_)).
udep_legal_dms([]).
udep_legal_dms([UDep|UDeps]) :-
    udep_legal_dms(UDep),
    udep_legal_dms(UDeps).
