%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: callgraph.ql,v 1.9 2000/11/15 06:29:49 itp Exp $

#include "../prover/lib/ergo.h"

database command
  callgraph(File, Exps)
doc
  'Build a call graph for File, expanding with Exps.',nl,
  'File should be a Unix path name (the Ergo path is *not* searched) of',nl,
  'a Prolog source file.',nl,
  'Exps should be one of:',nl,
  begin(list),
  item([]),'  ''global_tactic'' -- expand as a global tactic file.',
  item([]),'  ''tactic(Thy)''   -- expand as a tactic file for Thy.',
  item([]),'  ''source''        -- expand as an Ergo source file.',
  item([]),'  ''none''          -- no expansions.',nl,
  end(list),
  'The call graph is stored in the Prolog database as facts of the form ',nl,
  '  cg(CallerPred, CalledPred, File).',nl,
  'where CallerPred and CalledPred are typically Name/Arity,',nl,
  'and File is the source file name containing the call.',nl,
  'Call graphs can be examined by writing Prolog code; some examples are',nl,
  'listed below.'
mode
  callgraph(@filename, @nonvar)
extends
  misc_cmds.

callgraph(File, Kind) :-
    ( Kind = global_tactic -> Exps = [decl(global),macro]
    ; Kind = tactic(Thy)   -> Exps = [decl(Thy), macro, prefix(Thy), thy(Thy)]
    ; Kind = source        -> Exps = [decl(global),macro]
    ; Kind = none          -> Exps = []
    ; handle_error(['callgraph: bad Exps.'])
    ),
    unwind_protect((setup_expansions(Exps, File),
                    msg(callgraph, ['expanding file']),
                    fcompile(File, [expand_only(true)]),
                    atom_codes(File, Codes),
                    append(Root, [0'., 0'q, _], Codes),
                    append(Root, ".qg", QGCodes),
                    atom_codes(QG, QGCodes),
                    msg(callgraph, ['building call graph']),
                    callgraph_file(QG, File)),
                   remove_expansions(Exps)).

%% callgraph_file(File, SrcFile).
%%
%%  Write a call graph for the Prolog code in File.
%%
callgraph_file(File, Src) :-
    retractall(cg(_,_,Src)),
    open(File, read, Strm),
    callgraph_strm(Strm, Src),
    close(Strm).

callgraph_strm(Strm, File) :-
    read_term(Strm, Term, [op_table(user)]),
    ( Term == end_of_file ->
        true
    ;
        callgraph_term(Term, File),
        callgraph_strm(Strm, File)
    ).

callgraph_term(Term, File) :-
    ( Term = (Head :- Body) ->
        functor(Head, Proc, Arity),
        callgraph_body(Proc/Arity, Body, File)
    ; Term = (?- Query) ->
        callgraph_body(query, Query, File)
    ; Term = (:- Query) ->
        callgraph_body(query, Query, File)
    ;
        functor(Term, Proc, Arity),
        assert(cg(Proc/Arity, fact, File))
    ).

callgraph_body(C, V, F) :-
    var(V),
    !,
    assert(cg(C, variable, F)).
callgraph_body(C, (A,B), F) :-
    !,
    callgraph_body(C, A, F),
    callgraph_body(C, B, F).
callgraph_body(C, (A;B), F) :-
    !,
    callgraph_body(C, A, F),
    callgraph_body(C, B, F).
callgraph_body(C, (A->B), F) :-
    !,
    callgraph_body(C, A, F),
    callgraph_body(C, B, F).
callgraph_body(C, T, F) :-
    meta_predicate(T, Args),
    !,
    functor(T, N, A),
    assert(cg(C, N/A, F)),
    callgraph_meta(C, T, Args, F).
callgraph_body(C, T, F) :-
    functor(T, N, A),
    assert(cg(C, N/A, F)).

callgraph_meta(C, T, [], F).
callgraph_meta(C, T, [A|As], F) :-
    arg(A, T, Arg),
    callgraph_body(C, Arg, F),
    callgraph_meta(C, T, As, F).

database command
  cg_callers(Pred)
extends
  'database:callgraph/2'
doc
  'List all callers of Pred (Name/Arity) in the call graph database.',nl,
  'The database is built by ',
  xref(['callgraph(File, Exps)'], global, cmd(callgraph/2)), '.'
mode
  cg_callers(@compound).

cg_callers(Pred) :-
    (
        cg(Caller, Pred, File),
        writelist([left(24, [w(Caller)]), w(File), nl]),
        fail
    ;
        true
    ).

database command
  cg_find_callers(Preds),
  cg_find_callers_file(File)
extends
  'database:callgraph/2'
doc
  'List predicates together with the names of all files that contain',nl,
  'calls on those predicates.',nl,
  'Preds is a list of Name/Arity specs for predicates to list,',nl,
  'File is a filename (list all predicates defined in that file).'
mode
  cg_find_callers_file(@filename),
  cg_find_callers(@list(compound)).

cg_find_callers_file(File) :-
    findall(Name/Arity, cg(Name/Arity, _X, File), Preds1),
    sort(Preds1, Preds),
    cg_find_callers(Preds).

cg_find_callers(Preds) :-
    member(N/A, Preds),
    findall(CFile, cg(_X, N/A, CFile), CFiles1),
    sort(CFiles1, CFiles),
    ( CFiles = [] ->
        writelist([left(24, [w(N), '/', w(A)]), tab, 'has no callers', nl])
    ;
        member(CF, CFiles),
        writelist([left(24, [w(N), '/', w(A)]), tab, 'has callers in ',
                   w(CF), nl])
    ),
    fail.
cg_find_callers(_).
