%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: check.ql,v 1.15 2001/06/04 01:24:10 itp Exp $

#include "../prover/lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file contains procedures for checking the quality of theories.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

database command
  check_thy
extends
  misc_cmds
mode
  check_thy
implementation
  global_state_lookup(current_dir,Thy),
  check_thy(Thy)
doc
  'Check the current theory for its quality (e.g. duplicated rules).'.

database command
  check_thy(Thy)
extends
  misc_cmds
mode
  check_thy(@theory)
doc
  'Check Thy for its quality (e.g. duplicated rules).'.

database command
  check_thys
extends
  misc_cmds
mode
  check_thys
doc
  'Check all the theories for their quality (e.g. duplicated rules).'.

check_thys :-
    findall(Thy,
            query_global(ancestor(Thy,_)),
            Thys),
    map(check_thys1, [Thys]).

%% private check_thys1(Thy).
%%
%%  Cd into Thy and check it---always succeed.
%%
%% mode check_thys1(@thy).
%%
check_thys1(Thy) :-
    cd(Thy),
    check_thy(Thy).

check_thys1(_).

%% check_thy(Thy).
%%
%% Check the theory Thy for its quality (e.g. duplicated rules).
%% The following checks are made for each rule in Thy:
%%
%%   * That no rule that is identical or more general (in the sense of
%%     incomplete unification) than it appears in the theory or any ancestor.
%%   * That no rule is just another with the antecedant weakened or
%%     consequent strengthened.
%%   * That no rule is an existential version of a universal rule or a
%%     existential version of an unique existential rule.
%%   * That no equivalence is just the reverse of another.
%%
%% mode check_thy(@atom).

check_thy(Thy) :-
	cd(Thy),
        query_theory(rule(Name,Kind,Shape,Stmt,Body), Thy),
        freeze_vars(rule(Name,Kind,Shape,Stmt,Body)),
        findall(Thy1:Name1,
  	       (query_ancestors(rule(Name1,_,Shape,Stmt1,Body), Thy1),
                Name \= Name1,
                subsumes_or_subsumedby_rule(Stmt, Stmt1),
                incomplete_retry_delays),
               Rules),
        (   Rules \== []
	    ->
		writelist(['The following rules subsume or are subsumed by ',
                           Name, ':', nl]),
                map(write_conflicting_rule, [Rules])
	    ;
		true
	),
	fail.

check_thy(Thy).

%% private subsumes_or_subsumedby_rule(A,B).
%%
%% Check to see if rule A subsumes rule B or rule B subsumes rule A.
%%
%% mode subsumes_or_subsumedby_rule(@thy(term), @thy(term)).

subsumes_or_subsumedby_rule(A,B) :-
    subsumes_rule(A,B),
    !.
subsumes_or_subsumedby_rule(A,B) :-
    subsumes_rule(B,A),

subsumes_rule(A,A).

subsumes_rule(ERGO_IMPLIES(L,R), ERGO_IMPLIES(L1,R1)) :-
    weakens_andor_strengthens(L,R, L1,R1).

subsumes_rule(ERGO_EQUIV(L,R), ERGO_IMPLIES(L1,R1)) :-
    weakens_andor_strengthens(L,R, L1,R1).

subsumes_rule(ERGO_EQUIV(L,R), ERGO_IMPLIES(R1,L1)) :-
    weakens_andor_strengthens(L,R, L1,R1).

subsumes_rule(ERGO_EQUIV(L,R), ERGO_EQUIV(R1,L1)) :-
    once(conjunction_to_list(L,  LList)),
    once(conjunction_to_list(L1, L1List)),
    once(conjunction_to_list(R,  RList)),
    once(conjunction_to_list(R1, R1List)),
    diff_list(LList,  L1List, []),  % L and L1 same set.
    diff_list(L1List, LList,  []),
    diff_list(RList,  R1List, []),  % R and R1 same set.
    diff_list(R1List, RList,  []).

subsumes_rule(ERGO_EQUALS(L,R), ERGO_EQUALS(R,L)).

subsumes_rule(!!'jprop.all' !x A, !!'jprop.ex' !x A).

subsumes_rule(!!'jprop.ex_one' !x A, !!'jprop.ex' !x A).

weakens_andor_strengthens(L,R, L1,R1) :-
    once(conjunction_to_list(L,  LList)),
    once(conjunction_to_list(L1, L1List)),
    once(conjunction_to_list(R,  RList)),
    once(conjunction_to_list(R1, R1List)),
    diff_list(LList,  L1List, []),  % Weakens antecedant.
    diff_list(R1List, RList,  []).  % Strengthens consequent.
                    

%% private write_conflicting_rule(Thy:Name).
%%
%% Write an indented line containing Thy:Name.
%%
%% mode write_conflicting_rule(@ground).

write_conflicting_rule(Thy:Name) :-
	writelist(['   ', w(Thy:Name), nl]).

general command
  check_unification
mode
  check_unification
doc
  'Test Qu-Prolog''s unification algorithm.',nl,
  'It unifies every theorem in every ancestor theory against a',nl,
  'copy of itself.  It also unifies some large and complex terms',nl,
  'against copies of themselves'.

check_unification :-
    % CD into every theory.
    findall(Thy,
            query_global(ancestor(Thy,_)),
            Thys),
    map(check_unification1, [Thys]),
    
    %Attempt to unify every theorem against itself.
    findall(Stmt,
            (
                query_global(ancestor(Thy,_)),
                query_theory(rule(_Name,_Kind,_Shape,Stmt,_Body), Thy)
            ),
            Thms),
    map(check_unification2, [Thms]),

   %Attempt to unify some special test cases against themselves.
   findall(Term,
           check_unification3(Term),
           Terms),
   map(check_unification2, [Terms]).


%% private check_unification1(Thy).
%%
%%  Cd into Thy---always succeed.
%%
%% mode check_unification1(@atom).
%%
check_unification1(Thy) :-
    cd(Thy).

check_unification1(_).

check_unification2(Term) :-
    copy_term(Term, Term1),
    (
        Term = Term1
    ;
        writelist(['The term:', nl,
                   Term, nl,
                   'fails to unify correctly with itself.', nl])
    ),
    !.

check_unification3([!x1/!x2,!x3/!x4,!x5/!x6,!x7/!x8,!x9/!x10]
                   !!q [!y1,!y2,!y3,!y4,!y5,!y6,!y7,!y8,!y9,!y10]
                   F(!x1,!x2,!x3,!x4,!x5,!x6,!x7,!x8,!x9,!x10,
                     !y1,!y2,!y3,!y4,!y5,!y6,!y7,!y8,!y9,!y10)).

check_unification3([[[[[!x1/!x2]!x3/!x4]!x5/!x6]!x7/!x8]!x9/!x10]
                   !!q [!y1,!y2,!y3,!y4,!y5,!y6,!y7,!y8,!y9,!y10]
                   F(!x1,!x2,!x3,!x4,!x5,!x6,!x7,!x8,!x9,!x10,
                     !y1,!y2,!y3,!y4,!y5,!y6,!y7,!y8,!y9,!y10)).

check_unification3([N(M(L(K(J(I(H(G(F(!x1),!x2),!x3),!x4),!x5),!x6),!x7),!x8),!x9)/!x10]
                   !!q [!y1,!y2,!y3,!y4,!y5,!y6,!y7,!y8,!y9,!y10]
                   F(!x1,!x2,!x3,!x4,!x5,!x6,!x7,!x8,!x9,!x10,
                     !y1,!y2,!y3,!y4,!y5,!y6,!y7,!y8,!y9,!y10)).

check_unification3(F(X,!x1,!x2,!x3,!x4,!x5,!x6,!x7,!x8,!x9,!x10,
                     !!q [!x1,!x2,!x3,!x4,!x5,!x6,!x7,!x8,!x9,!x10] X)).
