%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% This file defines a user interface for Ergo theory management.
%% It allows users to invoke commands from the 'general' and 'database'
%% command tables.
%%
%% The name of the user interface is 'DBMS' and its top level help
%% node is called 'dbms' (this is linked to 'master_help_node').
%%
%% CVS $Id: dbms.ql,v 1.105 2006/02/05 22:15:24 itp Exp $

#include "../prover/lib/ergo.h"

?- require_global_tactic_file(abbrev).

%% Silent command
%% ==============

general command
  silent_cmd(Cmd)
extends
  general_cmds
doc
  'Call the general command Cmd silently - that is no prompt is produced', nl,
  'after the command is executed. This can be useful in, for example, GUI',nl,
  'interfaces where a general command should not cause any I/O on the',nl,
  'usual user interaction window.'
mode
  silent_cmd(@term).


%% The top level node of the DBMS help tree.
%% =========================================

dbms header
  'Theory/Database Commands (The DBMS user interface)'
extends
  master_help_node
doc
  'This section describes the `DBMS'' user interface, which allows you to',nl,
  'build theories and manage the Ergo database.',
  para,
  'Most DBMS commands operate on the `current'' theory, which is the one',nl,
  'whose name is displayed in the prompt.  You can set the current theory',nl,
  'by typing: [TheoryName].',
  para,
  'The following groups of commands are available:',nl,
  xref('General Commands (viewing data, setting parameters etc.)',
       global,
       general_cmds),nl.


%% Second level nodes of the DBMS help tree.
%% =========================================

misc_cmds header
  'Miscellaneous Commands.'
extends
  dbms
doc
  'The following miscellaneous commands are available:',nl.


save_exit_cmds header
  'Save/Exit Commands.'
extends
  dbms
doc
  'This set of commands allow changes made to theories and other',nl,
  'facts to be saved to disk in various ways.',nl.


theory_construction_cmds header
  'Theory Construction Commands.'
extends
  dbms
doc
  'This set of commands allows new theories and subtheories',nl,
  'to be created.  Each theory can inherit from one or more other',nl,
  'theories, provided that the resulting theory graph does not',nl,
  'contain cycles.',
  para,
  'Each subtheory is associated with some ''owner'' theory.',nl,
  'Conceptually, all the facts in the subtheory can be regarded as',nl,
  'belonging to its owner theory.  The only difference is that',nl,
  'the subtheory has its own name space.',nl.


theory_extension_cmds header
  'Theory Extension Commands.'
extends
  dbms
doc
  'This set of commands allow new facts to be added to the current',nl,
  'theory.',nl.


theory_deletion_cmds header
  'Theory Deletion Commands.'
extends
  dbms
doc
  'This set of commands allows theories and subtheories to be deleted.',nl,
  'However, no theory or subtheory can be deleted if other non-deleted',nl,
  'theories would still depend upon it.',nl.


tactic_file_cmds header
  'Tactic File Commands.'
extends
  dbms
doc
  'This set of commands allows global and theory-specific tactic',nl,
  'files to be managed. These commands use the ',
  xref(load_path,global,'parameter:load_path'),' parameter',nl,
  'to search for the tactic files.'.


sugar_cmds header
  'I/O Syntactic Sugar Commands.'
extends
  dbms
doc
  'This set of commands allows sugared syntax to be defined for',nl,
  'the terms of the current theory.  Each theory can have a different',nl,
  'set of abbreviations, operators etc.',nl.



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Commands that are directly related to entering/exiting the
%% DBMS User Interface go here.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

database command
  bye
extends
  save_exit_cmds
mode
  bye
implementation
  bye
doc
  'If called from the outermost theory-level user interface, this',nl,
  'command saves all theories (and global data) and exits the prover.',
  para,
  'If called from within a nested theory-level user interface, this',nl,
  'command will fail with an error message.  You should use',nl,
  'the resume command instead.'.


database command
  quit_proof
extends
  misc_cmds
mode
  quit_proof
implementation
  true
doc
  'This does nothing when called from the theory management level.',nl,
  'It exists simply so that a quit_proof command is available within',nl,
  'all of Ergo''s command-line user interfaces.'.


database command
  resume
extends
  save_exit_cmds
mode
  resume
implementation
  resume
doc
  'Closes the current (nested) theory-level user interface and',nl,
  'returns to the previous user interface (typically a suspended proof).',nl,
  'This command cannot be used to exit the outermost theory-level user',nl,
  'interface.  Use the bye/0 command to do that.'.



%% dbms_interface.
%%
%%  This is the main entry point for the DBMS user interface.
%%
%%  The set_dbms_interface(dbms_interface) call at the top of this
%%  file causes Ergo to call this procedure each time a theory
%%  management interface is required.
%%
%%  mode dbms_interface.
%%
dbms_interface :-
    catch(
        dbms_interface2,
        Template,
        ( % functor(Template,F,A),
          writelist([nl,'Continuing after catching ',q(Template),' event.',nl]),
          dbms_interface)).

dbms_interface2 :-
    repeat,
%% TO DO: In dbms_interfact2,
%% enable signals once we've made all dbms commands atomic
%% or checked that they clean up the database after a throw?
%% For now, we repeatedly disable signals, because the elves
%% proof interface enables them and has no way of restoring
%% the original enable/disable flag.
%% Ray's last comment: Needs major rethink.
    dbms_prompt(P, OpTable),
    get_command_top(P, OpTable, C),
    dbms_interface(C).

%%
%% This interface delas with running silent commands
%% without prompting.
%%
get_command_top(P, OpTable, C0) :-
    get_command(P, OpTable, C),
    (  C = silent_cmd(SC) ->
         ( dbms_interface(SC) ->
             C0 = true
           ;
             get_command_top([], OpTable, C0)
         )
       ;
         C0 = C
    ).            
   

%% auxillary routine for dbms_interface/0.
%%  mode dbms_interface(+nonvar, +theory).
%%
dbms_interface(C) :-
    \+ std_nonvar(C),
    !,
    handle_error(['illegal dbms command: ',q(C)]).
dbms_interface(true) :-
    !.
dbms_interface(end_of_file) :-
    !.
dbms_interface(C) :-
    functor(C, F, A),
    functor(C1, F, A),
    member(Table, [general, database]),
    find_command(C1, Table, Modes, C2, _DocNode),
    !,
    ( deabbreviate_cmd(Modes, C, C1) ->
        true
    ;
        handle_error(['bad arguments in command: ', q(C)])
    ),
    ( in_macro,get_verbosity(V),V > 1 ->
        writelist([nl,q(C),'.',nl])
    ;
        true
    ),
    ( C2 == end_of_file ->
        true
    ; C2 == resume ->
        check_or_error(
            ( ctr_get(dbms_interfaces, Num), Num > 1 ),
            ['not a nested theory-level interface.  Use bye/0 to exit.'])
    ; C2 == bye ->
        check_or_error(
            ( ctr_get(dbms_interfaces, Num), Num =:= 1 ),
            ['bye cannot exit a nested interface.  Use resume/0.'])
    ; call_predicate(C2) ->
        fail
    ; in_macro ->
        increment_error_count,
        writelist([q(C),nl,'(failed)',nl]),
        fail
    ;
        increment_error_count,
        writelist(['(failed)',nl]),
        fail
    ).
dbms_interface(C) :-
    handle_error(['unknown dbms command: ',q(C)]).



%% dbms_prompt(Prompt, OpTable).
%%
%%  Determines the prompt for the dbms user interface and
%%  which operator table should be used when reading commands.
%%
%%  mode dbms_prompt(-list(term), -atom).
%%
dbms_prompt(Prompt, OpTable) :-
    global_state_lookup(current_dir,CD), CD \= none,
    !,
    ctr_get(dbms_interfaces, Num),
    dbms_prompt2(Num, CD, List),
    Prompt = [nl,q(List),': '],
    OpTable = CD.
dbms_prompt('No theory: ', global).

dbms_prompt2(0, Thy, Thy).
dbms_prompt2(N, Thy, [List]) :-
    N > 0,
    N1 is N - 1,
    dbms_prompt2(N1, Thy, List).


?- set_dbms_interface(dbms_interface).

