%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##


%% $Id: dv.ql,v 1.13 2004/07/25 04:16:13 itp Exp $
%% General support for interfacing to daVinciV2.0

#include "../prover/lib/ergo.h"

general command
  daVinci_init
doc
  'Prepare for da Vinci interaction (but do not start da Vinci now).'
extends
  general_cmds
mode
  daVinci_init.

daVinci_init :-
    add_handler(reader_special_execute_hook(emacs(daVinci(Term))),
                daVinci_special_hook(Term)),
    ( query_global(announcement(daVinci(_),_)) ->
        true
    ;
        add_announcement(
            daVinci(_),
            ['daVinci(State): The daVinci connection has changed state.',nl,
             'State is one of [''connected'', ''closed''].'
            ])
    ),
    add_handler(signal('SIGPIPE',_),
                true).

daVinci_special_hook(connect(Pipe)) :-
    open(Pipe, write, DVStrm),
    atom_concat2('/bin/rm -f ', Pipe, Command),
    os(system(Command)),
    writelist(DVStrm, [nothing,nl]),
    erase_all(daVinci_stream),
    recorda(daVinci_stream, DVStrm, _),
    announce_event(daVinci(connected)).
daVinci_special_hook(close) :-
    (
        recorded(daVinci_stream, Strm, Ref),
        close(Strm),
        erase(Ref),
        fail
    ;
        true
    ),
    announce_event(daVinci(closed)).
daVinci_special_hook(selected(Nodes)) :-
    announce_event(daVinci(selected(Nodes))).
daVinci_special_hook(menu(Menu,Selection)) :-
    announce_event(daVinci(menu(Menu,Selection))).

general command
  daVinci_start
doc
  'Start a daVinci process.',nl,
  'Do ''daVinci_init'' first to set up the communication handlers.',nl,
  'Wait for a daVinci(connected) announcement before sending requests.'
extends
  general_cmds
mode
  daVinci_start.

daVinci_start :-
    current_predicate(emacs_magic/1),
    !,
    emacs_magic('DAVINCI'),
    write(start), nl,
    emacs_magic('NORMAL').    
daVinci_start :-
    writelist(['Emacs interface must be active to use daVinci.',nl]),
    announce_event(daVinci(disabled)).

library command
  daVinci_active
doc
  'True if there is an active da Vinci connection',nl,
  '(though the process may not necessarily be listening).'
extends
  general_cmds
mode
  daVinci_active.

daVinci_active :-
    recorded(daVinci_stream, _, _).

general command
  daVinci_command(Command)
doc
  'Send Command to da Vinci.',nl,
  'May be a string or a command term.',nl,
  'If a string, it must be in exactly the daVinci API format.',nl,
  'If a term, the notation is slightly ''prologized'':',
  begin(list),
  item(['Node, edge and menu IDs']), 'are atomics, not strings',
  item(['Node labels and status messages']), 'are writelists, not strings',
  item(['Node updates']),
    'can be new_node(Id,Label,Attrs) or delete_node(Id)',nl,
    tab, ' -- object(Label) is added to the explicit attributes;',
  item(['Edge updates']),
    'can be new_edge(Id,Src,Tgt,Attrs) or delete_edge(Id)',
  item(['Attribute updates']), 'can be node(Id, Attrs) or edge(Id, Attrs)',
  item(['Attributes']), 'are lists of terms of the following forms:',
  begin(list),
  item(['   object(Label):']), 'the OBJECT node attribute',
  item(['   font_family(Fam):']), 'the FONTFAMILY node attribute',
  item(['   font_style(Style):']), 'the FONTSTYLE node attribute',
  item(['   color(Colour):']), 'the COLOR node attribute',
  item(['   shape(Shape):']), 'the _GO node attribute',
  item(['   icon(File):']),
    'the ICONFILE node attribute (shape(icon) is assumed)',
  item(['   hidden(Flag):']), 'the HIDDEN node attribute',
  item(['   border(Style):']), 'the BORDER node attribute',
  item(['   edgecolor(Colour):']), 'the EDGECOLOR edge attribute',
  item(['   edgepattern(Pattern):']), 'the EDGEPATTERN edge attribute',
  item(['   direction(Dir):']), 'the _DIR edge attribute',
  end(list),
  end(list)
mode
  daVinci_command(@term).

daVinci_command(Command) :-
    %%writelist([w(daVinci_command(Command)), nl]),
    recorded(daVinci_stream, S, _),
    daVinci_write_command(S, Command),
    nl(S),
    flush_output(S).

daVinci_write_command(S, Command) :-
    is_chars(Command),
    !,
    writelist(S, [chars(Command)]).
daVinci_write_command(S, graph(new([]))) :-
    !,
    writelist(S, ['graph(new([]))']).
daVinci_write_command(S, graph(update(Nodes, Edges))) :-
    !,
    writelist(S, ['graph(update([']),
    daVinci_write_updates(S, Nodes),
    writelist(S, ['],[']),
    daVinci_write_updates(S, Edges),
    writelist(S, [']))']).
daVinci_write_command(S, graph(change_attr(Changes))) :-
    !,
    writelist(S, ['graph(change_attr([']),
    daVinci_write_change_attrs(S, Changes),
    writelist(S, [']))']).
daVinci_write_command(S, window(title(Title))) :-
    !,
    writelist(S, ['window(title("' | Title]),
    writelist(S, ['"))']).
daVinci_write_command(S, window(show_status(Status))) :-
    !,
    writelist(S, ['window(show_status("' | Status]),
    writelist(S, ['"))']).
daVinci_write_command(S, special(select_nodes(Nodes))) :-
    !,
    writelist(S, ['special(select_nodes([']),
    daVinci_write_ids(S, Nodes),
    writelist(S, [']))']).
daVinci_write_command(S, app_menu(create_menus(Menus))) :-
    !,
    writelist(S, ['app_menu(create_menus([']),
    daVinci_write_menus(S, Menus),
    writelist(S, [']))']).
daVinci_write_command(S, app_menu(activate_menus(MenuIds))) :-
    !,
    writelist(S, ['app_menu(activate_menus([']),
    daVinci_write_ids(S, MenuIds),
    writelist(S, [']))']).
daVinci_write_command(S, quit) :-
    !,
    writelist(S, [quit]).
daVinci_write_command(S, Cmd) :-
    write(S, Cmd).

daVinci_write_updates(_, []).
daVinci_write_updates(S, [Update]) :-
    !,
    daVinci_write_update(S, Update).
daVinci_write_updates(S, [Update|Updates]) :-
    daVinci_write_update(S, Update),
    write(S, ','),
    daVinci_write_updates(S, Updates).

daVinci_write_update(S, delete_node(Id)) :-
    writelist(S, ['delete_node("', w(Id), '")']).
daVinci_write_update(S, new_node(Id, Label, Attrs)) :-
    writelist(S, ['new_node("', w(Id), '", "", [']),
    daVinci_write_attributes(S, [object(Label)|Attrs]),
    writelist(S, ['])']).
daVinci_write_update(S, delete_edge(Id)) :-
    writelist(S, ['delete_edge("', w(Id), '")']).
daVinci_write_update(S, new_edge(Id, Source, Target, Attrs)) :-
    writelist(S, ['new_edge("', w(Id), '", "", [']),
    daVinci_write_attributes(S, Attrs),
    writelist(S, ['], "', w(Source), '", "', w(Target), '")']).

daVinci_write_change_attrs(_, []).
daVinci_write_change_attrs(S, [Change]) :-
    !,
    daVinci_write_change_attr(S, Change).
daVinci_write_change_attrs(S, [Change|Changes]) :-
    daVinci_write_change_attr(S, Change),
    write(S, ','),
    daVinci_write_change_attrs(S, Changes).

daVinci_write_change_attr(S, node(Id, Attrs)) :-
    writelist(S, ['node("', w(Id), '", [']),
    daVinci_write_attributes(S, Attrs),
    writelist(S, ['])']).
daVinci_write_change_attr(S, edge(Id, Attrs)) :-
    writelist(S, ['edge("', w(Id), '", [']),
    daVinci_write_attributes(S, Attrs),
    writelist(S, ['])']).

daVinci_write_attributes(_, []).
daVinci_write_attributes(S, [Attr]) :-
    !,
    daVinci_write_attribute(S, Attr).
daVinci_write_attributes(S, [Attr|Attrs]) :-
    daVinci_write_attribute(S, Attr),
    write(S, ','),
    daVinci_write_attributes(S, Attrs).

%%% Node attributes
daVinci_write_attribute(S, object(Label)) :-
    !,
    writelist(S, ['a("OBJECT", "' | Label]),
    writelist(S, ['")']).
daVinci_write_attribute(S, font_family(Fam)) :-
    !,
    writelist(S, ['a("FONTFAMILY", "', w(Fam), '")']).
daVinci_write_attribute(S, font_style(Style)) :-
    !,
    writelist(S, ['a("FONTSTYLE", "', w(Style), '")']).
daVinci_write_attribute(S, color(Colour)) :-
    !,
    writelist(S, ['a("COLOR", "', w(Colour), '")']).
daVinci_write_attribute(S, shape(Shape)) :-
    !,
    writelist(S, ['a("_GO", "', w(Shape), '")']).
daVinci_write_attribute(S, icon(File)) :-
    !,
    writelist(S, ['a("_GO","icon"), a("ICONFILE", "', w(File), '")']).
daVinci_write_attribute(S, hidden(Flag)) :-
    !,
    writelist(S, ['a("HIDDEN", "', w(Flag), '")']).
daVinci_write_attribute(S, border(Style)) :-
    !,
    writelist(S, ['a("BORDER", "', w(Style), '")']).
%%% Edge Attributes
daVinci_write_attribute(S, edgecolor(Colour)) :-
    !,
    writelist(S, ['a("EDGECOLOR", "', w(Colour), '")']).
daVinci_write_attribute(S, edgepattern(Pattern)) :-
    !,
    writelist(S, ['a("EDGEPATTERN", "', w(Pattern), '")']).
daVinci_write_attribute(S, direction(Dir)) :-
    !,
    writelist(S, ['a("_DIR", "', w(Dir), '")']).

daVinci_write_ids(S, []).
daVinci_write_ids(S, [Node]) :-
    !,
    daVinci_write_id(S, Node).
daVinci_write_ids(S, [Node|Nodes]) :-
    daVinci_write_id(S, Node),
    write(S, ','),
    daVinci_write_ids(S, Nodes).

daVinci_write_id(S, Node) :-
    writelist(S, ['"', w(Node), '"']).

daVinci_write_menus(S, []).
daVinci_write_menus(S, [Menu]) :-
    !,
    daVinci_write_menu(S, Menu).
daVinci_write_menus(S, [Menu|Menus]) :-
    daVinci_write_menu(S, Menu),
    write(S, ','),
    daVinci_write_menus(S, Menus).

daVinci_write_menu(S, menu_entry(Id,Label)) :-
    writelist(S, ['menu_entry("', w(Id), '","', w(Label), '")']).
daVinci_write_menu(S, menu_entry_mne(Id,Label,Mne,Mod,Acc)) :-
    writelist(S, ['menu_entry_mne(', '"', w(Id), '",',
                                     '"', w(Label), '",',
                                     '"', w(Mne), '",',
                                     w(Mod), ',',
                                     '"', w(Acc), '"',
                                 ')']).
