%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: elves.ql,v 1.69 2004/07/25 04:16:13 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file defines the 'elves' user interface for proofs.
%%
%% The main features of the elves interface are:
%%
%%    * it defines some abbreviations for common commands like focus.
%%
%%    * it consults the current list of `elves' (proof assistants)
%%      before prompting the user for commands.
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../prover/lib/ergo.h"

?- require_global_tactic_file(abbrev).
?- require_global_tactic_file(focus).
?- require_global_tactic_file(show).
?- require_global_tactic_file(showwin).


%% setup_elves.
%%
%%  Initialises the database for the elves interface.
%%  Call this when this tactic file is first added to the system.
%%
%%  mode setup_elves.
%%
setup_elves :-
    %% We must declare local (non-saved) facts every session.
    (theory_fact(elf,2) ->
        true
    ;
        new_theory_fact(elf,2,[save(false)])
    ),
    add_handler(deabbreviate_with_type(tactic,T0,T),
                elves_deabbrev_cmd(T0,T)),
    fail.
setup_elves :-
    query_global(request(proof_output(_,_),_)),
    %% Saved facts/events have already been declared.
    !.
setup_elves :-
    %% Declare all events and saved facts.
    new_theory_fact(default_elves_list,1,[save(true)]),
    new_theory_fact(default_active_elves,1,[save(true)]),
    add_request(proof_output(_,_),
                ['proof_output(Kind,Term).', nl,
                 'Kind is: assumption(No), newvars, focus, reln, goal,',
                 ' hyp(HypKind,No)',nl,
                 'HypKind is: discharged, dependent, unused, used.']).

?- setup_elves.

%% These asserted facts store the current elf settings during a proof.
?- dynamic(elves_list/1).
?- dynamic(active_elves/1).

?- dynamic(elves_suspended_state/4).

%% This is called automatically by the elf declarations in tactic files.
add_elf_name(Thy, Name, DocNode) :-
    is_theory_name(Thy),
    tactic_retractall_theory(elf(Name, _), Thy),
    tactic_assert_theory(elf(Name, DocNode), Thy).

    
?- Msg=[_|_],
   add_listener(wrap_tac(Id,Msg,_,call),
                msg(call(Id), 1, Msg)).
?- Msg=[_|_],
   add_listener(wrap_tac(Id,Msg,_,exit),
                msg(exit(Id), 1, Msg)).
?- Msg=[_|_],
   add_listener(wrap_tac(Id,Msg,_,fail),
                msg(fail(Id), 1, Msg)).


?- add_listener(proof(Thy,Name,Action),
                elves_proof_setup(Action,Thy,Name)).

elves_proof_setup(started, _, _) :-
    !,
    get_default_elves_list(EL),
    set_elves_list(EL),
    get_default_active_elves(N),
    set_active_elves(N).
elves_proof_setup(suspended, Thy, Name) :-
    !,
    get_elves_list(EL),
    get_active_elves(N),
    asserta(elves_suspended_state(Thy, Name, EL, N)).
elves_proof_setup(resumed, Thy, Name) :-
    !,
    retract(elves_suspended_state(Thy, Name, EL, N)),
    set_elves_list(EL),
    set_active_elves(N).
elves_proof_setup(_, _, _).


%% show_elves.
%%
%%  Displays the current list of elves and the number of active elves.
%%
%% mode show_elves.
%%
show_elves :-
    get_active_elves(N),
    get_elves_list(EL),
    writelist([q(N),' elves are active.',nl]),
    show_numbered_elves(EL,1).

show_numbered_elves([],_).
show_numbered_elves([E|EL],N) :-
    writelist(['    ',q(N),'.  ',q(E),nl]),
    N1 is N+1,
    show_numbered_elves(EL, N1).


default_elves_list parameter
  [set_default_elves_list, get_default_elves_list]
extends
  parameters
doc
  'The default_elves_list parameter is a list of elf names.',nl,
  'The ',
    xref('elves list',global,'parameter:elves_list'),
  ' parameter is set',nl,
  'to this default list each time a proof is started.'.

set_default_elves_list(EL) :-
    check_current_theory(CD),
    check_good_elves_list(EL, EL2),
    tactic_retractall_theory(default_elves_list(_), CD),
    tactic_assert_theory(default_elves_list(EL2), CD).

get_default_elves_list(EL) :-
    query_ancestors(default_elves_list(EL1)),
    !,
    EL = EL1.
get_default_elves_list([]).


default_active_elves parameter
  [set_default_active_elves, get_default_active_elves]
extends
  parameters
doc
  'The ',
    xref(active_elves,global,'parameter:active_elves'),
  ' parameter is set',nl,
  'to this default number each time a proof is started.'.

set_default_active_elves(N) :-
    check_current_theory(CD),
    check_active_elves(N),
    tactic_retractall_theory(default_active_elves(_), CD),
    tactic_assert_theory(default_active_elves(N), CD).

get_default_active_elves(N) :-
    query_ancestors(default_active_elves(N1)),
    !,
    N = N1.
get_default_active_elves(0).


elves_list parameter
  [set_elves_list, get_elves_list]
extends
  parameters
doc
  'The elves_list parameter determines which elves may participate in',nl,
  'the current proof.  Its value is a list of names of elves.',nl,
  nl,
  'Elves can be added to or removed from this list at any time',nl,
  'during a proof.  At the beginning of each proof the list is',nl,
  'initialised from the ',
    xref(default_elves_list,global,'parameter:default_elves_list'),
    ' parameter.',nl,
  nl,
  'Elves are called in the order that they appear in elves_list.',nl,
  'The active_elves parameter determines how many of the elves will',nl,
  'be called automatically before each user interaction.  Any suggestions',nl,
  'that they make will be tried before the user is given control.',nl,
  nl,
  'An alternative way of calling elves is via the ',
    xref('hints(N)',global,'proof:hints/1'),' command',nl,
  'which displays all suggestions from the first N elves in this list',nl,
  'without trying to execute those suggestions.'.

set_elves_list(EL) :-
    check_good_elves_list(EL, EL2),
    ( elves_list(EL2) ->
        %% no changes needed.
        true
    ;
        retractall(elves_list(_)),
        asserta(elves_list(EL2))
    ).

get_elves_list(EL) :-
    elves_list(L),
    !,
    L = EL.
get_elves_list(EL) :-
    get_default_elves_list(EL).


check_good_elves_list([], []).
check_good_elves_list([E0|EL0], [E|EL]) :-
    check_good_elf(E0,E),
    check_good_elves_list(EL0,EL).

check_good_elf(E0, E) :-
    split_atom(E0, Thy, ':', Name),
    query_theory(elf(Name, _), Thy),
    !,
    check_or_error(
        current_predicate(E0/4),
        ['there is no Prolog procedure for elf: ',q(E0/4)]),
    E = E0.
check_good_elf(E0, E) :-
    atom(E0),
    query_ancestors(elf(E0, _), Thy),
    !,
    concat_atom([Thy,':',E0], E),
    check_or_error(
        current_predicate(E/4),
        ['there is no Prolog procedure for elf: ',q(E/4)]).
check_good_elf(E0, _) :-
    handle_error(['no such elf: ',q(E0)]).


proof command
  add_elf(Elf,Pos)
extends
  elves_cmds
doc
  'Inserts Elf into the current ',
    xref('elves list',global,'parameter:elves_list'),
  ' list before number Pos.'
mode
  add_elf(@atom, @integer).

add_elf(Elf0, Pos) :-
    check_good_elf(Elf0, Elf),
    get_elves_list(EL0),
    check_or_error(
        0 < Pos,
        ['illegal position: ',q(Pos),'.  Must be greater than zero.']),
    Pos1 is Pos - 1,
    insert_after_nth1(EL0, Pos1, [Elf], EL),
    set_elves_list(EL).

insert_after_nth1([], _, New, Out) :-
    !,
    Out = New.
insert_after_nth1(L, 0, New, Out) :-
    !,
    append(New, L, Out).
insert_after_nth1([E|EL], Pos, New, [E|Out2]) :-
    Pos > 0,
    Pos1 is Pos - 1,
    insert_after_nth1(EL, Pos1, New, Out2).


proof command
  delete_elf(Elf)
extends
  elves_cmds
doc
  'Delete Elf from the current ',
    xref('elves list',global,'parameter:elves_list'),
  '.'
mode
  delete_elf(@atom).

delete_elf(Elf0) :-
    check_good_elf(Elf0,Elf),
    get_elves_list(EL0),
    delete(Elf, EL0, EL),
    set_elves_list(EL).


active_elves parameter
  [set_active_elves, get_active_elves]
extends
  parameters
doc
  'The active_elves parameter determines the number of elves in the',nl,
  'current elves list that are active.  This parameter is an alias',nl,
  'for the first component of the `elves'' parameter.',nl,
  nl,
  'For example, if the value of active_elves is 3, then the first',nl,
  'three elves in the elves list will be called automatically',nl,
  'whenever Ergo is about to prompt for the next proof command.'.

set_active_elves(N) :-
    check_active_elves(N),
    ( active_elves(N) ->
        %% no changes needed
        true
    ;
        retractall(active_elves(_)),
        asserta(active_elves(N))
    ).

get_active_elves(N) :-
    active_elves(N1),
    !,
    N = N1.
get_active_elves(N) :-
    get_default_active_elves(N).

check_active_elves(N) :-
    check_or_error(
        ( integer(N), N >= 0 ),
        ['active elves must be set to a non-negative integer, not: ', q(N)]
    ).



proof command
  hints,
  hints(N)
extends
  elves_cmds
doc
  'Displays all suggestions from the first N elves in the current',nl,
  xref('elves list',global,'parameter:elves_list'),
  ', without trying to execute those',nl,
  'suggestions.  If N is not given, then all elves in the current',nl,
  'elves list are tried.'
mode
  hints,
  hints(@integer).

hints :-
    get_active_elves(Active),
    OneMore is Active + 1,
    hints(OneMore).

hints(N) :-
    check_or_error(
        ( integer(N), N >= 0 ),
        ['The number of elves you ask for hints must be an non-negative integer, not ', q(N)]
    ),
    get_elves_list(EL),
    is_focus(F),
    is_goal(G),
    nth1(Num,EL,Elf),
    Num =< N,
    call_predicate(Elf(F,G,Msg,Action)),
    writelist_indent([left(20,[w(Elf),':'])|Msg],[tab(20)]),
    writelist([nl,tab(20),q(Action),nl]),
    fail.
hints(_) :-
    nl.

proof command
  try_elves,
  try_elves(N)
extends
  elves_cmds
doc
  'Tries to execute the first N elves in the current',nl,
  xref('elves list',global,'parameter:elves_list'), '.',nl,
  'If N is not given, then all the active elves are tried:',nl,
  'this may be useful from a tactic.'
mode
  try_elves,
  try_elves(@integer).

try_elves :-
    get_active_elves(N),
    try_elves(N).

try_elves(N) :-
    check_or_error(
        ( integer(N), N >= 0 ),
        ['Number of active elves to try must be an integer, not ', q(N)]
    ),
    get_elves_list(EL),
    try_elves(N, EL, Done),
    !,
    nonvar(Done).

try_elves(N, EL, done) :-
    is_focus(F),
    is_goal(G),
    nth1(Num,EL,Elf),
    Num =< N,
    is_focus(F),
    is_goal(G),
    call_predicate(Elf(F,G,Msg,Action0)),
    Action0 \== close,
    msg(elf, [q(Elf),' suggests: '|Msg]),
    elves_deabbrev_cmd(Action0, Action),
    elves_do_cmd(Action, Action0, _),
    !,
    try_elves(N, EL, _).
try_elves(_, _, _).

proof command
  call_elf(Elf),
  call_elf(Elf,Action)
extends
  elves_cmds
doc
  'This command provides a way of executing a command suggested by a',nl,
  'given elf.  It is typically used after the hints command has',nl,
  'displayed a list of elf-generated suggestions.',nl,
  nl,
  'Elf must be the name of a elf (which need not be active).',nl,
  'If Action is specified, it must match the suggested action returned',nl,
  'by Elf.'
mode
  call_elf(@atom),
  call_elf(@atom,?code).

call_elf(Elf) :-
    call_elf(Elf, _).

call_elf(Elf, _) :-
    %% call_elf should have been expanded by elves_deabbrev_cmd,
    %% so it is strange to have reached here!
    %% The above declaration should perhaps not be a command declaration
    %% at all, to discourage other proof interfaces from calling call_elf.
    handle_error(program, ['call_elf(',q(Elf),' ...) was not expanded']).


%% elves(Tac).
%% elves(Tac, MinElf).
%% elves_user_only(Tac).
%%
%%  Instantiates Tac to a series of proof commands that will prove
%%  the current window, with help from the current list of `elves'
%%  (proof assistants) and from the user!
%%
%%  elves/1 tries elves MinElf..ActiveElves, then the user.
%%  elves/2 tries elves 1..ActiveElves, then the user.
%%  elves_user_only/1 tries only the user.
%%
%%  mode elves(?term).
%%  mode elves(?term, @integer).
%%  mode elves_user_only(?term).
%%
elves(Tac) :-
    elves(Tac, 1).

elves(Tac, _) :-
    nonvar(Tac),
    !,
    call_tac(Tac).
elves(Tac, MinElf) :-
    PRE('elves', var(Tac)),
    PRE('elves: min elf', (integer(MinElf),MinElf > 0)),
    catch(
        elves2(Tac, MinElf),
%        signal(Sig,_),
         abort_tactic(_),
        ( writelist([nl,'Continuing after abort_tac interrupt.',nl]),
          elves_user_only(Tac))).

elves_user_only(Tac) :-
    get_active_elves(Active),
    OneMore is Active + 1,
    elves(Tac, OneMore).


elves2(Tac, MinElf) :-
    ensure_constraints_accepted,
    debug(elves, ['calling: ', q(elves_get_cmd(MinElf,...))]),
    elves_get_cmd(MinElf, Cmd),
    debug(elves, ['calling: ', q(elves_deabbrev_cmd(Cmd,...))]),
    elves_deabbrev_cmd(Cmd, Cmd2),
    debug(elves, ['doing cmd: ',q(Cmd2)]),
    elves_do_cmd(Cmd2, Cmd, Tac),
    !,
    Tac \== fail.


%% elves_get_cmd(MinElf, Cmd).
%%
%%  Returns an infinite stream of proof commands, one by one,
%%  in Cmd.  The first few commands may be suggested by elves
%%  MinElf..ActiveElf, while all following commands will come
%%  from the user.
%%
%%  mode elves_get_cmd(@integer, -term).
%%
elves_get_cmd(MinElf, Cmd) :-
    PRE('elves_get_cmd', integer(MinElf)),
    %% try all elves with numbers in MinElf..Active.
    get_active_elves(Active),
    MinElf =< Active,
    is_focus(F),
    is_goal(G),
    get_elves_list(EL),
    nth1(I, EL, Elf),
    MinElf =< I, I =< Active,
    call_predicate(Elf(F,G,Msg,Cmd)),
    msg(elf, [q(Elf),' suggests: '|Msg]).
elves_get_cmd(_, Cmd) :-
    repeat,
    show_window,
    elves_prompt(Prompt, Thy),
    get_command(Prompt, Thy, Cmd),
    check_is_legal_cmd(Cmd).


%% elves_deabbrev_cmd(Cmd0, Cmd).
%%
%%  Checks that Cmd0 is a legal elves command and deabbreviates
%%  the command itself (eg. [2] becomes (focus([2]) ---> _)),
%%  then deabbreviates all its arguments according to the mode
%%  declaration of the command.
%%
%%  Some/most of these commands should probably be moved out
%%  into appropriate command tables rather than being built-in
%%  like this.
%%
%%  mode elves_deabbrev_cmd(+cmd, -code).
%%
elves_deabbrev_cmd(Cmd0, Cmd) :-
    var(Cmd0),
    !,
    Cmd = Cmd0.
elves_deabbrev_cmd(call_elf(Elf0), Impl) :-
    !,
    elves_deabbrev_cmd(call_elf(Elf0,_), Impl).
elves_deabbrev_cmd(call_elf(Elf0,Action), Impl) :-
    !,
    check_good_elf(Elf0, Elf),
    is_focus(F),
    is_goal(G),
    Elf(F,G,Msg,Action),
    msg(elf, [q(Elf),' suggests: '|Msg]),
    elves_deabbrev_cmd(Action, Impl).
elves_deabbrev_cmd(close, close) :-
    !.
elves_deabbrev_cmd(undo, undo) :-
    !.
elves_deabbrev_cmd(quit_proof, quit_proof) :-
    !.
elves_deabbrev_cmd(end_of_file, Tac) :-
    !,
    %% We don't want end_of_file to be too drastic (eg. abandon), since
    %% it is fairly easy to generate it by mistake.  However, if we get 
    %% it on standard input, we must eventually exit the proof (and the
    %% prover), otherwise we will just process an infinite stream of
    %% end_of_file commands.  So, we make it equivalent to `quit_proof'.
    %% (but don't advertise this fact).
    Tac = quit_proof.
elves_deabbrev_cmd((A0,B0), (A,B)) :-
    !,
    elves_deabbrev_cmd(A0, A),
    elves_deabbrev_cmd(B0, B).
elves_deabbrev_cmd(Cmd0 ---> SubProof0, Cmd) :-
    !,
    elves_deabbrev_cmd(Cmd0, Cmd),
    check_or_error(
        Cmd = (_ ---> SubProof),
        [q(Cmd0),' has no subproofs.']),
    %% Allow the [...] brackets to be omitted from singleton subproofs.
    ( var(SubProof0) ->
        SubProof1 = SubProof0
    ; list(SubProof0) ->
        SubProof1 = SubProof0
    ;
        SubProof1 = [SubProof0]
    ),
    deabbreviate_with_type(list(tactic), SubProof1, SubProof2),
    check_or_error(
        SubProof = SubProof2,
        ['wrong number of subproofs.']).
elves_deabbrev_cmd(hyp:Num, hyp(Num) ---> [_]) :-
    integer(Num),
    !.
elves_deabbrev_cmd(hyp:Term0, hyp(Term) ---> [_]) :-
    !,
    deabbreviate_with_type(thy(term), Term0, Term).
elves_deabbrev_cmd(Path, focus(Path) ---> [_]) :-
    list(Path),
    !.
elves_deabbrev_cmd(Thy:Cmd0, Cmd) :-
    !,
    ( Thy == global ->
        true
    ;
        check_is_theory_name(Thy)
    ),
    elves_deabbrev_cmd2(Thy, Cmd0, Cmd).
elves_deabbrev_cmd(Cmd0, Cmd) :-
    elves_deabbrev_cmd2(_, Cmd0, Cmd).


%% private elves_deabbrev_cmd2(Thy, Cmd, Impl).
%%
%%  Searches Thy for a unique command that matches the
%%  outermost functor of Cmd.  Then it deabbreviates Cmd according
%%  to the mode declarations of that command and returns the resulting
%%  executable Prolog code in Impl.
%%
%%  If Thy is nonvar, it must be 'global' or the name of a
%%  specific ancestor theory to search in.
%%
%%  mode elves_deabbrev_cmd2(?atom, +cmd, -code).
%%
elves_deabbrev_cmd2(Thy, Cmd0, Impl) :-
    check_is_legal_cmd(Cmd0),
    functor(Cmd0, F, A),
    functor(Cmd1, F, A),
    debug(elves, ['Trying to find: ',q(F/A),' in theory: ',q(Thy)]),
    findall(cmd(Cmd1,M,I),
            (member(Tab, [general,proof]),
             find_command(Cmd1, Tab, Thy, M, I, _)
            ),
            Cmds),
    debug(elves, ['Found proof commands: ', q(Cmds)]),
    ( Cmds = [] ->
        handle_error(['unknown command: ',q(F/A),'.  Try ''help.'''])
    ; Cmds = [C] ->
        C = cmd(Cmd1,Mode,Impl),
        debug(elves, ['deabbreviating: ', q(Cmd0),nl,'with mode: ',q(Mode)]),
        check_or_error(
            deabbreviate_cmd(Mode, Cmd0, Cmd1),
            ['bad arguments in command: ',q(Cmd0)])
    ;
        length(Cmds,Len),
        handle_error(['there are ',w(Len),' ',q(F/A),' commands.',nl,
                      'Use `Thy:Command...'' to specify which one you mean.'])
    ).


%% elves_do_cmd(Cmd, MsgCmd, Tac).
%%
%%  Executes Cmd and instantiates Tac to (Cmd,_) if it succeeds.
%%  If Cmd fails, a message is printed using MsgCmd, which should
%%  be a prettier equivalent of Cmd.
%%
%%  mode elves_do_cmd(+code, +code, -code).
%%
elves_do_cmd(undo, _, Tac) :-
    !,
    Tac = fail.
elves_do_cmd(abandon, _, _) :-
    !,
    abandon.
elves_do_cmd(quit_proof, _, _) :-
    in_proof(Thy:Name),
    !,
    announce_event(proof(Thy, Name, aborted)),
    throw(abort_proof(Thy:Name)).
elves_do_cmd(quit_tactic, _, _) :-
    in_proof(Thy:Name),
    !,
%    announce_event(proof(Thy, Name, aborted)),
    throw(abort_tactic(Thy:Name)).
elves_do_cmd(close, _, Tac) :-
    ( is_subproof_depth(1) ->
        check_focus_is_true
    ;
        true
    ),
    check_is_justified_focus,
    !,
    Tac = true.
elves_do_cmd(Cmd, MsgCmd, _) :-
    elves_non_proof_command(MsgCmd),
    !,
    debug(elves, ['calling non proof command: ', q(Cmd)]),
    call_tac(Cmd),
    !,
    % fail silently, so that we do not leave a choicepoint.
    fail.
elves_do_cmd(Cmd, _, Tac) :-
    Cmd \== close,
    debug(elves, ['calling command: ', q(Cmd)]),
    once(wrap_tac(Cmd)),
    Tac = (Cmd,Tac2),
    elves(Tac2).
elves_do_cmd(_, MsgCmd, _) :-
    writelist(['failed: ', q(MsgCmd),nl]),
    fail.


%% TO DO: It would be nice to be able to ??? commands in the table
%% elves_not_proof_command, perhaps by asking the core prover whether the
%% current proof has been changed at all by the execution of this command.
%% Other commands should be here - e.g., most of the emacs interface commands.
%% What is really needed is a way of dynamically extending the table.
%%
%% This table identifies those commands that do not change the
%% current proof state and therefore should not leave choicepoints.
%%
elves_non_proof_command(xyzzy).
elves_non_proof_command(help).
elves_non_proof_command(help(_)).
elves_non_proof_command(help(_,_)).
elves_non_proof_command(rule(_)).
elves_non_proof_command(rule(_,_)).
elves_non_proof_command(runtime).
elves_non_proof_command(show).
elves_non_proof_command(show(_)).
elves_non_proof_command(show(_,_)).
elves_non_proof_command(statistics).
elves_non_proof_command(hints).
elves_non_proof_command(hints(_)).

check_focus_is_true :-
    is_focus(F),
    atom(F),
    F = 'ergo.true',
    !.
check_focus_is_true :-
    handle_error(['Cannot close until proof is complete.']).



%% elves_prompt(Prompt, OpTable).
%%
%%  Determines the prompt for the elves user interface and
%%  which operator table should be used when reading commands.
%%
%%  mode elves_prompt(-list(term), -atom).
%%
elves_prompt([q(Depth),'% '], Thy) :-
    global_state_lookup(current_dir,Thy), Thy \= none,
    is_subproof_depth(Depth).
 
%% Listener for automatically splitting hypotheses as soon as they are added.
%% The listener is enabled by default, and can be disabled by setting the
%% implicit parameter split_hyp_disable to 'disable'.
?- add_listener(hyp(add(_),_,Num,'ergo.and'(_,_)),
                ( ip_lookup(split_hyp_disable, ASH),
                  ASH \== disable,
                  split_then_delete_hyp(Num)
               )).

split_then_delete_hyp(Num) :-
    split_hyp(Num),
    ( delete_hyp(Num) ->
        true
    ;
        true
    ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%  Documentation Nodes
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

elves_interface header
  'Proof commands (The ''Elves'' User Interface)'
extends
  master_help_node
doc
  'This section describes the `Elves'' user interface, which allows you to',nl,
  'build proofs.  In addition to using basic proof steps and tactics, this',nl,
  'interface provides a set of `elves'' for automating simple proof steps.',nl,
  nl,
  'The following groups of commands are available:',nl.


proof_info_cmds header
  'Proof Information Commands'
extends
  elves_interface
doc
  'The following commands display information about the current proof.',nl.


open_close_proofcmds header
  'Open/Close Subproof Commands'
extends
  elves_interface
doc
  'These commands are for creating new subproofs and exiting',nl,
  'the current subproof, either successfully or unsuccessfully.',nl,
  nl,
  '[Num,...].     Transform a subterm of the focus.',nl,
  nl,
  'hyp:Num.       Transform hypothesis Num.',nl,
  nl,
  'hyp:Pattern.   Transform the hypothesis that matches Pattern.',nl,
  nl,
  'close.         Try to close the current subproof.  This may fail',nl,
  '                 if there are outstanding conjectures or if you are',nl,
  '                 in the initial subproof and the focus is not ''true''.',nl,
  nl,
  'undo.          Undo the last command.  After all commands have been',nl,
  '                 undone, this will fail out of the current subproof.',nl,
  nl,
  'suspend.       This suspends the current proof and starts a new theory',nl,
  '                 management user interface.  To resume the current',nl,
  '                 proof, exit that interface using the resume/0 command.',nl,
  nl,
  'quit_proof.    This discards the current proof and returns to the',nl,
  '                 theory management user interface.',nl,
  nl,
  'abandon.       This exits the whole prover without saving anything!',nl,
  '                 WARNING: This discards all theory changes and proofs',nl,
  '                 done since the last save command.',nl.


focus_proofcmds header
  'Focus Transformation Commands'
extends
  elves_interface
doc
  'These commands are for transforming the focus to some new term.',nl.

hyp_proofcmds header
  'Hypothesis Commands'
extends
  elves_interface
doc
  'These commands are for adding/deleting/transforming hypotheses.',nl.

var_proofcmds header
  'Variable Instantiation Commands'
extends
  elves_interface
doc
  'These commands control the instantiation of proof variables',nl,
  '(both meta-variables and object-variables).  The proof variables',nl,
  'are split into two classes: `fixed variables'' and `new variables''.',nl,
  nl,
  'Fixed variables are those that appear in the original theorem.',nl,
  'They can only be instantiated manually, not automatically.',nl,
  nl,
  'New variables have typically been introduced during the proof.',nl,
  'They may be instantiated automatically during proof steps.',nl.

nfi_proofcmds header
  'Constraint Management Commands'
extends
  elves_interface
doc
  'These commands are for managing the not_free_in constraints',nl,
  'associated with the current theorem.',nl.

meta_proofcmds header
  'Meta-Commands (for combining other commands)'
extends
  elves_interface
doc
  'These are higher-order commands which take other commands as',nl,
  'arguments and apply them to the current proof in various ways.',nl.

elves_cmds header
  'Elf Management Commands'
extends
  elves_interface
doc
  'Elves are procedures that look at the current proof, especially',nl,
  'the current focus and goal, and suggest tactics and proof steps',nl,
  'that might be helpful.',nl,
  nl,
  'The following commands and parameters allow you to control which',nl,
  'elves may help with the current proof.',nl,
  'The ',xref('active elves',global,'parameter:active_elves'),
  ' will attempt to help',nl,
  'automatically, whereas other elves in the current',nl,
  xref('elves list',global,'parameter:elves_list'),
  ' will only offer suggestions when',nl,
  'the hints command is executed.',nl.

elves header
  'Elves (Proof Assistants)'
extends
  elves_interface
doc
  'The following elves are available for use within proofs.',nl.
