%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: emacs5.ql,v 1.1 2002/07/02 23:26:32 itp Exp $
%%
%% Support for Emacs interface for ergo5.

emacs_magic parameter
  [set_emacs_magic, current_emacs_magic]
doc
  'Control the generation of emacs magic commands.',nl,
  'Value is a list saying what kinds of magic to enable:',
  begin(list),
  item('  help:'),' help goes in own window, with clickable xrefs.',
  item('  xref:'),' help goes in transcript window, with clickable xrefs.',
  end(list),
  nl,
  'Alternatively, Value is the atom ''default'', in which case the',nl,
  'value is obtained from the parameter ',
  xref([default_emacs_magic],global,'parameter:default_emacs_magic'), '.', nl,
  '`set emacs_magic === default'' is generated automatically whenever the',nl,
  'Emacs interface starts Ergo.'
extends
  parameters.

default_emacs_magic parameter
  [set_default_emacs_magic, get_default_emacs_magic]
doc
  'Default value for ',
  xref([emacs_magic],global,'parameter:emacs_magic'), '.'
extends
  parameters.

?- dynamic(current_emacs_magic/1),
   ( global_fact(default_emacs_magic,1) ->
       true
   ;
       new_global_fact(default_emacs_magic,1,[save(true)]),
       assert_global(default_emacs_magic([help]))
   ).

set_emacs_magic(Default) :-
    Default == default,
    !,
    get_default_emacs_magic(Magic),
    set_emacs_magic(Magic).
set_emacs_magic(Kinds) :-
    diff_list(Kinds, [help,xref], []),
    !,
    retractall(current_emacs_magic(_)),
    assert(current_emacs_magic(Kinds)),
    delete_handlers(showing_help(_,_), emacs_showing_help(_,_)),
    delete_handlers(write(_, _, _), emacs_write_xref(_,_,_,_)),
    delete_handlers(write(_, _, _), emacs_write_emph(_,_)),
    ( member(help, Kinds) ->
        add_handler(write(Strm, _, xref(WL, Thy, Name)),
                    emacs_write_xref(Strm, WL, Thy, Name)),
        add_handler(write(Strm, _, emph(WL)),
                    emacs_write_emph(Strm, WL)),
        add_handler(showing_help(Action,Node),
                    emacs_showing_help(Action,Node))
    ; member(xref, Kinds) ->
        add_handler(write(Strm, _, xref(WL, Thy, Name)),
                    emacs_write_xref(Strm, WL, Thy, Name))
    ;
        true
    ).
set_emacs_magic(Kinds) :-
    handle_error(['Bad value ''', w(Kinds), '''',
                  ' for parameter ''emacs_magic''.']).

set_default_emacs_magic(Magic) :-
    retractall_global(default_emacs_magic(_)),
    assert_global(default_emacs_magic(Magic)).

get_default_emacs_magic(Magic) :-
    query_global(default_emacs_magic(Magic)).


%% emacs_magic(Command).
%%
%%  Send MAGIC COMMAND to emacs.
%%
%%  mode emacs_magic(@atom).
%% 
emacs_magic(Command) :-
    write('MAGIC '), write(Command), nl.


%%general command
%%  send_theories
%%doc
%%  'Send the names of all theories to emacs.'
%%mode
%%  send_theories.

send_theories :-
    emacs_magic('THEORIES'),
    ( query_global(ancestor(Thy,_)),
      writeq(Thy),
      nl,
      fail
    ;
      true
    ),
    write(global), nl,
    emacs_magic('NORMAL').

%% emacs_showing_help(Action, Doc).
%%
%%  Start/end output to help buffer with heading Doc.
%%
emacs_showing_help(start, Doc) :-
    emacs_magic('HELP'),
    writelist(['%%%%node: ' | Doc]),
    nl.
emacs_showing_help(end, _) :-
    emacs_magic('NORMAL').
    
%% emacs_write_xref(WriteList, Thy, Name).
%% emacs_write_xref(Stream, WriteList, Thy, Name).
%%
%%  Write and mouse-sensitize a help cross-reference.
%%  If Stream is specified, fail unless Strem is the standard output.
%%
%%  mode emacs_write_xref(@writelist, @thyname, @term).
%%  mode emacs_write_xref(@stream, @writelist, @thyname, @term).
%%
emacs_write_xref(Strm, WL, Thy, Name) :-
    stream_property(Strm, alias(user_output)),
    emacs_write_xref(WL, Thy, Name).

emacs_write_xref(WL, Thy, Name) :-
    emacs_magic('XREF'),
    writelist(WL), nl,
    write('COMMAND'), nl,
    writelist(['help(', q(Name), ', ', q(Thy), ')']),
    emacs_magic('END').

%% emacs_write_emph(Stream, WriteList).
%%
%%  Write and emphasize some text.
%%
%%  mode emacs_write_emph(@stream, @writelist).
%%
emacs_write_emph(Strm, WriteList) :-
    stream_property(Strm, alias(user_output)),
    !,
    emacs_magic('EMPH'),
    writelist(WriteList),
    emacs_magic('END').

general command
  synchronize_emacs(Tag)
doc
  'Used by the emacs interface to synchronize the processes.'
mode
  synchronize_emacs(@integer).

synchronize_emacs(Tag) :-
    emacs_magic('SYNC'),
    write(Tag),
    emacs_magic('NORMAL').

%% emacs_echo_input(Caller, Input).
%%
%%  Caller has just read Input.
%%  Echo it.
%%
%%  mode emacs_echo_input(@term, @term).
%%
emacs_echo_input(get_command(_Prompt, OpTable, _Term), Term) :-
    write_term(user_output, Term,
               [remember_name(true),
                op_table(OpTable),
                obvar_prefix_table(OpTable)]),
    write(user_output, '.'),
    nl.
emacs_echo_input(get_ergo_line(_,_), Line) :-
    write_chars(Line, user_output, noquotes),
    nl.

%% emacs_reader_special_execute_hook(Cmd).
%%
%%  Handler for reader special execute commands.
%%  These commands will be generated by the Emacs interface.
%%
emacs_reader_special_execute_hook(send_theories) :-
    send_theories.
emacs_reader_special_execute_hook(help(T)) :-
    do_help(T,_).
emacs_reader_special_execute_hook(help(T,Thy)) :-
    do_help(T,Thy).
emacs_reader_special_execute_hook(set_parameter(P,V)) :-
    set_parameter(P, V).
emacs_reader_special_execute_hook(synchronize_emacs(Tag)) :-
    synchronize_emacs(Tag).
emacs_reader_special_execute_hook(emacs_do_deferred_output) :-
    emacs_do_deferred_output.

emacs_do_deferred_output.


general command
  emacs_button(Label, Command)
extends
  general_cmds
doc
  'Add or remove a button from the Emacs interface.',nl,
  ' Label is the label of the button to add or remove.',nl,
  ' Command is an atom giving the command the new button should generate',nl,
  '  or a variable if a button should be deleted.'
mode
  emacs_button(@atom, @atom),
  emacs_button(@atom, @var).

emacs_button(Label, Command) :-
    retractall_global(emacs_button(Label, _)),
    ( var(Command) ->
        Cmd = ''
    ;
        Cmd = Command,
        tactic_assert_global(emacs_button(Label, Cmd))
    ),
    emacs_add_button(Label, Cmd).

emacs_add_button(Label, Cmd) :-
    emacs_magic('BUTTON'),
    writelist([w(Label), :, w(Cmd), nl]),
    emacs_magic('NORMAL').

emacs_add_saved_buttons :-
    (
        query_global(emacs_button(Label, Cmd)),
        emacs_add_button(Label, Cmd),
        fail
    ;
        true
    ).

%% emacs_log(WriteList).
%%
%%  Write WriteList into the proof script buffer, if there is one and logging
%%  is enabled.
%%
%%  mode emacs_log(@writelist).
%%
emacs_log(WL) :-
    emacs_magic('LOG'),
    writelist(WL),
    emacs_magic('NORMAL').

emacs_prover_hook(exit).
emacs_prover_hook(resume) :-
    emacs_prover_hook(startup).
emacs_prover_hook(startup) :-
    add_handler(reader_special_execute_hook(emacs(Call)),
                emacs_reader_special_execute_hook(Call)).

library command
  emacs_send_string(WL)
extends
  proof_info_cmds
doc
  'Send WL as an echoed command to Ergo, via Emacs.'
mode
  emacs_send_string(@writelist).

emacs_send_string(WL) :-
    emacs_magic('SENDSTRING'),
    writelist(WL),
    emacs_magic('NORMAL').  

?- add_listener(prover(Event), emacs_prover_hook(Event)).
