%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##


%% Support for Emacs interface: write a term with an index.
%% $Id: emacs_index.ql,v 1.29 2000/11/15 06:29:49 itp Exp $

#include "../prover/lib/ergo.h"

?- require_global_tactic_file(pp).

% fail over write to emacs in order to save memory
indprint(Focus) :-
    once(do_indprint(Focus)),
    fail.

indprint(_).

do_indprint(Focus) :-
    check_current_theory(CD),
    get_write_depth(D),
    abbreviate(Focus, CD, [], D, Focus1),
    tokenize(Focus1, CD, TokTree, TokChars),
    typeset(TokTree, 60),
    !,
    emacs_magic('FOCUS'),
    write_tokens(TokTree, TokChars, 8, Index),
    nl,
    emacs_magic('INDEX'),
    emacs_write_index(Index),
    emacs_magic('NORMAL').

emacs_write_index([]).
emacs_write_index([i(Path,Start,End)|Is]) :-
    write_index_line(user_output, Start, End, Path),
    emacs_write_index(Is).

%% fail after doing the write to save memory (Luke)
noindprint(Focus) :-
    once(do_noindprint(Focus)),
    fail.

noindprint(_).

do_noindprint(Focus) :-
    check_current_theory(CD),
    get_write_depth(D),
    abbreviate(Focus, CD, [], D, Focus1),
    tokenize(Focus1, CD, TokTree, TokChars),
    typeset(TokTree, 1000),
    write_tokens(TokTree, TokChars, 8, _).

%    write_term(Focus1, [op_table(CD), obvar_prefix_table(CD),
%                        quoted(true), remember_name(true)]).

%% write_index_line(Strm, Start, End, Path).
%%
%%  Removes zeroes from paths.
%%  Writes nothing if Strm is 'null'.
%% 
write_index_line(null, _, _, _) :-
    !.
write_index_line(Strm, Start, End, Path0) :-
    PRE('write_index_line', Strm \== null),
    delete_all(0, Path0, Path),
    !,
    write_encoded_number(2, Strm, Start),
    write_encoded_number(2, Strm, End),
    write_encoded_path(Strm, Path),
    %% writelist(Strm, [w(Start), ' ', w(End), ' ', w(Path)]),
    nl(Strm).
write_index_line(_, _, _, _).  

%% reverse_nonzero(In, Out0, Out).
%%
%%  Out-Out0 is In reversed, and In contains no 0.
%%
reverse_nonzero([], Out, Out).
reverse_nonzero([In|Ins], Out0, Out) :-
    In \== 0,
    !,
    reverse_nonzero(Ins, [In|Out0], Out).

%% write_encoded_number(NumChars, Strm, Number).
%% write_encoded_path(Strm, Path).
write_encoded_path(_Strm, []).
write_encoded_path(Strm, [Pos|Path]) :-
    write_encoded_number(1, Strm, Pos),
    write_encoded_path(Strm, Path).

write_encoded_number(1, Strm, Num) :-
    Num < 8'100,
    !,
    Enc is Num + 0'0,
    put_code(Strm, Enc).
write_encoded_number(2, Strm, Num) :-
    Num < 8'10000,
    !,
    Enc1 is (Num >> 6)  + 0'0,
    Enc2 is (Num /\ 63) + 0'0,
    put_code(Strm, Enc1),
    put_code(Strm, Enc2).
write_encoded_number(3, Strm, Num) :-
    Num < 8'1000000,
    !,
    Enc1 is (Num >> 12)        + 0'0,
    Enc2 is ((Num >> 6) /\ 63) + 0'0,
    Enc3 is (Num /\ 63)        + 0'0,
    put_code(Strm, Enc1),
    put_code(Strm, Enc2),
    put_code(Strm, Enc3).
write_encoded_number(4, Strm, Num) :-
    Num < 8'100000000,
    !,
    Enc1 is (Num >> 18)         + 0'0,
    Enc2 is ((Num >> 12) /\ 63) + 0'0,
    Enc3 is ((Num >> 6) /\ 63)  + 0'0,
    Enc4 is (Num /\ 63)         + 0'0,
    put_code(Strm, Enc1),
    put_code(Strm, Enc2),
    put_code(Strm, Enc3),
    put_code(Strm, Enc4).
write_encoded_number(_NumChars, Strm, Num) :-
    write(Strm, #),
    write_encoded_number(4, Strm, Num).
