%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: focus.ql,v 1.48 2000/11/15 06:29:50 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file defines the 'focus' and 'hyp' tactics, which provide higher
%% level interfaces to the opening procedures open_pos and open_hyp.
%% They also maintain an `explanation' stack which records the
%% most recent focus of each user-visible subproof and the reason
%% for opening each subproof.  The show_subproof_stack commands
%% display this stack.
%%
%% It also defines procedures for managing rule groups and using
%% rule groups to perform simple simplifications.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../prover/lib/ergo.h"

?- require_global_tactic_file(abbrev).

?- add_listener(proof(_,_,started), ip_set(focus_stack,[])).
 

%% setup_focus.
%%
%%  Initialises the Ergo database to allow rulegroups.
%%  This is typically only called when building a new Ergo
%%  from an empty global.erg file.  However, repeated calls
%%  do no harm.
%%
%%  mode setup_focus.
%%
setup_focus :-
    (theory_tactic_fact(rulegroup,7) ->
        true
    ;
        new_theory_fact(rulegroup,7,[save(true)])
    ),
    (query_global(announcement(rulegroup(_, _, _, _, _, _:_, _), _)) ->
        true
    ;
        add_announcement(
          rulegroup(_, _, _, _, _, _:_, _),
          ['rulegroup(Group, From, Conds, Reln, Out, ThmThy:Name, Dir).',nl,
           'The rulegroup has been added to the current theory.',nl])
    ).

?- setup_focus.

subterm_path header
  'Subterm Paths'
extends
  concepts
doc
  'A subterm path is a list of numbers and t(_) and b(_) terms that',nl,
  'specifies a subterm of some term, typically the current focus.',nl,
  nl,
  'The following table illustrates how paths specify subterms.',nl,
  nl,
  begin(verb),nl,
  '    Term             Subterm        Path',nl,
  '    ----             -------        ----',nl,
  '    X                X              []',nl,
  '    f(X,Y)           X              [1]',nl,
  '    f(X,Y)           Y              [2]',nl,
  '    f(a(c),f(d,e))   a(c)           [1]',nl,
  '    f(a(c),f(d,e))   c              [1,1]',nl,
  '    f(a(c),f(d,e))   f(d,e)         [2]',nl,
  '    f(a(c),f(d,e))   d              [2,1]',nl,
  '    f(a(c),f(d,e))   e              [2,2]',nl,
  '    q [] B           B              [2] or [b([])]',nl,
  '    q [x] B          B              [2] or [b([x])]',nl,
  '    q [x:T,y] B      B              [2] or [b([x,y])]',nl,
  '    q [x:T,y] a(c)   c              [2,1] or [b([x,y]),1]',nl,
  '    q [x:T] B        T              [t(1)] or [1,1,2]',nl,
  '    q [x:T,y:U] B    U              [t(2)] or [1,2,2]',nl,
  '    q [x,y,z:U] B    U              [t(3)] or [1,3,2].',
  end(verb).


%% TO DO: Should the parameter simplification_rules be made more persistent
%% or should we add default_simplification_rules parameter???

%% TO DO: Should be support a list of rule groups in the parameter
%% simplification_rules?

simplification_rules parameter
  [set_simplification_rules, get_simplification_rules]
doc
  'The name of the rulegroup used by ''focus'' for automatic simplifications.',
      nl,
  'The value of this parameter persists only within current proof.'
extends
  parameters.

%% set_simplification_rules(RuleGroup).
%% get_simplification_rules(RuleGroup).
%%
%%  Set/get the name of the rule group used by focus/2 for automatic
%%  simplifications.
%%
%%  mode set_simplification_rules(@rulegroup).
%%  mode get_simplification_rules(-rulegroup).
%%
set_simplification_rules(G) :-
    ip_set(simplification_rules, G).

get_simplification_rules(G) :-
    ip_lookup(simplification_rules, G).


proof command
  focus(Path)
implementation
  focus(Path) ---> [_]
command
  focus(Path,Rulegroup)
implementation
  focus(Path,Rulegroup) ---> [_]
extends
  open_close_proofcmds
doc
  'Opens a subwindow on the subterm of the current focus',nl,
  'that is specified by ',xref('Path',global,subterm_path),'.',nl,
  nl,
  'It automatically applies simplification rules from the',nl,
  'given rulegroup after the subproof has been completed.',nl,
  'By default, the ''simplify'' rulegroup is used.',nl,
  'Specify the ''empty'' rulegroup if you want no simplifications.'
mode
  focus(+subterm_path),
  focus(+subterm_path,@rulegroup).

focus(Path, Code) :-
    get_simplification_rules(RuleGroup),
    nonvar(RuleGroup),
    !,
    focus(Path, RuleGroup, [q(focus(Path,RuleGroup))], Code).
focus(Path, Code) :-
    focus(Path, simplify, [q(focus(Path,simplify))], Code).

focus(Path, RuleGroup, Code) :-
    focus(Path, RuleGroup, [q(focus(Path,RuleGroup))], Code).

focus(Path, RuleGroup, Explanation, [Tac]) :-
    !,
    ip_lookup(focus_stack, Stk),
    is_subproof_depth(Depth),
    is_subproof_id(ID),
    is_focus(F),
    ip_set(focus_stack, [focus(Depth,ID,F,Path,Explanation) | Stk]),
    focus2(Path, RuleGroup, Tac),
    ip_set(focus_stack, Stk).
focus(_, _, _, Code) :-
    handle_error(['focus/4, illegal tactic argument: ',code(Code)]).


focus2([], _, Tac) :-
    call_tac(Tac).
focus2([A|B], RuleGroup, Tac) :-
    open_pos([A|B], RestOfPath, 
             ( %% split_all_hyps,
               focus2(RestOfPath, RuleGroup, Tac))
            ),
    do_simplifications(RuleGroup).


proof command
  hyp(Hyp)
implementation
  hyp(Hyp) '--->' [_]
extends
  open_close_proofcmds
doc
  'Opens a subwindow on the hypothesis specified by Hyp.',nl,
  'Hyp may be an integer, or a term to match against the',nl,
  'hypotheses of the current window.'
mode
  hyp(@integer),
  hyp(?thy(term)).

hyp(Hyp, [Tac]) :-
    integer(Hyp),
    !,
    hyp(Hyp, ['hyp(',q(Hyp),')'], _, [Tac]).
hyp(Hyp, [Tac]) :-
    hyp(Hyp, ['hyp(',thy(Hyp),')'], _, [Tac]).

%%  The following forms of hyp can be useful when writing tactics.
%%
%%  mode hyp(?hypname, @writelist, ?list(tactic)).
%%  mode hyp(?hypname, @writelist, -integer, ?list(tactic)).
%%
hyp(Hyp, Explanation, [Tac]) :-
    hyp(Hyp, Explanation, _, [Tac]).

hyp(Hyp, Explanation, NewHypNum, [Tac]) :-
    ip_lookup(focus_stack, Stk),
    is_subproof_depth(Depth),
    is_subproof_id(ID),
    is_focus(F),
    ip_set(focus_stack, [focus(Depth,ID,F,hyp:Hyp,Explanation) | Stk]),
    open_hyp(hyp:Hyp, Tac, NewHypNum, _),
    %% split_all_hyps,
    ip_set(focus_stack, Stk).


database command
  rulegroup(Group,Thm,Direction),
  rulegroup(Group,Thm,Direction,Input)
implementation
  deabbreviate(Input, Input1),
  rulegroup(Group,Thm,Direction,Input1)
extends
  theory_extension_cmds
mode
  rulegroup(@atom, @rulename, @direction),
  rulegroup(@atom, @rulename, @direction, @thy(term))
doc
  'Add Thm to the rulegroup called Group.',nl,
  'Direction must be ''f'', ''r'', ''t'' or ''rt''.',nl,
  'Input must unify with the input of the rule when it is',nl,
  'split into components according to Direction.  Typically,',nl,
  'Input is used to decide whether or not the rule should be',nl,
  'a conditional one.'.

rulegroup(Group, Theorem, Dir) :-
    rulegroup(Group, Theorem, Dir, _).

rulegroup(Group, Theorem, Dir, From) :-
    check_current_theory(Thy),
    check_atom(Group),
    check_or_error(
        Group \== empty,
        ['cannot add rules to rulegroup empty.']),
    check_or_error(
        member_var(Dir, ['f','r','t','rt']),
        ['rulegroup direction ',q(Dir),' should be f/r/t/rt.']),
    decode_rule(Theorem, ThmThy, Name),
    (var(ThmThy) ->
        ThmThy = Thy
    ;
        true
    ),
    find_rule(ThmThy:Name, Form, _),
    freeze_vars(Form),
    !,
    choose_conditional(_, Form, Conds, Concl),
    choose_direction(Dir, Concl, From, Reln, Out),
    is_preorder(Reln),
    !,
    tactic_assert_theory(rulegroup(Group, From, Conds, Reln, Out, ThmThy:Name, Dir)),
    announce_event(rulegroup(Group, From, Conds, Reln, Out, ThmThy:Name, Dir)).


%% do_simplifications(RuleGroup).
%%
%%  Try applying all any unconditional rules in RuleGroup that
%%  match the current focus.
%%
%%  mode do_simplifications(?atom).
%%
do_simplifications(empty) :-
    !.
do_simplifications(Group) :-
    is_focus(From),
    nonvar(From),
    \+ From == 'ergo.true',
    query_ancestors(rulegroup(Group, From, 'ergo.true', _, _, Name, Dir)),
    trans(Name, Dir, [], _),
    !,
    do_simplifications(Group).
do_simplifications(_).


%% split_all_hyps.
%%
%%  Recursively split all conjunct hypotheses.
%%
%%  Known problems: The current implementation splits non-deletable
%%        conjunct hypotheses every time it is called.  It should
%%        have some way of realising that they have already been split.
%% 
%%  mode split_all_hyps.
%%
split_all_hyps :-
    split_all_hyps([]).

split_all_hyps(Done) :-
    split_hyp(Num),
    \+ member(Num,Done),
    !,
    split_all_hyps([Num|Done]).
split_all_hyps(_).

%%%%%%%%%%%%%%%%% OBSOLETE demo3.x proof commands. %%%%%%%%%%%%%%%%%%

trans_r(Rule)                   :- obsolete(trans_r(Rule), 
                                            trans(Rule,r)).
trans_r(Rule,Ant)               :- obsolete(trans_r(Rule,Ant),
                                            trans(Rule,r,Ant)).
trans_dir(Rule,Dir)             :- obsolete(trans_dir(Rule,Dir),
                                            trans(Rule,Dir)).
trans_dir(Rule,Dir,Ant)         :- obsolete(trans_dir(Rule,Dir,Ant),
                                            trans(Rule,Dir,Ant)).
trans_true(Rule)                :- obsolete(trans_true(Rule),
                                            trans(Rule,t)).
trans_true(Rule,Ant)            :- obsolete(trans_true(Rule,Ant),
                                            trans(Rule,t,Ant)).

trans_goal(Rule,G)              :- obsolete(trans_goal(Rule,G)).
trans_goal(Rule,G,Ant)          :- obsolete(trans_goal(Rule,G,Ant)).

h_trans_r(Name,Rule)            :- obsolete(h_trans_r(Name,Rule)).
h_trans_r(Name,Rule,Ant)        :- obsolete(h_trans_r(Name,Rule,Ant)).
h_trans_dir(Name,Rule,Dir)      :- obsolete(h_trans_dir(Name,Rule,Dir)).
h_trans_dir(Name,Rule,Dir,Ant)  :- obsolete(h_trans_dir(Name,Rule,Dir,Ant)).
h_trans_goal(Name,Rule,G)       :- obsolete(h_trans_goal(Name,Rule,G)).
h_trans_goal(Name,Rule,G,Ant)   :- obsolete(h_trans_goal(Name,Rule,G,Ant)).
h_trans_true(Name,Rule)         :- obsolete(h_trans_true(Name,Rule)).
h_trans_true(Name,Rule,Ant)     :- obsolete(h_trans_true(Name,Rule,Ant)).

d_trans_r(Name,Rule)            :- obsolete(d_trans_r(Name,Rule)).
d_trans_r(Name,Rule,Ant)        :- obsolete(d_trans_r(Name,Rule,Ant)).
d_trans_dir(Name,Rule,Dir)      :- obsolete(d_trans_dir(Name,Rule,Dir)).
d_trans_dir(Name,Rule,Dir,Ant)  :- obsolete(d_trans_dir(Name,Rule,Dir,Ant)).
d_trans_true(Name,Rule)         :- obsolete(d_trans_true(Name,Rule)).
d_trans_true(Name,Rule,Ant)     :- obsolete(d_trans_true(Name,Rule,Ant)).

%% TO DO: Tidy up the command focus_substring.
%% Ray's last comment: Looks pretty good to me?!

proof command
  focus_substring(String)
extends
  open_close_proofcmds
mode
  focus_substring(+chars)
wrapper
  ['Focussing on ',chars(String)] ---> [[]]
doc
  'focus_substring(String) ---> SubProof',nl,
  ' -- focus on the smallest subterm of the focus whose printed',nl,
  '    representation completely contains String.'.

focus_substring(Str, Tac) :-
    is_focus(F),
    subterm_substring(F, Str, Pos),
    focus(Pos, Tac).

% subterm_substring(+Term, +Str, -Pos)
% Pos is the position in Term of the smallest subterm whose printed
% representation contains Str.
subterm_substring(Term, Str, Pos) :-
    ( subterm_substring(Term, Str, [], Pos1) ->
        reverse(Pos1, Pos)
     
    ;
        handle_error(['No subterm of ', w(Term), ' contains ', chars(Str), nl])
    ).

subterm_substring(Term, Str, Pos0, Pos) :-
    abbreviate(Term, Term1),
    term_to_chars(Term1, Chars),
    substring(Str, Chars),
    ( subterm_path(Term, [N], Sub),
      subterm_substring(Sub, Str, [N|Pos0], Pos) ->
          true
    ;
        Pos = Pos0
    ).

term_to_chars(Term, Chars) :-
    open_string(write, Strm),
    write(Strm, Term),
    stream_to_chars(Strm, Chars).
%   close(Strm).

substring(Sub, Str) :-
    append([_, Sub, _], Str).
