%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: gumtree_compiler.ql,v 1.15 2006/02/06 00:52:13 itp Exp $
%%
%% Gumtree tactic compiler.
%%
%% The compiler converts gumtree tactics into Prolog code
%% and carries out some partial evaluation on the converted code.
%%

%% Compile a file of tactics to Prolog clauses.

gum_compile(File, OutFile) :-
    load_tactic_declarations(File),
    open(File, read, Str),
    open(OutFile, write, OutStr),
    once((global_state_lookup(current_dir,Thy) ; Thy = global)),
    repeat,
    read_term(Str, T, [op_table(Thy), variable_names(VNames)]),
    ip_set('@gumtree_consult_vlist', VNames),
    (   T = end_of_file
	->
	    true
	;
	T = tactic(_)
	->
	    (
	        gumtree_comp(T, Code) 
		->
		    out_code(OutStr, Code)
	        ;
		    write_term_list(['cannot translate: ', wR(T), nl]),
		    fail
	    )
	;
	    encoded_write_term(OutStr, '?-'(T), []),
	    fail
    ),
    !,
    close(Str),
    close(OutStr).

%%
%% Write the Prolog clauses for a gumtree tactic to the stream.
%%
out_code(OutStr, Code) :-
    member(C, Code),
    encoded_write_term(OutStr, C, []),
    fail.

%%
%% gumtree_comp(Tac, Code) compiles
%% the tactic declaration/definition Tac into Code.
%%
gumtree_comp(tactic(delete(_)), []) :- !.
gumtree_comp(Tac, Code) :-
    Tac = tactic(doc(Head, mode(Doc, arity(Mode, Arity)))),
    !,
    ip_lookup('@gumtree_consult_vlist', VNames),
    Code = ['?-'(user_tactic_and_doc_decl(VNames, Head, Doc, Mode, Arity))].
gumtree_comp(Tac, Code) :-
    Tac = tactic(mode(Head, arity(Mode, Arity))),
    !,
    Code = ['?-'(tactic(mode(Head, arity(Mode, Arity))))].
gumtree_comp(tactic('==='(Head, T)), Final) :-
    define_tactic0(Head, T, NewHead, Body),
    Body = expand_tac_body_and_exec(Tac, In, InT, Out, OutT),
    query_global(gumtree_tactic_declaration(Head, _, '--->'(InA, OutA))),
    build_arity_code(InX, InTX, OutX, OutTX, InA, OutA, ICode, OCode),
    gumtree_trans(Tac, InX, InTX, OutX, OutTX, Code1, OCode),
    build_unify_goals([In, InT, Out, OutT], [InX, InTX, OutX, OutTX], UG),
    append([gum_unify(UG)|ICode], Code1, Code2),
    gumtree_optimize(Code2, Code3),
    gumtree_to_prolog_list(Code3, Code4),
    optimize_prolog([[NewHead|Code4]], Code5),
    fix_gum_unify(Code5, Code6),
    !,
    build_clauses(Code6, Final).

%%
%% build_arity_code(In, InT, Out, OutT, InA, OutA, ICode, OCode)
%% gererates code to be added to generated code when either the
%% input or output arity is given.
%%
build_arity_code(In, InT, Out, OutT, InA, OutA, ICode, OCode) :-
    ( integer(InA) ->
        ICode = [gumtree_length(In, InT, InA)]
    ;
        ICode = []
    ),
    ( integer(OutA) ->
        OCode = [gumtree_unify(Out, OutX), gumtree_unify(OutT, OutTX), 
                 gumtree_length(OutX, OutTX, OutA)]
    ;
        OCode = []
    ).
        
%%
%% A DCG for generating a list of Prolog calls
%% This is used to translate the tactic into equivalent Prolog code.
%%
gumtree_trans({G}, In, InT, Out, OutT) -->
    !,
    {
      prolog_code_to_list(G,GList)
    },
    [
     GList,
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT)
    ].
gumtree_trans(term(T), In, InT, Out, OutT) -->
    !,
    [
     gumtree_unify(In, [Id|InT]),
     nonvar(Id),
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT),
     proof_node_term(Id,T)
    ].
gumtree_trans(name(Id), In, InT, Out, OutT) -->
    !,
    [
     gumtree_unify(In, [Id1|InT]),
     nonvar(Id1),
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT),
     record_proof_node_names([Id1], [Id]),
     Id1=Id
    ].
gumtree_trans(terms(T), In, InT, Out, OutT) -->
    !,
    [
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT),
     diff_list_to_list(In, InT, List),
     proof_node_terms(List,T)
    ].
gumtree_trans(names(T), In, InT, Out, OutT) -->
    !,
    [
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT),
     diff_list_to_list(In, InT, List),
     record_proof_node_names(List, T),
     List = T
    ].
gumtree_trans(premises(T), In, InT, Out, OutT) -->
    !,
    [
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT),
     gumtree_get_abbrev_premises(T1),
     structural_unify(T, T1)
    ].
gumtree_trans(context(K,I,T), In, InT, Out, OutT) -->
    !,
    [
     gumtree_unify(In, [Id|InT]),
     nonvar(Id),
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT),
     context_search(K, Id, I, T)
    ].
gumtree_trans(context_terms(CT), In, InT, Out, OutT) -->
    !,
    [
     gumtree_unify(In, [Id|InT]),
     nonvar(Id),
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT),
     collect_context_terms(Id, CT)
    ].
gumtree_trans(zero, In, InT, Out, OutT) -->
    !,
    [
     gumtree_unify(In, InT),
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT)
    ].
gumtree_trans(one, In, InT, Out, OutT) -->
    !,
    [
     gumtree_unify(In, [Id|InT]),
     nonvar(Id),
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT)
    ].
gumtree_trans(two, In, InT, Out, OutT) -->
    !,
    [
     gumtree_unify(In, [Id1, Id2|InT]),
     nonvar(Id1),
     nonvar(Id2),
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT)
    ].
gumtree_trans(fix(N), In, InT, Out, OutT) -->
    !,
    [
     gumtree_length(In, InT, N),
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT)
    ].
gumtree_trans(skip, In, InT, Out, OutT) -->
    [
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT)
    ].
gumtree_trans(use(Q), In, InT, Out, OutT) -->
    [
     gumtree_unify(In, [X|InT]),
     nonvar(X),
     quse(Q, X, Y),
     gumtree_append(Y, OutT, Out)
    ].
gumtree_trans(use(Q, O), In, InT, Out, OutT) -->
    [
     gumtree_unify(In, [X|InT]),
     nonvar(X),
     quse(Q, X, O, Y),
     gumtree_append(Y, OutT, Out)
    ].
gumtree_trans(iuse(Q), In, InT, Out, OutT) -->
    [
     gumtree_unify(In, [X|InT]),
     nonvar(X),
     qiuse(Q, X, Y),
     gumtree_append(Y, OutT, Out)
    ].
gumtree_trans(iuse(Q, O), In, InT, Out, OutT) -->
    [
     gumtree_unify(In, [X|InT]),
     nonvar(X),
     qiuse(Q, X, O, Y),
     gumtree_append(Y, OutT, Out)
    ].
gumtree_trans(rules(Q), In, InT, Out, OutT) -->
    [
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT),
     rulebrowse(Q, In, InT)
    ].
gumtree_trans(irules(Q), In, InT, Out, OutT) -->
    [
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT),
     qirule(Q, In, InT)
    ].
gumtree_trans(fails(T), In, InT, Out, OutT) -->
    !,
    {
     gumtree_trans(T, In1, InT1, _Out1, _OutT1, Code, [])
    },
    [
     fails(['gum='(In, In1), 'gum='(InT, InT1)|Code]),
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT)
    ].
gumtree_trans(succs(T), In, InT, Out, OutT) -->
    !,
    {
     gumtree_trans(T, In, InT, _Out1, _OutT1, Code, [])
    },
    [
     succs(Code),
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT)
    ].
gumtree_trans(rule(R), In, InT, Out, OutT) -->
    !,
    %
    % Note that if the rule is if the form Thy:Name then 
    % a call to decode_rule is avoided
    %
    {
      (   var(R)
	  ->
              PremLen = true,
	      Decode = decode_rule(R, Thy, Name)
	  ;
          R = Thy:Name, atom(Thy), (atom(Name); std_compound(Name))
	  ->
	      Decode = true,
              ( rule_query(Thy, Name, _, _, Seq, _) ->
                  length(Seq, Len),
                  PremLen = gumtree_length(Out, OutT, Len)
              ;
                  PremLen = true,
                  writelist(['WARNING: Can''t find rule ', q(R), nl])
              )
	  ;
              PremLen = true,
	      Decode = decode_rule(R, Thy, Name)
      )
    },
    [
     gumtree_unify(In, [InGoal|InT]),
     nonvar(InGoal),
     Decode,
     ip_set('rule@unify@flag', unify),
     proof_step(Thy, Name, InGoal, Children),
     PremLen,
     gumtree_append(Children, OutT, Out)
    ].
gumtree_trans(srule(R), In, InT, Out, OutT) -->
    !,
    %
    % Note that if the rule is if the form Thy:Name then 
    % a call to decode_rule is avoided
    %
    {
      (   var(R)
	  ->
              PremLen = true,
	      Decode = decode_rule(R, Thy, Name)
	  ;
          R = Thy:Name, atom(Thy), (atom(Name); std_compound(Name))
	  ->
	      Decode = true,
              ( rule_query(Thy, Name, _, _, Seq, _) ->
                  length(Seq, Len),
                  PremLen = gumtree_length(Out, OutT, Len)
              ;
                  PremLen = true,
                  writelist(['WARNING: Can''t find rule ', q(R), nl])
              )
	  ;
              PremLen = true,
	      Decode = decode_rule(R, Thy, Name)
      )
    },
    [
     gumtree_unify(In, [InGoal|InT]),
     nonvar(InGoal),
     Decode,
     ip_set('rule@unify@flag', sunify),
     proof_step(Thy, Name, InGoal, Children),
     PremLen,
     gumtree_append(Children, OutT, Out)
    ].
gumtree_trans((T1;T2), In, InT, Out, OutT) -->
    !,
    gumtree_trans(T1, In, InT, MOut, MOutT),
    gumtree_trans(T2, MOut, MOutT, Out, OutT).
gumtree_trans('|'(T1,T2), In, InT, Out, OutT) -->
    !,
    {
     gumtree_trans(T1, In1, InT1, Out1, OutT1, Code1, []),
     gumtree_trans(T2, In2, InT2, Out2, OutT2, Code2, [])
    },
    [
     alt(
         ['gum='(In,In1), 'gum='(InT,InT1), 
          'gum='(Out,Out1), 'gum='(OutT,OutT1)],
         ['gum='(In,In2), 'gum='(InT,InT2), 
          'gum='(Out,Out2), 'gum='(OutT,OutT2)],
	 Code1, Code2
        )
    ].
gumtree_trans(if_tac(Cond, Then,Else), In, InT, Out, OutT) -->
    !,
    {
     gumtree_trans(Cond, In1, InT1, MOut1, MOutT1, Code1, []),
     gumtree_trans(Then, MOut1, MOutT1, Out1, OutT1, Code2, []),
     gumtree_trans(Else, In2, InT2, Out2, OutT2, Code3, [])
    },
    [
     ite(['gum='(In,In1), 'gum='(InT,InT1)], 
         ['gum='(In,In2), 'gum='(InT,InT2)],
         ['gum='(Out,Out1), 'gum='(OutT,OutT1)],
         ['gum='(Out,Out2), 'gum='(OutT,OutT2)],
         Code1, Code2, Code3)
    ].
gumtree_trans([], In, InT, Out, OutT) -->
    !,
    [gumtree_unify(In, InT), gumtree_unify(Out, In), gumtree_unify(OutT, InT)].
gumtree_trans([T1|Rest], In, InT, Out, OutT) -->
    !,
    [
      gumtree_split_open_lists(In, InT, In1, InT1, In2, InT2) 
    ],
    gumtree_trans(T1, In1, InT1, Out, OutLink),
    gumtree_trans(Rest, In2, InT2, OutLink, OutT).
gumtree_trans(!(T), In, InT, Out, OutT) -->
    !,
    {
     gumtree_trans(T, In, InT, Out1, OutT1, Code, [])
    },
    [
      gum_once(Code), 'gum='(Out, Out1), 'gum='(OutT, OutT1)
    ].
gumtree_trans(trace(Caller,Level,Msg), In, InT, Out, OutT) -->
    !,
    [
     gumtree_unify(Out, In),
     gumtree_unify(OutT, InT),
     msg(Caller(X), Level, Msg)
    ].
gumtree_trans(deabbrev_and_call_gumtree_tactic(Cmd), In, InT, Out, OutT) -->
    !,
    {
      build_unify_goals([InX, InTX, OutX, OutTX], [In, InT, Out, OutT], UG)
    },
    [
     gum_unify(UG), 
     deabbrev_and_call_gumtree_tactic(Cmd, InX, InTX, OutX, OutTX)
    ].
gumtree_trans(T, In, InT, Out, OutT) -->
    {
      find_command(T, gumtree, _, call_gumtree_tactic(Tac), _),
      gumtree_deabbrev_tactic_args(Tac, Tac1)
    },
    !,
    gumtree_trans(Tac1, In, InT, Out, OutT).
gumtree_trans(T, In, InT, Out, OutT) -->
    !,
    {
      query_global(gumtree_tactic_declaration(T, _, '--->'(InA, OutA))),
      ( integer(InA) ->
          ICode = gumtree_length(In, InT, InA)
      ;
          ICode = true
      ),
      ( integer(OutA) ->
          OCode = gumtree_length(Out, OutT, OutA)
      ;
          OCode = true
      ),
      build_unify_goals([InX, InTX, OutX, OutTX], [In, InT, Out, OutT], UG),
      gumtree_2_predicate(T, InX, InTX, OutX, OutTX, NewT)
    },
    [ICode, gum_unify(UG), NewT, OCode].


%%
%% The optimizer for the generated code lists.
%% This optimizer keeps trying until no more progress
%% is made (i.e. when the Flag remains unbound).
%%
gumtree_optimize(Code0, Code) :-
    gumtree_optimize(Code0, Code1, Flag),
    nonvar(Flag),
    !,
    gumtree_optimize(Code1, Code).
gumtree_optimize(Code, Code1) :-
    post_opt_alt(Code, Code1).

%%
%% Individual steps for the optimizer
%%
gumtree_optimize([], [], _) :-
    !.
gumtree_optimize([fail], [fail], _) :-
    !.
%% a conjunction with a fail - simplify to fail
%gumtree_optimize(C, Os, Flag) :-
%    member(fail, C),
%    !,
%    Os = [fail],
%    Flag = yes.
%% remove true conjuncts
gumtree_optimize([C|Cs], Os, Flag) :-
    C == true,
    !,
    Os = Cs,
    Flag = yes.
%% flatten sublists
gumtree_optimize([C|Cs], Os, Flag) :-
    list(C),
    !,
    append(C, Cs, Os),
    Flag = yes.
%% optimize conjuncts individually
gumtree_optimize([C|Cs], [O|Os], Flag) :-
    !,
    gumtree_optimize(C, O, Flag),
    gumtree_optimize(Cs, Os, Flag).
%% If the first disjunct simplifies to false then throw it away
gumtree_optimize(alt(_, Unif2, [fail], C2), Opt, Flag) :-
    !,
    append(Unif2, C2, Opt),
    Flag = yes.
%% If the unifications for the first disjunct cannot be solved 
%% then throw it away
gumtree_optimize(alt(Unif1, Unif2, _C1, C2), Opt, Flag) :-
    not_unif(Unif1),
    !,
    append(Unif2, C2, Opt),
    Flag = yes.
%% Ditto second disjunct
gumtree_optimize(alt(Unif1, _, C1, [fail]), Opt, Flag) :-
    !,
    append(Unif1, C1, Opt),
    Flag = yes.
%% Ditto second disjunct unifications
gumtree_optimize(alt(Unif1, Unif2, C1, _), Opt, Flag) :-
    not_unif(Unif2),
    !,
    append(Unif1, C1, Opt),
    Flag = yes.
%% Optimize each part of a disjunct - care is taken
%% when optimizing the unifiers to stop cross infection.
gumtree_optimize(alt(Unif1, Unif2, C1, C2), Opt, Flag) :-
    !,
    gumtree_optimize(C1, Opt1, Flag),
    gumtree_optimize(C2, Opt2, Flag),
    opt_unif(Unif1, Unif2, OptUnif1, OptUnif2, Flag),
    Opt = alt(OptUnif1, OptUnif2, Opt1, Opt2).
%% Optimize the gum_once code
gumtree_optimize(gum_once([fail]), Opt, Flag) :-
    !,
    Opt = fail,
    Flag = yes.
gumtree_optimize(gum_once([]), Opt, Flag) :-
    !,
    Opt = [],
    Flag = yes.
gumtree_optimize(gum_once(Code), gum_once(Opt), Flag) :-
    !,
    gumtree_optimize(Code, Opt, Flag).
%% Optimize the fails code
gumtree_optimize(fails([fail]), Opt, Flag) :-
    !,
    Opt = [],
    Flag = yes.
gumtree_optimize(fails([]), Opt, Flag) :-
    !,
    Opt = fail,
    Flag = yes.
gumtree_optimize(fails(Code), fails(Opt), Flag) :-
    !,
    gumtree_optimize(Code, Opt, Flag).
%% Optimize the succs code
gumtree_optimize(succs([fail]), Opt, Flag) :-
    !,
    Opt = fail,
    Flag = yes.
gumtree_optimize(succs([]), Opt, Flag) :-
    !,
    Opt = [],
    Flag = yes.
gumtree_optimize(succs(Code), succs(Opt), Flag) :-
    !,
    gumtree_optimize(Code, Opt, Flag).
%% Optimize the ite structure
gumtree_optimize(ite(InU1, InU2, OutU1, OutU2, Code1, Code2, Code3), 
                 Opt, Flag) :-
    !,
    gumtree_optimize(Code1, Opt1, Flag),
    gumtree_optimize(Code2, Opt2, Flag),
    gumtree_optimize(Code3, Opt3, Flag),
    opt_unif(InU1, InU2, OptInU1, OptInU2, Flag),
    opt_unif(OutU1, OutU2, OptOutU1, OptOutU2, Flag),
    Opt = ite(OptInU1, OptInU2, OptOutU1, OptOutU2, Opt1, Opt2, Opt3).
%% Apply unifications
gumtree_optimize(gumtree_unify(X,Y), Opt, Flag) :-
    !,
    gumtree_opt_now(X=Y, Opt, Flag).
%% Evaluate gumtree_length where possible
gumtree_optimize(gumtree_length(L,LT,N), Opt, Flag) :-
    !,
    ( \+ \+ (LT = [], closed_list(L)) ->
        gumtree_opt_now(open_length(L,N), Opt, Flag)
    ; integer(N) ->
        gumtree_opt_now((length(LB,N), append(LB, LT, L)), Opt, Flag)
    ;
        Opt = gumtree_length(L,LT,N)
    ).
%% Evaluate gumtree_split_open_lists where possible
gumtree_optimize(gumtree_split_open_lists(In, InT, In1, InT1, In2, InT2), 
                 Opt, Flag) :-
    !,
    ( In1 == InT1 ->
         gumtree_opt_now((In = In2, InT = InT2), Opt, Flag)
    ; In2 == InT2 ->
         gumtree_opt_now((In = In1, InT = InT1), Opt, Flag)
    ; In == InT ->
         gumtree_opt_now((InT1 = In1, InT2 = In2), Opt, Flag)
    ; nonvar(In1) ->
         gumtree_opt_now((In1 = [H|I1], In = [H|I]), Opt1, Flag),
         ( Opt1 = false ->
             Opt = false
         ;
             Opt = gumtree_split_open_lists(I, InT, I1, InT1, In2, InT2)
         )
    ; \+ \+ (InT = [], closed_list(In), InT2 = [], closed_list(In2)) ->
         gumtree_opt_now(split_open_lists(In, InT, In1, InT1, In2, InT2), 
                         Opt, Flag)
    ;
       Opt = gumtree_split_open_lists(In, InT, In1, InT1, In2, InT2)
    ).
%% Evaluate gumtree_append where possible
gumtree_optimize(gumtree_append(L1, L2, L3), Opt, Flag) :-
    !,
    ( closed_list(L1) ->
        gumtree_opt_now(append(L1, L2, L3), Opt, Flag)
    ; closed_list(L2),closed_list(L3) ->
        gumtree_opt_now(append(L1, L2, L3), Opt, Flag)
    ; L2 == [] ->
        gumtree_opt_now(L1 = L3, Opt, Flag)
    ; L3 == [] ->
        gumtree_opt_now((L1 = [], L2 = []), Opt, Flag)
    ; is_tail(L2, L3) ->
        gumtree_opt_now(append(L1, L2, L3), Opt, Flag)
    ;
        Opt = gumtree_append(L1, L2, L3)
    ).
%% Simplify 'gum=' unifications where possible.
gumtree_optimize('gum='(X,Y), Opt, Flag) :-
    !,
    ( X == Y ->
	Opt = true,
	Flag = yes
    ;
      \+ (X = Y) ->
	Opt = fail,
	Flag = yes
    ;
      list(X), list(Y) ->
	X = [HX|TX],
	Y = [HY|TY],
	Opt = ['gum='(HX, HY), 'gum='(TX, TY)],
	Flag = yes
    ;
	Opt = 'gum='(X,Y)
    ).
%%Catch all - no optimizations possible.
gumtree_optimize(Code, Code, _).

%%
%% is_tail(L1, L2) is true iff L1 is the tail of the list L2
%%
is_tail(L1, L2) :- L1 == L2,!.
is_tail(L1, L2) :-
    nonvar(L2),
    L2 = [_|L3],
    is_tail(L1, L3).

%% gumtree_opt_now(Call, Opt, Flag).
%%
%%  Execute Call; if it succeeds, return Opt=true, Flag=yes.
%%  Otherwise, Opt=fail, Flag=yes.
%%
gumtree_opt_now(Call, Opt, Flag) :-
    ( call_predicate(Call) ->
        Opt = true,
        Flag = yes
    ;
        Opt = fail,
	Flag = yes
    ).

%% Optimize (evaluate) unifiers used in a disjunct (alt)
%% The invariant that makes this simple is that the two
%% input lists "pair up" element by element - i.e.
%% the LHS of each pair refers to the same outer variable.
%% For each pair of unification problems
%%    if they are identical then throw them away
%%    if they can be simplified as a pair then do so
%%    otherwise leave alone
%% Outer optimizations can move in but not the other way around
%% unless both branches do the same thing.
%%
opt_unif([], [], [], [], _).
opt_unif([U1|R1], [U2|R2], OptUnif1, OptUnif2, Flag) :-
    (   U1 == U2
	->
	    Flag = yes,
	    call_predicate(U1),
	    opt_unif(R1, R2, OptUnif1, OptUnif2, Flag)
	;
	    (   simp_unifys(U1, U2, SU1, SU2, Flag)
		->
		    opt_unif(R1, R2, OR1, OR2, Flag),
		    append(SU1, OR1, OptUnif1),
		    append(SU2, OR2, OptUnif2)
		;
		    opt_unif(R1, R2, OR1, OR2, Flag),
		    OptUnif1 = [U1|OR1],
		    OptUnif2 = [U2|OR2]
	    )
    ).

%%
%% A call on 'gum=' is just unification.
%%
'gum='(X,X).

%% Attempts to simplify pairs of unification problems from an alt goal.
%% If the outer variable of each pair is a nonvar then each inner var
%% can be unified with separate copies of the outer term. More unification
%% problems are added to maintain the connection between the outer variables
%% and the inner variables.
%% If both inner variables have the same skeletal structure then that
%% skeleton can move outside the alt.
%%
simp_unifys('gum='(L, R1), 'gum='(L, R2), S1, S2, Flag) :-
    nonvar(L),
    !,
    Flag = yes,
    copy_term(L, R1),
    copy_term(L, R2),
    collect_vars(L, LV),
    collect_vars(R1, R1V),
    collect_vars(R2, R2V),
    build_unify_goals(LV, R1V, S1),
    build_unify_goals(LV, R2V, S2).
simp_unifys('gum='(L, R1), 'gum='(L, R2), S1, S2, Flag) :-
    nonvar(R1),
    nonvar(R2),
    !,
    Flag = yes,
    skeletal(R1, L, S1),
    skeletal(R2, L, S2).

skeletal(A, L, S) :-
    atom(A),
    !,
    L = A,
    S = [].
skeletal(C, L, S) :-
    list(C),
    !,
    L = [X|Y],
    C = [CH|CT],
    S = ['gum='(X, CH), 'gum='(Y, CT)].


%% optimize_prolog(ClauseList, OptClauseList) is true if OptClauseList
%% is an optimized version of ClauseList. 
%% In this case clause lists are lists whose first element is the head of
%% the clause.
%%
optimize_prolog([], []).
%% If the last clause in the clause list has gum_once in the neck then 
%% the gum_once can be replaced by a cut following the body of the gum_once.
optimize_prolog([[Head|Body]], Final) :-
    Body = [gum_once(B)|R],
    !,
    append(B, ['!'|R], B1),
    optimize_prolog([[Head|B1]], Final).
optimize_prolog([[Head|Body]], Final) :-
    Body = [gum_unify(UG),gum_once(B)|R],
    !,
    append([gum_unify(UG)|B], ['!'|R], B1),
    optimize_prolog([[Head|B1]], Final).
%% A neck cut can be removed from the last clause.
optimize_prolog([[Head|Body]], Final) :-
    Body = ['!'|R],
    !,
    optimize_prolog([[Head|R]], Final).
optimize_prolog([[Head|Body]], Final) :-
    Body = [gum_unify(UG),'!'|R],
    !,
    optimize_prolog([[Head,gum_unify(UG)|R]], Final).
%% Remove clauses with fail body.
optimize_prolog([[_Head|Body]|Rest], Final) :-
    Body = [fail|_],
    !,
    optimize_prolog(Rest, Final).
%% Neck unifications can be evaluated.
optimize_prolog([[Head|Body]|Rest], Final) :-
    Body = ['gum='(X,Y)|R],
    !,
    (   X=Y
	->
            optimize_prolog([[Head|R]|Rest], Final)
	;
	    optimize_prolog(Rest, Final)
    ).
optimize_prolog([[Head|Body]|Rest], Final) :-
    Body = [gum_unify(UG), 'gum='(X,Y)|R],
    !,
    (   X=Y
	->
            optimize_prolog([[Head,gum_unify(UG)|R]|Rest], Final)
	;
	    optimize_prolog(Rest, Final)
    ).
optimize_prolog([[_Head|Body]|Rest], Final) :-
    Body = [gum_unify(_UG), fail|_R],
    !,
    optimize_prolog(Rest, Final).
%% Neck disjuncts cause the clause to split into 2 clauses.
optimize_prolog([[Head|Body]|Rest], Final) :-
    Body = [gum_unify(UG),(B1;B2)|R],
    !,
    (   B1 = (_;_)
	->
	    Code1 = [B1|R]
	;
	    append([gum_unify(UG)|B1], R, Code1)
    ),
    (   B2 = (_;_)
	->
	    Code2 = [B2|R]
	;
	    append([gum_unify(UG)|B2], R, Code2)
    ),
    copy_term([Head|Code2], Cl2),
    optimize_prolog([[Head|Code1], Cl2|Rest], Final).
optimize_prolog([[Head|Body]|Rest], Final) :-
    Body = [(B1;B2)|R],
    !,
    (   B1 = (_;_)
	->
	    Code1 = [B1|R]
	;
	    append(B1, R, Code1)
    ),
    (   B2 = (_;_)
	->
	    Code2 = [B2|R]
	;
	    append(B2, R, Code2)
    ),
    copy_term([Head|Code2], Cl2),
    optimize_prolog([[Head|Code1], Cl2|Rest], Final).
%% If none of the above optimize the body.
optimize_prolog([[Head|Body]|Rest], Final) :-
    (   optimize_body(Body, [Head], OptBody)
	->
	    optimize_prolog([[Head|OptBody]|Rest], Final)
	;
	    optimize_prolog(Rest, RestF),
	    Final = [[Head|Body]|RestF]
    ).


%% optimize_body(BodyList, SoFar, OptBodyList) is true if
%% OptBodyList is an optimization of BodyList.
%% SoFar is a list of body calls process so far - this is used to
%% determine if a variable has appeared before - if not then a unification
%% involving the variable can be carried out.
%%
%% Optimize length - note that calls to length are replaced by unifications
%% where possible and dealt with later.
optimize_body([G|Rest], SoFar, OptBody) :-
    gumtree_can_remove_repeats(G),
    member_eq(G, SoFar),
    !,
    OptBody = Rest.
optimize_body([(X=Y)|Rest], _SoFar, OptBody) :-
    !,
    optimize_body(['gum='(X,Y)|Rest],_SoFar, OptBody).
optimize_body([gumtree_length(L,N)|Rest], _SoFar, OptBody) :-
    (   integer(N)
	->
	    length(L1, N),
	    OptBody = ['gum='(L, L1)|Rest]
	;
	    closed_list(L),
	    length(L, N1),
	    OptBody = ['gum='(N,N1)|Rest]
    ),
    !.
%% ditto append
optimize_body([gumtree_append(L1,L2,L3)|Rest], _SoFar, OptBody) :-
    (   closed_list(L1)
	->
	    append(L1, L21, L31),
	    OptBody = ['gum='(L2, L21), 'gum='(L3, L31)|Rest]
	;
	L2 == []
	->
	    OptBody = ['gum='(L1, L3)|Rest]
	;
	L3 == []
	->
	    OptBody = ['gum='(L1, []), 'gum='(L2, [])|Rest]
	;
	    closed_list(L2), 
	    closed_list(L3),
	    append(L11, L2, L3),
	    OptBody = ['gum='(L1, L11)|Rest]
    ),
    !.
%% ditto split
optimize_body([gumtree_split_open_lists(In, InT, In1, InT1, In2, InT2)|Rest], 
               _SoFar, OptBody) :-
    ( In1 == InT1 ->
         OptBody = ['gum='(In, In2), 'gum='(InT, InT2)|Rest]
    ; In2 == InT2 ->
         OptBody = ['gum='(In, In1), 'gum='(InT, InT1)|Rest]
    ; In == InT ->
         OptBody = ['gum='(InT1, In1), 'gum='(InT2, In2)|Rest]
    ; nonvar(In1) ->
         OptBody = ['gum='(In1, [H|I1]), 'gum='(In, [H|I]), 
                    split_open_lists(I, InT, I1, InT1, In2, InT2)|Rest]
    ; 
         \+ \+ (InT = [], closed_list(In), InT2 = [], closed_list(In2)),
         split_open_lists(In, InT, In1X, InT1X, In2, InT2), 
         OptBody = ['gum='(In1, In1X), 'gum='(InT1, InT1X)|Rest]
    ).
%% useless unifications can be removed.
optimize_body(['gum='(X,Y)|Rest], _SoFar, OptBody) :-
    X == Y,
    !,
    OptBody = Rest.
%% Failing unifications optimize the remainder of the body to fail.
optimize_body(['gum='(X,Y)|_Rest], _SoFar, OptBody) :-
    \+ (X = Y),
    !,
    OptBody = [fail].
%% List unifications are expanded.
optimize_body(['gum='(X,Y)|Rest], _SoFar, OptBody) :-
    list(X),
    list(Y),
    X = [XH|XT],
    Y = [YH|YT],
    !,
    OptBody = ['gum='(XH, YH), 'gum='(XT, YT)|Rest].
%% If a variable is involved in a unification and that variable has
%% not occurred before in the  body then it can be instantiated.
%% Uses QP's occurs check to test for occurrences.
optimize_body(['gum='(X,Y)|Rest], SoFar, OptBody) :-
    var(X),
    \+ \+ (X = SoFar),
    !,
    X = Y,
    OptBody = Rest.
optimize_body(['gum='(X,Y)|Rest], SoFar, OptBody) :-
    var(Y),
    \+ \+ (Y = SoFar),
    !,
    X = Y,
    OptBody = Rest.
optimize_body([H|Rest], _SoFar, OptBody) :-
    H = (B1 ; B2),
    ( B1 == [fail] ->
        append(B2, Rest, OptBody)
    ; B2 == [fail] ->
        append(B1, Rest, OptBody)
    ; optimize_disj(B1, OB1) ->
        OptBody = [(OB1;B2)|Rest]
    ;
        optimize_disj(B2, OB2),
        OptBody = [(B1;OB2)|Rest]
    ),
    !.
optimize_body([H|Rest], SoFar, [H|OptRest]) :-
    optimize_body(Rest, [H|SoFar], OptRest).

%% A list of goals for which repeats can be removed.
gumtree_can_remove_repeats(G) :- var(G), !,fail.
gumtree_can_remove_repeats(G) :- compound(G), functor(G, F, _), var(F),!,fail.
gumtree_can_remove_repeats(_ = _).
gumtree_can_remove_repeats('gum='(_,_)).
gumtree_can_remove_repeats(gumtree_split_open_lists(_, _, _, _, _, _)).
gumtree_can_remove_repeats(gumtree_append(_,_,_)).
gumtree_can_remove_repeats(gumtree_length(_,_)).
gumtree_can_remove_repeats(nonvar(_)).

%%
%% Remove unification problems from given disjunct that don't bind variables.
%%
optimize_disj([(X=Y)|Rest], Opt) :-
    ( X == Y ->
        Opt = Rest
    ; 
        \+ (X = Y),
        Opt = [fail]
    ),
    !.
optimize_disj([(B1;B2)|Rest], Opt) :-
    (
      optimize_disj(B1, OB1),
      Opt = [(OB1;B2)|Rest]
    ;
      optimize_disj(B2, OB2),
      Opt = [(B1;OB2)|Rest]
    ),
    !.
optimize_disj([H|Rest], [H|Opt]) :-
    optimize_disj(Rest, Opt).
      
%%
%% fix_gum_unify(Clauses, FixedClauses) is true iff
%% FixedClauses is Clauses with the calls to gum_unify evaluated as
%% much as possible without generating repeated compound terms.
%%
fix_gum_unify([], []).
fix_gum_unify([[Head|Body]|Rest], [[Head|NewBody]|NewRest]) :-
    fix_gum_unify_body(Body, NewBody),
    fix_gum_unify(Rest, NewRest).

%%
%% Extract all the 'gum=' goals of all the gum_unify goals
%% in the same order as they appear in the clause - IMPORTANT!
%%
fix_gum_unify_body(Body, NewBody) :-
    extract_gum_unifys(Body, Unifs),
    gumtree_flatten(Unifs, All),
    fix_gum_unify_body1(Body, NewBody, All).

%%
%% Extract all the 'gum=' goal lists.
%%
extract_gum_unifys([], []).
extract_gum_unifys([gum_unify(UG)|Body], [UG|Unifs]) :-
    !,
    extract_gum_unifys(Body, Unifs).
extract_gum_unifys([gum_once(B)|Body], Unifs) :-
    !,
    extract_gum_unifys(B, BU),
    extract_gum_unifys(Body, U),
    append(BU, U, Unifs). 
extract_gum_unifys([(B1;B2)|Body], Unifs) :-
    !,
    extract_gum_unifys(B1, B1U),
    extract_gum_unifys(B2, B2U),
    extract_gum_unifys(Body, U),
    append(B2U, U, U2), 
    append(B1U, U2, Unifs). 
extract_gum_unifys([ite(A,B,C)|Body], Unifs) :-
    extract_gum_unifys(A, A1),
    extract_gum_unifys(B, B1),
    extract_gum_unifys(C, C1),
    extract_gum_unifys(Body, U),
    append(C1, U, U1),
    append(B1, U1, U2),
    append(A1, U2, Unifs).
extract_gum_unifys([_|Body], Unifs) :-
    extract_gum_unifys(Body, Unifs).

%%
%% Flatten a list of list of tems into a list of terms - preserving order.
%%
gumtree_flatten([], []).
gumtree_flatten([H|T], R) :-
    gumtree_flatten(T, RT),
    append(H, RT, R).

%%
%% Go through the clause and evalaute the 'gum=' goals where reasonable.
%%
fix_gum_unify_body1([],[],_).
fix_gum_unify_body1([gum_unify(UG)|R], Body, Unifys) :-
    !,
    gum_unify_solve(UG, Left, Unifys),
    fix_gum_unify_body1(R, ROut, Unifys),
    append(Left, ROut, Body).
fix_gum_unify_body1([gum_once(G)|R], Body, Unifys) :-
    !,
    fix_gum_unify_body1(G, GF, Unifys),
    fix_gum_unify_body1(R, ROut, Unifys),
    Body = [gum_once(GF)|ROut].
fix_gum_unify_body1([ite(A,B,C)|R], Body, Unifys) :-
    fix_gum_unify_body1(A, AF, Unifys),
    fix_gum_unify_body1(B, BF, Unifys),
    fix_gum_unify_body1(C, CF, Unifys),
    fix_gum_unify_body1(R, ROut, Unifys),
    Body = [ite(AF,BF,CF)|ROut].
fix_gum_unify_body1([(G1;G2)|R], Body, Unifys) :-
    !,
    fix_gum_unify_body1(G1, GF1, Unifys),
    fix_gum_unify_body1(G2, GF2, Unifys),
    fix_gum_unify_body1(R, ROut, Unifys),
    Body = [(GF1;GF2)|ROut].
fix_gum_unify_body1([H|R], [H|ROut], Unifys) :-
    fix_gum_unify_body1(R, ROut, Unifys).

%%
%% Evaluate the 'gum=' goals for a single list.
%%
%% The point of all this is when we have multiple occurrences of the same 
%% compound term. We want to avoid building/checking the same structure
%% as much as possible.
%% E.g. consider simplifying 'gum='(X,[C|D]) and later 'gum='(Y, [C|D])
%% If we execute both unifications then we will find [C|D] appears
%% at least twice in the final clause.
%% If however we unified X and Y and left the first unification problem
%% then X will occur where the [C|D] would have occurred (and of course
%% we will end up with X = [C|D] added to the final clause).
%%
gum_unify_solve([], [], _).
%% Two variables can be unified without trouble.
gum_unify_solve(['gum='(X,Y)|R], Out, Unifys) :-
    var(X), var(Y), 
    !,
    X = Y,
    gum_unify_solve(R, Out, Unifys).
%% The Y is not a variable (and is therefore a list)
%% If the Y occurs at least twice as the RHS of unifys
%% then we keep the first unify as is but for subsequent ones
%% we remove the goal but unify the LHS of the first and subsequent ones.
gum_unify_solve(['gum='(X,Y)|R], Out1, Unifys) :-
    var(X),  
    multiple_occurs(Y, Unifys, Z),
    !,
    ( Z == X ->
        gum_unify_solve(R, Out, Unifys),
        Out1 = ['gum='(X,Y)|Out]
      ;
        X = Z,
        gum_unify_solve(R, Out1, Unifys)
    ).
%% Only a single occurrence and so it is OK to do the unification.
gum_unify_solve(['gum='(X,Y)|R], Out, Unifys) :-
    var(X),!,
    X = Y,
    gum_unify_solve(R, Out, Unifys).
gum_unify_solve([H|R], [H|Out], Unifys) :-
    gum_unify_solve(R, Out, Unifys).

multiple_occurs(Y, Unifys, Z) :-
    append(_, ['gum='(A,B)|Rest], Unifys),
    Y == B,
    !,
    Z = A,
    member('gum='(_,D), Rest),
    Y == D.

%% Convert gumtree goals into prolog goals.
%% fails and succs goals become fails and succs goals.
%%
gumtree_to_prolog_list(GList, PList) :-
    ( GList == [] ->
        PList = []
    ; GList = true ->
	PList = []
    ; GList = [L] ->
        gumtree_to_prolog_list(L, C),
        PList = C
    ; GList = [LH|LT] ->
        gumtree_to_prolog_list(LH, CH),
        gumtree_to_prolog_list(LT, CT),
	append(CH, CT, PList)
    ; GList = gum_once(L1) ->
	gumtree_to_prolog_list(L1, C1),
	PList = [gum_once(C1)]
    ; GList = fails(L1) ->
	gumtree_to_prolog_list(L1, C1),
	PList = [fails(C1)]
    ; GList = succs(L1) ->
	gumtree_to_prolog_list(L1, C1),
	PList = [succs(C1)]
    ; GList = ite(In1, In2, Out1, Out2, Cond, Then, Else) ->
        gumtree_to_prolog_list(In1, CIn1),
        gumtree_to_prolog_list(In2, CIn2),
        gumtree_to_prolog_list(Out1, COut1),
        gumtree_to_prolog_list(Out2, COut2),
        gumtree_to_prolog_list(Cond, CCond),
        gumtree_to_prolog_list(Then, CThen),
        gumtree_to_prolog_list(Else, CElse),
	append(CIn1, CCond, CondCode),
	append(COut1, CThen, ThenCode),
	append(COut2, CElse, ElseCode1),
        append(CIn2, ElseCode1, ElseCode),
        PList = [ite(CondCode, ThenCode, ElseCode)]
    ; GList = alt(U1, U2,L1,L2) ->
        gumtree_to_prolog_list(U1, UC1),
        gumtree_to_prolog_list(L1, C1),
        gumtree_to_prolog_list(U2, UC2),
        gumtree_to_prolog_list(L2, C2),
	append(UC1, C1, Code1),
	append(UC2, C2, Code2),
        PList = [(Code1 ; Code2)]
    ;
        PList = [GList]
    ).      

%% Turn clauselists into lists of Prolog clauses.
%%
build_clauses([], []).
build_clauses([H|T], [(Head :- BodyCode)|TC]) :-
    H = [Head|Body],
    build_body(Body, BodyCode),
    build_clauses(T, TC).

build_body([], true) :- !.
build_body([(A;B)], Code) :-
    \+ (A = '->'(_,_)),
    !,
    build_body(A, ACode),
    build_body(B, BCode),
    Code = (ACode ; BCode).
build_body([fails(A)], Code) :-
    !,
    build_body(A, ACode),
    Code = (\+ ACode).
build_body([succs(A)], Code) :-
    !,
    build_body(A, ACode),
    Code = (\+ \+ ACode).
build_body([gum_once(A)], Code) :-
    !,
    build_body(A, ACode),
    Code = once(ACode).
build_body([true], C) :- 
    !,
    C = true.
build_body([C], C1) :- 
    !,
    build_body([C,true], (C1, true)).
build_body([(A;B)|R], Code) :-
    \+ (A = '->'(_,_)),
    !,
    build_body(A, ACode),
    build_body(B, BCode),
    build_body(R, RCode),
    Code = ((ACode ; BCode) , RCode).
build_body([ite(A,B,C)|R], Code) :-
    !,
    build_body(A, ACode),
    build_body(B, BCode),
    build_body(C, CCode),
    build_body(R, RCode),
    Code = ((ACode -> BCode ; CCode), RCode).
build_body([fails(A)|R], Code) :-
    !,
    build_body(A, ACode),
    build_body(R, RCode),
    Code = ((\+ ACode), RCode).
build_body([succs(A)|R], Code) :-
    !,
    build_body(A, ACode),
    build_body(R, RCode),
    Code = ((\+ \+ ACode), RCode).
build_body([gum_once(A)|R], Code) :-
    !,
    build_body(A, ACode),
    build_body(R, RCode),
    Code = (once(ACode) , RCode).
build_body([gumtree_append(A,B,C)|R], Code) :-
    !,
    build_body(R, RCode),
    Code = (append(A,B,C), RCode).
build_body(['gum='(A,B)|R], Code) :-
    !,
    build_body(R, RCode),
    Code = ((A=B), RCode).
build_body([gumtree_split_open_lists(In, InT, In1, InT1, In2, InT2)|R], Code) :-
    !,
    build_body(R, RCode),
    Code = (split_open_lists(In, InT, In1, InT1, In2, InT2), RCode).
build_body([C|R], Code) :-
    build_body(R, RCode),
    Code = (C, RCode).

%% After gumtree optimization has finished a pass is made over the
%% goals seeing if further optimizations are possible for alt goals.
%% NOTE: We cannot re-enter normal optimization after this optimization
%% because the invariant that the RHS of the unification involve only
%% local variables is lost.
%%
post_opt_alt([], []).
post_opt_alt([alt(U1, U2, C1, C2)|Rest], 
	     [alt(U1Opt, U2Opt, C1Opt, C2Opt)|RestOpt]) :-
    !,
    post_opt_alt(C1, C1Opt),
    post_opt_alt(C2, C2Opt),
    gumtree_alt_opt(U1, U2, U1Opt, U2Opt, []),
    post_opt_alt(Rest, RestOpt).
post_opt_alt([gum_once(C)|Code], [gum_once(OptC)|OptCode]) :-
    list(C),
    !,
    post_opt_alt(C, OptC),
    post_opt_alt(Code, OptCode).
post_opt_alt([ite(I1, I2, O1, O2, A,B,C)|Code], 
             [ite(I1Opt, I2Opt, O1Opt, O2Opt, OptA, OptB, OptC)|OptCode]) :-
    !,
    post_opt_alt(A, OptA),
    post_opt_alt(B, OptB),
    post_opt_alt(C, OptC),
    gumtree_alt_opt(I1, I2, I1Opt, I2Opt, []),
    gumtree_alt_opt(O1, O2, O1Opt, O2Opt, []),
    post_opt_alt(Code, OptCode).
post_opt_alt([C|Code], [C|CodeOpt]) :-
    post_opt_alt(Code, CodeOpt).

%% Each unification pair is considered. 
%% If the pairs are identical then they can be evaluated and removed.
%% If both sides of both pairs are variables then all the variables
%% can be unified provided an earlier optimization has not already done this
%% to one of the "local" variables - in that case it is no longer a local 
%% variable.
%% This optimization relies on the property that if one disjunct contains
%% X = Y and the other contains X = Z and both Y and Z are local to
%% their disjuncts and X is a variable then all 3 vars can be unified 
%% without changing the semantics.
%%
gumtree_alt_opt([], [], [], [], _).
gumtree_alt_opt(['gum='(V11,V12)|Rest1], ['gum='(V21,V22)|Rest2], 
                Opt1, Opt2, Vars) :-
    (   V11 == V21, var(V11),
	var(V12), \+member_eq(V12, Vars), \+member_eq(V22, Vars), var(V22)
	->
	    V11 = V12,
	    V11 = V22,
	    NewVars = [V11|Vars],
	    gumtree_alt_opt(Rest1, Rest2, Opt1, Opt2, NewVars)
	;
	    gumtree_alt_opt(Rest1, Rest2, Opt1Rest, Opt2Rest, Vars),
	    (   'gum='(V11,V12) == 'gum='(V21,V22)
		->
		    V11 = V12,
		    Opt1 = Opt1Rest,
		    Opt2 = Opt2Rest
		;
	            Opt1 = ['gum='(V11,V12)|Opt1Rest],
	            Opt2 = ['gum='(V21,V22)|Opt2Rest]
	    )
    ).

%% Change prolog conjuncts to lists. Used to turn embedded Prolog code
%% into goal lists for uniform processing in the optimizer.
%%
prolog_code_to_list(C, List) :-
    var(C),
    !,
    List = [call_predicate(C)].
prolog_code_to_list((C1;C2), List) :-
    \+ (nonvar(C1), C1 = '->'(_,_)),
    !,
    prolog_code_to_list(C1, L1),
    prolog_code_to_list(C2, L2),
    List = (L1 ; L2).
prolog_code_to_list((C1,C2), List) :-
    !,
    prolog_code_to_list(C1, L1),
    prolog_code_to_list(C2, L2),
    append(L1, L2, List).
prolog_code_to_list(C, [C]).


%%
%% not_unif(UnifList) succeeds iff UnifList contains a unification that fails.
%%
not_unif(Unif) :- member('gum='(X,Y),Unif), \+(X = Y).

%%
%% build_unify_goals is really a zip using 'gum=' as the pairing operator.
%%
build_unify_goals([], [], []).
build_unify_goals([X|TVars], [Y|CopyTVars], ['gum='(X,Y)|Rest]) :-
    build_unify_goals(TVars, CopyTVars, Rest).

%%
%% Load all the gumtree tactic declarations in File and any file loaded within
%% File.
%%
?- dynamic(already_loaded_tac_decls/1).

load_tactic_declarations(File) :-
    already_loaded_tac_decls(File),!.
load_tactic_declarations(File) :-
    assert(already_loaded_tac_decls(File)),
    open(File, read, Str),
    once((global_state_lookup(current_dir,Thy) ; Thy = global)),
    repeat,
    read_term(Str, T, [op_table(Thy), variable_names(VNames)]),
    (   T = end_of_file
	->
	    true
	;
	T = tactic(doc(Head, mode(Doc, arity(Mode, Arity))))
	->
          ip_set('@gumtree_consult_vlist', VNames),
          tactic_and_doc_decl(Head, Doc, Mode, Arity),
          fail
        ;
        T = tactic(mode(Head, arity(Mode, Arity)))
        ->
          ip_set('@gumtree_consult_vlist', VNames),
          tactic(mode(Head, arity(Mode, Arity))),
          fail
        ;
        T = load_global_tactic_file(F)
        ->
          check_atom(F),
          do_load_tactic_declarations(F),
          fail
        ;
        T = ensure_loaded_global_tactic_file(F)
        ->
          check_atom(F),
          do_load_tactic_declarations(F),
          fail
        ;
        T = require_global_tactic_file(F)
        ->
          check_atom(F),
          do_load_tactic_declarations(F),
          fail
        ;
        T = ensure_loaded_tactic_file(F)
        ->
          check_atom(F),
          global_state_lookup(current_dir,Thy),
          concat_atom([Thy, '/', F], FName),
          do_load_tactic_declarations(FName),
          fail
        ;
        T = ensure_loaded_tactic_file(F, Thy)
        ->
          check_atom(F),
          concat_atom([Thy, '/', F], FName),
          do_load_tactic_declarations(FName),
          fail
        ;
        T = load_tactic_file(F)
        ->
          check_atom(F),
          global_state_lookup(current_dir,Thy),
          concat_atom([Thy, '/', F], FName),
          do_load_tactic_declarations(FName),
          fail
        ;
        T = load_tactic_file(F, Thy)
        ->
          check_atom(F),
          concat_atom([Thy, '/', F], FName),
          do_load_tactic_declarations(FName),
          fail
        ;
        T = require_tactic_file(F)
        ->
          check_atom(F),
          global_state_lookup(current_dir,Thy),
          concat_atom([Thy, '/', F], FName),
          do_load_tactic_declarations(FName),
          fail
        ;
        T = require_tactic_file(F, Thy)
        ->
          check_atom(F),
          concat_atom([Thy, '/', F], FName),
          do_load_tactic_declarations(FName),
          fail
        ;
          fail
    ),
    !,
    close(Str).
          
do_load_tactic_declarations(File) :-
    get_load_path(PathList),
    member(P, PathList),
    full_file_name(P, File, _, PSFile),
    atom_append(PSFile, '.gum', FullName),
    access(FullName, 4, 0),
    msg(loading, 2, ['  ',w(FullName)]),
    !,
    load_tactic_declarations(FullName).


