%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: gumtree_tactics.ql,v 1.13 2002/06/27 05:10:25 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file contains support for user defined gumtree tactics.
%%
%% Support includes: hooks for finding and loading .gum files,
%% declarations and defns of gumtree tactics, type checker and arity checker.
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../prover/lib/ergo.h"

setup_gumtree_tactics :-
    ( global_fact(gumtree_tactic_declaration,3) ->
        true
    ;
        new_global_fact(gumtree_tactic_declaration,3,[])
    ),
    ( query_global(request(fileoperation(_,_,_),_)) ->
        true
    ;
        add_request(fileoperation(_Extension,_Operation,_FileName),
         ['Does the requested Operation (load/compile/save)',nl,
           'on the file called FileName, of type Extension.'])
    ),
    add_handler(fileoperation('.gum',consult,F), consult_gumtree_file(F)),
    add_handler(fileoperation('.gum',reconsult,F), reconsult_gumtree_file(F)),
    add_handler(fileoperation('.gum',compile(A),F), 
                 compile_gumtree_file(F, A, '.gum')),
    add_handler(fileoperation('.qo',compile(A),F), 
                 compile_gumtree_file(F, A, '.qo')),
    ( query_global(request(filevariant(_,_,_),_)) ->
        true
    ; 
        add_request(filevariant(_File, _Extension, _FullName),
         ['Determine if the file called File with Extension is the',nl,
          'appropriate variant to apply a file operation to.'])
    ),
    add_handler(filevariant(F, '.gum',FF), gumtree_file_variant(F,FF)).

?- setup_gumtree_tactics.

%% Read in a .gum file and add tactic declarations and definitions.
%%
consult_gumtree_file(F) :-
   concat_atom([F,'.gum'],Full),
   open(Full, read, Stream),
   once((global_state_lookup(current_dir,Thy), Thy \= none ; Thy = global)),
   read_term(Stream, Term, [variable_names(VNames), op_table(Thy)]),
   gumtree_consult(Term, Stream, [], Thy, VNames),
   close(Stream),
   fail.
consult_gumtree_file(_).

reconsult_gumtree_file(F) :-
   open(F, read, Stream),
   once((global_state_lookup(current_dir,Thy) ; Thy = global)),
   read_term(Stream, Term, [variable_names(VNames), op_table(Thy)]),
   gumtree_consult(Term, Stream, [], Thy, VNames),
   close(Stream),
   fail.
reconsult_gumtree_file(_).

gumtree_consult(end_of_file, _, _,_,_):- !.
gumtree_consult(tactic(delete(T)), Stream, Clauses, Thy, _) :-
   writelist([w(tactic(delete(T))), ' ignored.',nl]),
   !,
   read_term(Stream, Term, [variable_names(VNames), op_table(Thy)]),
   gumtree_consult(Term, Stream, Clauses, Thy, VNames).
gumtree_consult(tactic(doc(Head, mode(Doc, arity(Mode, Arity)))), 
                Stream, Clauses, Thy, VNames0) :-
   ip_set('@gumtree_consult_vlist', VNames0),
   tactic_and_doc_decl(Head, Doc, Mode, Arity),
   !,
   read_term(Stream, Term, [variable_names(VNames), op_table(Thy)]),
   gumtree_consult(Term, Stream, Clauses, Thy, VNames).
gumtree_consult(tactic(mode(Head, arity(Mode, Arity))), 
                Stream, Clauses, Thy, VNames0) :-
   ip_set('@gumtree_consult_vlist', VNames0),
   tactic(mode(Head, arity(Mode, Arity))),
   !,
   read_term(Stream, Term, [variable_names(VNames), op_table(Thy)]),
   gumtree_consult(Term, Stream, Clauses, Thy, VNames).
gumtree_consult(tactic('==='(Tac, Defn)), Stream, Clauses, Thy, _) :-
   functor(Tac, F, N),
   N4 is N+4,
   ( member('/'(F,N4), Clauses) ->
       Clauses1 = Clauses
   ;  
       gumtree_2_predicate(F, FPred),
       functor(Skel, FPred, N4),
       retractall(Skel),
       Clauses1 = ['/'(F,N4)|Clauses]
   ),
   tactic('==='(Tac, Defn)),
   !,
   read_term(Stream, Term, [variable_names(VNames), op_table(Thy)]),
   gumtree_consult(Term, Stream, Clauses1, Thy, VNames).
gumtree_consult(Goal, Stream, Clauses, Thy, _) :-
   call_predicate(Goal),
   !,
   read_term(Stream, Term, [variable_names(VNames), op_table(Thy)]),
   gumtree_consult(Term, Stream, Clauses, Thy, VNames).
gumtree_consult(Term0, Stream, Clauses, Thy, VNames0) :-
   gumtree_consult_failed(Term0, VNames0),
   !,
   read_term(Stream, Term, [variable_names(VNames), op_table(Thy)]),
   gumtree_consult(Term, Stream, Clauses, Thy, VNames).

gumtree_consult_failed(Term0, VList) :-
  tactic_try_unify(VList),
  writelist(['Failed: ', nl, w(Term0),nl]),
  fail.
gumtree_consult_failed(_, _).
    
compile_gumtree_file(F, A, '.qo') :-
   A == yes,
   !,
   msg(compile, [w(F),' does not need recompiling.']).
compile_gumtree_file(F, _, _) :-
   atom_append(F, '.gum', FullName),
   access(FullName, 4, 0),
   atom_append(F, '.qge', FullName1),
   gum_compile(FullName, FullName1),
   atom_concat2('$QPPATH/qc -c ', FullName1, Cmd),
   os(system(Cmd, 0)).

%% Used by the filevariant handler to determine if a .gum file
%% exists and is newer than the corresponding .qo file
%%
gumtree_file_variant(RootName,FullName) :-
    atom_append(RootName, '.gum', FullName),
    access(FullName, 4, 0),   % Read OK?
    atom_append(RootName, '.qo', FullName1),
    ( access(FullName1, 4, 0) ->
        file_newer_than(FullName, FullName1) %% .gum newer than .qo
    ;   
        true
    ),
    atom_append(RootName, '.ql', FullName2),
    ( access(FullName2, 4, 0) ->
        writelist(['Warning: ', w(FullName2), ' exists but is ignored - ',
                   w(FullName), ' used.', nl])
    ;   
        true
    ).

database command
  declare_tactic(PredType, Mode, InOutArity)
doc
  'Declare a gumtree tactic with its argument types and',nl,
  'its input-output arity.'
extends
  gumtree_decl_cmds
mode
  declare_tactic(@term, @term, @term).

declare_tactic(_Head, Mode, _InOutArity) :-
  \+ ground(Mode),
  !,
  handle_error(['The type part of the tactic declaration must be ground.', nl]).
declare_tactic(Head, Mode, _InOutArity) :-
  functor(Head, F, N),
  \+ functor(Mode, F, N),
  !,
  handle_error([q(Head), ' and ', q(Mode), ' are inconsistent.',nl]).
declare_tactic(Head, _Mode, _InOutArity) :-
  \+ skel_check(Head),
  !,
  handle_error([q(Head), ' should be a skeletal term.',nl]).
declare_tactic(Head, PredType, InOutArity) :-
  query_global(gumtree_tactic_declaration(Head, Mode, IO)),
  !,
  ( Mode = PredType, 
    collect_vars(InOutArity, IOVars),
    collect_vars(IO, IOVars),
    InOutArity == IO ->
      true
  ;
      handle_error(['New declaration different from existing declaration',
                    nl, q(Mode), ' : ', q(IO), nl])
  ).
declare_tactic(Head, PredType, InOutArity) :-
  assert_global(gumtree_tactic_declaration(Head, PredType, InOutArity)).
   
skel_check(Term) :-
  functor(Term, _, N),
  skel_check_args(N, Term).
  
skel_check_args(0,_) :- !.
skel_check_args(N, Term) :-
  arg(N, Term, '$skel'(N)),
  N1 is N-1,
  skel_check_args(N1, Term).

general command
  tactic(Decl)
header
  'tactic Decl'
doc
  'This command is used to declare, define and delete tactics.',nl,
  'Examples:',br,
  'tactic tac1(X) mode tac1(thy(term)) arity 1 ---> N',br,
  'tactic tac1(X) doc doc1 mode tac1(thy(term)) arity 1 ---> N',br,
  'declares tac1/1 as a gumtree tactic whose argument is of type thy(term)',nl,
  'with arity 1 ---> N.',nl,'A doc node containing doc1 is added for the tactic.',para,
  'tactic tac1(X) === term(X)',br,
  'defines a ''clause'' for tac1/1',para,
  'tactic delete tac1(Pattern)',br,
  'deletes all clauses for tac1/1 whose head argument matches Pattern.',para,
  'Tactics must be declared before defined or deleted.',nl,
  'Multiple identical declarations of tactics are allowed.',nl,
  'Only dynmaic tactics (i.e. tactics that were not loaded from compiled',nl,
  'files) can be deleted.'
extends
  general_cmds
mode
  tactic(@compound).

tactic('==='(Head, Tac)) :-
  define_tactic(Head, Tac).
tactic(doc(Head, mode(Doc, arity(Mode, Arity)))) :-
  tactic_and_doc_decl(Head, Doc, Mode, Arity).
tactic(mode(Head, arity(Mode, Arity))) :-
  functor(Head, F, N),
  ( functor(Mode, F, N) ->
      true
  ;
      handle_error([q(Head), ' and ', q(Mode), ' are inconsistent.',nl])
  ),
  gumtree_parse_head(Mode, Mode1),
  declare_tactic(Head, Mode1, Arity).
tactic(delete(Head)) :-
  delete_tactic(Head).
  
%
% A variant so users can do the declaration in code - VList is a
% list of variable, name pairs to be used in error messages and
% in generating documentation.

user_tactic_and_doc_decl(VList, Head, Doc, Mode, Arity) :-
   ip_set('@gumtree_consult_vlist', VList),
   tactic_and_doc_decl(Head, Doc, Mode, Arity).

%
% Declare the tactic and add documentation.
%
tactic_and_doc_decl(Head, Doc, Mode, Arity) :-
  tactic(mode(Head, arity(Mode, Arity))),
  functor(Head, F, N),
  make_doc_name(gumtree, F, N, Name),
  add_tactic_doc(Head, Doc, Name, F, N).

add_tactic_doc(_Head, _Doc, Name, _F, _N) :-
  query_global(doc_node(Name,_,_,_)),
  !.
add_tactic_doc(Head, Doc, Name, F, N) :-
  open_string(write, Strm),
  tactic_doc_header(Strm, Head),
  stream_to_atom(Strm, CallA),
  gumtree_comma_to_list(Doc, DocList),
  check_is_legal_writelist(DocList),
  add_doc_extension(global, gumtree_user_tactics, 
                    [xref([CallA], global, Name)]),
  add_doc(global, Name, [CallA], DocList, [cmd(global, gumtree, F, N)]).
  
tactic_doc_header(Strm, Head) :-
  ip_lookup('@gumtree_consult_vlist', VList),
  tactic_try_unify(VList),
  write_term(Strm, Head, []),
  fail.
tactic_doc_header(_, _).
  
tactic_try_unify([]) :- !.
tactic_try_unify([G|R]) :-
  once((call_predicate(G) ; true)),
  tactic_try_unify(R).
  
gumtree_comma_to_list(D, L) :-
  var(D),
  !,
  L = [D].
gumtree_comma_to_list((D,Doc), DocList) :-
  !,
  gumtree_comma_to_list(Doc, DocL),
  DocList = [D|DocL].
gumtree_comma_to_list(D, [D]).


gumtree_parse_head(Head, Mode) :-
    functor(Head, Fun, N),
    functor(Mode, Fun, N),
    gumtree_parse_head(N, Head, Mode).

gumtree_parse_head(0, _, _) :-
    !.
gumtree_parse_head(N, Head, Mode) :-
    PRE('gumtree_parse_head', N > 0),
    arg(N, Head, H),
    gumtree_check_dir(H, DH),
    arg(N, Mode, DH),
    N1 is N - 1,
    gumtree_parse_head(N1, Head, Mode).

gumtree_check_dir('-'(H), R) :- !, R = '-'(H).
gumtree_check_dir('?'(H), R) :- !, R = '?'(H).
gumtree_check_dir('@'(H), R) :- !, R = '@'(H).
gumtree_check_dir(H, '?'(H)).

%
% Add 'clause' of the definition of a user defined gumtree tactic.
%
define_tactic(Head0, Tac0) :-
    define_tactic0(Head0, Tac0, NewHead, Body),
    assert(':-'(NewHead, Body)),
    fail.
define_tactic(_, _).

define_tactic0(Head0, Tac0, NewHead, Body) :-
    ( query_global(gumtree_tactic_declaration(Head0, Mode, '--->'(In,Out))) ->
        true
    ;
        !,
        functor(Head0, F, N),
        handle_error(['Tactic ',q(F), '/', q(N), ' has no type declaration.'])
    ),
    ( gumtree_type_check(Head0, Mode, Tac0) ->
        true
    ;
        !,
        handle_error(['Tactic ', q(Head0), ' === ', q(Tac0),' has bad type.'])
    ),
    ( \+ \+ gumtree_check_arity(Tac0, In, Out) ->
        true
    ;
        !,
        handle_error(['Tactic ', q(Head0), ' === ', q(Tac0),' has bad arity.'])
    ),
    %% We deabbreviate the Head, just in case it contains patterns
    %% (like theory terms) that need deabbreviating.
    ( deabbreviate_cmd_relaxed(Mode, Head0, Head) ->
        true
    ;
        !,
        handle_error(['bad patterns in definition of tactic: ',q(Head0)])
    ),
    ( gumtree_deabbrev_tactic_args(Tac0, Tac) ->
        true
    ;
        !,
        handle_error(['unable to deabbreviate tactic body: ',q(Tac0)])
    ),
    %% Finally, we add the new definition.
    gumtree_2_predicate(Head, X,XT,Y,YT,NewHead),
    Body = expand_tac_body_and_exec(Tac, X, XT, Y, YT).

%
% delete a 'clause' of a user defined gumtree tactic.
%
delete_tactic(Head0) :-
    ( query_global(gumtree_tactic_declaration(Head0, Mode, _)) ->
        true
    ;
        !,
        functor(Head0, F, N),
        handle_error(['Tactic ',q(F), '/', q(N), ' has no type declaration.'])
    ),
    %% We deabbreviate the Head, just in case it contains patterns
    %% (like theory terms) that need deabbreviating.
    ( deabbreviate_cmd_relaxed(Mode, Head0, Head) ->
        true
    ;
        !,
        handle_error(['bad patterns in definition of tactic: ',q(Head0)])
    ),
    gumtree_2_predicate(Head, _X,_XT,_Y,_YT,NewHead),
    retract(':-'(NewHead, _)),
    fail.
delete_tactic(_).

%
% Carry out type checking  of user define gumtree tactic.
%
gumtree_type_check(Head, Mode, Tac) :-
  functor(Head, F, N),
  functor(Mode, F, N),
  gumtree_type_check_head(N, Head, Mode, VTypes),
  gumtree_check_type(tactic, Tac, VTypes).

gumtree_type_check_head(0, _, _, _) :- !.
gumtree_type_check_head(N, Head, Mode, VTypes) :-
  arg(N, Head, ArgN),
  arg(N, Mode, _M(ModeN)),
  gumtree_check_type(ModeN, ArgN, VTypes),
  N1 is N-1,
  gumtree_type_check_head(N1, Head, Mode, VTypes).

gumtree_check_type(Type, Term, VTypes) :-
  var(Term),!,
  gumtree_lookup_vtypes(VTypes, Term, TermType),
  gumtree_types_compatible(Type, TermType, Term).
gumtree_check_type(integer, Term, _) :-
  ( integer(Term) ->
      true
  ;
    ip_lookup('@gumtree_consult_vlist', VList),
    tactic_try_unify(VList),
    handle_error(['Type: ', q(Term), ' should be of type integer',nl])
  ).
gumtree_check_type(ctermid, Term, _) :-
  ( integer(Term) ->
      true
  ;
    ip_lookup('@gumtree_consult_vlist', VList),
    tactic_try_unify(VList),
    handle_error(['Type: ', q(Term), ' should be of type ctermid',nl])
  ).
gumtree_check_type(cterm, Term, VList) :-
  ( Term = c(ID, T) ->
      gumtree_check_type(ctermid, ID, VList),
      gumtree_check_type(thy(term), T, VList)
  ;
    ip_lookup('@gumtree_consult_vlist', VList),
    tactic_try_unify(VList),
    handle_error(['Type: ', q(Term), ' should be of type cterm',nl])
  ).
gumtree_check_type(cname, Term, _) :-
  ( atom(Term) ->
      true
  ;
    ip_lookup('@gumtree_consult_vlist', VList),
    tactic_try_unify(VList),
    handle_error(['Type: ', q(Term), ' should be of type cname',nl])
  ).
gumtree_check_type(cmdtable, Term, _) :-
  ( atom(Term) ->
      true
  ;
    ip_lookup('@gumtree_consult_vlist', VList),
    tactic_try_unify(VList),
    handle_error(['Type: ', q(Term), ' should be of type cmdtable',nl])
  ).
gumtree_check_type(nodeid, Term, _) :-
  ( integer(Term) ->
      true
  ;
    ip_lookup('@gumtree_consult_vlist', VList),
    tactic_try_unify(VList),
    handle_error(['Type: ', q(Term), ' should be of type nodeid',nl])
  ).
gumtree_check_type(atom, Term, _) :-
  ( atom(Term) ->
      true
  ;
    ip_lookup('@gumtree_consult_vlist', VList),
    tactic_try_unify(VList),
    handle_error(['Type: ', q(Term), ' should be of type atom',nl])
  ).
gumtree_check_type(obvar, Term, _) :-
  ( obvar(Term) ->
      true
  ;
    ip_lookup('@gumtree_consult_vlist', VList),
    tactic_try_unify(VList),
    handle_error(['Type: ', q(Term), ' should be of type obvar',nl])
  ).
gumtree_check_type(term, Term, VTypes).
gumtree_check_type(list(Type), Term, VTypes) :-
  gumtree_check_list_type(Type, Term, VTypes).
gumtree_check_type(thy(term), Term, VTypes) :-
  gumtree_check_sub_term_types(thy(term), Term, VTypes).
gumtree_check_type(tactic, Term, VTypes) :-
  gumtree_check_tactic_type(Term, VTypes).
gumtree_check_type(rulename, Term, VTypes).
   

gumtree_lookup_vtypes(Var, V, Type) :-
  var(Var),
  !,
  Var = [(V=Type)|_].
gumtree_lookup_vtypes([(X=T)|_], V, Type) :-
  V == X,
  !,
  gumtree_types_compatible(T, Type, V).
  
gumtree_lookup_vtypes([_|Types], V, Type) :-
  gumtree_lookup_vtypes(Types, V, Type).

gumtree_types_compatible(Type, TermType, _) :-
  var(TermType),
  !,
  TermType = Type.
gumtree_types_compatible(Type, TermType, _) :-
  Type = TermType,
  !.
gumtree_types_compatible(integer, nodeid, _) :- !.
gumtree_types_compatible(integer, ctermid, _) :- !.
gumtree_types_compatible(nodeid, integer, _) :- !.
gumtree_types_compatible(nodeid, ctermid, _) :- !.
gumtree_types_compatible(ctermid, nodeid, _) :- !.
gumtree_types_compatible(ctermid, integer, _) :- !.
gumtree_types_compatible(list(T), list(T1), Term) :-
  !,
  gumtree_types_compatible(T, T1, Term).
gumtree_types_compatible(thy(term), obvar, _) :- !.
gumtree_types_compatible(Type, TermType, Term) :-
    ip_lookup('@gumtree_consult_vlist', VList),
    tactic_try_unify(VList),
    handle_error(['Type: ', q(Term), ' have incompatible types ',q(Type), ' and ', q(TermType),nl]).

gumtree_check_list_type(Type, Term, VTypes) :-
  var(Term),
  !,
  gumtree_lookup_vtypes(VTypes, Term, TermType),
  gumtree_types_compatible(list(Type), TermType, Term).
gumtree_check_list_type(_, [], _) :- !.
gumtree_check_list_type(Type, [H|T], VTypes) :-
  !,
  gumtree_check_type(Type, H, VTypes),
  gumtree_check_list_type(Type, T, VTypes).
gumtree_check_list_type(Type, Term, _) :-
  ip_lookup('@gumtree_consult_vlist', VList),
  tactic_try_unify(VList),
  handle_error(['Type: ', q(Term), ' is not a list of type ',q(Type),nl]).

gumtree_check_sub_term_types(Type, Term, VTypes) :-
  compound(Term),
  !,
  functor(Term, _, N),
  gumtree_check_arg_types(N, Type, Term, VTypes).
gumtree_check_sub_term_types(_, _, _).

gumtree_check_arg_types(0, _, _, _) :- !.
gumtree_check_arg_types(N, Type, Term, VTypes) :-
  arg(N, Term, Arg),
  gumtree_check_type(Type, Arg, VTypes),
  N1 is N-1,
  gumtree_check_arg_types(N1, Type, Term, VTypes).

gumtree_check_tactic_type(Term, VTypes) :-
  functor(Term, F, N),
  once((
     query_global(gumtree_tactic_declaration(Term, Mode, _)),
     M = N
   ;
     functor(Cmd, F, N),
     find_command(Cmd, gumtree, _, [arity(Mode,_)], _, _),
     M = N
   ;
     M = 0
  )),
  gumtree_type_check_head(M, Term, Mode, VTypes).
  
  
%
% arity checker for user defined gumtree tactics.
%
gumtree_check_arity(V, _, _) :-
  var(V), !.
gumtree_check_arity(';'(T1, T2), I, O) :-
  !,
  gumtree_check_arity(T1, I1, O1),
  gumtree_check_arity(T2, I2, O2),
  ( O1 = I2  ->
      true
  ;
      ip_lookup('@gumtree_consult_vlist', VList),
      tactic_try_unify(VList),
      handle_error(['Inconsistent arities for', nl,
                    q(';'(T1, T2)), nl,
                    q(T1), ' has output arity ', q(O1),nl,
                    q(T2), ' has input arity ', q(I2), nl])
  ),
  I1 = I,
  O2 = O.
gumtree_check_arity('|'(T1, T2), I, O) :-
  !,
  gumtree_check_arity(T1, I1, O1),
  gumtree_check_arity(T2, I2, O2),
  gumtree_arity_compatible(I1, I2, I, input, T1, T2, '|'(T1, T2)),
  gumtree_arity_compatible(O1, O2, O, outout, T1, T2, '|'(T1, T2)).
gumtree_check_arity(if_tac(Cond, Then, Else), I, O) :-
  !,
  gumtree_check_arity(Cond, ICond, OCond),
  gumtree_check_arity(Then, IThen, OThen),
  gumtree_check_arity(Else, IElse, OElse),
  ( OCond = IThen ->
      true
  ;
      ip_lookup('@gumtree_consult_vlist', VList),
      tactic_try_unify(VList),
      handle_error(['Inconsistent arities for', nl,
                    q(if_tac(Cond, Then, Else)), nl,
                    q(Cond), ' has output arity ', q(OCond),nl,
                    q(Then), ' has input arity ', q(IThen), nl])
  ),
  gumtree_arity_compatible(ICond, IElse, I, input, Cond, Else, 
                           if_tac(Cond, Then, Else)),
  gumtree_arity_compatible(OThen, OElse, O, output, Then, Else, 
                           if_tac(Cond, Then, Else)).
      
  
gumtree_check_arity([], I, O) :-
  !,
  I = 0,
  O = 0.
gumtree_check_arity([T|Ts], I, O) :-
  !,
  gumtree_check_arity(T, IT, OT),
  gumtree_check_arity(Ts, ITs, OTs),
  gumtree_arity_add(IT, ITs, I, input, [T|Ts]),
  gumtree_arity_add(OT, OTs, O, output, [T|Ts]).
gumtree_check_arity(rule(Name), I, O) :-
  compound(Name),
  Name = (Thy:Nm),
  rule_query(Thy, Nm, _, _, Seq, _),
  !,
  length(Seq, Len),
  ( I = 1, O = Len ->
      true
  ;
      ip_lookup('@gumtree_consult_vlist', VList),
      tactic_try_unify(VList),
      handle_error(['Inconsistent arities for', nl,q(rule(Name)),nl,
                    'Arity is: ', '1 ---> ', q(Len),nl,
                    'Required arity is: ', q(I), ' ---> ', q(O),nl])
  ).
gumtree_check_arity(T, I, O) :-
  once((
     query_global(gumtree_tactic_declaration(T, _, '--->'(In, Out)))
   ;
     functor(T, F, N),
     functor(Cmd, F, N),
     find_command(Cmd, gumtree, _, [arity(_,'--->'(In, Out))], _, _)
  )),
  !,
  ( I = In, O = Out ->
      true
  ;
      ip_lookup('@gumtree_consult_vlist', VList),
      tactic_try_unify(VList),
      handle_error(['Inconsistent arities for', nl,q(T),nl,
                    'Arity is: ', q(In), ' ---> ', q(Out),nl,
                    'Required arity is: ', q(I), ' ---> ', q(O),nl])
  ).
gumtree_check_arity(_, _, _).

gumtree_arity_add(I1, I2, I, Dir, Tac) :-
  integer(I1),
  integer(I2),
  !,
  ( I is I1 + I2 ->
      true
   ;
      gumtree_arity_add_error(I1, I2, I, Dir, Tac)
  ).
gumtree_arity_add(I1, I2, I, Dir, Tac) :-
  integer(I1),
  integer(I),
  !,
  ( I2 is I - I1 ->
      true
   ;
      gumtree_arity_add_error(I1, I2, I, Dir, Tac)
  ).
gumtree_arity_add(I1, I2, I, Dir, Tac) :-
  integer(I),
  integer(I2),
  !,
  ( I1 is I - I2 ->
      true
   ;
      gumtree_arity_add_error(I1, I2, I, Dir, Tac)
  ).
gumtree_arity_add(I1, I2, I, Dir, Tac) :-
  integer(I),
  !,
  delay_until(or(ground(I1), ground(I2)), 
              gumtree_arity_add(I1, I2, I, Dir, Tac)).
gumtree_arity_add(I1, I2, I, Dir, Tac) :-
  integer(I1),
  !,
  delay_until(or(ground(I), ground(I2)), 
              gumtree_arity_add(I1, I2, I, Dir, Tac)).
gumtree_arity_add(I1, I2, I, Dir, Tac) :-
  integer(I2),
  !,
  delay_until(or(ground(I1), ground(I)), 
              gumtree_arity_add(I1, I2, I, Dir, Tac)).
gumtree_arity_add(I1, I2, I, Dir, Tac) :-
  var(I1),var(I2),var(I),
  !,
  delay_until(or(ground(I1), or(ground(I2),ground(I))), 
              gumtree_arity_add(I1, I2, I, Dir, Tac)).

gumtree_arity_add_error(I1, I2, I, Dir, Tac) :-
  ip_lookup('@gumtree_consult_vlist', VList),
  tactic_try_unify(VList),
  handle_error(['Inconsistent ', q(Dir), ' arity for', nl,q(Tac),nl,
                'Arity is: ', q(I1 + I2),nl,
                'Required arity is: ', q(I),nl]).
  
gumtree_arity_compatible(I1, I2, I, Dir, T1, T2, Term) :-
  I1 == I2,
  !,
  ( I = I1 ->
      true
  ;
      gumtree_arity_compatible_error(I1, I2, I, Dir, T1, T2, Term)
  ).
gumtree_arity_compatible(I1, I2, I, Dir, T1, T2, Term) :-
  integer(I),
  !,
  ( I1 = I, I2 = I ->
      true
  ;
      gumtree_arity_compatible_error(I1, I2, I, Dir, T1, T2, Term)
  ).
gumtree_arity_compatible(I1, I2, I, Dir, T1, T2, Term) :-
  var(I1),
  var(I2),
  !,
  delay_until(or(bound(I1), or(bound(I2), ground(I))),
              gumtree_arity_compatible(I1, I2, I, Dir, T1, T2, Term)).
gumtree_arity_compatible(I1, I2, I, Dir, T1, T2, Term) :-
  integer(I1),
  integer(I2),
  !,
  delay_until(ground(I), gumtree_arity_compatible(I1, I2, I, Dir, T1, T2, Term)).
gumtree_arity_compatible(I1, I2, I, Dir, T1, T2, Term) :-
  integer(I1),
  !,
  delay_until(or(bound(I2),ground(I)), 
              gumtree_arity_compatible(I1, I2, I, Dir, T1, T2, Term)).
gumtree_arity_compatible(I1, I2, I, Dir, T1, T2, Term) :-
  integer(I2),
  !,
  delay_until(or(bound(I1),ground(I)), 
              gumtree_arity_compatible(I1, I2, I, Dir, T1, T2, Term)).

gumtree_arity_compatible_error(I1, I2, I, Dir, T1, T2, Term) :-
  ip_lookup('@gumtree_consult_vlist', VList),
  tactic_try_unify(VList),
  handle_error(['Incompatible ', q(Dir), ' arities for', nl,
                q(Term),nl,
                'Arity for ', q(T1), ' : ', q(I1),nl,
                'Arity for ', q(T2), ' : ', q(I2),nl,
                'Required arity is: ', q(I),nl]).
  

%%
%% Turn a gumtree tactic call into a predicate call
%% 
gumtree_2_predicate(Tac, A,B,C,D, Pred) :-
    Tac =.. [F|Args],
    append(Args, [A,B,C,D], AllArgs),
    atom_concat2(gumtree_tactic_, F, GF),
    Pred =.. [GF|AllArgs].
gumtree_2_predicate(Tac, Pred) :-
    Tac =.. [F|Args],
    atom_concat2(gumtree_tactic_, F, GF),
    Pred =.. [GF|Args].
