%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: help.ql,v 1.32 2007/03/27 23:40:57 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file provides commands for finding and displaying
%% on-line help information from doc_node's.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../prover/lib/ergo.h"

master_help_node header
  'Online Help Information for the Ergo Proof Tool'
doc
  'The commands that are available in Ergo depend upon which user',nl,
  'interface you are using.  For example, whether you are constructing',nl,
  'theories or are within a proof.  In addition to the commands supplied',nl,
  'by each user interface, there are also a number of `general'' commands',nl,
  'that are available from all user interfaces.  These allow you to',nl,
  'inspect the Ergo database, obtain help and so on.',para,
  'The following help topics are available:'.
  

general_cmds header
  'General commands (viewing data and setting parameters)'
extends
  master_help_node
doc
  'The following commands are available in all Ergo user interfaces.',nl,
  'They allow you to browse the contents of the Ergo database and the',nl,
  'online help system, inspect and set global ',
  xref(parameters,global,parameters),' etc.',nl.


concepts header
  'Concepts'
extends
  master_help_node
doc
  'This section describes major Ergo concepts that are common to',nl,
  'several commands or user interfaces.',nl.


general command
  help
extends
  general_cmds
implementation
  do_help(master_help_node,global)
doc
  'Entry point for help about dbms commands.'
mode
  help.

general command
  help(Topic)
header
  'help Topic'
implementation
  do_help(Topic, _)
command
  help(Topic, Thy)
header
  'help(Topic, Thy)'
implementation
  do_help(Topic, Thy)
extends
  general_cmds
doc
  xinc(global,help_topics)
mode
  help(@term),
  help(@term, @theory).


help_topics doc_fragment
  'Get help on Topic, which may be a command name, cmd/arity,',nl,
  'a node name, or a string (double-quoted).',nl,
  begin(list),
  item(' - '),'Cmd:       search for any node documenting Cmd of any arity.',
  item(' - '),'Cmd/Arity: search for any node documenting Cmd/Arity.',nl,
  '              (Arity may also be: parameter)',
  item(' - '),'Node:      display the named node.',
  item(' - '),'"String":  display all nodes containing String.',
  end(list),nl,
  'Thy may be a theory name or ''global''.',nl,
  'If Thy is specified, then only that theory is searched.'.


do_help(Topic, Thy) :-
    ( Topic = Cmd/N, integer(N) ->
        do_help_cmd(Cmd, N, Thy)
    ; Topic = Param/parameter ->
        do_help_param(Param, Thy)
    ; list(Topic) ->
        do_help_substring(Topic, Thy)
    ; do_help_node(Topic, Thy) ->
        true
    ; do_help_cmd(Topic, _, Thy) ->
        true
    ; do_help_param(Topic, Thy) ->
        true
    ;
        handle_error(['No help for ', w(Topic), '.'])
    ).

do_help_cmd(Cmd, N, Thy) :-
    findall([xref(Hdr,Thy,Node)],
            ( ( ancestor(Thy)
              ; Thy = global
              ),
              query_theory(doc_node(Node, Hdr, _Doc, Cmds), Thy),
              member(cmd(_, _, Cmd, N), Cmds)
            ),
            Xrefs),
    sort(Xrefs, Sorted),
    ( nonvar(N) ->
        true
    ;
        N = *
    ),
    ( Sorted = [[xref(_,_,Node)]] ->
        do_help_node(Node, Thy)
    ; Sorted = [_|_] ->
        showing_help(start, [w(Cmd), /, w(N)]),
        show_help(['The following documentation is available: ', br]),
        show_help_list(Sorted),
        showing_help(end, [w(Cmd), /, w(N)])
    ).

do_help_param(Param, Thy) :-
    atom_concat2('parameter:', Param, Node),
    do_help_node(Node, Thy).
    
do_help_node(Node, Thy) :-
    ( Thy = global
    ; ancestor(Thy)
    ),
    query_theory(doc_node(Node, Hdr, Doc, Cmds), Thy),
    !,
    showing_help(start, [w(Node)]),
    show_help(Hdr),
    show_help([para]),
    show_help(Doc),
    show_help([br]),
    ( Cmds = [] ->
        true
    ;
        show_help([br]),
        show_help_modes(Cmds),
        show_help([br])
    ),
    show_help_ext(Node),
    showing_help(end, [w(Node)]).

do_help_substring(Str, Thy) :-
    string_to_atom(Str, Sub),
    append(Str, _, StrOpen),
    showing_help(start, ['"', w(Sub), '"']),
    show_help(['Topics matching ', qchars(Str), ' are:', para]),
    ( ( ancestor(Thy)
      ; Thy = global
      ),
      query_theory(doc_node(Node, Hdr, Doc, _), Thy),
      ( find_substring_in_WL(Sub, StrOpen, Hdr) ->
          show_help([xref(Hdr,Thy,Node),br])
      ; find_substring_in_WL(Sub, StrOpen, Doc) ->
          show_help([xref(Hdr,Thy,Node),br])
      ;
          true
      ),
      fail
    ;
        showing_help(end, ['"', w(Sub), '"'])
    ).

find_substring_in_WL(Sub, StrOpen, WL) :-
    member(El, WL),
    nonvar(El),
    ( atom(El) ->
        atom_search(El, 1, Sub, _)
    ; El = chars(Chars) ->
        append(_, StrOpen, Chars)        
    ; El = file(F) ->
        get_load_path(LPath),
        member(Dir, [lib(lib)|LPath]),
        full_file_name(Dir, F, _, File),
        access(File, 4, 0),
        open(File, read, Strm),
        ( encoded_read_term(Strm, doc(_,WL1), []) ->
            true
        ;
            WL1 = []
        ),
        close(Strm),
        find_substring_in_WL(Sub, StrOpen, WL1)
    ;
        fail
    ).

%% show_help_modes(Cmds).
%%
%%  Display mode descriptors for each command in Cmds.
%%
%%  mode show_help_modes(@list(compound)).
%%
show_help_modes([]).
show_help_modes([cmd(Thy, Table, Name, Arity)|Cmds]) :-
    show_help_modes(Cmds),
    functor(Cmd, Name, Arity),
    (
        find_command(Cmd, Table, Thy, Modes, _, _),
        show_help_modes_modes(Modes),
        fail
    ;
        query_global(gumtree_tactic_declaration(Cmd, Mode, TArity)),
        show_help_modes_modes([arity(Mode,TArity)]),
        fail
    ;
        true
    ).

show_help_modes_modes([]).
show_help_modes_modes([Mode|Modes]) :-
    show_help(['mode ']),
    show_help_modes_mode(Mode),
    show_help(['.', br]),
    show_help_modes_modes(Modes).

show_help_modes_mode(arity(Mode,Arity)) :-
    !,
    show_help_modes_mode(Mode),
    show_help([' arity ' , w(Arity)]).
show_help_modes_mode(Mode) :-
    functor(Mode, Fun, Ar),
    show_help([w(Fun)]),
    ( Ar =:= 0 ->
        true
    ;
        show_help(['(']),
        show_help_modes_args(Mode, 1, Ar),
        show_help([')'])
    ).

show_help_modes_args(Mode, N, Ar) :-
    arg(N, Mode, Arg),
    Arg = Dir(Type),
    show_help([w(Dir), w(Type)]),
    ( N =:= Ar ->
        true
    ;
        show_help([', ']),
        N1 is N + 1,
        show_help_modes_args(Mode, N1, Ar)
    ).

%% show_help(Doc).
%% 
%%  Display a line of help.
%%
%%  mode show_help(@writelist).
%%
show_help(Doc) :-
    writelist(Doc).

%% showing_help(Action, Doc).
%%
%%  Start/finish showing a help node.
%%
%%  mode showing_help(@atom, @writelist).
%%
showing_help(Action, Doc) :-
    handle_event(showing_help(Action, Doc)),
    !.
showing_help(_, _).

%% show_help_list(Docs).
%% 
%%  Display some lines of help.
%%
%%  mode show_help_list(@list(writelist)).
%% 
show_help_list([]).
show_help_list([Doc|Docs]) :-
    show_help(Doc),
    show_help([br]),
    show_help_list(Docs).


%% show_help_ext(Name).
%%
%%  Sort and display all help extensions for the doc_node called Name.
%%
%%  mode show_help_ext(@atom).
%% 
show_help_ext(Node) :-
    get_doc_extns(Node, Docs),
    !,
    show_help_list(Docs).
show_help_ext(_).


%% get_doc_node(Name, Hdr, Doc, Cmds, Thy).
%% get_doc_extns(Name, Extns).
%%
%%  These return information about the doc_node called Name.
%%
%%  Extns is the list of doc_extension writelists
%%  for the doc_node called Name.
%%
%%  mode get_doc_node(@atom, ?writelist, ?writelist, ?list(compound), -theory).
%%  mode get_doc_extns(@atom, ?list(writelist)).
%% 
get_doc_node(Node, Hdr, Doc, Cmds, Thy) :-
    (
        Thy = global
    ;
        ancestor(Thy)
    ),
    query_theory(doc_node(Node,Hdr,Doc,Cmds), Thy),
    !.

get_doc_extns(Node, Extns) :-
    setof(Doc, Thy^
          ( ( ancestor(Thy)
            ; Thy = global
            ),
            query_theory(doc_extension(Node, Doc), Thy)
          ),
          Extns),
    !.
get_doc_extns(_, []).

?- (\+ query_global(request(showing_help(Action, Doc),_))
    ->
        add_request(showing_help(Action, Doc),
                    ['showing_help(Action, Doc).',nl,
                     'Action is ''start'' or ''end'',',nl,
                     'Doc is a label for the display.'])
   ;
       true
   ).
