%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: mktex.ql,v 1.35 2001/06/04 01:24:11 itp Exp $

#include "../prover/lib/ergo.h"

:- op(990, xfx, --->).

database command
  mktex
mode
  mktex
doc
  'Write the current theory (Thy) as LaTeX to Thy.tex.'
extends
  misc_cmds.

database command
  mktex(Thy)
mode
  mktex(@theory)
doc
  'Write the theory as LaTeX to Thy.tex.'
extends
  misc_cmds.

%% mktex.
%% mktex(Thy).
%%
%%  Write a LaTeX version of Thy (by default the current theory) to Thy.tex.
%%
%%  mode mktex.
%%  mode mktex(@atom).
%%
mktex :- 
    global_state_lookup(current_dir,Thy),
    mktex(Thy).

mktex(Thy) :-
    ( query_global(subthy_owner(Thy, Thy1)) ->
        cd(Thy1)        %% Cd to the owning theory.
    ;
        cd(Thy)
    ),
    atom_append(Thy, '.tex', OName),
    open(OName, write, OStream),
    current_output(OldOut), set_output(OStream),
    mktex_section(Thy),
    mktex_ancestors(Thy),
    mktex_owner(Thy),
    mktex_const(Thy),
    mktex_function(Thy),
    mktex_quant(Thy),
    mktex_op(Thy),
    mktex_abbrev(Thy),
    mktex_elf(Thy),
    mktex_tactic(Thy),
    mktex_openingrule(Thy),
    mktex_rule(Thy),
    mktex_rulegroup(Thy),
    set_output(OldOut),
    close(OStream).


%% private mktex_section(Thy).
%%
%%  Write a LaTeX section header for Thy.
%%
%%  mode mktex_section(@atom).
%%
mktex_section(Thy) :-
    mktex_name(Thy, Thy1),
    writelist(['\\section{The Theory \\theory{', Thy1, '}}',
               '\\label{sec:', Thy, '.gen}', nl]).


%% private mktex_ancestors(Thy).
%%
%%  If Thy has ancestors then write a LaTeX version of them.
%%
%%  mode mktex_ancestors(@atom).
%%
mktex_ancestors(Thy) :-
    ( (query_global(ancestor(Thy,Ancestors)),
      %% Any ancestors
      (\+ (Ancestors = []))) ->
        mktex_ancestors1(Thy,Ancestors)
    ;
        true
    ).


%% private mktex_ancestors1(Thy, Ancestors).
%%
%%  Write a LaTeX version of the ancestors of Thy.
%%
%%  mode mktex_ancestors1(@atom, @list(ground)).
%%
mktex_ancestors1(_, Ancestors) :-
    writelist(['\\subsection*{Ancestor Theories}', nl]),
    writelist([
               '\\tablefirsthead{\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Ancestor}', nl,
               ' &\\multicolumn{1}{|c|}{\\bf Visible}', nl,
               ' &\\multicolumn{1}{|c|}{\\bf Operators}', nl,
               ' &\\multicolumn{1}{|c|}{\\bf Abbreviations}',
               '\\\\\\hline\\hline}', nl,
               '\\tablehead{\\hline \\multicolumn{4}{|l|}%', nl,
               '{\\small\\sl continued from previous page}\\\\', nl,
               '\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Ancestor}', nl,
               ' &\\multicolumn{1}{|c|}{\\bf Visible}', nl,
               ' &\\multicolumn{1}{|c|}{\\bf Operators}', nl,
               ' &\\multicolumn{1}{|c|}{\\bf Abbreviations}',
               '\\\\\\hline}', nl,
               '\\tabletail{\\hline\\multicolumn{4}{|r|}%', nl,
               '{\\small\\sl continued on next page}\\\\\\hline}', nl,
               '\\tablelasttail{}', nl]),
    writelist(['\\begin{supertabular}{|l|c|c|c|}', nl]),
    sort(Ancestors,Ancestors1),
    member(Name1, Ancestors1),
    (query_theory(visible_theory(Name1)) ->
        Vis = '\\tick'
    ;
        Vis = '$\\times$'
    ),
    (query_theory(operator(_,_,_,Name1)) ->
        Ops = '\\tick'
    ;
        Ops = '$\\times$'
    ),
    (query_theory(abbrev(_,_,_,Name1)) ->
        Abbrevs = '\\tick'
    ;
        Abbrevs = '$\\times$'
    ),
    writelist(['\\verb"', Name1, '"', nl,
               ' &', Vis, nl,
               ' &', Ops, nl,
               ' &', Abbrevs, '\\\\\\hline', nl]),
    fail.
mktex_ancestors1(_, _) :-
    writelist(['\\end{supertabular}', nl]).


%% private mktex_owner(Thy).
%%
%%  If Thy is a subtheory then write a LaTeX description.
%%
%%  mode mktex_owner(@atom).
%%
mktex_owner(Thy) :-
    ( query_global(subthy_owner(Thy,Owner)) ->
        mktex_owner1(Thy, Owner)
    ;
        true
    ).


%% private mktex_owner1(Thy, Owner).
%%
%%  Write a LaTeX description of the subtheories owner.
%%
%%  mode mktex_const1(@atom,@atom).
%%
mktex_owner1(Thy, Owner) :-
    writelist(['\\subsection*{Owning Theory}', nl,
               'The theory \\verb"', Thy,
               '" is a subtheory of \\verb"', Owner, '".', nl]).


%% private mktex_const(Thy).
%%
%%  If Thy has constants then write a LaTeX version of them.
%%
%%  mode mktex_const(@atom).
%%
mktex_const(Thy) :-
    ( ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
      %% Any constants
      query_theory(symbol(_,[],_), Thy1)) ->
        mktex_const1(Thy)
    ;
        true
    ).


%% private mktex_const1(Thy).
%%
%%  Write a LaTeX version of the constants of Thy.
%%
%%  mode mktex_const1(@atom).
%%
mktex_const1(Thy) :-
    writelist(['\\subsection*{Constants}', nl]),
    writelist([
               '\\tablefirsthead{\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Name}\\\\\\hline\\hline}', nl,
               '\\tablehead{\\hline \\multicolumn{1}{|l|}%', nl,
               '{\\small\\sl continued from previous page}\\\\', nl,
               '\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Name}\\\\\\hline}', nl,
               '\\tabletail{\\hline\\multicolumn{1}{|r|}%', nl,
               '{\\small\\sl continued on next page}\\\\\\hline}', nl,
               '\\tablelasttail{}', nl]),
    writelist(['\\begin{supertabular}{|l|}', nl]),
    setof(Name,
          (Kind, Thy1) ^
          ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
            query_theory(symbol(Name,[],Kind), Thy1)),
          Names),
    member(Name1, Names),
    abbreviate(Name1, Name2),
    writelist(['\\verb"', Name2, '"\\\\\\hline', nl]),
    fail.
mktex_const1(_) :-
    writelist(['\\end{supertabular}', nl]).


%% private mktex_function(Thy).
%%
%%  If Thy has functions then write a LaTeX version of them.
%%
%%  mode mktex_function(@atom).
%%
mktex_function(Thy) :-
    ( ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
      query_theory(symbol(_,[Arity],_), Thy1),
      %% Any functions
      integer(Arity)) ->
        mktex_function1(Thy)
    ;
        true
    ).


%% private mktex_function1(Thy).
%%
%%  Write a LaTeX version of the functions of Thy.
%%
%%  mode mktex_function1(@atom).
%%
mktex_function1(Thy) :-
    writelist([nl, '\\subsection*{Operations}', nl]),
    writelist([
               '\\tablefirsthead{\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Name}', nl,
               '&\\multicolumn{1}{c|}{\\bf Arity}\\\\\\hline\\hline}', nl,
               '\\tablehead{\\hline \\multicolumn{2}{|l|}%', nl,
               '{\\small\\sl continued from previous page}\\\\', nl,
               '\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Name}', nl,
               '&\\multicolumn{1}{c|}{\\bf Arity}\\\\\\hline}', nl,
               '\\tabletail{\\hline\\multicolumn{2}{|r|}%', nl,
               '{\\small\\sl continued on next page}\\\\\\hline}', nl,
               '\\tablelasttail{}', nl]),
    writelist(['\\begin{supertabular}{|l|l|}', nl]),
    setof((Name, Arity),
          (Kind, Thy1) ^
           ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
            query_theory(symbol(Name,[Arity],Kind), Thy1), integer(Arity)),
          Names),
    member((Name1, Arity1), Names),
    functor(Func, Name1, Arity1),
    abbreviate(Func, Func1),
    functor(Func1, Name2, Arity1),
    writelist(['\\verb"', Name2, '"',
               ' &', w(Arity1), '\\\\\\hline', nl]),
    fail.
mktex_function1(_) :-
    writelist(['\\end{supertabular}', nl]).


%% private mktex_quant(Thy).
%%
%%  If Thy has quantifiers then write a LaTeX version of them.
%%
%%  mode mktex_quant(@atom).
%%
mktex_quant(Thy) :-
    ( ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
      %% Any constants
      query_theory(symbol(_,[quant(_)],_), Thy1)) ->
        mktex_quant1(Thy)
    ;
        true
    ).


%% private mktex_quant1(Thy).
%%
%%  Write a LaTeX version of the quantifiers of Thy.
%%
%%  mode mktex_quant1(@atom).
%%
mktex_quant1(Thy) :-
    writelist(['\\subsection*{Quantifiers}', nl]),
    writelist([
               '\\tablefirsthead{\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Name}',
               ' &\\multicolumn{1}{|c|}{\\bf Type}\\\\\\hline\\hline}', nl,
               '\\tablehead{\\hline \\multicolumn{2}{|l|}%', nl,
               '{\\small\\sl continued from previous page}\\\\', nl,
               '\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Name}',
               ' &\\multicolumn{1}{|c|}{\\bf Type}\\\\\\hline}', nl,
               '\\tabletail{\\hline\\multicolumn{2}{|r|}%', nl,
               '{\\small\\sl continued on next page}\\\\\\hline}', nl,
               '\\tablelasttail{}', nl]),
    writelist(['\\begin{supertabular}{|l|r|}', nl]),
    setof((Name, Type),
          (Kind, Thy1) ^
           ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
            query_theory(symbol(Name,[quant(Type)],Kind), Thy1)),
          Names),
    member((Name1, Type1), Names),
    abbreviate(quant(Name1, x, B), quant(Name2, x, B)),
    mktex_quant_types(Type1, Type2),
    writelist(['\\verb"', Name2, '"',
               ' &', w(Type2), '\\\\\\hline', nl]),
    fail.
mktex_quant1(_) :-
    writelist(['\\end{supertabular}', nl]).


%% private mktex_quant_types(In, Out).
%%
%%  Convert the non-empty list In into a single atom consisting of a comma
%%  separated list of the atoms of In.
%%
%%  mode mktex_quant_types1(@list(atom), -atom).
%%
mktex_quant_types([Type], Type) :-
    !.
mktex_quant_types([Type|Rest], Out) :-
    mktex_quant_types1([Type|Rest], Out).


%% private mktex_quant_types1(In, Out).
%%
%%  Convert non-empty, non-singular list In into a single atom consisting of a
%%  comma seperated list of the atoms of In.
%%
%%  mode mktex_quant_types1(@list(atom), -atom).
%%
mktex_quant_types1([Type], Out) :-
    !,
    atom_append(', ', Type, Out).
mktex_quant_types1([Type|Rest], Out) :-
    mktex_quant_types(Rest, Rest1),
    atom_append(Type, ', ', Type1),
    atom_append(Type1, Rest1, Out).


%% private mktex_op(Thy).
%%
%%  If Thy has operators then write a LaTeX version of them.
%%
%%  mode mktex_op(@atom).
%%
mktex_op(Thy) :-
    ( ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
      %% Any operators
      query_theory(operator(_,_,_,Thy1))) ->
        mktex_op1(Thy)
    ;
        true
    ).


%% private mktex_op1(Thy).
%%
%%  Write a LaTeX version of the operators of Thy.
%%
%%  mode mktex_op1(@atom).
%%
mktex_op1(Thy) :-
    writelist([nl, '\\subsection*{Operators}', nl]),
    writelist([
               '\\tablefirsthead{\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Name}', nl,
               '&\\multicolumn{1}{c|}{\\bf Associativity}', nl,
               '&\\multicolumn{1}{c|}{\\bf Precedence}\\\\\\hline\\hline}', nl,
               '\\tablehead{\\hline \\multicolumn{3}{|l|}%', nl,
               '{\\small\\sl continued from previous page}\\\\', nl,
               '\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Name}', nl,
               '&\\multicolumn{1}{c|}{\\bf Associativity}', nl,
               '&\\multicolumn{1}{c|}{\\bf Precedence}\\\\\\hline}', nl,
               '\\tabletail{\\hline\\multicolumn{3}{|r|}%', nl,
               '{\\small\\sl continued on next page}\\\\\\hline}', nl,
               '\\tablelasttail{}', nl]),
    writelist(['\\begin{supertabular}{|l|l|r|}', nl]),
    setof((Name,Assoc,Prec),
          Thy1 ^
           ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
            query_theory(operator(Name,Assoc,Prec,Thy1))),
          Names),
    member((Name1,Assoc1,Prec1), Names),
    abbreviate(Name1, Name2),
    operator_kind(Assoc2, Assoc1),
    writelist(['\\verb"', Name2, '"', 
               ' &', Assoc2,
               ' &', w(Prec1), '\\\\\\hline', nl]),
    fail.
mktex_op1(_) :-
    writelist(['\\end{supertabular}', nl]).


%% private mktex_abbrev(Thy).
%%
%%  If Thy has abbreviations then write a LaTeX version of them.
%%
%%  mode mktex_abbrev(@atom).
%%
mktex_abbrev(Thy) :-
    (   ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
      query_theory(abbrev(_,_,_,Thy1), Thy1)) %% Any abbreviations
      ->
          mktex_abbrev1(Thy)
    ;
        true
    ).


%% private mktex_abbrev1(Thy).
%%
%%  Write a LaTeX version of the abbreviations of Thy.
%%
%%  mode mktex_abbrev1(@atom).
%%
mktex_abbrev1(Thy) :-
    writelist([nl, '\\subsection*{Abbreviations}', nl]),
    writelist([
               '\\tablefirsthead{\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Abbreviation}', nl,
               '&\\multicolumn{1}{c|}{\\bf Expansion}\\\\\\hline\\hline}', nl,
               '\\tablehead{\\hline \\multicolumn{2}{|l|}%', nl,
               '{\\small\\sl continued from previous page}\\\\', nl,
               '\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Abbreviation}', nl,
               '&\\multicolumn{1}{c|}{\\bf Expansion}\\\\\\hline}', nl,
               '\\tabletail{\\hline\\multicolumn{2}{|r|}%', nl,
               '{\\small\\sl continued on next page}\\\\\\hline}', nl,
               '\\tablelasttail{}', nl]),
    writelist(['\\begin{supertabular}{|l|l|}', nl]),
    setof((Abbrev, Exp),
          (Cond, Thy1) ^
           ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
            query_theory(abbrev(Abbrev,Exp,Cond,Thy1),Thy1)),
          Names),
    member((Abbrev1, Exp1), Names),
    writelist(['\\verb"', thy(Abbrev1), '"',
               ' &\\verb"', thy(Exp1), '"\\\\\\hline', nl]),
    fail.
mktex_abbrev1(_) :-
    writelist(['\\end{supertabular}', nl]).


%% private mktex_openingrule(Thy).
%%
%%  If Thy has rules then write a LaTeX version of them.
%%
%%  mode mktex_rule(@atom).
%%
mktex_openingrule(Thy) :-
    ( ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
      %% Any opening rules
      query_theory(openrule(_,_,_,_,_,_,_,_), Thy1)) ->
        mktex_openingrule1(Thy)
    ;
        true
    ).


%% private mktex_openingrule1(Thy).
%%
%%  Write a LaTeX version of the opening rules of Thy.
%%
%%  mode mktex_openingrule1(@atom).
%%
mktex_openingrule1(Thy) :-
    writelist([nl, '\\subsection*{Opening Rules}', nl]),
    writelist(['\\begin{itemize}', nl]),
    setof((Func,Place,NewHyps,TransHyps,TransRel,TransGoal),
          (Thy1, Close) ^
           ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
            query_theory(openrule(Func,Place,_,NewHyps,
                                  TransHyps,TransRel,TransGoal, Close), Thy1)),
          Rules),
    member((Func1,Place1,NewHyps1,TransHyps1,TransRel1,TransGoal1), Rules),
    abbreviate(Func1, Func2),
    mktex_long_verb(Func2, Func3),
    write('\\item'),
    write_long_verb(Func3),
    writelist([' opened at \\verb"', w(Place1), '" gives:\\\\', nl]),
    mktex_newhyps(NewHyps1),
    mktex_hyptrans(TransHyps1),
    mktex_reltrans(TransRel1),
    mktex_goaltrans(TransGoal1),
    fail.
mktex_openingrule1(_) :-
    writelist(['\\end{itemize}', nl]).


%% private mktex_newhyps(NewHyps)
%%
%%  Write a LaTeX version of the new hypotheses of an opening rule.
%%
%%  mode mktex_newhyps(@list(term)).
%%
mktex_newhyps([]) :-
    !.
mktex_newhyps(L) :-
    writelist(['the new hypotheses:', nl,
               '\\begin{itemize}', nl]),
    mktex_newhyps1(L),
    writelist(['\\end{itemize}', nl]).


%% private mktex_newhyps1(NewHyps)
%%
%%  Write a LaTeX \item for each new hypotheses of an opening rule.
%%
%%  mode mktex_newhyps1(@list(term)).
%%
mktex_newhyps1([]).
mktex_newhyps1([Hyp | L]) :-
    abbreviate(Hyp, Hyp1),
    mktex_long_verb(Hyp1, Hyp2),
    write('\\item '),
    write_long_verb(Hyp2),
    nl,
    mktex_newhyps1(L).


%% private mktex_hyptrans(HypTrans).
%%
%%  Write a LaTeX version of the hypotheses transformations of an opening rule.
%%
%%  mode mktex_hyptrans(@list(term)).
%%
mktex_hyptrans([]) :-
    !.
mktex_hyptrans([(H '--->' H)]) :-
    !.
mktex_hyptrans(L) :-
    writelist(['the hypothesis transformations:', nl,
               '\\begin{itemize}', nl]),
    mktex_hyptrans1(L),
    writelist(['\\end{itemize}', nl]).


%% private mktex_hyptrans1(HypTrans).
%%
%%  Write a LaTeX \item for each new hypotheses transformation of an opening
%%  rule.
%%
%%  mode mktex_hyptrans1(@list(term)).
%%
mktex_hyptrans1([]) :-
    !.
mktex_hyptrans1([(H '--->' H)]) :-
    !.
mktex_hyptrans1([(OHyp '--->' NHyp) | L]) :-
    abbreviate(OHyp, OHyp1),
    mktex_long_verb(OHyp1, OHyp2),
    abbreviate(NHyp, NHyp1),
    mktex_long_verb(NHyp1, NHyp2),
    write('\\item '),
    write_long_verb(OHyp2),
    write(' $\\longrightarrow$ '),
    write_long_verb(NHyp2),
    nl,
    mktex_hyptrans1(L).


%% private mktex_reltrans(RelTrans).
%%
%%  Write a LaTeX version of the relation transformations of an opening rule.
%%
%%  mode mktex_reltrans(@list(ground)).
%%
mktex_reltrans([]) :-
    !.
mktex_reltrans(L) :-
    writelist(['the relation transformations:', nl,
               '\\begin{itemize}', nl]),
    mktex_reltrans1(L),
    writelist(['\\end{itemize}', nl]).


%% private mktex_reltrans1(HypTrans).
%%
%%  Write a LaTeX \item for each new relation transformation of an opening
%%  rule.
%%
%%  mode mktex_reltrans1(@list(ground), @list(ground)).
%%
mktex_reltrans1([]).
mktex_reltrans1([(ORel '--->' NRel) | L]) :-
    abbreviate(ORel, ORel1),
    mktex_long_verb(ORel1, ORel2),
    abbreviate(NRel, NRel1),
    mktex_long_verb(NRel1, NRel2),
    write('\\item '),
    write_long_verb(ORel2),
    write(' $\\longrightarrow$ '),
    write_long_verb(NRel2),
    nl,
    mktex_reltrans1(L).


%% private mktex_goaltrans(GoalTrans).
%%
%%  Write a LaTeX version of the goal transformations of an opening rule.
%%
%%  mode mktex_goaltrans(@list(term)).
%%
mktex_goaltrans([]) :-
    !.
mktex_goaltrans([(G '--->' G)]) :-
    !.
mktex_goaltrans(L) :-
    writelist(['the goal transformations:', nl,
               '\\begin{itemize}', nl]),
    mktex_goaltrans1(L),
    writelist(['\\end{itemize}', nl]).

%% private mktex_goaltrans1(GoalTrans).
%%
%%  Write a LaTeX \item for each new goal transformation of an opening rule.
%%
%%  mode mktex_goaltrans1(@list(term)).
%%
mktex_goaltrans1([]) :-
    !.
mktex_goaltrans1([(G '--->' G)]) :-
    !.
mktex_goaltrans1([(OGoal '--->' NGoal) | L]) :-
    abbreviate(OGoal, OGoal1),
    mktex_long_verb(OGoal1, OGoal2),
    abbreviate(NGoal, NGoal1),
    mktex_long_verb(NGoal1, NGoal2),
    write('\\item '),
    write_long_verb(OGoal2),
    write(' $\\longrightarrow$ '),
    write_long_verb(NGoal2),
    nl,
    mktex_goaltrans1(L).


%% private mktex_rule(Thy).
%%
%%  If Thy has rules then write a LaTeX version of them.
%%
%%  mode mktex_rule(@list(ground), @atom).
%%
mktex_rule(Thy) :-
    ( ((Thy1 = Thy ; query_global(subthy_owner(Thy1, Thy))),
      %% Any Rules
      query_theory(rule(_,_,_,_,_), Thy1)) ->
        mktex_rule1(Thy)
    ;
        true
    ).


%% private mktex_rule1(Thy).
%%
%%  Write a LaTeX version of the rules of Thy.
%%
%%  mode mktex_rule1(@atom).
%%
mktex_rule1(Thy) :-
    writelist([nl, '\\subsection*{Rules}', nl]),
    writelist(['\\begin{flushleft}', nl,
               '\\begin{description}', nl]),
    %% Try to conserve buffer area.
    findall(rule(Kind,Name,Stmt,Body),
            ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
              query_theory(rule(Name,Kind,_,Stmt,Body), Thy1)),
            Rules),
    sort(Rules, Rules1),
    member(rule(Kind1,Name1,Stmt1,Body1), Rules1),
    abbreviate(Stmt1, Stmt2),
    mktex_long_verb(Stmt2, Stmt3),
    mktex_rule_kind(Kind1, Kind2),
    writelist(['\\item[', Kind2, '] \\verb"', Name1, '":\\\\', nl]),
    write_long_verb(Stmt3),
    ( (\+ (Body1 = true)) ->
        write(' {\\bf provided }'),
        mktex_provided(Body1)
    ;
        true
    ),
    nl,
    fail.
mktex_rule1(_) :-
    writelist(['\\end{description}', nl,
               '\\end{flushleft}', nl]).


%% private mktex_rule_kind(Kind, PrintKind).
%%
%%  Convert the rule kind Kind into it's output format.
%%
%%  mode mktex_rule_kind(@atom, -atom).
%%
mktex_rule_kind(axiom,       'Axiom') :- !.
mktex_rule_kind(axiom(_),    'Axiom') :- !.
mktex_rule_kind(defn,        'Definition') :- !.
mktex_rule_kind(postulate,   'Postulate') :- !.
mktex_rule_kind(theorem,     'Theorem') :- !.
mktex_rule_kind(theorem(_),  'Theorem') :- !.
mktex_rule_kind(X,           X).


%% private mktex_provided(NFI, NFIs).
%%
%%  Convert a list of not free in conditions into LaTeX output.
%%
%%  mode mktex_provided(@term).
%%
mktex_provided(true).
mktex_provided((V not_free_in B)) :-
    mktex_long_verb(V, V1),
    mktex_long_verb(B, B1),
    write_long_verb(V1),
    write(' {\\bf not~free~in} '),
    write_long_verb(B1).
mktex_provided(((V not_free_in B), true)) :-
    !,
    mktex_long_verb(V, V1),
    mktex_long_verb(B, B1),
    write_long_verb(V1),
    write(' {\\bf not~free~in} '),
    write_long_verb(B1).
mktex_provided(((V not_free_in B), NFIs)) :-
    mktex_long_verb(V, V1),
    mktex_long_verb(B, B1),
    write_long_verb(V1),
    write(' {\\bf not~free~in} '),
    write_long_verb(B1),
    write(', '),
    mktex_provided(NFIs).


%% private mktex_rulegroup(Thy).
%%
%%  If Thy uses rule groups then write a LaTeX version of them.
%%
%%  mode mktex_rulegroup(@list(ground), @atom).
%%
mktex_rulegroup(Thy) :-
    ( ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
      %% Any operators
      query_theory(rulegroup(_,_,_,_,_,_,_))) ->
        mktex_rulegroup1(Thy)
    ;
        true
    ).


%% private mktex_rulegroup1(Thy).
%%
%%  Write a LaTeX version of the rule groups of Thy.
%%
%%  mode mktex_rulegroup1(@list(ground), @atom).
%%
mktex_rulegroup1(Thy) :-
    writelist([nl, '\\subsection*{Rule Groups}', nl]),
    writelist([
               '\\tablefirsthead{\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Group}', nl,
               '&\\multicolumn{1}{c|}{\\bf Theorem}', nl,
               '&\\multicolumn{1}{c|}{\\bf Direction}\\\\\\hline\\hline}', nl,
               '\\tablehead{\\hline \\multicolumn{3}{|l|}%', nl,
               '{\\small\\sl continued from previous page}\\\\', nl,
               '\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Group}', nl,
               '&\\multicolumn{1}{c|}{\\bf Theorem}', nl,
               '&\\multicolumn{1}{c|}{\\bf Direction}\\\\\\hline}', nl,
               '\\tabletail{\\hline\\multicolumn{3}{|r|}%', nl,
               '{\\small\\sl continued on next page}\\\\\\hline}', nl,
               '\\tablelasttail{}', nl]),
    writelist(['\\begin{supertabular}{|l|l|l|}', nl]),
    setof((Group,Name,Dir),
          (Thy1, In, Conds, Reln, Out, ThmThm) ^
           ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
            query_theory(rulegroup(Group, In, Conds, Reln, Out, ThmThm:Name, Dir))),
          Groups),
    member((Group1,Name1,Dir1), Groups),
    writelist(['\\verb"', Group1, '"', 
               ' &\\verb"', Name1, '"',
               ' &', w(Dir1), '\\\\\\hline', nl]),
    fail.
mktex_rulegroup1(_) :-
    writelist(['\\end{supertabular}', nl]).

%% private mktex_name(In, Out).
%%
%%  Convert the atom In into a format suitable to be read in by LaTeX.
%%
%%  mode mktex_name(@atom, -atom).
%%
mktex_name(In, Out) :-
    name(In,N1),
    mktex_name1(N1, N2),
    name(Out,N2).


%% private mktex_name1(In, Out).
%%
%%  Convert the name In (represented as a list of characters) into a format
%%  suitable to be read in by TeX.
%%
%%  mode mktex_name1(@list(atom), -list(atom)).
%%
mktex_name1([], []) :-
    !.
mktex_name1([0'_|In], [0'\\, 0'_ |Out]) :-
    !,
    mktex_name1(In, Out).
mktex_name1([L|In], [L |Out]) :-
    mktex_name1(In, Out).

%% private mktex_long_verb(Term, Chars).
%%
%%  Transform a term into a series of space delimited LaTeX \verb's.
%%  This will allow long terms to be broken across lines.
%%
%%  mode mktex_long_verb(@term, -list(integer)).
%%
mktex_long_verb(Term, Chars) :-
    open_string(write, Stream),
    global_state_lookup(current_dir,Thy),
    write_term(Stream, Term, [op_table(Thy),numbervars(true),
                              obvar_prefix_table(Thy)]),
    stream_to_chars(Stream, Chars1),
%    close(Stream),
    mktex_long_verb1(Chars1, Chars).


%% private mktex_long_verb1(In, Out).
%%
%%  This is an auxilary predicate for mktex_long_verb/2.  Transform a
%%  list of characters into a series of space delimited LaTeX \verb's.
%%  This will allow long terms to be broken across lines.
%%
%%  mode mktex_long_verb1(@list(integer), -list(integer)).
%%
mktex_long_verb1([], []) :-
    !.
mktex_long_verb1([C], [0'\\,0'v,0'e,0'r,0'b,0'",C,0'"]) :-
    !.
mktex_long_verb1([0' |L], L1) :-
    !,
    mktex_long_verb2([0' |L], L1).
mktex_long_verb1(L, [0'\\,0'v,0'e,0'r,0'b,0'"|L1]) :-
    !,
    mktex_long_verb2(L, L1).


%% private mktex_long_verb2(In, Out).
%%
%%  This is an auxilary predicate for mktex_long_verb/2.  Transform a list of
%%  characters into a series of space delimited LaTeX \verb's.  This will allow
%%  long terms to be broken across lines.  This predicate assumes that the
%%  original list was longer than one character.
%%
%%  mode mktex_long_verb2(@list(integer), -list(integer)).
%%
mktex_long_verb2([0' ], [0' ]) :-
    !.
mktex_long_verb2([C], [C, 0'"]) :-
    !.
mktex_long_verb2([0' , 0' |L], [0' ,L1]) :-
    !,
    mktex_long_verb2([0' |L], L1).
mktex_long_verb2([0' , C|L], [0' ,0'\\,0'v,0'e,0'r,0'b,0'"|L1]) :-
    !,
    mktex_long_verb2([C|L], L1).
mktex_long_verb2([C, 0' | L], [C, 0'"|L1]) :-
    !,
    mktex_long_verb2([0' |L], L1).
mktex_long_verb2([C|L], [C|L1]) :-
    !,
    mktex_long_verb2(L, L1).


%% private write_long_verb(Chars)
%%
%% Send the characters in list to the current output.
%%
%% mode write_long_verb(@list(integer)).
%%
write_long_verb(Chars) :-
    map(put_code, [Chars]).

database command
  mksignature
mode
  mksignature
doc
  'Write the current theory''s (Thy) signature, as LaTeX, to Thy.tex.'
extends
  misc_cmds.

database command
  mksignature(Thy)
mode
  mksignature(@theory)
doc
  'Write the theory''s signature, as LaTeX, to Thy.tex.'
extends
  misc_cmds.

%% mksignature.
%% mksignature(Thy).
%%
%%  Write a LaTeX version of the signature of Thy (by default the current
%%  theory) to Thy.tex.
%%
%%  mode mksignature.
%%  mode mksignature(@atom).
%%
mksignature :- 
    global_state_lookup(current_dir,Thy),
    mksignature(Thy).

mksignature(Thy) :-
    ( query_global(subthy_owner(Thy, Thy1)) ->
        cd(Thy1)        %% Cd to the owning theory.
    ;
        cd(Thy)
    ),
    atom_append(Thy, '.tex', OName),
    open(OName, write, OStream),
    current_output(OldOut), set_output(OStream),
    mktex_section(Thy),
    mktex_signature(Thy),
    set_output(OldOut),
    close(OStream).

%% private mktex_signature(Thy).
%%
%%  If Thy has functions or constants then write a LaTeX version of them.
%%
%%  mode mktex_signature(@atom).
%%
mktex_signature(Thy) :-
    ( ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
      (   query_theory(symbol(Name,[Arity],Kind), Thy1)
       ;
          query_theory(symbol(Name,[],Kind), Thy1)
      ),
      %% Any functions
      integer(Arity)) ->
        mktex_signature1(Thy)
    ;
        true
    ).


%% private mktex_signature1(Thy).
%%
%%  Write a LaTeX version of the signature of Thy.
%%
%%  mode mktex_signature1(@atom).
%%
mktex_signature1(Thy) :-
    writelist([
               '\\tablefirsthead{\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Name}', nl,
               '&\\multicolumn{1}{c|}{\\bf Arity}\\\\\\hline\\hline}', nl,
               '\\tablehead{\\hline \\multicolumn{2}{|l|}%', nl,
               '{\\small\\sl continued from previous page}\\\\', nl,
               '\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Name}', nl,
               '&\\multicolumn{1}{c|}{\\bf Arity}\\\\\\hline}', nl,
               '\\tabletail{\\hline\\multicolumn{2}{|r|}%', nl,
               '{\\small\\sl continued on next page}\\\\\\hline}', nl,
               '\\tablelasttail{}', nl]),
    writelist(['\\begin{supertabular}{|l|l|}', nl]),
    setof((Name, Arity),
          (Kind, Thy1) ^
           ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
            (  (query_theory(symbol(Name,[Arity],Kind), Thy1),
               integer(Arity))
            ;
               query_theory(symbol(Name,[],Kind), Thy1),
               Arity = 0
            )),
          Names),
    member((Name1, Arity1), Names),
    functor(Func, Name1, Arity1),
    abbreviate(Func, Func1),
    functor(Func1, Name2, Arity1),
    writelist(['\\verb"', Name2, '"',
               ' &', w(Arity1), '\\\\\\hline', nl]),
    fail.
mktex_signature1(_) :-
    writelist(['\\end{supertabular}', nl]).


%% private mktex_elf(Thy).
%%
%%  If Thy has elves then write a LaTeX version of them.
%%
%%  mode mktex_elf(@atom).
%%
mktex_elf(Thy) :-
    ( ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
      query_theory(elf(_, _), Thy1)) ->
        mktex_elf1(Thy)
    ;
        true
    ).


%% private mktex_elf1(Thy).
%%
%%  Write a LaTeX version of the elves of Thy.
%%
%%  mode mktex_elf1(@atom).
%%
mktex_elf1(Thy) :-
    writelist([nl, '\\subsection*{Elves}', nl]),
    writelist([
               '\\tablefirsthead{\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Name}', nl,
               '&\\multicolumn{1}{c|}{\\bf Active}\\\\\\hline\\hline}', nl,
               '\\tablehead{\\hline \\multicolumn{2}{|l|}%', nl,
               '{\\small\\sl continued from previous page}\\\\', nl,
               '\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Name}', nl,
               '&\\multicolumn{1}{c|}{\\bf Active}\\\\\\hline}', nl,
               '\\tabletail{\\hline\\multicolumn{2}{|r|}%', nl,
               '{\\small\\sl continued on next page}\\\\\\hline}', nl,
               '\\tablelasttail{}', nl]),
    writelist(['\\begin{supertabular}{|l|l|}', nl]),
    setof((Name,FullName),
          (Body, Thy1, Thy2) ^
           ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
            query_theory(elf(Name, Body), Thy1),
            atom_concat2(Thy1, ':', Thy2),
            atom_concat2(Thy2, Name, FullName)),
          Names),
    member((Name1, FullName1), Names),
    get_elves_list(Elves),
    get_active_elves(NrActive),
    ((between(1, NrActive, ElfNr),
      nth1(ElfNr, Elves, FullName1)) ->
        Active = '\\tick'
    ;
        Active = '$\\times$'
    ),
    writelist(['\\verb"', Name1, '"',
               ' &', Active, '\\\\\\hline', nl]),
    fail.

mktex_elf1(_) :-
    writelist(['\\end{supertabular}', nl]).


%% private mktex_tactic(Thy).
%%
%%  If Thy has proof tactics then write a LaTeX version of them.
%%
%%  mode mktex_elf(@atom).
%%
mktex_tactic(Thy) :-
    ( ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
      query_theory(proof(_,_,_,_), Thy1)) ->
        mktex_tactic1(Thy)
    ;
        true
    ).


%% private mktex_tactic1(Thy).
%%
%%  Write a LaTeX version of the tactics of Thy.
%%
%%  mode mktex_elf1(@atom).
%%
mktex_tactic1(Thy) :-
    writelist([nl, '\\subsection*{Tactics}', nl]),
    writelist([
               '\\tablefirsthead{\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Name}', nl,
               '&\\multicolumn{1}{c|}{\\bf Arity}\\\\\\hline\\hline}', nl,
               '\\tablehead{\\hline \\multicolumn{2}{|l|}%', nl,
               '{\\small\\sl continued from previous page}\\\\', nl,
               '\\hline', nl,
               '\\multicolumn{1}{|c|}{\\bf Name}', nl,
               '&\\multicolumn{1}{c|}{\\bf Arity}\\\\\\hline}', nl,
               '\\tabletail{\\hline\\multicolumn{2}{|r|}%', nl,
               '{\\small\\sl continued on next page}\\\\\\hline}', nl,
               '\\tablelasttail{}', nl]),
    writelist(['\\begin{supertabular}{|l|l|}', nl]),
    setof((Name, Arity),
          (Tactic, Modes, Impl, DocNode, Thy1) ^
           ((Thy1 = Thy; query_global(subthy_owner(Thy1, Thy))),
            query_theory(proof(Tactic,Modes,Impl,DocNode), Thy1),
           functor(Tactic, Name, Arity)),
          Tactics),
    member((Name1, Arity1), Tactics),
    writelist(['\\verb"', Name1, '"',
               ' &', w(Arity1), '\\\\\\hline', nl]),
    fail.

mktex_tactic1(_) :-
    writelist(['\\end{supertabular}', nl]).
