%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: prettyrule.ql,v 1.34 2002/06/27 05:10:26 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file defines the concrete syntax for inference rules.
%% It also handles deabbreviation of inference rules.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../prover/lib/ergo.h"

?- require_global_tactic_file(abbrev).

?- ( global_fact(theorem_action,1) ->
       true
   ;
       new_global_fact(theorem_action,1,[save(false),key(1),retract(true)])
   ).


database command
  axiom('==='(Name,RHS))
header
  'axiom Name === RHS'
extends
  theory_extension_cmds
implementation
  pretty_rule(Name,RHS),
  rule_make_fundamental(Name)
doc
  'Adds a fundamental inference rule called Name to the current theory.',
  para,
  'RHS can be just a predicate, or an inference rule of the form:',
  para,
  begin(verb),
  '    Seq1, Seq2, ... SeqN',nl,
  '    --------------------',nl,
  '    ConclSeq',nl,
  '  provided',nl,
  '    Constraints.',nl,
  end(verb),
  'The meaning of the rule is that any instance of ConclSeq that',nl,
  'satisfies Constraints can be proved by proving the corresponding',nl,
  'instances of Seq1...SeqN.  Operationally, when this rule is applied',nl,
  'to an open proof node P, ConclSeq is unified with that proof node,',nl,
  'any Constraints are evaluated, then new proof nodes corresponding to',nl,
  'Seq1..SeqN are created as children of P (the `parent'').',
  para,
  'Now we shall describe each of the components in more detail.',
  para,
  'Name must be an atom, or a compound term whose functor will be',nl,
  'treated as the name of the rule and whose arguments can be used',nl,
  'as parameters to the rule, typically to control how the rule',nl,
  'will be instantiated when it is applied.',
  para,
  'ConclSeq must be of the form: IdVar ::: Term.  However, the',nl,
  '`IdVar :::'' part is optional -- it is mainly used when one of the',nl,
  'constraint predicates needs to refer to the name of the proof node',nl,
  'to which the rule is being applied.',
  para,
  'Each of the Seq_i sequents before the line must be of the',nl,
  'form:  IdVar ::: (CName===CSpec) &&& ... ---> Pred.',nl,
  'However, the `IdVar :::'' part is optional, and so is everything',nl,
  'before the ''--->'' part.',
  para,
  'When the rule is applied each of these sequents becomes a node',nl,
  'in the proof tree.  IdVar is instantiated to the name of the node,',nl,
  '(it must be a variable when the rule is defined), Pred becomes the',nl,
  'predicate of the node and CSpec determines what CName context terms',nl,
  '(eg. hypotheses) are available at the node.',
  para,
  'Some logics may define several different classes of context',nl,
  'terms.  Each class of context terms has a name (eg. CName) and',nl,
  'contains a sequence of theory terms (with optional names).',nl,
  'Rules that manipulate multiple context lists will have several',nl,
  '(CName===CSpec) clauses, separated by ''&&&''.  Each of the',nl,
  '`CName''s must have been declared using ',
  xref(add_context_name, global, cmd(add_context_name/1)),'.',nl,
  'Note that parentheses are needed around each === expression.',
  para,
  'The syntax of each CSpec is:',
  para,
  begin(verb),
  '  CSpec ::= [Id:::Term,...]',nl,
  '         |  CName(Node)',nl,
  '         |  CSpec +++ [Id:::Term,...]',nl,
  '         |  CSpec --- [Id:::Term,...]',nl,
  '         |  CSpec ??? Filter.',nl,
  '         |  CSpec >>> Transformer.',nl,
  end(verb),
  'The semantics of most of these should be fairly obvious.',nl,
  'An empty list can be used to specify an empty context,',nl,
  'or [Id ::: f=2] specifies a context containing only f=2.',nl,
  'The CName(Node) specification returns the CName context from',nl,
  'Node (which must be the identity of the conclusion of the rule).',nl,
  'The +++ and --- modifiers add and remove context terms.',nl,
  'The ??? modifier returns the subset of the terms from CSpec',nl,
  'that satisfy Filter.  The most common Filter function is ',nl,
  'nfi(ObjectVar), which returns only those terms that do not',nl,
  'contain free occurences of ObjectVar.  It adds not_free_in',nl,
  'constraints to metavariables to ensure this if necessary.',nl,
  'The >>> modifier can be used to transform every context term',nl,
  'from CSpec, but suitable Transformer functions must be defined',nl,
  'first, and there is currently little support for doing this.',
  para,
  'Most sequents set their context lists to be identical or similar',nl,
  'to the context lists of the rules conclusion sequent.  Several',nl,
  'shorthands are provided to make this more convenient.  Firstly,',nl,
  'omitting a CName===CSpec entry is equivalent to CName===CName(ConclId),',nl,
  'where ConclId is the identity variable of the conclusion sequent.',nl,
  'Secondly, the ConclId defaults to the conclusion sequent, so the',nl,
  'above example can be written as just CName===CName.',nl,
  'Thirdly, `CName === CName(ConclId) Modifiers'', where Modifiers',nl,
  'is one or more +++, --- or >>> clauses, can be abbreviated to just',nl,
  'CName Modifiers.  Finally, context term identities (Id:::) can be',nl,
  'omitted if desired.',
  para,
  'With these shorthands, hyp+++[A,B] is equivalent to:',para,
  begin(verb),
  '    (hyp === hyp(ConclId) +++ [_ ::: A, _ ::: B]).',nl,
  end(verb),
  para,
  'This concrete syntax is unfolded into the lower-level syntax',nl,
  'used by ',xref(rule_add,global,'database:rule_add/5'),'.',
  para,
  'The line must contain at least 4 and at most 60 minus characters.',nl,
  'If there are no Seq_i sequents (N=0), then the line must be omitted.',
  para,
  'Constraints is a comma-separated series of ',nl,
    xref('constraint predicates',global,rule_constraint_cmds),
  ' which will be evaluated when',nl,
  'the rule is used to extend a proof. All of them must succeed (possibly',nl,
  'by adding delayed constraints) for an application of the rule to succeed.',
  para,
  'Examples:',
  para,
  begin(verb),
  '    axiom imp_reflex === A => A.',nl,
  nl,
  '    axiom imp_trans ===   A => B,',nl,
  '                          B => C',nl,
  '                        ----------',nl,
  '                          A => C.',nl,
  nl,
  '    axiom or_elim ===  I1:::                A or B,',nl,
  '                       I2::: hyp+++[A] ---> C,',nl,
  '                       I3::: hyp+++[B] ---> C,',nl,
  '                       ------------------------------',nl,
  '                       I :::                C',nl,
  '        provided',nl,
  '            annotate(I,cases(A or B)).',nl,
  nl,
  '    axiom all_intro ===   hyp ??? nfi(x) +++ [x:T] ---> C',nl,
  '                         ---------------------------------',nl,
  '                                  (all x:T C).',nl,
  nl,
  '    axiom assump === Id::: C',nl,
  '        provided',nl,
  '            context_search(hyp, Id, _, C).',
  end(verb)
mode
  axiom(@compound).


theorem_action parameter
  [set_theorem_action, get_theorem_action]
doc
  'This parameter allows a user-specified action to be triggered each',nl,
  'time a theorem is added to a theory via the `theorem'' command.',
  para,
  'If the parameter is set to `none'', then no action is performed.',nl,
  'Otherwise, the parameter must be set to a pattern that matches',nl,
  'a `database'' or `general'' command whose mode is one of the',nl,
  'following:',nl,
  begin(list),
  item(0),' Name -- no arguments',nl,
  item(1),' Name(M rulename) -- where M is @, + or ?.',nl,
  item(2),' Name(M1 rulename, M2 thy(term)) -- where M1 and M2',nl,
  '    are each @, + or ?.',nl,
  end(list),
  nl,
  'The Name command will be called immediately after the theorem has',nl,
  'been added.  The short name of the new theorem will be passed as',nl,
  'its first argument (if required) and the conclusion of the theorem',nl,
  'will be passed as its second argument (if required).',
  para,
  'A typical use is to set the parameter to `prove_rule(_,_)'', so',nl,
  'that each theorem is proved as soon as it is added.  Using',nl,
  'prove_rule/2 rather than prove_rule/1 means that the proof will',nl,
  'use the same variable names as the original theorem.',
  para,
  'Why not just start up a proof directly, rather than add a postulate',nl,
  'then prove it, you ask?  Well, an advantage of the postulate approach',nl,
  'is that the resulting theorem is guaranteed to be exactly what you',nl,
  'specify with this command, whereas an ordinary proof allows you to',nl,
  'instantiate variables and add constraints, which may result in a less',nl,
  'general theorem or one with the constraints expressed in a slightly',nl,
  'different form.'
extends
  parameters.

set_theorem_action(Action) :-
    check_legal_theorem_action(Action),
    tactic_retractall_global(theorem_action(_)),
    tactic_assert_global(theorem_action(Action)).

get_theorem_action(Action) :-
    query_global(theorem_action(Action)),
    !.
get_theorem_action(none).

check_legal_theorem_action(none) :-
    !.
check_legal_theorem_action(Action) :-
    std_nonvar(Action),
    member(Table, [database,general]),
    find_command(Action, Table, Modes, Impl, _),
    !,
    check_or_error(
        Impl = Action,
        ['implementation of command ',q(Action),' is too complex.']),
    check_or_error(
        ( Modes = [Mode],
          theorem_action_mode(Mode, F, A) ),
        [q(Action),' has wrong modes for a theorem action: ',q(Modes)]),
    check_or_error(
        current_predicate(F/A),
        ['predicate ',q(F/A),' is not loaded.']).
check_legal_theorem_action(Action) :-
    handle_error(['illegal theorem action: ',q(Action)]).


theorem_action_mode(F, F, 0) :-
    atom(F).
theorem_action_mode(F(M(rulename)), F, 1) :-
    atom(F),
    member(M, [@, +, ?]).
theorem_action_mode(F(M(rulename),M2(thy(term))), F, 2) :-
    atom(F),
    member(M, [@, +, ?]),
    member(M2, [@, +, ?]).


database command
  theorem('==='(Name,RHS))
header
  'theorem Name === RHS'
extends
  theory_extension_cmds
implementation
  pretty_rule(Name,RHS)
doc
  'Adds a theorem or inference rule called Name to the current theory',nl,
  'as a postulate.  Depending on the value of the ',
  xref(theorem_action,global,'parameter:theorem_action'),nl,
  'parameter, some further action may take place after the postulate',nl,
  'has been added, such as an interactive proof of the postulate or a',nl,
  'rerun of an existing proof.',
  para,
  'The arguments are the same as for the ',
  xref(rule,global,'database:rule/1'),' command.',nl,
  'However, this ''theorem'' command should be used for rules',nl,
  'that are believed to follow from the other rules in the theory.',nl,
  'Such rules should eventually be proven, or changed into fundamental',nl,
  'rules using the ',
  xref(rule_make_fundamental,global,'database:rule_make_fundamental/1'),
  ' command.'
mode
  theorem(@compound).


pretty_rule(Name, RHS) :-
    std_compound(RHS),
    RHS = provided(A,B),
    !,
    pretty_rule2(Name,A,B).
pretty_rule(Name, RHS) :-
    pretty_rule2(Name,RHS,true).

pretty_rule2(Name,Rule,Constraints) :-
    std_compound(Rule),
    Rule = Line(Seqs,Concl),
    rule_line(Line),
    !,
    deabbrev_rule_concl(Concl, ConclId, ConclPred),
    deabbrev_rule_seqs(Seqs, ConclId, Seqs2),
    deabbrev_rule_constraints(Constraints, Constraints2),
    rule_add(Name, ConclPred, ConclId, Seqs2, Constraints2),
    theorem_action(Name, ConclPred).
pretty_rule2(Name,Concl,Constraints) :-
    deabbrev_rule_concl(Concl, ConclId, ConclPred),
    deabbrev_rule_constraints(Constraints, Constraints2),
    rule_add(Name, ConclPred, ConclId, [], Constraints2),
    theorem_action(Name, ConclPred).


deabbrev_rule_concl(Concl, ConclId, ConclPred) :-
    std_compound(Concl),
    Concl = ':::'(ConclId,ConclPred0),
    !,
    deabbreviate(ConclPred0, ConclPred).
deabbrev_rule_concl(Concl, _, ConclPred) :-
    deabbreviate(Concl, ConclPred).


?- add_handler(deabbreviate_with_type(rule_constraint,C0,C),
               deabbrev_rule_constraints(C0,C)).

deabbrev_rule_constraints(true, Out) :-
    !,
    Out = true.
deabbrev_rule_constraints((A0,B0), (A,B)) :-
    !,
    deabbrev_rule_constraints(A0, A),
    deabbrev_rule_constraints(B0, B).
deabbrev_rule_constraints(C0, C) :-
    query_global_or_ancestors(rule_constraint(C0, Modes, _, _)),
    !,
    deabbreviate_cmd_relaxed(Modes, C0, C).
deabbrev_rule_constraints(C0, _) :-
    handle_error(['unknown rule constraint: ',q(C0)]).

    
theorem_action(Name, Concl) :-
    get_theorem_action(A),
    ( A = none ->
        true
    ; atom(A) ->
        do_theorem_action(A)
    ; functor(A, _, 1) ->
        arg(1, A, Name),
        do_theorem_action(A)
    ; functor(A, _, 2) ->
        arg(1, A, Name),
        arg(2, A, Concl),
        do_theorem_action(A)
    ;
        handle_error(program, ['illegal theorem action'])
    ).

do_theorem_action(A) :-
    call_predicate(A),
    !.
do_theorem_action(A) :-
    writelist(['WARNING: ',q(A),' failed.  Postulate is unchanged.',nl]).


deabbrev_rule_seqs(Seqs0, ParentId, Seqs) :-
    std_compound(Seqs0),
    Seqs0 = (L0, R0),
    !,
    deabbrev_rule_seqs(L0, ParentId, L),
    deabbrev_rule_seqs(R0, ParentId, R),
    append(L, R, Seqs).
deabbrev_rule_seqs(Seq0, ParentId, [seq(Id,CSpecs,Pred)]) :-
    deabbrev_rule_seq(Seq0, ParentId, Id, CSpecs, Pred).


deabbrev_rule_seq(Seq0, ParentId, Id, CSpecs, Pred) :-
    std_compound(Seq0),
    Seq0 = ':::'(Id0, RHS),
    !,
    Id = Id0,
    deabbrev_rule_seq_rhs(RHS, ParentId, CSpecs, Pred).
deabbrev_rule_seq(RHS, ParentId, _, CSpecs, Pred) :-
    deabbrev_rule_seq_rhs(RHS, ParentId, CSpecs, Pred).


deabbrev_rule_seq_rhs(RHS, ParentId, CSpecs, Pred) :-
    std_compound(RHS),
    RHS = '--->'(CSpecs0, Pred0),
    !,
    deabbrev_rule_cspecs(CSpecs0, ParentId, CSpecs),
    deabbreviate(Pred0, Pred).
deabbrev_rule_seq_rhs(Pred0, _, [], Pred) :-
    deabbreviate(Pred0, Pred).


%% This assumes left associative +++, --- etc. operators.
deabbrev_rule_cspecs('&&&'(A0,B0), ParentId, CSpecs) :-
    !,
    deabbrev_rule_cspecs(A0, ParentId, A),
    deabbrev_rule_cspecs(B0, ParentId, B),
    append(A, B, CSpecs).
deabbrev_rule_cspecs('==='(CName,Expr0), ParentId, CSpec) :-
    atom(CName),
    !,
    deabbrev_rule_cexpr(Expr0, ParentId, Expr),
    CSpec = [CName=Expr].
deabbrev_rule_cspecs(CSpec0, ParentId, CSpec) :-
    cexpr_starting_point(CSpec0, CName),
    !,
    deabbrev_rule_cexpr(CSpec0, ParentId, Expr),
    CSpec = [CName=Expr].
deabbrev_rule_cspecs(Spec0, _, _) :-
    handle_error(['bad context specification: ',q(Spec0)]).


cexpr_starting_point(CName, CName) :-
    atom(CName),
    CName \== [].
cexpr_starting_point(F(L,_), CName) :-
    member(F, [+++,---,???,>>>]),
    cexpr_starting_point(L, CName).


deabbrev_rule_cexpr(L, _, Out) :-
    list(L),
    !,
    deabbrev_rule_ctermlist(L, empty, add, Out).
deabbrev_rule_cexpr(CName(PId), PId, Out) :-
    atom(CName),
    !,
    Out = use(CName, PId).
deabbrev_rule_cexpr(CName, PId, Out) :-
    atom(CName),
    !,
    Out = use(CName, PId).
deabbrev_rule_cexpr('+++'(E0, L), PId, Out) :-
    list(L),
    !,
    deabbrev_rule_cexpr(E0, PId, E),
    deabbrev_rule_ctermlist(L, E, add, Out).
deabbrev_rule_cexpr('---'(E0, L), PId, Out) :-
    list(L),
    !,
    deabbrev_rule_cexpr(E0, PId, E),
    deabbrev_rule_ctermlist(L, E, delete, Out).
deabbrev_rule_cexpr('???'(E0, F), PId, Out) :-
    std_nonvar(F),
    !,
    deabbrev_rule_cexpr(E0, PId, E),
    Out = filter(F, E).
deabbrev_rule_cexpr('>>>'(E0, F), PId, Out) :-
    std_nonvar(F),
    !,
    deabbrev_rule_cexpr(E0, PId, E),
    Out = transform(F, E).
deabbrev_rule_cexpr(E0, _, _) :-
    handle_error(['bad context specification expression: ',q(E0)]).


deabbrev_rule_ctermlist([], E, _, E).
deabbrev_rule_ctermlist([T|TS], E, Op, Out) :-
    ( std_compound(T), T = ':::'(Id,Term0) ->
        true
    ;
        Term0 = T
    ),
    deabbreviate(Term0, Term),
    deabbrev_rule_ctermlist(TS, Op(Id,Term,E), Op, Out).



general command
  show_rule(Rule),
  show_rule(Rule, Kind),
  show_rule(Name, Thy, Kind)
doc
  'Display inference rule Rule.',nl,
  'If Rule is not fully specified, displays all rules on backtracking.',nl,
  'If Kind is specified, restrict display to that kind of rule',nl,
  '(postulate, theorem, fundamental_rule, definition).'
extends
  general_cmds
mode
  show_rule(?rulename),
  show_rule(?rulename, ?atom),
  show_rule(?atom, ?theory, ?atom).

show_rule(Rule) :-
    show_rule(Rule, _).

show_rule(Rule, Kind) :-
    decode_rule(Rule, Thy, Name),
    show_rule(Name, Thy, Kind).

show_rule(Name, Thy, Kind) :-
    rule_query(Thy, Name, FullName, Concl, ConclId, Ants, Constraints),
    get_rule_kind(Thy, Name, Kind),
    set_var_name(ConclId, 'C'),
    show_rule_name_sequents(Ants),
    writelist([w(Kind),' ', w(FullName), ' ===', nl]),
    simplify_rule_constraints(Constraints, [], Constraints1),
    show_rule_ants(Ants, ConclId),
    show_rule_sequent(ConclId, _, [], Concl),
    show_rule_constraints(Constraints1),
    writelist(['.',nl]),
    fail.
show_rule(_, _, _).

%% show_rule_1(Thy, Ignored).
%%
%%  This interface is used by show(rule).
%%
show_rule_1(Thy, _) :-
    show_rule(_, Thy, _).

get_rule_kind(Thy, Name, Kind) :-
    query_theory(infrule_status(Name, Deps), Thy),
    !,
    ( Deps = fundamental ->
        Kind = axiom
    ;
        Kind = theorem
    ).
get_rule_kind(_, _, postulate).


show_rule_name_sequents([]).
show_rule_name_sequents([seq(Id,Contexts,_)|Seqs]) :-
    set_var_name(Id, 'A'),
    transform_subterms(show_rule_name_ctermid, Contexts, _),
    show_rule_name_sequents(Seqs).

show_rule_name_ctermid(ST, _) :-
    compound(ST),
    ST = Op(Id,_,_),
    nonvar(Op),
    member(Op, [add,delete]),
    var(Id),
    \+get_var_name(Id,_),
    !,
    set_var_name(Id, 'N').
show_rule_name_ctermid(_, _).

show_rule_ants([], _).
show_rule_ants([seq(Id,Contexts,Concl)], PId) :-
    !,
    show_rule_sequent(Id,PId,Contexts,Concl),
    writelist([nl,'    ------------------------------',nl]).
show_rule_ants([seq(Id,Contexts,Concl)|Seqs], PId) :-
    show_rule_sequent(Id,PId,Contexts,Concl),
    writelist([',', nl]),
    show_rule_ants(Seqs, PId).

simplify_rule_constraints(true, L0, L) :-
    !,
    L = L0.
simplify_rule_constraints((C1,C2), L0, L) :-
    !,
    simplify_rule_constraints(C2, L0, L1),
    simplify_rule_constraints(C1, L1, L).
simplify_rule_constraints(C, L0, L) :-
    L = [C|L0].

show_rule_constraints([]) :-
    !.
show_rule_constraints([C|Cs]) :-
    writelist([nl, '  provided ', code(C)]),
    show_rule_constraints1(Cs).

show_rule_constraints1([]) :-
    !.
show_rule_constraints1([C|Cs]) :-
    writelist([', ', code(C)]),
    show_rule_constraints1(Cs).

show_rule_sequent(Id, PId, Contexts, Concl) :-
    writelist(['    ', w(Id), '::: ']),
    show_rule_sequent_contexts(Contexts, PId),
    writelist([thy(Concl)]).

show_rule_sequent_contexts([], _) :-
    !.
show_rule_sequent_contexts(Cs, PId) :-
    write_rule_spec(Cs, PId),
    write(' ---> ').

%% write_rule_spec(CDiff, Parent).
%%
%%  Abbreviate and write a context-difference (e.g. as returned by
%%  context_diff).
%%
%%  mode(@cdiff, @nodeid).
%% 
write_rule_spec(Spec, PId) :-
    !,
    map(abbrev_rule_cspec(PId), [Spec,CSpec]),
    cspec_to_writelist(CSpec, WL),
    writelist(WL).

%% cspec_to_writelist(CSpec, WL).
%% cspec_to_writelist(CSpec) --> WL.
%%
%%  Convert a context-difference specification (e.g. as returned by
%%  abbrev_rule_cspec) to a writelist.
%%
%%  mode cspec_to_writelist(@cspec, ?writelist).
%%  mode cspec_to_writelist(@cspec, ?writelist, ?writelist).
%%
cspec_to_writelist(CS, WL) :-
    cspec_to_writelist(CS, WL, []).

cspec_to_writelist([]) -->
    !,
    [].
cspec_to_writelist([CS]) -->
    !,
    cspec_to_writelist(CS).
cspec_to_writelist([CS1|CS]) -->
    !,
    cspec_to_writelist(CS1),
    [' &&& '],
    cspec_to_writelist(CS).
cspec_to_writelist(===(CN,CE)) -->
    !,
    [q(CN), ' === '],
    cspec_to_writelist(CE).
cspec_to_writelist(+++(E1,E2)) -->
    !,
    cspec_to_writelist(E1),
    ['+++['],
    cspec_list_to_writelist(E2),
    [']'].
cspec_to_writelist(---(E1,E2)) -->
    !,
    cspec_to_writelist(E1),
    ['---['],
    cspec_list_to_writelist(E2),
    [']'].
cspec_to_writelist(???(E,F)) -->
    !,
    cspec_to_writelist(E),
    [' ??? ', code(F)].
cspec_to_writelist(>>>(E,F)) -->
    !,
    cspec_to_writelist(E),
    [' >>> ', code(F)].
cspec_to_writelist(:::(Id,Term)) -->
    !,
    [proof_node(Id), '::: ', thy(Term)].
cspec_to_writelist(E) -->
    [w(E)].

cspec_list_to_writelist([Exp]) -->
    !,
    cspec_to_writelist(Exp).
cspec_list_to_writelist([Exp|Exps]) -->
    cspec_to_writelist(Exp),
    [', '],
    cspec_list_to_writelist(Exps).

%% abbrev_rule_cspec(PId, CDiff, CSpec).
%%
%%  Convert a context-difference to a context-difference specification.
%%  (Implementation still not complete: does not yet handle
%%  transformation specifications).
%% 
%%  mode abbrev_rule_cspec(@nodeid, +cdiff, ?cspec).
%% 
abbrev_rule_cspec(PId, =(CName,Expr), CSpec) :-
    abbrev_rule_expr(PId, Expr, CExp),
    ( CExp = _(CName,_) ->
        CSpec = CExp
    ;
        CSpec = ===(CName, CExp)
    ).
abbrev_rule_expr(PId, use(CName,Id), CExp) :-
    !,
    ( Id == PId ->
        CExp = CName
    ;
        CExp = CName(Id)
    ).
abbrev_rule_expr(_, empty, CExp) :-
    !,
    CExp = [].
abbrev_rule_expr(PId, Exp, CExp) :-
    Exp = Op(_,_,_),
    ( Op = add ->
        COp = (+++)
    ; Op = delete ->
        COp = ---
    ;
        fail
    ),
    !,
    abbrev_rule_expr(Op, PId, Exp, CExp1, CList),
    CExp = COp(CExp1, CList).
abbrev_rule_expr(PId, filter(Fun, Expr), CExp) :-
    !,
    abbrev_rule_expr(PId, Expr, CExp1),
    CExp = ???(CExp1, Fun).
abbrev_rule_expr(PId, transform(Fun, Expr), CExp) :-
    !,
    abbrev_rule_expr(PId, Expr, CExp1),
    CExp = >>>(CExp1, Fun).
abbrev_rule_expr(_, Exp, _) :-
    handle_error(['abbrev_rule_expr: cannot (yet) handle ', w(Exp)]).

abbrev_rule_expr(Op, PId, Exp, CExp, CList) :-
    Exp = Op(CId,CTerm,Exp1),
    !,
    CList = [:::(CId,CTerm)|CList1],
    abbrev_rule_expr(Op, PId, Exp1, CExp, CList1).
abbrev_rule_expr(_, PId, Exp, CExp, []) :-
    abbrev_rule_expr(PId, Exp, CExp).
