%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##


%% Indexes for rules.
%% CVS $Id: rule_index.ql,v 1.17 2001/06/04 01:24:11 itp Exp $

#include "../prover/lib/ergo.h"

database command
  make_rule_index(Thys, Options)
doc
  'Make a rule index for each theory in Thys.',nl,
  'Thys may be a single theory, a list of theories,',nl,
  '  or the atom ''all'' (meaning all ancestors).',nl,
  'Options is a list of options allowing immediate processing of',nl,
  '  the generated index.',nl,
  'Options can include:',nl,
  '  `compile'',',nl,
  '  `load'' and',nl,
  '  `add''.'
mode
  make_rule_index(@term, @list(atom))
extends
  misc_cmds.

make_rule_index(all, Options) :-
    !,
    \+ ( ancestor(Thy),
         \+ ( build_rule_index(Thy, Name),
              rule_index_options(Thy, Name, Options)
           )).
make_rule_index(Thys, Options) :-
    \+ ( elem(Thy, Thys),
         \+ ( build_rule_index(Thy, Name),
              rule_index_options(Thy, Name, Options)
           )).

build_rule_index(Thy, Name) :-
    check_ancestor(Thy),
    concat_atom([Thy, '_rule_index'], Name),
    concat_atom([Thy, '/', Name, '.ql'], File),
    msg(rule_index, ['Building rule index for ', w(Thy)]),
    ( file_type(Thy, dir) ->
        true
    ;
        make_directory(Thy)
    ),
    concat_atom(['(echo "%% Automatically generated rule index for ',
                 Thy, '"; ',
                 ' echo "%% Generated: `date`")',
                 ' > ', File],
                Cmd),
    os(system(Cmd)),
    open(File, append, Strm),
    enable_signals,
    with_local_exception_handler(
        ( findall(-(Rule, In, Out, InVar, Dir, Cond),
                  ( query_theory(rule(Rule, _Kind, _Shape, Stmt, _Body), Thy),
                    rule_index_entry(Stmt, 1, In, 1, Out, InVar, Dir, Cond)
                  ),
                  Rules),
          write_rule_index(Rules, Strm, nonvar, 0, 0, NLinks),
          write_rule_index(Rules, Strm, var, 0, 0, VLinks),
          write_rule_index_links(NLinks, rule_index_nonvar, Strm),
          write_rule_index_links(VLinks, rule_index_var, Strm),
          ( NLinks = [] ->
              true
          ;
              writelist(Strm, ['rule_index(I, O, nonvar, R, D, C) :- ', nl,
                               tab, 'rule_index_nonvar(I, O, R, D, C).', nl])
          ),
          ( VLinks = [] ->
              true
          ;        
              writelist(Strm, ['rule_index(I, O, var, R, D, C) :- ', nl,
                               tab, 'rule_index_var(I, O, R, D, C).', nl])
         )),
%        signal(_,_,_,'SIGINT'),
        sigint_exception(_,_,_),
        ( close(Strm),
          handle_error([w(build_rule_index(Thy)), ': aborted.'])
       )),
    close(Strm).

%% TO DO: split long indexes in write_rule_index more sensibly!
%% At the moment, we just increment a numeric counter Idx each 100 entries.
%% It'd be much better to make use of some structure,
%% to get additional indexing!
%% Ray's last comment: obsolete.

%% write_rule_index(Rules, Strm, Kind, Idx, Count, Links).
%%
%%  Write a rule index for the nonvar/var entries in Rules.
%%  Count is the maximum number we can write to that index.
%%  Links are the links necessary to link tables together.
%% 
write_rule_index([], _, _, _, _, []).
write_rule_index([-(R,I,O,Kind,D,C)|Rules], Strm, Kind, T, N, L) :-
    !,
    ( N > 0 ->
        T = T1, L = L1, N1 is N - 1
    ;
        T1 is T + 1, L = [Tab|L1], N1 = 100
    ),
    concat_atom(['rule_index_', Kind, '_', T1], Tab),
    writelist(Strm, [w(Tab), '(',
                               canon(I), ',', canon(O), ',', canon(R), ',',
                               canon(D), ',', canon(C),
                             ').', nl]),
    write_rule_index(Rules, Strm, Kind, T1, N1, L1).
write_rule_index([_|Rules], Strm, Kind, T, N, L) :-
    write_rule_index(Rules, Strm, Kind, T, N, L).

write_rule_index_links([], _, _).
write_rule_index_links([Link|Links], Pred, Strm) :-
    writelist(Strm, [w(Pred), '(I, O, R, D, C) :- ', nl,
               tab, w(Link), '(I, O, R, D, C).', nl]),
    write_rule_index_links(Links, Pred, Strm).

%% rule_index_options(Thy, File, Options).
%%
%%  Process the rule index Options for File in Thy.
%% 
rule_index_options(_, _, []).
rule_index_options(Thy, File, [compile|Opts]) :-
    operate_tactic_file(Thy, compile, File),
    rule_index_options(Thy, File, Opts).
rule_index_options(Thy, File, [load|Opts]) :-
    operate_tactic_file(Thy, load, File),
    rule_index_options(Thy, File, Opts).
rule_index_options(Thy, File, [add|Opts]) :-
    add_tactic_file(Thy, File),
    rule_index_options(Thy, File, Opts).

%% rule_index_entry(RuleStmt, InLevels, Input, OutLevels, Output,
%%                  InputType, Direction, Conditionality).
%%
%%  Enumerate the directions/conditionalities for the rule.
%% 
%%  mode rule_index_entry(@term, +integer, -term, +integer, -term,
%%                        -atom, -atom, -atom).
rule_index_entry(Stmt, InN, In, _OutN, 'ergo.true', Var, t, u) :-
    copy_term_levels(Stmt, InN, In, Var).
rule_index_entry(Stmt, _, 'ergo.true', OutN, Out, nonvar, rt, u) :-
    copy_term_levels(Stmt, OutN, Out, _).
rule_index_entry(_Reln(In0, Out0), InN, In, OutN, Out, Var, f, u) :-
    copy_term_levels(In0, InN, In, Var),
    copy_term_levels(Out0, OutN, Out, _).
rule_index_entry(_Reln(Out0, In0), InN, In, OutN, Out, Var, r, u) :-
    copy_term_levels(In0, InN, In, Var),
    copy_term_levels(Out0, OutN, Out, _).
rule_index_entry('ergo.=>'(_Cond, In0), InN, In, _, 'ergo.true', Var, t, c) :-
    copy_term_levels(In0, InN, In, Var).
rule_index_entry('ergo.=>'(_Cond, Out0), _, 'ergo.true', OutN, Out,
                 nonvar, rt, c) :-
    copy_term_levels(Out0, OutN, Out, _).
rule_index_entry('ergo.=>'(_Cond, _Reln(In0, Out0)), InN, In, OutN, Out,
                 Var, f, c) :-
    copy_term_levels(In0, InN, In, Var),
    copy_term_levels(Out0, OutN, Out, _).
rule_index_entry('ergo.=>'(_Cond, _Reln(Out0, In0)), InN, In, OutN, Out,
                 Var, r, c) :-
    copy_term_levels(In0, InN, In, Var),
    copy_term_levels(Out0, OutN, Out, _).

%% copy_term_levels(Term0, Levels, Term, Var).
%%
%%  Term is a copy of Term0, to a depth of Levels levels.
%%  Var is `var' if Term0 is any_variable, otherwise nonvar.
%%
%%  mode copy_term_levels(@term, +integer, -term, -atom).
%% 
copy_term_levels(Term0, _Levels, _Term, Var) :-
    any_variable(Term0),
    !,
    %% Note that we create a fresh variable (_Term) for each such Term0.
    Var = var.
copy_term_levels(Term0, Levels, Term, nonvar) :-
    copy_term_levels(Term0, Levels, Term).

copy_term_levels(_, 0, _) :-
    !.
copy_term_levels(Term0, _Levels, _Term) :-
    any_variable(Term0),
    !.
copy_term_levels(Term0, Levels, Term) :-
    PRE('copy_term_levels: atomic', Levels > 0),
    %% Term0 not var,
    atomic(Term0),
    !,
    Term = Term0.
copy_term_levels(Term0, Levels, Term) :-
    PRE('copy_term_levels: compound', Levels > 0),
    %% Term0 not var or atomic,
    compound(Term0),
    !,
    functor(Term0, Fun, Ar),
    functor(Term, Fun, Ar),
    Levels1 is Levels - 1,
    copy_term_levels_cpd(Ar, Term0, Levels1, Term).
copy_term_levels(Term0, Levels, Term) :-
    PRE('copy_term_levels: quant', Levels > 0),
    %% Term0 not var, atomic or compound,
    quant(Term0),
    !,
    quantify(Term0, Quant, Vars0, Body0),
    quantify(Term, Quant, Vars, Body),
    copy_term_levels_bv(Vars0, Vars),
    Levels1 is Levels - 1,
    copy_term_levels(Body0, Levels1, Body).
%% TO DO: Should
%% copy_term_levels(_, _, _) :- handle_error( ... ) be added?
%% MB 7/9/1998.

copy_term_levels_cpd(0, _Term0, _Levels, _Term) :-
    !.
copy_term_levels_cpd(N, Term0, Levels, Term) :-
    PRE('copy_term_levels_cpd', N > 0),
    arg(N, Term0, Arg0),
    arg(N, Term, Arg),
    copy_term_levels(Arg0, Levels, Arg),
    N1 is N - 1,
    copy_term_levels_cpd(N1, Term0, Levels, Term).

%% TO DO: Change the following behaviour to return a metavariable
%% once the Qu-prolog compiler can cope with that.
%% Ergo puts more restrictions on bound variable lists than does Qu-Prolog.
%% These restrictions are exploited here:
%%   -- bound variable list is a closed list.
%%   -- each entry in the list is either !x or !x:T.
%% Ray's last comment: obsolete.
copy_term_levels_bv([], []).
copy_term_levels_bv([!_x0|Vs0], [!_x|Vs]) :-
    copy_term_levels_bv(Vs0, Vs).
copy_term_levels_bv([_:_|Vs0], [!_x:_|Vs]) :-
    copy_term_levels_bv(Vs0, Vs).

%% Look up a rule index
rule_index_lookup(Index, Input, Output, Var, Rule, Dir, Cond) :-
    call_predicate(Index(Input, Output, Var, Rule, Dir, Cond)).

rule_index_exists(Thy, Index) :-
    atom_concat2(Thy, ':rule_index', Index),
    current_predicate(Index/6).
