%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: show.ql,v 1.42 2002/02/20 02:08:29 itp Exp $

#include "../prover/lib/ergo.h"

general command
  show
extends
  general_cmds
mode
  show
doc
  'Lists available topics for show/1 and show/2.'.

general command
  show(Topic)
extends
  general_cmds
mode
  show(?topics)
doc
  'Lists all facts in the Ergo database about topic(s) Topic.',nl,
  'Theory specific topics use the current theory.'.

general command
  show(Topic,Thy)
extends
  general_cmds
mode
  show(?topics, @theory),
  show(?topics, @list(theory))
doc
  'Lists all facts in theory Thy about topic Topic.',nl,
  'Thy may be `global'', `parameter'' or the name of a loaded theory.'.


%% special_topic(?Topic, -Handler).
%% theory_topic(?Topic, -Arity_or_Global).
%% global_topic(?Topic, -Arity).
%% parameter_topic(?Topic).
%%
%%   These predicates can be used to check whether a given
%%   topic is valid, or to backtrack through all possible topics
%%   (though some topics are hidden when Topic is not instantiated).
%%
%%   Arity is an integer, giving the arity of Topic in global.erg.
%%
%%   Arity_or_Global is either an integer, which is the arity of
%%   Topic in each theory file, or a term of the form global/N,
%%   which means that Topic is actually stored in global.erg and
%%   has arity N.
%%
%%  The Handler for a special topic is a predicate that takes a single
%%  argument: the name of the topic.
%% 
% Ergo 5 special topics:
special_topic(rule,      show_rule_1) :- current_predicate(show_rule_1/2).

theory_topic(ancestor, global/2).
theory_topic(parent,   global/2).
theory_topic(T, N) :-
    atom(T),
    !,
    theory_fact(T, N).
theory_topic(T, N) :-
    % \+ atom(T)
    theory_fact(T, N),
    \+ hidden_theory_topic(T,N).

%% Topics that should be hidden from normal users because
%% they are more usefully displayed in some other way.
hidden_theory_topic(doc_node,_).
hidden_theory_topic(doc_fragment,_).
hidden_theory_topic(doc_extension,_).
hidden_theory_topic(compiled_rule,_).
hidden_theory_topic(save_clauses,_).
hidden_theory_topic(infrule_apply,_).

global_topic(theory, 2).
global_topic(nfi, 0) :- in_proof.
global_topic(allnfi, 0) :- in_proof.
global_topic(proof_names, 1).
global_topic(prover, 2) :- in_proof.
global_topic(T, N) :-
    atom(T),
    !,
    %% allow user to see hidden ones only when specified explicitly.
    global_fact(T, N),
    \+ theory_topic(T, global/_).
global_topic(T, N) :-
    PRE('global_topic', \+ atom(T)),
    global_fact(T, N),
    \+ hidden_global_topic(T, N),
    \+ theory_topic(T, global/_).

%% Topics that should be hidden from normal users because
%% they are more usefully displayed in some other way.
hidden_global_topic(database,_) :- !.
hidden_global_topic(doc_node,_) :- !.
hidden_global_topic(doc_fragment,_) :- !.
hidden_global_topic(doc_extension,_) :- !.
hidden_global_topic(emacs_auto_rules,_) :- 
    ergo_version(ergo),
    !.
hidden_global_topic(emacs_auto_hints,_) :- 
    ergo_version(ergo),
    !.
hidden_global_topic(Anc,_) :- 
    atom(Anc),
    atom_concat(ancestor_of_, _, Anc).


parameter_topic(Name) :-
    query_global(parameter(Name,_,_,_)).
parameter_topic(Name) :-
    query_ancestors(parameter(Name,_,_,_)).


%% private show_topic(+Head, +Body, +Indent).
%%
%%   This predicate determines how each clause of information is displayed.
%%   The last case can handle everything, but the earlier cases are there
%%   to pretty print tactics, heuristics etc.
%%

%% TO DO: Make this asserted (global?) information in show_topic so that
%% it can be extended easily.

show_topic(rule(Name,axiom,_Shape,Stmt,Body), true, Indent) :-
    !,
    Ass = [],
    writelist(Indent),
    writelist(['axiom ', q(Name), ' === ']),
    show_topic_rule(Stmt,Ass,Indent),
    show_topic_body(Body, Indent).
show_topic(rule(Name,axiom(_RealName),_Shape,Stmt,Body), true, Indent) :-
    !,
    Ass = [],
    writelist(Indent),
    writelist(['param_axiom(', q(Name), ', ',
               w(Ass), ', ', thy(Stmt), ', ',
               q(Body), ').', nl]).
show_topic(rule(Name,theorem(_Deps),_Shape,Stmt,Body), true, Indent) :-
    !,
    Ass = [],
    writelist(Indent),
    writelist(['theorem ', q(Name), ' === ']),
    show_topic_rule(Stmt,Ass,Indent),
    show_topic_body(Body, Indent).
show_topic(rule(Name,postulate,_Shape,Stmt,Body), true, Indent) :-
    !,
    Ass = [],
    writelist(Indent),
    writelist(['postulate ', q(Name), ' === ']),
    show_topic_rule(Stmt,Ass,Indent),
    show_topic_body(Body, Indent).
show_topic(rule(Name,defn,_Shape,Stmt,Body), true, Indent) :-
    ergo_version(ergo4),
    !,
    Ass = [],
    writelist(Indent),
    writelist(['definition ', q(Name), ' === ']),
    show_topic_rule(Stmt,Ass,Indent),
    show_topic_body(Body, Indent).
show_topic(rule(_Name,defn,_Shape,_EQ(Term, Defn),_Body), _, _) :-
    %% Ergo 5 defn
    !,
    writelist(['definition ', thy(Term), ' === ', thy(Defn), nl]).
show_topic(infrule(Name,_,_,_,_,_,_), _, _) :-
    %% Ergo 5 inference rule
    !,
    show_rule(Name,_),
    nl.
show_topic(compiled_rule(_,Call,Form,_,Status,PreCond,_,_), Body, Indent) :-
    !,
    writelist(Indent),
    writelist(['compiled_rule ', q(Call), ' === ', q(Form),nl]),
    writelist(Indent),
    writelist([tab(14),'status:        ', q(Status),nl]),
    writelist(Indent),
    writelist([tab(14),'precondition:  ', q(PreCond)]),
    show_topic_body(Body, Indent).
show_topic(request(Patt,Doc), true, Indent) :-
    !,
    writelist(Indent),
    writelist_indent(['request ',q(Patt),nl], Indent),
    writelist_indent(['    '|Doc], ['    '|Indent]),
    nl.
show_topic(announcement(Patt,Doc), true, Indent) :-
    !,
    writelist(Indent),
    writelist_indent(['announcement ',q(Patt),nl], Indent),
    writelist_indent(['    '|Doc], ['    '|Indent]),
    nl.
show_topic(Cmd, true, Indent) :-
    is_cmd_table(Cmd),
    !,
    Cmd =.. [Table, Call, Mode, Impl, DocNode],
    functor(Call, Name, Arity),
    writelist(Indent),
    writelist_indent([q(Table),' command: ',
                      xref([q(Name),'/',q(Arity)],_,DocNode),'.',nl,
                      '  mode: ',q(Mode),nl,
                      '  impl: ',q(Impl),'.'],
                     Indent),
    nl.
show_topic(Head, Body, Indent) :-
    writelist(Indent),
    writelist([q(Head)]),
    show_topic_body(Body, Indent).


is_cmd_table(tactic(_,_,_,_)).
is_cmd_table(proof(_,_,_,_)).
is_cmd_table(dbms(_,_,_,_)).
is_cmd_table(library(_,_,_,_)).
is_cmd_table(general(_,_,_,_)).


%% private show_topic_rule(+Stmt, +Assumptions, +Indent).
%% 
show_topic_rule(Stmt, [], _) :-
    !,
    writelist([thy(Stmt)]).
show_topic_rule(Stmt, Ass, _) :-
    writelist([q(Ass),'  /-  ', thy(Stmt)]).


%% private show_topic_body(+Body, +Indent).
%%
show_topic_body(true, _) :-
    !,
    write('.'),
    nl.
show_topic_body(Body, Indent) :-
    writelist_indent([nl, '    :- ', code(Body), '.'], Indent),
    nl.



%% show.
%%
%%  Lists all the available (global and theory specific) topics.
%%
%%  mode show.
%%
show :- 
    write('Usage: show(Topic)'), nl,
    write('       show(Topic, Theories)'), nl,
    nl,
    write('Available global topics are: '),
    setof(Topic, X ^ (global_topic(Topic, X),
                      \+ hidden_global_topic(Topic, X)), Topics),
    member(Topic1, Topics),
    writelist([' ', q(Topic1)]),
    fail.
show :-
    nl,nl,write('Theory specific topics are: '),
    setof(Topic, X ^ (theory_topic(Topic, X),
                      \+ hidden_theory_topic(Topic,X)), Topics),
    member(Topic1, Topics),
    writelist([' ', q(Topic1)]),
    fail.
show :-
    nl,nl,write('Parameter names are: '),
    setof(Name,
          parameter_topic(Name),
          Names),
    member(N, Names),
    writelist([' ', q(N)]),
    fail.
show :-
    nl.


%% show(Topics).
%%
%%  Lists all information about the given Topics.
%%
%%  mode show(@atom).
%%  mode show(@list(atom)).
%%
show(Topics) :-
    elem(T, Topics),
    (
        special_topic(T, Handler),
        global_state_lookup(current_dir,Thy), Thy \= none,
        call_predicate(Handler(Thy,T))
    ;
        global_topic(T,N),
        do_global_topic(T,N)
    ;
        theory_topic(T,N),
        global_state_lookup(current_dir,Thy), Thy \= none,
        do_theory_topic(T,N,Thy)
    ;
        parameter_topic(T),
        do_parameter_topic(T)
    ;
        atom(T),
        theory_topic(T,_),
        \+ global_topic(T,_),
        \+ parameter_topic(T),
        global_state_lookup(current_dir, none),
        handle_error(['no current theory'])
    ;
        \+ special_topic(T,_),
        \+ global_topic(T,_),
        \+ theory_topic(T,_),
        \+ parameter_topic(T),
        writelist(['Unknown topic: ',q(T),nl])
    ),
    fail.
show(_).



%% show(Topics, Theories).
%%
%%  Lists all information about the given Topics in each of
%%  the given Theories.
%%
%%  mode show(@atom,       @atom).
%%  mode show(@list(atom), @atom).
%%  mode show(@atom,       @list(atom)).
%%  mode show(@list(atom), @list(atom)).
%%
show(Topics, all) :-
    !,
    check_current_theory(CD),
    (
        ancestor_of(CD, Thy),
        show(Topics, [Thy]),
        fail
    ;
        true
    ),
    !.
show(Topics, global) :-
    !,
    elem(T, Topics),
    global_topic(T,N),
    do_global_topic(T,N),
    fail.
show(Topics, parameter) :-
    !,
    elem(T, Topics),
    parameter_topic(T),
    do_parameter_topic(T),
    fail.
show(Topics, Theories) :-
    elem(Thy, Theories),
    check_or_error(
        is_theory_name(Thy),
        ['unknown theory: ',q(Thy)]),
    check_or_error(
        (query_theory(status(_), Thy)),
        ['theory ',q(Thy),' is not loaded yet.']),
    writelist([nl,'========== Theory: ',q(Thy),' ==========',nl]),
    elem(T, Topics),
    ( special_topic(T, Handler) ->
        call_predicate(Handler(Thy,T))
    ; theory_topic(T,N) ->
        do_theory_topic(T, N, Thy)
    ;
        handle_error(['unknown theory topic: ',q(T)])
    ),
    fail.
show(_, _).


%% private do_global_topic(+Topic, +Arity).
%%
%%   This writes out global information about Topic.
%%
do_global_topic(prover, _) :-
    !,
    get_prover(Name,Arity),
    show_topic(prover(Name,Arity), true, ['    ']).
do_global_topic(theory, _) :-
    !,
    query_global(ancestor(Thy, Ancestors)),
    show_topic(theory(Thy, Ancestors), true, ['    ']).
do_global_topic(nfi, _) :-
    !,
    in_proof,
    show_fixedvar_constraints.
do_global_topic(allnfi, _) :-
    !,
    in_proof,
    show_proofvar_constraints.
do_global_topic(proof_names, _) :-
    !,
    get_current_proofs(Names),
    list(Names),
    member(N, Names),
    show_topic(proof(N), true, ['    ']).
do_global_topic(Topic, Arity) :-
    functor(Fact, Topic, Arity),
    query_global(Fact),
    show_topic(Fact, true, ['    ']).


%% private do_parameter_topic(Name).
%%
%%  This writes out the current value of the parameter called Name.
%%
%%  mode do_parameter_topic(@atom).
%%
do_parameter_topic(T) :-
    get_parameter(T, Val),
    writelist(['    ',w(T),'  ===  ',q(Val),nl]).


%% private do_theory_topic(Topic, Arity, Theory).
%%
%%  This writes out theory-specific information about Topic
%%  (or about all theory-specific topics, if Topic is uninstantiated).
%%  Theory must be instantiated to the name of a theory.
%%
%%  mode do_theory_topic(@atom, @ground, +atom).
%%
do_theory_topic(Topic, global/Arity, Theory) :-
    functor(Fact, Topic, Arity),
    arg(1, Fact, Theory),
    query_global(Fact),
    show_topic(Fact, true, ['    ']).
do_theory_topic(Topic, Arity, Theory) :-
    integer(Arity),
    functor(Fact, Topic, Arity),
    query_theory(Fact, Theory),
    show_topic(Fact, true, ['    ']).
