%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: showtree.ql,v 1.23 2003/03/19 06:59:33 itp Exp $
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file defines commands for viewing Ergo 5 proof trees.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include "../prover/lib/ergo.h"


proof command
  show_node(Id)
extends
  prooftree_info_cmds
doc
  'Prints details about the given proof node.'
implementation
  call_passive_cmd(show_node(Id))
mode
  show_node(@nodeid).

show_node(Id) :-
    proof_node_term(Id, Term),
    get_write_depth(D),
    writelist([tab(4), proof_node(Id),'::: ',
               max_depth(D, [thy(Term)]),nl]),
    ( proof_node_rule_and_children(Id, Thy, Rule, _, Children) ->
        ( Children = [] ->
            writelist([nl, 'proved by ',nl,tab(4),q(Thy),':',thy(Rule),nl])
          ;
            writelist([nl, 'proved by ',nl,tab(4),q(Thy),':',thy(Rule),
                       nl,'from ',proof_nodes(Children),nl])
        )
    ;
        true
    ).


proof command
  show_open_nodes
mode
  show_open_nodes
implementation
  call_passive_cmd(show_open_nodes)
command
  show_open_nodes(Id)
mode
  show_open_nodes(@nodeid)
implementation
  call_passive_cmd(show_open_nodes(Id))
extends
  prooftree_info_cmds
doc
  'Display the names and conclusions of all open nodes.',nl,
  'If the Id argument is supplied, it must be the name of a node in the',nl,
  'current proof, and just the open nodes within that subtree of the',nl,
  'proof will be displayed.'.

show_open_nodes :-
    ip_lookup(gumtree_root_node, Root),
    show_open_nodes(Root).

show_open_nodes(Id) :-
    proof_open_nodes(Id,Open),
    map(show_node, [Open]).


proof command
  show_context(Id)
extends
  prooftree_info_cmds
doc
  'Prints details of context of the given proof node.'
implementation
  call_passive_cmd(show_context(Id))
mode
  show_context(@nodeid).

show_context(c(CName,CId,Id)) :-
    !,
    context_retrieve(CName, Id, CId, CTerm),
    get_write_depth(D),
    writelist([tab(4), w(CName), ' ', w(CId), '::: ', ' ', 
               max_depth(D, [thy(CTerm)]), nl]).
show_context(Id) :-
    atomic(Id),
    is_context_name(CName),
    context_search(CName, Id, CId, _),
    show_context(c(CName,CId,Id)),
    fail.
show_context(_).


proof command
  show_node_and_context(Id)
mode
  show_node_and_context(@nodeid)
implementation
  call_passive_cmd(show_node_and_context(Id))
command
  show_nodes_and_context(Ids)
mode
  show_nodes_and_context(@list(nodeid))
implementation
  call_passive_cmd(show_nodes_and_context(Ids))
extends
  prooftree_info_cmds
doc
  'Print details of given proof node(s) and their context.'.

show_node_and_context(Id) :-
    show_nodes_and_context([Id]).

proof command
  show_open_nodes_and_context
mode
  show_open_nodes_and_context
implementation
  call_passive_cmd(show_open_nodes_and_context)
command
  show_open_nodes_and_context(Id)
mode
  show_open_nodes_and_context(@nodeid)
implementation
  call_passive_cmd(show_open_nodes_and_context(Id))
extends
  prooftree_info_cmds
doc
  'Prints details of all open proof nodes and their contexts.',nl,
  'If the Id argument is supplied, it must be the name of a node in the',nl,
  'current proof, and just the open nodes within that subtree of the',nl,
  'proof will be displayed.'.

show_open_nodes_and_context :-
    ip_lookup(gumtree_root_node, Root),
    show_open_nodes_and_context(Root).

show_open_nodes_and_context(Id) :-
    proof_open_nodes(Id, Open),
    show_nodes_and_context(Open).

show_nodes_and_context(Nodes) :-
    ( setof(c(CName,CId,Node), Node^CTerm^
            ( is_context_name(CName),
              member(Node, Nodes),
              context_search(CName, Node, CId, CTerm)
            ),
            Contexts) ->
        show_contexts(Contexts),
        ( Nodes = [SingleID] ->
            writelist([nl]),
            show_node(SingleID)
          ;  
            map(show_node_with_context, [Nodes])
        )
    ;
        map(show_node, [Nodes])
    ).

show_contexts([]).
show_contexts([c(CName,CId,_),C|Cs]) :-
    C = c(CName,CId,_),
    !,
    show_contexts([C|Cs]).
show_contexts([C|Cs]) :-
    show_context(C),
    show_contexts(Cs).

/*
show_single_node(ID) :-
    writelist([nl,tab(4), w(ID), '::: ']),
    proof_node_term(ID, Term),
    get_write_depth(D),
    writelist([max_depth(D, [thy(Term)]), nl]).
*/

show_node_with_context(Id) :-
    proof_node_term(Id, Term),
    writelist([tab(4), w(Id), '::: ']),
    ( setof(CName=CIds,
             ( is_context_name(CName),
               setof(CId, CTerm^context_search(CName, Id, CId, CTerm), CIds)
             ),
             CDescs) ->
        write_context_descs(CDescs)
    ;
        write([])
    ),
    get_write_depth(D),
    writelist([' ---> ', max_depth(D, [thy(Term)]), nl]).
    
write_context_descs([N=I]) :-
    !,
    writelist([w(N), '=', w(I)]).
write_context_descs([N=I | Ds]) :-
    writelist([w(N), '=', w(I), ', ']),
    write_context_descs(Ds).


proof command
  show_annotations(Ids)
extends
  prooftree_info_cmds
doc
  'Prints all the annotations of the given proof node(s).'
implementation
  call_passive_cmd(show_annotations(Ids))
mode
  show_annotations(@nodeid),
  show_annotations(@list(nodeid)).

show_annotations([]) :-
    !.
show_annotations([Id|Ids]) :-
    !,
    show_annotations(Id),
    show_annotations(Ids).
show_annotations(Id) :-
    annotation(Id, Ann),
    ( Ann = usercmd(Num, Node, In, Out, Cmd) ->
        writelist(['    ','usercmd(',q(Num),', ',proof_node(Node),', ',
                       proof_nodes(In),', ',proof_nodes(Out),', ',nl,
                       tab(12),qchars(Cmd),tab(12),')',nl])
    ;
        writelist(['    ',q(Ann),nl])
    ),
    fail.
show_annotations(_).


proof command
  show_constraints
mode
  show_constraints
implementation
  call_passive_cmd(show_constraints)
command
  show_constraints(Vars)
mode
  show_constraints(@list(anyvar))
implementation
  call_passive_cmd(show_constraints(Vars))
extends
  prooftree_info_cmds
doc
  'Prints out the constraints on the current proof.',nl,
  'These include any not_free_in constraints on the proof variables,',nl,
  'plus any user-defined constraint predicates and context assumptions.',nl,
  'The show_constraints(Vars) version only prints constraints that',nl,
  'are related to the variables in Vars, whereas show_constraints/0',nl,
  'shows constraints associated with all the variables in the root',nl,
  'node and all the open nodes of the proof',nl,
  'Note that if you want to see whether or not there are any',nl,
  'constraints between two variables, you typically need to mention',nl,
  emph(both),' those variables in the Vars list.'.

show_constraints :-
    ip_lookup(gumtree_root_node, Root),
    proof_open_nodes(Root, OpenNodes),
    proof_constraints([Root|OpenNodes], Dist, NFI, User, Context),
    show_constraints2(Dist, NFI, User, Context).

show_constraints(Vars) :-
    ip_lookup(gumtree_root_node, Root),
    proof_constraints(Root, Vars, Dist, NFI, User, Context),
    show_constraints2(Dist, NFI, User, Context).

show_constraints2(Dist, NFI, User, Context) :-
    show_constraints3('Distinctness constraints: ', Dist),
    show_constraints3('Freeness constraints:     ', NFI),
    show_constraints3('User-defined constraints: ', User),
    show_constraints3('Context assumptions:      ', Context),
    nl.

show_constraints3(Kind, []) :-
    !,
    writelist([tab(4),w(Kind),'none',nl]).
show_constraints3(Kind, L) :-
    !,
    writelist([tab(4),w(Kind),nl]),
    map(show_constraint4, [L]).

show_constraint4(C) :-
    %% TODO: implement abbreviation of these constraint predicates.
    %%   This probably requires defining an abbreviate_with_type/3
    %%   procedure, with hooks for user-defined types.
    %%   Then we could do this:
    %%
    %%     abbreviate_with_type(rule_constraint,C, C2),
    %%     writelist([tab(8),q(C2),nl]).
    %%
    %%   Maybe, there should be a way of doing this from within
    %%   the writelist.  Eg.  type(Type,Term)?
    %%
    %%   For the moment, we just use code(...), which blindly
    %%   applies abbreviate_and_write/1 to the whole term.
    %%   (The danger is that some code operators will get treated
    %%    as theory terms, but that is a lesser evil than not abbreviating
    %%    any theory terms within C).
    %%
    writelist([tab(8),code(C),nl]).

%%% This declaration should be a `general' declaration, subsuming one currently
%%% in dbms.ql.
 
write_depth parameter
  [set_write_depth, get_write_depth]
doc
  'Global parameter that controls the depth of proof terms',nl,
  'that are displayed.',nl,
  'The default depth is 4.'
extends
  parameters.  

%% set_write_depth(D).
%% get_write_depth(D).
%% local_write_depth(D, Tactic).
%%
%%  Get/set the global parameter that controls the depth of proof terms that
%%  are displayed.
%%  The default depth is 4.
%%
%%  mode get_write_depth(-integer).
%%  mode local_write_depth(+integer, +tactic).
%%
set_write_depth(New) :-
    retractall(write_depth(_)),
    assert(write_depth(New)).
 
get_write_depth(D) :-
    ip_lookup(local_write_depth, D),
    nonvar(D),
    !.
get_write_depth(D) :-
    write_depth(D).

local_write_depth(D, Code) :-
    ip_set(local_write_depth, D),
    call_predicate(Code),
    ip_set(local_write_depth, _).
 
?- set_write_depth(4).
