%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: showwin.ql,v 1.52 2001/06/13 03:21:20 itp Exp $

#include "../prover/lib/ergo.h"

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% This file contains procedures for printing various parts of
%% the current proof window in various ways.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

?- require_global_tactic_file(abbrev).


%% show_window.
%% show_window_all.
%%
%%  Displays the current window of the current proof, including
%%  hypotheses, new variables, focus, relation and goal.
%%
%%  show_window shows just hypotheses that have changed since the last time.
%%  show_window_all shows all the hypotheses.
%%
%%  mode show_window.
%%   
show_window :-
    show_window(changed).

show_window_all :-
    show_window(all).

show_window(Hyps) :-
    in_proof(Prf),
    is_subproof_depth(Dp),
    is_subproof_id(Id),
    %%
    %% This is a bit excessive.  It could probably be moved into
    %% prove_postulate and promote/demote_var.
    ip_lookup(fixedvars, FixedVars),
    name_vars(FixedVars),
    ip_lookup(newvars,  NewVars),
    name_vars(NewVars),
    \+ \+ proof_output(window, [Prf,Dp,Id]),
    debug(show_window, ['starting hyps']),
    ( Hyps == all ->
        \+ \+ show_all_hyps
    ;
        \+ \+ show_changed_hyps
    ),
    debug(show_window, ['starting newvars']),
    \+ \+ proof_output(newvars, NewVars),
    debug(show_window, ['starting focus']),
    \+ \+ show_focus_reln_goal,
    debug(show_window, ['done']).


%% show_assumptions.
%%
%%  Display the list of proof assumptions.
%%
%%  mode show_assumptions.
%%
show_assumptions :-
    (
        is_ass(N,T,_),
        proof_output(assumption(N), T),
        fail
    ;
        true
    ),
    !.

proof command
  show_all_hyps
extends
  proof_info_cmds
doc
  'Display the full hypothesis list.'
mode
  show_all_hyps.

%% show_all_hyps.
%% show_changed_hyps.
%%
%%  Display (part of) the current hypothesis list.
%%
show_all_hyps :-
    proof_output(allhyps, 0).

show_changed_hyps :-
    is_subproof_id(CurSubP),
    ctr_get(displayedsubproof, DispSubP),
    ( CurSubP =:= DispSubP ->
        ctr_get(changedhyps, Change0)
    ;
        ctr_set(displayedsubproof, CurSubP),
        Change0 = 0
    ),
    number_of_hyps(N),
    N1 is N + 1,
    ( N1 < Change0 ->
        Change is N1
    ;
        Change = Change0
    ),
    ctr_set(changedhyps, N1),
    proof_output(changedhyps, Change).

%% show_focus_reln_goal.
%%
%%  Display the Focus, Relation and Goal of the current window.
%%
%%  mode show_focus_reln_goal.
%%
show_focus_reln_goal :-
    is_focus(Focus),
    is_reln(Reln),
    show_pretty_reln(Reln, Reln2),
    is_goal(Goal),
    proof_output(focus, Focus),
    proof_output(reln, Reln2),
    proof_output(goal, Goal).


%% show_pretty_reln(Reln0, Reln).
%%
%%  Reln is a prettier version of Reln0, which should be a legal
%%  preorder relation.
%%
%%  mode show_pretty_reln(@ground, -ground).
%%
show_pretty_reln(rev('ergo.=>'), '<=') :-
    % perhaps this special case should go away
    !.
show_pretty_reln(Reln, Pretty) :-
    global_state_lookup(current_dir,Thy),
    simplify_relation(Reln, Simp),
    abbreviate(Simp, Thy, [2], Pretty).

%% show_hyps_ascending(MinHypNr).
%%
%%  Displays, in ascending order, the hypotheses with Hypnr >= MinHypNr.
%%  If there are hypotheses with Hypnr < MinHypNr, a '...' is displayed
%%  before the new hypotheses are displayed.
%%
%%  mode show_hyps_ascending(@integer).
%%
show_hyps_ascending(Min) :-
    ip_lookup(hyplist, Hyps), %% we assume *descending* numeric order.
    ip_lookup(focusdep, FD),
    show_hyps_ascending(Hyps, FD, Min).

show_hyps_ascending([], _, _).
show_hyps_ascending([HREC|Rest], FD, Min) :-
    hrec_num(HREC, Num),
    ( hrec_deleted(HREC) ->
        show_hyps_ascending(Rest, FD, Min)
    ; Num >= Min ->
        show_hyps_ascending(Rest, FD, Min),
        hrec_term(HREC, Hyp),
        hrec_just(HREC, Just),
        hrec_cdep(HREC, CDep),
        show_hyp(Num, Hyp, Just, CDep, FD) 
    ;
        %% Don't recurse any further, because all hyps in Rest
        %% must have smaller numbers.
        writelist(['    ...', nl])
    ).


%% show_hyp(HypNr, Term, Just, CDep, FocusCDep).
%%
%%  Display a hypothesis, indicating whether it is 
%% 
%%      (*) an undischarged conjecture on which the focus depends
%%      (?) an undischarged conjecture from which the focus is independant
%%      (%) a hypothesis which relies on an undischarged conjecture
%%      (:) any other hypothesis
%%  
%%  mode show_hyp(@integer, @term, @term, @term, @term).
%%
show_hyp(Num, Hyp, cj(Just), Dep, FD) :-
    var(Just),  % is it an undischarged conjecture?
    !,
    (cdep_depends_on(FD, Dep) ->
        proof_output(hyp(used, Num), Hyp)
    ;
        proof_output(hyp(unused, Num), Hyp)
    ).
show_hyp(Num, Hyp, _, Dep, _) :-
    %% nonvar(Dep).
    cdep_depends_on_nothing(Dep),
    !,
    proof_output(hyp(discharged, Num), Hyp).
show_hyp(Num, Hyp, _, _Dep, _) :-
    %% nonvar(_Dep).
    %% and _Dep depends on other conjectures.
    proof_output(hyp(dependent, Num), Hyp).

%% proof_output(Kind, Term).
%%
%%  Display a term during a proof.
%%  Kind says what the term is for; Term is the term.
%% 
%%  mode proof_output(@nonvar, @term).
%%

proof_output(Kind, Term) :-
    handle_event(proof_output(Kind, Term)),
    !.
proof_output(window, _).
proof_output(assumption(N), Ass) :-
    get_write_depth(D),
    writelist([' ',q(N), ': ',max_depth(D,[thy(Ass)]), nl]).
proof_output(newvars, Vars) :-
    ( Vars == [] ->
        true
    ;
        writelist(['    new vars: ',q(Vars), nl])
    ).
proof_output(focus, Focus) :-
    get_write_depth(D),
    writelist([tab, 'focus:  ', max_depth(D,[thy(Focus)]), nl]).
proof_output(reln, Reln) :-
    writelist([tab, 'reln :  ', q(Reln), nl]).
proof_output(goal, Goal) :-
    get_write_depth(D),
    writelist([tab, 'goal :  ', max_depth(D,[thy(Goal)]), nl]).
proof_output(allhyps, _) :-
    show_hyps_ascending(0).
proof_output(changedhyps, Change) :-
    Retained is Change-1,
    ( Retained > 0 ->
        writelist(['hyps 1..', w(Retained), '  <unchanged>', nl])
    ;
        true
    ),
    show_hyps_ascending(Change).
proof_output(hyp(Kind, Num), Hyp) :-
    get_write_depth(D),
    hyp_marker(Kind, Marker),
    writelist(['    ',q(Num),chars(Marker),' ',max_depth(D,[thy(Hyp)]),nl]).
proof_output(proof_step(Num,_,new,_), F) :-
    !,
    get_write_depth(D),
    writelist([q(Num),':',tab,max_depth(D,[thy(F)]),nl]).
proof_output(proof_step(Num,_,Just,Reln), F) :-
    pstep_msg(Just, JMsg),
    get_write_depth(D),
    writelist([max_depth(D,[tab(4),q(Reln),tab,tab|JMsg])]),
    writelist([nl,q(Num),':',tab,max_depth(D,[thy(F)]),nl]).
proof_output(subproof_stack(Level,_ID,F,_Path), Explanation) :-
    get_write_depth(D),
    writelist([q(Level),':',tab,max_depth(D,[thy(F)]),nl,
               tab,tab,tab|Explanation]),
    nl.

%%% This declaration should be a `general' declaration, subsuming one currently
%%% in dbms.ql.

write_depth parameter
  [set_write_depth, get_write_depth]
doc
  'Global parameter that controls the depth of proof terms',nl,
  'that are displayed.',nl,
  'The default depth is 10.',nl,
  'This parameter may be set locally using local_write_depth(D, Tactic).'
extends
  parameters.


%% set_write_depth(D).
%% get_write_depth(D).
%% local_write_depth(D, Tactic).
%% 
%%  Get/set the global parameter that controls the depth of proof terms that
%%  are displayed.
%%  The default depth is 4.
%%
%%  mode get_write_depth(-integer).
%%  mode local_write_depth(+integer, +tactic).
%% 
set_write_depth(New) :-
    retractall(write_depth(_)),
    assert(write_depth(New)).

get_write_depth(D) :-
    ip_lookup(local_write_depth, D),
    nonvar(D),
    !.
get_write_depth(D) :-
    write_depth(D).

local_write_depth(D, Code) :-
    ip_set(local_write_depth, D),
    call_predicate(Code),
    ip_set(local_write_depth, _).

?- set_write_depth(10).


%% hyp_marker(Kind, Marker).
%%
%%  Marker is used to mark hypotheses of Kind.
%%
%%  mode hyp_marker(?atom, ?chars).
%%
hyp_marker(discharged, ":").
hyp_marker(dependent, "%").
hyp_marker(unused, "?").
hyp_marker(used, "*").

%% handle_changed_hyp(Action, SubProof, Number).
%%
%%  Handle a hypothesis change event, by recording the number of the
%%  smallest-numbered hypothesis that has changed since we last displayed.
%%
%%  mode handle_changed_hyp(@nonvar, @integer, @integer).
%%
handle_changed_hyp(add(Kind), Num) :-
    Kind \== inherited,
    ctr_get(changedhyps, Old),
    ( Num < Old ->
        ctr_set(changedhyps, Num)
    ;
        true
    ).
handle_changed_hyp(delete, Num) :-
    ctr_get(changedhyps, Old),
    ( Num < Old ->
        ctr_set(changedhyps, Num)
    ;
        true
    ).
handle_changed_hyp(discharge, Num) :-
    ctr_get(changedhyps, Old),
    ( Num < Old ->
        ctr_set(changedhyps, Num)
    ;
        true
    ).

?- add_listener(hyp(Action, _SubProof, Num, _Term),
                handle_changed_hyp(Action, Num)).

proof command
  show_subproof,
  show_subproof(SubProofId),
  show_subproof(SubProofId, StartId, StartNum, FinalNum)
extends
  proof_info_cmds
doc
  'Displays lines of the subproof SubProofId (default: current subproof).',nl,
  'Displays all lines whose ids are strictly greater than StartId,',nl,
  'numbers those lines sequentially from StartNum+1 upwards,',nl,
  'and returns the final line number in FinalNum.',nl,
  'The zero and one argument versions write all the lines.'
mode
  show_subproof,
  show_subproof(@integer),
  show_subproof(@integer, @integer, @integer, -integer).

show_subproof :-
    is_subproof_id(Id),
    show_subproof(Id, 0, 0, _).

show_subproof(Id) :-
    show_subproof(Id, 0, 0, _).

show_subproof(SubProofId, StartId, StartNum, FinalNum) :-
    ( is_subproof_id(SubProofId) ->
        %% A special case for the current subproof, because
        %% find_subproof takes *ages* in a big proof!
        %% Also, we don't need the other outputs of find_subproof.
        ip_lookup(proof, Steps),
        is_reln(Reln0)
    ;
        find_subproof(SubProofId, _Ancs, Steps, _Hyps, Reln0)
    ),
    show_pretty_reln(Reln0, Reln),
    show_subproof(Steps, StartId, StartNum, Reln, FinalNum).


%% private show_subproof(PSteps, Start, StartCount, Reln, CurrCount).
%%
%%  PSteps is a list of pstep(Id,Focus,Just) terms, ordered
%%  by decreasing Id's.  This outputs all psteps that satisfy
%%  Start < Id.  It also generates a sequentially increasing line
%%  number with each output.  This number equals StartCount at
%%  the line whose Id is Start (even though this line is not output).
%%  After each recursive call, CurrCount is the line number of the
%%  preceeding line.
%%
%%  mode show_subproof(@list(pstep), @integer, @integer, @ground, -integer).
%%
show_subproof([pstep(Id,F,J)|PS], Start, StartCount, Reln, Count) :-
    Start < Id,
    !,
    show_subproof(PS, Start, StartCount, Reln, Count1),
    Count is Count1 + 1,
    proof_output(proof_step(Count,Id,J,Reln), F).
show_subproof(_, _, StartCount, _, StartCount).


pstep_msg(tr(T:N,D,_,A),
          ['by ',q(T),':',q(N),', ',q(D),', ',q(A)]).
pstep_msg(op(P,subprf(Id,_,_,_)),
          ['by subproof ', w(Id), ' on ',q(P)]).
pstep_msg(wr(wf(TacId,Msg,_),_,_,subprf(Id,_,_,_)),
          ['by subproof ', w(Id), ' using tactic ', q(TacId),': '|Msg]) :-
    Msg = [_|_].
pstep_msg(wr(wf(TacId,[],Tac),_,_,subprf(Id,_,_,_)),
          ['by subproof ', w(Id), ' using tactic ', q(TacId),': ',q(Tac)]).
pstep_msg(wr(wa(TacId,Msg,Arg),_,_,subprf(Id,_,_,_)),
      ['by subproof ', w(Id), ' using tactic ', q(TacId),'.',q(Arg),'  '|Msg]).


library command
  in_curr_subproof(Id)
extends
  proof_info_cmds
doc
  'Succeeds if the proof step with identify=Id is in the current subproof.'
mode
  in_curr_subproof(@integer).

in_curr_subproof(Id) :-
    integer(Id),
    ip_lookup(proof, Steps),
    member(pstep(Id,_,_), Steps).


proof command
  show_subproof_stack,
  show_subproof_stack(Depth)
extends
  proof_info_cmds
doc
  'Display information about the current stack of subproofs.',nl,
  'It displays the most recent focus of each subproof that the',nl,
  'user has visited, from level Depth (defaults to 1) up to, ',nl,
  'but not including, the current subproof.'
mode
  show_subproof_stack,
  show_subproof_stack(@integer).

show_subproof_stack :-
    show_subproof_stack(1).

show_subproof_stack(Min) :-
    ip_lookup(focus_stack, Stk),
    show_subproof_stack2(Stk, Min).

show_subproof_stack2([focus(D,ID,F,Path,Explanation)|Stk], Min) :-
    D >= Min,
    !,
    show_subproof_stack2(Stk,Min),
    proof_output(subproof_stack(D,ID,F,Path), Explanation).
show_subproof_stack2(_, _).
