%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: tables.ql,v 1.17 2002/06/30 05:33:59 itp Exp $
%%
%% This file defines two kinds of tables.
%% Tactic tables are indexed, extensible sets of related tactics.
%% Tables associate two terms and are also are indexed and extensible.


%%  Initialise the Ergo database to allow tables.
%%
?- require_global_tactic_file(operator),
   ( query_global(global_operator(declare_table, _, _)) ->
       true
   ;
       add_global_operator(1190, prefix, declare_table)
   ),
   (global_tactic_fact(table_name,3) ->
        true
    ;
        new_global_fact(table_name,3,[save(false)])
   ),
   ( query_global(global_operator(declare_tactic_table, _, _)) ->
       true
   ;
       add_global_operator(1190, prefix, declare_tactic_table)
   ),
   (global_tactic_fact(tactic_table_name,1) ->
        true
    ;
        new_global_fact(tactic_table_name,1,[save(false)])
    ).

tactic_tables header
  'Rule/Tactic Table Commands'
extends
  dbms
doc
  'A `tactic table'' is a set of rules and tactics.',nl,
  'For example, one tactic table might contain various simplification rules,',nl,
  'while another might contain rule and tactics for type checking.',nl,
  'The three main advantages of grouping rules and tactics together',nl,
  'into tactic tables are:',
  begin(list),
  item('Abstraction:'),'  You can apply a tactic table to a proof node without',nl,
  '    caring exactly which rule or tactic within the tactic table will be used.',
  item('Extensibility:'),'  You can add extra rules/tactics to a tactic table at,',nl,
  '    any time, even from a different theory.  This is more flexible than',nl,
  '    writing a tactic which tries a fixed set of rules.',
  item('Efficiency:'),'  Tables are designed to allow efficient indexing',nl,
  '    so that the term of the current proof node, and an optional',nl,
  '    argument to table/2 are used to speed up the search for an',nl,
  '    appropriate rule/tactic within the table.  Currently, the indexing',nl,
  '    is rather naive (just standard Prolog clause indexing), but we',nl,
  '    have plans to use discrimination trees as fast whole-term indexes.',
  end(list),
  'The idea is that you define a `tactic table'', using declare_tactic_table',nl,
  'then use in_tactic_table/{2,3,4} to add tactics to that table.',nl,
  'Each tactic that you add should accept one proof node as input.',nl,
  'To add individual rules to the table, just use `rule(Thy:Name)''',nl,
  'as the value for Tactic.',nl,
  nl,
  'After setting up a table like this, the tactics tactic_table/1 and tactic_table/2',nl,
  'can be called.  They execute one of the tactics in the table.',nl,
  nl.

user_tactic_table header
  'User Defined Tactic Tables'
extends
  tactic_tables
doc
  'The current user defined tactic tables are:',nl.

database command
  declare_tactic_table(Decl)
header
  'declare_tactic_table Decl'
doc
  'This tactic is used to declare tables.',nl,
  'Each table stores a set of tactics, indexed by a theory term.',nl,
  'The typical reason for defining a table is to group together',nl,
  'a set of related tactics or rules, then use the ',
  xref('tactic_table/{1,2}',global,'gumtree:tactic_table/2'),nl,
  'command to call one of the tactics in a given table.', nl,nl,
  'declare_tactic_table Name - declare Name as the name of a tactic table.',nl,
  'declare_tactic_table Name doc Doc - declare Name as the name of a tactic table',nl,
  'and associate Doc as the documentation for Name.'
extends
  tactic_tables
mode
  declare_tactic_table(@term).

declare_tactic_table(doc(Name, Doc)) :-
    !,
    declare_tactic_table_aux(Name),
    add_tactic_table_doc(Name, Doc).
declare_tactic_table(Name) :-
    declare_tactic_table_aux(Name).
    
declare_tactic_table_aux(Name) :-
    query_global(tactic_table_name(Name)),
    !,
    msg(tactic_table, [q(Name), ' already declared.']).
declare_tactic_table_aux(Name) :-
    theory_fact(Name, 3),
    !,
    handle_error([q(Name/3),' is already used for some theory fact.']).
declare_tactic_table_aux(Name) :-
    check_atom(Name),
    atom_concat2(tactic_table_entry_, Name, G),
    tactic(mode(G(_,_), arity(G(thy(term), term), '--->'(1,_)))),
    tactic_assert_global(tactic_table_name(Name)),
    new_theory_fact(Name, 3, [save(false), assert(z)]).

add_tactic_table_doc(Name, Doc) :-
    make_doc_name(tactic_tables, Name, DocName),
    add_tactic_table_doc_aux(Name, DocName, Doc).

add_tactic_table_doc_aux(_Name, DocName, _Doc) :-
    query_global(doc_node(DocName,_,_,_)),
    !.
add_tactic_table_doc_aux(Name, DocName, Doc) :-
    gumtree_comma_to_list(Doc, DocList),
    check_is_legal_writelist(DocList),
    add_doc_extension(global, user_tactic_table,
                    [xref([Name], global, DocName)]),
    add_doc(global, DocName, [Name], DocList, []).
    
database command
  erase_tactic_table(Name)
extends
  tactic_tables
doc
  'Erase all entries in the current theory for the tactic table called Name.'
mode
  erase_tactic_table(@atom).

erase_tactic_table(Name) :-
    query_global(tactic_table_name(Name)),
    !,
    atom_concat2(gumtree_tactic_table_entry_, Name, G),
    retractall(G(_,_,_,_,_,_)),
    check_current_theory(Thy),
    retractall_theory(Name(_,_,_), Thy).

%
% in_tactic_table records 2 versions of the entry - one is so that users can
% call show_tactic_table/1 to list the entries. The other is a gumtree compiled 
% (but asserted) user define gumtree tactic that is used when tactic_table/{1,2}
% is used as a gumtree tactic.
%
database command
  in_tactic_table(Name, Tactic),
  in_tactic_table(Name, Tactic, Term),
  in_tactic_table(Name, Tactic, Term, Arg)
extends
  tactic_tables
doc
  'This adds Tactic into the tactic table called Name.',nl,
  'Note that rules can be added by specifying `rule(Thy:Name)''',nl,
  'as the value for Tactic.',
  para,
  'The tactic table (Name) must have been declared using declare_tactic_table/2.',
  para,
  'Term and Arg can be used to restrict the times when Tactic will',nl,
  'be called.  If tactic_table/1 is called on a proof node containing',nl,
  'a term T, then Tactic will only be called when T unifies with Term.',nl,
  'The tactic_table/2 tactic is similar, but in addition, it requires its',nl,
  'second argument to unify with Arg.'  
mode
  in_tactic_table(@atom, @tactic),
  in_tactic_table(@atom, @tactic, ?thy(term)),
  in_tactic_table(@atom, @tactic, ?thy(term), ?term).

in_tactic_table(Table, Tac) :-
    in_tactic_table(Table, Tac, _, _).

in_tactic_table(Table, Tac, Term) :-
    in_tactic_table(Table, Tac, Term, _).

in_tactic_table(Table, Tac, Term, Arg) :-
    query_global(tactic_table_name(Table)),
    !,
    check_current_theory(Thy),
    gumtree_deabbrev_tactic_args(Tac, DTac),
    deabbreviate(Term, DTerm),
    tactic_assert_theory(Table(DTerm, Arg, DTac), Thy),
    atom_concat2(tactic_table_entry_, Table, G),
    gumtree_comp(tactic('==='(G(DTerm, Arg), DTac)), Code),
    assert_table_entry_code(Code).
in_tactic_table(Table, _, _, _) :-
    handle_error(['unknown table: ',q(Table)]).
  

assert_table_entry_code(Code) :-
    member(Clause, Code),
    assert(Clause),
    fail.
assert_table_entry_code(_).

%
% Declare tactic_table/{1,2} as user defined gumtree tactics.
%
?- user_tactic_and_doc_decl([Name='Name'],
 tactic_table(Name),
 (
  'This proof command executes one of the tactics in the',nl,
  'tactic table called Name.',nl,
  nl,
  'The term of the current proof node are used to restrict the set',nl,
  ' of matching tactics in Table.' 
 ),
 tactic_table(cmdtable),
 '--->'(1,_)
).
?- user_tactic_and_doc_decl([Name='Name', Val='Val'],
 tactic_table(Name, Val),
 (
  'This proof command executes one of the tactics in the',nl,
  'tactic table called Name.',nl,
  nl,
  'The term of the current proof node and Val are',nl,
  'used to restrict the set of matching tactics in Table.' 
 ),
 tactic_table(cmdtable, term),
 '--->'(1,_)
).

%
% User defined gumtree tactics are defined as predicated with
% 'gumtree_tactic_' prepended to the name and 4 extra arguments
% added to represent 2 difference lists - one for the input nodes
% and one for the output nodes.
%
% The entries in the table are converted to user defined gumtree tactics
% with 'table_entry_' prepended to the table name - and hence the 
% compiled clauses have 'gumtree_tactic_table_entry_' prepended to
% the table name.

gumtree_tactic_tactic_table(Name, A, B, C, D) :-
    query_ancestors(Name(_,_,_),_),!, % The table exists and has an entry
    A = [E|B],
    proof_node_term(E, F),  % F is the goal of the (only) proof node
    atom_concat2(gumtree_tactic_tactic_table_entry_, Name, G),
    call_predicate(G(F, _, A, B, C, D)).

gumtree_tactic_tactic_table(Name, Val, A, B, C, D) :-
    query_ancestors(Name(_,_,_),_),!,
    A = [E|B],
    proof_node_term(E, F),
    atom_concat2(gumtree_tactic_tactic_table_entry_, Name, G),
    call_predicate(G(F, Val, A, B, C, D)).


general command
  show_tactic_tables
extends
  tactic_tables
extends
  general_cmds
doc
  'List all the currently loaded tactic tables.'
mode
  show_tactic_tables.

show_tactic_tables :-
    query_global(tactic_table_name(Table)),
    writelist([q(Table),nl]),
    fail.
show_tactic_tables.
    
general command
  show_tactic_table(Table)
extends
  tactic_tables
extends
  general_cmds
doc
  'List all the currently loaded entries in the tactic table.'
mode
  show_tactic_table(?atom).

show_tactic_table(Table) :-
    query_global(tactic_table_name(Table)),
    query_ancestors(Table(T,Val,Tac),_),
    writelist([thy(in_tactic_table(Table, Tac, T, Val)),nl]),
    fail.
show_tactic_table(_).
    
  

tables header
  'Table Commands'
extends
  dbms
doc
  'A `table'' is an association between terms.',nl,
  'For example, one table might be used to associate terms with operator names',nl,
  'to aid in the construction of rules for window inference,',nl,
  'while another table might associate terms with rules, with the intention',nl,
  'of using the rules in a more flexible way than with tactic tables.',nl,
  'The idea is that you define a `table'', using declare_table',nl,
  'then use in_table/3 to add term pairs to that table.',
  para,
  'After setting up a table like this, the call table/3',nl,
  'can be used to extract information from the table.'.

user_table header
  'User Defined Tables'
extends
  tables
doc
  'The current user defined tables are:',nl.

database command
  declare_table(Decl)
header
  'declare_table Decl'
doc
  'This tactic is used to declare tables.',nl,
  'Each table stores a set of term pairs.',nl,
  'The typical reason for defining a table is to group together',nl,
  'a set of related terms, then use the ',
  xref('table/2',global,'dbms:table/2'),nl,
  'command to extract information from the table.', nl,nl,
  'declare_table Name(Type1, Type2) - declare Name as the name of a table,',nl,
  'where Type1 and Type2 are the types of the two terms in the table entries.',nl,
  'declare_table Name(Type1, Type2) doc Doc - declare Name as the name of a table',nl,
  'and associate Doc as the documentation for Name.'
extends
  tables
mode
  declare_table(@term).

declare_table(doc(Name(Type1, Type2), Doc)) :-
    !,
    declare_table_aux(Name, Type1, Type2),
    add_table_doc(Name, Doc).
declare_table(Name(Type1, Type2)) :-
    declare_table_aux(Name, Type1, Type2).
    
declare_table_aux(Name, _Type1, _Type2) :-
    query_global(table_name(Name, _, _)),
    !,
    msg(table, [q(Name), ' already declared.']).
declare_table_aux(Name, _Type1, _Type2) :-
    theory_fact(Name, 3),
    !,
    handle_error([q(Name/3),' is already used for some theory fact.']).
declare_table_aux(Name, Type1, Type2) :-
    check_atom(Name),
    check_or_error((ground(Type1),ground(Type2)),
          ['Types of ', q(Name(Type1, Type2)), 'must be ground']),
    tactic_assert_global(table_name(Name, Type1, Type2)),
    new_theory_fact(Name, 2, [save(false), assert(z)]).

add_table_doc(Name, Doc) :-
    make_doc_name(tables, Name, DocName),
    add_table_doc_aux(Name, DocName, Doc).

add_table_doc_aux(_Name, DocName, _Doc) :-
    query_global(doc_node(DocName,_,_,_)),
    !.
add_table_doc_aux(Name, DocName, Doc) :-
    gumtree_comma_to_list(Doc, DocList),
    check_is_legal_writelist(DocList),
    add_doc_extension(global, user_table,
                    [xref([Name], global, DocName)]),
    add_doc(global, DocName, [Name], DocList, []).
    
database command
  erase_table(Name)
extends
  tables
doc
  'Erase all entries in the current theory for the table called Name.'
mode
  erase_table(@atom).

erase_table(Name) :-
    query_global(table_name(Name, _, _)),
    !,
    check_current_theory(Thy),
    retractall_theory(Name(_,_), Thy).

%
% in_table records table entries.
% call show_table/1 to list the entries. 
%
database command
  in_table(Name, Term1, Term2)
extends
  tables
doc
  'This adds the pair of terms Term1, Term2 into the table called Name.',
  para,
  'The table (Name) must have been declared using declare_table/2.'
mode
  in_table(@atom, @term, @term).

in_table(Table, Term1, Term2) :-
    query_global(table_name(Table, Type1, Type2)),
    !,
    deabbreviate_with_type(Type1, Term1, DTerm1),
    deabbreviate_with_type(Type2, Term2, DTerm2),
    check_current_theory(Thy),
    tactic_assert_theory(Table(DTerm1, DTerm2), Thy).
in_table(Table, _, _) :-
    handle_error(['unknown table: ',q(Table)]).
  

%
% table/3 looks up entries in the table.
%
database command
  table(Name, Term1, Term2)
extends
  tables
doc
  'This is used to look up the pair of terms Term1, Term2 in the table called Name.',
  para,
  'The table (Name) must have been declared using declare_table/2.'
mode
  table(@atom, ?term, ?term).

table(Table, Term1, Term2) :-
    query_ancestors(Table(Term1, Term2),_).


general command
  show_tables
extends
  tables
extends
  general_cmds
doc
  'List all the currently loaded tables.'
mode
  show_tables.

show_tables :-
    query_global(table_name(Table, Type1, Type2)),
    writelist([q(Table(Type1, Type2)),nl]),
    fail.
show_tables.
    
general command
  show_table(Table)
extends
  tables
extends
  general_cmds
doc
  'List all the currently loaded entries in table.'
mode
  show_table(?atom).

show_table(Table) :-
    query_global(table_name(Table, _, _)),
    query_ancestors(Table(T1, T2),_),
    writelist([thy(in_table(Table, T1, T2)),nl]),
    fail.
show_table(_).
    
  
