%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% This file defines theory-level commands, parameters and
%% documentation that are common to all versions of Ergo
%% (Ergo 4 and Ergo 5 at least).
%%
%% CVS $Id: thycmds.ql,v 1.13 2002/06/27 05:10:28 itp Exp $

#include "../prover/lib/ergo.h"

?- require_global_tactic_file(operator).
?- require_global_tactic_file(abbrev).

database command
  add_theory(T)
extends
  theory_construction_cmds
mode
  add_theory(@theory)
implementation
  add_theory(T),
  add_visible_theory(T)
doc
  'Create a new empty theory called T.'.

database command
  add_subtheory(S)
extends
  theory_construction_cmds
mode
  add_subtheory(@theory)
doc
  'Create a new (empty) subtheory of the current theory called S.',nl,
  'Subtheories are typically created as the result of a theory',nl,
  'interpretation.  However, they can also be used to group a set',nl,
  'of theorems under the new name S, which can then be used to',nl,
  'restrict searching later.'.

database command
  delete_theory(T)
extends
  theory_deletion_cmds
mode
  delete_theory(@theory)
doc
  'Delete a theory called T and all its subtheories.',nl,
  'This will fail if it would cause the current theory',nl,
  'to be deleted.'.

database command
  delete_subtheory(S)
extends
  theory_deletion_cmds
mode
  delete_subtheory(@theory)
doc
  'Delete a subtheory of the current theory called S.'.

database command
  delete_theory_and_dependents(T)
extends
  theory_deletion_cmds
mode
  delete_theory_and_dependents(@theory)
doc
  'Delete T and all the dependent theories of T.',nl,
  'This will fail if it would cause the current theory',nl,
  'to be deleted.'.

database command
  cd(Thy)
header
  'cd Thy.',nl
mode
  cd(@theory)
doc
  'Change to theory Thy and load it if necessary.',nl,
  'This command has been superseded by the command ',
  xref('[Thy]',global,'database:./2'),'.'.

database command
  [Thy]
header
  '[Thy].'
extends
  misc_cmds
mode
  '.'(@theory,@emptylist)
implementation
  cd(Thy)
doc
  'Change to theory Thy, and load it if it is not already loaded.',nl,
  'This will also cause all its ancestor theories to be loaded.',
  para,
  'The directories that the theories are loaded from will be determined',nl,
  'by the ',xref(load_path,global,'parameter:load_path'),' parameter.'.

database command
  inherit_theory(Thy)
extends
  theory_construction_cmds
mode
  inherit_theory(@theory)
doc
  'Inherit theory Thy into the current theory, which must be open.',nl,
  nl,
  'This command does not copy abbreviations or operators',nl,
  'of Thy into the current theory or even make Thy visible.',nl,
  'This means that full names (eg. ''Thy.Name'') must be used',nl,
  'for all symbols that come from Thy.',
  para,
  'The add_abbrevs, add_operators and visible_theory commands',nl,
  'can be used to define more readable notation for Thy terms.',nl,
  'Alternatively, use the ',
       xref(include_theory,global,'database:include_theory/2'), 
       ' command, which',nl,
  'provides a higher level interface for inheriting theories,',nl,
  'complete with all their abbreviations and operators etc.'.

database command
  close_theory
extends
  theory_construction_cmds
mode
  close_theory
doc
  'Marks the current theory as being "closed".',nl,
  'Once a theory is closed, its axiomatic basis cannot be extended.',nl,
  'That is, no further constants or fundamental inference rules (axioms)',nl,
  'can be added to it, and no additional theories can be inherited into',nl,
  'it.  However, commands that extend the theory in a conservative way,',nl,
  'such as definitions and proofs of theorems, are still allowed.'.

%%% A bit of a hack: just call the declaration expansion routine.
database command
  doc(Name, Rest)
header
  'Name doc DocStrings extends Nodes.',nl
extends
  theory_extension_cmds
mode
  doc(@atom, @compound)
implementation
  (global_state_lookup(current_dir,Thy) ; Thy = global ),
  expand_decls(Thy, doc(Name, Rest), Qs, []),
  call_queries(Qs)
doc
  'Create a doc concept node called Name.',nl,
  'The DocStrings are as described in ',
     xref([writelist], global, writelist), '.',nl,
  'Nodes is a comma seperated list of doc nodes.',
  para,
  'An optional ''extends'' clause adds cross-references in the',nl,
  'named Nodes, pointing to this one.  It is also possible to include',nl,
  '''header'' clauses to generate a heading for the resulting doc node.',nl,
  'In fact, ''header'' can be used as the first infix keyword, as an',nl,
  'alternative to ''doc''.',
  para,
  'For example:',
  para, begin(verb),
  'modes_help_ergo4',nl,
  'doc',nl,
  '  ''Types specific to Ergo4 are:''',nl,
  'header',nl,
  '  ''Ergo4 types''',nl,
  'extends',nl,
  '  modes_help.',nl,
  end(verb).


database command
  declare_object_vars(A)
extends
  sugar_cmds
mode
  declare_object_vars(@atom)
doc
  'Declare an object variable prefix for the current theory.',
  para,
  'Eg. declare_object_vars(tim) declares tim, tim_0, tim_1...'.

database command
  declare(Symbol)
header
  'declare Symbol.',nl
extends
  theory_extension_cmds
mode
  declare(@term)
implementation
  declare_name(Symbol)
doc
  'Declare Symbol as a function/quantifier.',
  para,
  'Eg 1. declare a. declares the constant a.',
  para,
  'Eg 2. declare f(A,B). declares the function f with arity [2].',
  para,
  'Eg 3. declare (!!q [x:_,x_2] _). declares the quantifier q',
  '       with arity [quant([typed,untyped])].'.

database command
  declare(Name,Arity)
header
  'declare(Name,Arity).',nl
extends
  theory_extension_cmds
mode
  declare(@atom, @arity)
implementation
  declare_name(Name,Arity)
doc
  'Declare Name as a function/quantifier with the given Arity.'.

database command
  declare_param_const(Name,Value,Check)
extends
  theory_extension_cmds
mode
  declare_param_const(@atom, @term, @code)
doc
  'Declare a set of constants called Name, parameterized by Value.',nl,
  'Check is Prolog code that determines allowable parameter values.',
  para,
  'Eg: declare_param_const(intconst,V,integer(V)).'.

database command
  define('==='(F,Term))
header
  'define F === Term.',nl
extends
  theory_extension_cmds
mode
  define(@compound)
implementation
  deabbreviate(Term,Term2),
  define_name(F,Term2)
doc
  'Define a new constant/function/quantifier F to be equal to Term.',
  para,
  'Eg 1. define a === b+1.',
  para,
  'Eg 2. define f(X,Y) === X + g(Y).',
  para,
  'Eg 3. define (!!ex [x] B) === not (all [x] (not B)).',
  para,
  'The rule name given by default is "def_"F, a convention which allows',nl,
  'tactics such as ',
      xref(unfold,jprop,'proof:unfold/1'),' and ',
      xref(fold,jprop,'proof:fold/1'),
      ' to find the relevant definition.'.


database command
  include_theory(Thys),
  include_theory(Thys, Syntax)
extends
  theory_construction_cmds
mode
  include_theory(@theory),
  include_theory(@list(theory)),
  include_theory(@theory, @atom),
  include_theory(@list(theory), @atom)
doc
  'Thys may be a theory name, or a list of theory names.',nl,
  'This command ensures that the current theory inherits (all of)',nl,
  'Thys and that they are all visible in the current theory.',nl,
  'Note that this does not automatically make all ancestors',nl,
  'of Thys visible.',
  para,
  'The Syntax argument determines how much sugared syntax is',nl,
  'copied into the current theory from Thys.',
  para,
  'Syntax=none means that no operators or abbreviations are copied',nl,
  'and Thys does not become visible within the current theory.',nl,
  '(this means that full ''Thy.Name'' names must be used for all',nl,
  'symbols from Thys).',
  para,
  'Syntax=local means that operators and abbreviations that were defined',nl,
  'within Thys are copied into the current theory and Thys becomes',nl,
  'visible within the current theory (which allows ''Thys.'' prefixes',nl,
  'to be dropped, when no ambiguity arises).',
  para,
  'Syntax=all means that all operators and abbreviations from Thys',nl,
  '(both those that were defined within Thys and those that were ',nl,
  'copied into it from other theories) are copied into the current theory',nl,
  'and all theories that were visible in Thys become visible in the',nl,
  'current theory.',
  para,
  'Syntax=share means that the current theory will SHARE all the',nl,
  'syntax (operators, abbreviations and visible theory facts) of Thys.',nl,
  'This sharing is not transitive.',nl,
  'The effect of Syntax=share is similar to Syntax=all, except that',nl,
  'future changes to the syntax of Thys will affect the current theory',nl,
  'and sharing gives less freedom to customize the syntax of the current',nl,
  'theory, because inherited operators etc. cannot be deleted.',
  para,
  'For backwards compatibility, include_theory/1 sets Syntax to ''local''.'.

include_theory(Thys) :-
    include_theory(Thys, local).

include_theory([], _) :-
    !.
include_theory([Parent|L], Syntax) :-
    !,
    include_theory(Parent, Syntax),
    include_theory(L, Syntax).
include_theory(Parent, Syntax) :-
    %% \+ list(Parent).
    ( global_state_lookup(current_dir,Parent) ->
        handle_error(['you cannot include ',q(Parent),' into itself!'])
    ; ancestor(Parent) ->
	true
    ;
        msg(include_theory, ['inheriting theory: ',q(Parent)]),
	inherit_theory(Parent)
    ),
    include_theory2(Syntax, Parent).

include_theory2(none, _) :-
    !.
include_theory2(local, Parent) :-
    !,
    msg(include_theory, ['making theory ',q(Parent),' visible.']),
    add_visible_theory(Parent),
    add_abbrevs(Parent),
    add_operators(Parent).
include_theory2(all, Parent) :-
    !,
    msg(include_theory, ['including syntax of theory ',q(Parent),'.']),
    add_all_visible_theories(Parent),
    add_all_abbrevs(Parent),
    add_all_operators(Parent).
include_theory2(share, Parent) :-
    !,
    msg(include_theory, ['sharing syntax of theory ',q(Parent),'.']),
    inherit_all_visible_theories(Parent),
    inherit_all_abbrevs(Parent),
    inherit_all_operators(Parent).
include_theory2(Syntax, _) :-
    handle_error(['Bad Syntax argument to include_theory: ',q(Syntax),nl,
                  'Legal values are: none, local, all, share.']).

