%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% This file defines theory-level commands, parameters and
%% documentation that are specific to Ergo 4.x, where x > 0.
%%
%% CVS $Id: thycmds4.ql,v 1.9 2001/06/04 01:24:11 itp Exp $

#include "../prover/lib/ergo.h"

?- require_global_tactic_file(abbrev).


init_thycmds4 :-
    %% We must declare local (non-saved) facts every session.
    (global_fact(theorem_action,1) ->
        true
    ;
        new_global_fact(theorem_action,1,[save(false)])
    ),
    (global_fact(save_proof_record,1) ->
        true
    ;
        new_global_fact(save_proof_record,1,[save(false)])
    ),
    %% Now check that saved facts are declared.
    ( theory_fact(default_prover,2) ->
        true
    ;
        new_theory_fact(default_prover,2,[save(true)])
    ).

?- init_thycmds4.


default_prover parameter
  [set_default_prover, get_default_prover]
doc
  'Ergo allows new proof interfaces to be defined by users.',nl,
  'This `default prover'' parameter determines which proof interface',nl,
  'will be invoked each time a theorem is started in the current theory',nl,
  'and all descendent theories that do not override this default.',nl,
  nl,
  'The value of the parameter must be of the form Command/Arity,',nl,
  'where Arity is 0 or 1.  Command/Arity should be an executable',nl,
  'tactic.'
extends
  parameters.

%% set_default_prover(Name/Arity).
%% set_default_prover(Name, Arity).
%% get_default_prover(Name/Arity).
%% get_default_prover(Name, Arity).
%%
%%  Set/Get the default prover for the current theory and all descendent
%%  theories that do not set their own prover.
%%
%%  mode set_default_prover(@atom, @integer).
%%  mode get_default_prover(@atom, @integer).
%%
set_default_prover(Name/Arity) :-
    set_default_prover(Name, Arity).

set_default_prover(Name, Arity) :-
    check_current_theory(Thy),
    check_prover_name(Name, Arity),
    tactic_retractall_theory(default_prover(_,_), Thy),
    tactic_assert_theory(default_prover(Name,Arity), Thy).

get_default_prover(Name/Arity) :-
    get_default_prover(Name, Arity).

get_default_prover(Name, Arity) :-
    check_current_theory(Thy),
    query_theory(default_prover(Name, Arity), Thy).




database command
  preorder(Reln)
extends
  theory_extension_cmds
mode
  preorder(@atom)
implementation
  deabbreviate_relation(Reln,Reln2),
  add_preorder(Reln2)
doc
  'Inform Ergo that Reln is a reflexive-transitive relation.',nl,
  'Reln must be a binary function that is already accessible in the',nl,
  'current theory and there must be axioms/theorems of the form:',nl,
  nl,
  '    Reln(A,A)',nl,
  nl,
  '    Reln(A,B) and Reln(B,C) => Reln(A,C).'.

database command
  equivalence(Reln)
extends
  theory_extension_cmds
mode
  equivalence(@atom)
implementation
  deabbreviate_relation(Reln,Reln2),
  add_equivalence(Reln2)
doc
  'Inform Ergo that Reln is an equivalence relation.',nl,
  'Reln must be a binary function that is already accessible in the',nl,
  'current theory.  The add_preorder(Reln) command will be called',nl,
  'if Reln is not already a preorder.  In addition, there must be',nl,
  'an axiom/theorem of the form:',nl,
  nl,
  '     Reln(A,B) => Reln(B,A).',nl,
  nl,
  'This command also automatically does reverse_of(Reln,Reln)',nl,
  'so that the relation rev(Reln) can be simplified to Reln.',nl,
  'This corresponds to the theorem, Reln(A,B) <=> Reln(B,A),',nl,
  'which is equivalent to the above theorem.'.

database command
  stronger_than(R1,R2)
extends
  theory_extension_cmds
mode
  stronger_than(@atom, @atom)
implementation
  deabbreviate_relation(R1,Rel1),
  deabbreviate_relation(R2,Rel2),
  add_stronger_than(Rel1,Rel2)
doc
  'Inform Ergo that preorder/equivalence relation R1 implies R2.',nl,
  'For this to succeed, the following theorem/axiom must be available:',nl,
  nl,
  '    R1(A,B) => R2(A,B).'.

database command
  reverse_of(R1,R2)
extends
  theory_extension_cmds
mode
  reverse_of(@atom, @atom)
implementation
  deabbreviate_relation(R1,Rel1),
  deabbreviate_relation(R2,Rel2),
  add_reverse_of(Rel1,Rel2)
doc
  'Inform Ergo that R1(A,B) is equivalent to R2(B,A).',nl,
  'R1 and R2 must both be preorders in the current theory and',nl,
  'there must an axiom/theorem of the form:',nl,
  nl,
  '    R1(A,B) => R2(A,B).'.

database command
  openrule(Term,Path,NewHyps,HypFn,RelFn,GoalFn)
extends
  theory_extension_cmds
mode
  openrule(@thy(term), @subterm_path, @list(thy(term)), @term, @term, @term)
implementation
  deabbreviate(Term,Term1),  %% This had better not modify the path...
  deabbreviate_list(NewHyps,NewHyps1),
  deabbreviate_term_map(HypFn,HypFn1),
  deabbreviate_relation_map(RelFn,RelFn1),
  deabbreviate_term_map(GoalFn,GoalFn1),
  openrule(primitive,Term1,Path,NewHyps1,HypFn1,RelFn1,GoalFn1)
doc
  'Add a primitive window inference rule, which can be used',nl,
  'to create a new subwindow whose focus is the subterm of Term',nl,
  'specified by Path.',nl,
  nl,
  'NewHyps is a list of hypotheses to be added in the subwindow.',nl,
  nl,
  'HypFn is a list of ParentHyp--->SubHyp mappings that specifies',nl,
  'how parent window hypotheses are translated into the subwindow.',nl,
  'The special mapping ''id'' copies all hypotheses unchanged.',nl,
  nl,
  'RelFn is a list of mappings ParentReln--->SubReln, where',nl,
  nl,
  'ParentReln is the relation of the parent window and SubReln',nl,
  'is the relation of the subwindow.',nl,
  nl,
  'GoalFn is a list of ParentGoal--->SubGoal mappings, to specify',nl,
  'what the goal of the subwindow should be.  If the goal of the',nl,
  'parent window does not match any of these mappings, the goal',nl,
  'of the subwindow is set to ''unknown''.'.



database command
  axiom('==='(Name, RHS))
header
  'axiom Name === Stmt.',nl
extends
  theory_extension_cmds
mode
  axiom(@compound)
implementation
  ( RHS = provided(Stmt,Body) ->
      true
  ;
      RHS = Stmt, Body = true
  ),
  deabbreviate(Stmt,Stmt2),
  add_axiom(Name,[],Stmt2,Body)
%% TO DO: Deabbreviate Body in the command axiom.
doc
  'Add Stmt as an axiom, called Name.',nl,
  'Stmt may have the form: Term provided Body, where Body is a',nl,
  'comma-separated sequence of ObVar not_free_in Term constraints,',nl,
  'or ''true''.'.

database command
  axiom(Name,Ass,Stmt,Body)
extends
  theory_extension_cmds
mode
  axiom(@rulename, @list(thy(term)), @thy(term), @code)
implementation
  deabbreviate(Stmt,Stmt2),
  deabbreviate_list(Ass,Ass2),
  add_axiom(Name,Ass2,Stmt2,Body)
%% TO DO: Deabbreviate Body in the command axiom.
doc
  'Add Ass |- Stmt as an axiom, called Name.',nl,
  'WARNING: Axioms with assumptions (Ass /= []) are',nl,
  'not fully supported by this release of Ergo.'.

database command
  param_axiom(Head,Ass,Stmt,Body)
extends
  theory_extension_cmds
mode
  param_axiom(@nonvar, @list(thy(term)), @thy(term), @code)
implementation
  deabbreviate(Stmt,Stmt2),
  deabbreviate_list(Ass,Ass2),
  functor(Head, Name, _),
  add_param_axiom(Name,Head,Ass2,Stmt2,Body)
%% TO DO: Deabbreviate Body in the command param_axiom.
doc
  'Add Ass |- Stmt as a parameterized axiom.',nl,
  nl,
  'WARNING: This command should not be used by most users.',nl,
  '         It is for experienced theory hackers only!',nl,
  nl,
  'Head is a compound term; its principal functor is the name.',nl,
  'The arguments of Head must contain enough information so',nl,
  'that the axiom can be instantiated sufficiently for the Body',nl,
  'to succeed when it is executed.  For example, an axiom that',nl,
  'proves A=<B for integer constants A and B would need to have',nl,
  'both A and B in Head.  Body is unconstrained, except that it',nl,
  'should be designed to ensure the above property.',nl,
  'WARNING: Axioms with assumptions (Ass /= []) are',nl,
  'not fully supported by this release of Ergo.'.

database command
  postulate('==='(Name, RHS))
header
  'postulate Name === Stmt.',nl
extends
  theory_extension_cmds
mode
  postulate(@compound)
implementation
  ( RHS = provided(Stmt,Body) ->
      true
  ;
      RHS=Stmt, Body=true
  ),
  deabbreviate(Stmt,Stmt2),
  add_postulate(Name,[],Stmt2,Body)
%% TO DO: Deabbreviate Body in the command postulate.
doc
  'Add Stmt as a postulate, called Name.',nl,
  'Stmt may be of the form: Term provided Body, where Body is a',nl,
  'comma-separated sequence of ObVar not_free_in Term constraints,',nl,
  'or ''true''.'.

database command
  rerun(Name)
extends
  theory_extension_cmds
mode
  rerun(@rulename)
implementation
  rerun_proof(Name)
doc
  'Prove theorem Name by rerunning its existing proof.'.

database command
  prove_postulate(Name)
extends
  theory_extension_cmds
mode
  prove_postulate(@rulename),
  prove_postulate(@list(rulename))
implementation
  query_ancestors(default_prover(PN,PA)),
  prove_postulate(Name, (set_prover(PN,PA),_))
doc
  'Discharge the postulate(s) called Name.',nl,
  'After the proof is complete, it is checked against all theorems',nl,
  'in the current theory (and its subtheories) to ensure that no',nl,
  'circular reasoning has been introduced.  If circularities have been',nl,
  'introduced, the proof will be rejected and the postulate will remain',nl,
  'unproven.'.

database command
  prove_postulate(Name, Stmt)
extends
  theory_extension_cmds
mode
  prove_postulate(@rulename,  @thy(term) ),
  prove_postulate(@list(rulename),  @thy(term))
implementation
  query_ancestors(default_prover(PN,PA)),
  prove_postulate(Name, (set_prover(PN,PA),_), Stmt)
doc
  'Discharge the postulate(s) called Name.',nl,
  'Unify the result with Stmt to name MetaVariables.',nl,
  'After the proof is complete, it is checked against all theorems',nl,
  'in the current theory (and its subtheories) to ensure that no',nl,
  'circular reasoning has been introduced.  If circularities have been',nl,
  'introduced, the proof will be rejected and the postulate will remain',nl,
  'unproven.'.

theorem_action parameter
  [set_theorem_action, get_theorem_action]
doc
  'This parameter controls the action of the ',
      xref([theorem], global, 'database:theorem/1'),
      ' command.',nl,
  'The value of theorem_action must be one of prove, rerun or postulate.'
extends
  parameters.

set_theorem_action(Action) :-
    check_or_error(
        member_var(Action, [prove,rerun,postulate]),
        ['invalid theorem action: ',q(Action)]),
    tactic_retractall_global(theorem_action(_)),
    tactic_assert_global(theorem_action(Action)).

get_theorem_action(Action) :-
    ( query_global(theorem_action(Act)) ->
        Act = Action
    ;
        Action = prove
    ).

database command
  theorem('==='(Name, RHS))
header
  'theorem Name === Stmt.',nl
extends
  theory_extension_cmds
mode
  theorem(@compound)
implementation
  ( RHS = provided(Stmt,Body) ->
      deabbreviate(Stmt,Stmt2)
  ;
      deabbreviate(RHS,Stmt2),
      Body = true
  ),
  get_theorem_action(Action),
  theorem_action(Action, Name, [], Stmt2, Body)
%% TO DO: Deabbreviate Body in the command theorem.
doc
  'Prove/Rerun/Postulate the given theorem, according to the',nl,
  'current setting of the theorem_action mode',nl,
  '(See the set_theorem_action/1 command).',nl,
  nl,
  'This will either initiate an interactive proof of Stmt,',nl,
  'attempt to prove Stmt by rerunning an existing proof',nl,
  '(which will be read from the file: ''CurrentThy/Name.pf'')',nl,
  'or add Stmt as a postulate that should be proved later.',nl,
  nl,
  'Stmt may be of the form: Term provided Body, where Body is',nl,
  'a comma-separated sequence of ObVar nfi Term constraints, or',nl,
  '''true'', or any prover commands.'.



%% private theorem_action(Action, Name, Ass, Stmt, Body).
%%
%%  Handles the theorem Name === Stmt provided Body
%%  according to Action, which must be one of:
%%
%%     prove     -- start up a new proof using the default prover.
%%     rerun     -- try to rerun an existing proof of Name.
%%     postulate -- postulate the theorem and leave proof for later.
%%
%%  mode theorem_action(+atom, +rulename, +list(term), +term, +nonvar).
%%
theorem_action(prove, Name, Ass, Stmt, Body) :-
    query_ancestors(default_prover(PN,PA)),
    !,
    (Body == true ->
        Tactic = _
    ;
        Tactic = (Body,accept_constraints,_)
    ),
    add_theorem(Name, Ass, Stmt, ask, (set_prover(PN,PA),Tactic)).
theorem_action(postulate, Name, Ass, Stmt, Body) :-
    add_postulate(Name, Ass, Stmt, Body).
theorem_action(rerun, Name, Ass, Stmt, Body) :-
    add_theorem(Name, Ass, Stmt, ask, (call_predicate(Body),rerun(false))).

