%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: bag.thy,v 1.5 2002/09/09 00:43:55 itp Exp $

%%%%%%%%%%%%%%
%% ZFC Bags %%
%%%%%%%%%%%%%%

%% NOTE: This is an old version of this theory and needs to be redone
%% using the new zfc theory.

%% This theory is a theory based on Z's theory of bags.


add_theory(bag).
include_theory(stdops, share). 
include_theory(rel, all).
include_theory(func, all).
include_theory(undef, all).
include_theory(localness, all).
include_theory(zfc, all).


%% Add some extra object variables.
declare_object_vars(w).
declare_object_vars(y).
declare_object_vars(z).

%%%%%%%%%%%%%%%
%% Constants %%
%%%%%%%%%%%%%%%

define empty_bag ===
  empty_set.

%%%%%%%%%%%%%%%
%% Functions %%
%%%%%%%%%%%%%%%

define bag(T) ===
  (T '-|->' pos_ints) provided_that
   T:sets.

%%%%%%%%%%%%%%%
%% Constants %%
%%%%%%%%%%%%%%%

define bags ===
  class_of x (ex y:sets x:bag(y)).


%%%%%%%%%%%%%%%
%% Functions %%
%%%%%%%%%%%%%%%

abbreviation is_empty_bag(B) ===
  B = empty_bag
  provided input.

declare the_bag(X).

define singleton_bag(X) ===
  single(ordered_pair(X,1)) provided_that
   X:sets.

define count(B,X)  ===
  if(X in domain(B),B at X,0) provided_that
  B:bags and X:sets.

define bag_union(B_1,B_2) ===
  (set_of x ex y:(domain(B_1) union domain(B_2)) (
              x = (y maps_to (count(B_1,y) + count(B_2,y)))))
  provided_that B_1:bags and B_2:bags.

add_operator(525, leftassoc, bag_union).

define in_bag(X,B) ===
  count(B,X) > 0 provided_that
  B:bags and X:sets.

add_operator(550, nonassoc, in_bag).

abbreviation not_in_bag(X,B) ===
  not (X in_bag B).

add_operator(550, nonassoc, not_in_bag).

%%%%%%%%%%%%
%% Axioms %%
%%%%%%%%%%%%

axiom bags_subtype_of_sets ===
  bags subtype_of sets.

axiom bag_subtype_of_bags ===
  T:sets <=>
  bag(T) subtype_of bags.

axiom bag_is_undefined_A ===
  not (T:sets) <=> is_undefined(bag(T)).

%% TO DO: What does ``Somewhat mysterious!'' about the following
%% theorem mean? MB 21/9/1998.
axiom bag_dom_bags_is_bags ===
  bag(sets) = bags.  

axiom subtypes_bag_subtype_of_bag ===
  T_1 subtype_of T_2 and T_2:sets <=> 
  bag(T_1) subtype_of bag(T_2).

axiom empty_bag_is_bag ===
  T:sets <=> empty_bag:bag(T).

axiom singleton_bag_equal ===
  singleton_bag(X) = singleton_bag(Y) <=> X = Y.

axiom singleton_bag_not_equal_empty_bag ===
  singleton_bag(X) /= empty_bag.

axiom singleton_bag_is_undefined_A ===
  not (X:sets) <=> is_undefined(singleton_bag(X)).

axiom bag_def_base ===
  the_bag([]) = empty_bag.

axiom bag_def_ind ===
  the_bag([X|L]) = singleton_bag(X) bag_union the_bag(L).

%%%%%%%%%%%%%%
%% Theorems %%
%%%%%%%%%%%%%%

%% Simplification rules
theorem bag_type_A ===
  is_type(bags).

theorem bag_undefined ===
  bag(?) = (?).

theorem not_is_bag_undefined_A ===
  not (?):bags.

theorem not_is_bag_undefined_B ===
  (?):bags <=> false.

theorem not_is_bag_undefined_C ===
  not (?):bag(T).

theorem not_is_bag_undefined_D ===
  (?):bag(T) <=> false.

theorem count_undefined_A ===
  count(?,_) = (?).

theorem count_undefined_B ===
  count(_,?) = (?).

theorem bag_union_undefined_A ===
  ? bag_union _ = (?).

theorem bag_union_undefined_B ===
  _ bag_union ? = (?).

theorem in_bag_undefined_A ===
  ? in_bag _ = (?).

theorem in_bag_undefined_B ===
  _ in_bag ? = (?).

theorem not_in_bag_empty_bag_A ===
  _ not_in_bag empty_bag.

theorem not_in_bag_empty_bag_B ===
  _ in_bag empty_bag <=> false.

%% Localness theorems
theorem bag_local ===
  local x bag(x).

theorem the_bag_local ===
  local x the_bag(x).

theorem singleton_bag_local ===
  local x singleton_bag(x).

theorem count_local_2 ===
  local x count(A,x)
    provided x not_free_in A.

theorem count_local_1 ===
  local x count(x,A)
    provided x not_free_in A.

theorem bag_union_local_2 ===
  local x A bag_union x
    provided x not_free_in A.

theorem bag_union_local_1 ===
  local x x bag_union A
    provided x not_free_in A.

theorem in_bag_local_2 ===
  local x A in_bag x
    provided x not_free_in A.

theorem in_bag_local_1 ===
  local x x in_bag A
    provided x not_free_in A.


%% Typing rules.
theorem bag_type_B ===
  T:sets => is_type(bag(T)).

theorem bag_is_defined ===
  T:sets <=> is_defined(bag(T)).

theorem bag_is_undefined_B ===
 is_undefined(T) => is_undefined(bag(T)).

theorem empty_bag_is_bags ===
 empty_bag : bags.

theorem is_bag_is_defined ===
  B:bags => is_defined(B).

theorem is_bag_is_bag ===  %% TO DO: Fix it! (Found MB 14/9/1998)
  B:bag(T) => B:bags.

theorem singleton_bag_is_bag_A ===
  X:sets <=> singleton_bag(X):bags.

theorem singleton_bag_is_bag_B ===
  T:sets and X:T <=> singleton_bag(X):bag(T).

theorem singleton_bag_is_undefined_B ===
 is_undefined(X) => is_undefined(singleton_bag(X)).

theorem count_is_defined ===
  B:bags and X:sets <=> is_defined(count(B,X)).

theorem count_is_nat ===
  B:bags and X:sets <=> count(B,X):nats.

theorem count_is_pos_int ===
  B:bags and X:sets and X in_bag B <=> count(B,X):pos_ints.

theorem count_is_undefined_B ===
  is_undefined(B) => is_undefined(count(B,_)).

theorem count_is_undefined_C ===
  is_undefined(X) => is_undefined(count(_,X)).

theorem bag_union_is_defined ===
  B_1:bags and B_2:bags <=> is_defined((B_1 bag_union B_2)).

theorem bag_union_is_bag_B ===
  B_1:bags and B_2:bags <=> (B_1 bag_union B_2):bags.

theorem bag_union_is_undefined_B ===
  is_undefined(B) => is_undefined((B bag_union _)).

theorem bag_union_is_undefined_C ===
  is_undefined(B) => is_undefined((_ bag_union B)).

theorem in_bag_is_defined ===
  X:sets and B:bags <=> is_defined((X in_bag B)).

theorem in_bag_is_bool ===
  X:sets and B:bags <=> (X in_bag B):bools.

theorem not_in_bag_is_bool ===
  X:sets and B:bags <=> (X not_in_bag B):bools.

theorem in_bag_is_undefined_A ===
  is_undefined(X) => is_undefined((X in_bag _)).

theorem in_bag_is_undefined_B ===
  is_undefined(B) => is_undefined((_ in_bag B)).

theorem in_bag_is_bag_type ===
  B:bag(T) and X in_bag B => X:T.

%% Other theorems
theorem count_empty_bag ===
  X:sets => count(empty_bag,X) = 0.

theorem count_bag_A ===
  X:sets => count(singleton_bag(X),X) = 1.

theorem count_bag_B ===
  X:sets and Y:sets and X /= Y => count(singleton_bag(X),Y) = 0.

theorem count_is_undefined_A ===
  not (B:bags and X:sets) <=> is_undefined(count(B, X)).

theorem bag_union_is_undefined_A ===
  not (B_1:bags and B_2:bags) <=> is_undefined(B_1 bag_union B_2).

theorem bag_union_empty_bag_A ===
  B:bags => B bag_union empty_bag = B.

theorem count_bag_union ===
  B_1:bags and B_2:bags and X:sets =>
  count(B_1 bag_union B_2,X) = count(B_1,X) + count(B_2,X).

theorem bag_union_commutes ===
  B_1 bag_union B_2 = B_2 bag_union B_1.

theorem bag_union_associates ===
  B_1 bag_union (B_2 bag_union B_3) = (B_1 bag_union B_2) bag_union B_3.

theorem bag_def === 
  T:sets => (
    B:bag(T) <=>
    is_empty_bag(B) or
    (ex x:T B = singleton_bag(x)) or
    (ex x:bag(T) (ex y:bag(T) B = x bag_union y))).

theorem bag_union_empty_bag_B ===
  B:bags => empty_bag bag_union B = B.

theorem count_not_in_bag ===
  X:sets and B:bags and X not_in_bag B <=>
  count(B,X) = 0.

theorem count_in_bag ===
  X in_bag B <=>
  count(B,X) > 0.

theorem in_bag_bag ===
  X:sets =>
  X in_bag singleton_bag(X).

theorem in_bag_bag_union_A ===
  B_1:bags and X in_bag B_2 =>
  X in_bag (B_1 bag_union B_2).

theorem in_bag_bag_union_B ===
  B_2:bags and X in_bag B_1 =>
  X in_bag (B_1 bag_union B_2).

%%%%%%%%%%%%%%%%%%%
%% Opening Rules %%
%%%%%%%%%%%%%%%%%%%


theorem open_bag_1_eq ===
    X = A
    ----------
    bag(X) = bag(A).

theorem open_singleton_bag_1_eq ===
    X = A
    ----------
    singleton_bag(X) = singleton_bag(A).

theorem open_count_1_eq ===
    B = A
    ----------
    count(B, X) = count(A, X).

theorem open_count_2_eq ===
    X = A
    ----------
    count(B, X) = count(B, A).

theorem open_bag_union_1_eq ===
    B_1 = A
    ----------
    (B_1 bag_union B_2) = (A bag_union B_2).

theorem open_bag_union_2_eq ===
    B_2 = A
    ----------
    (B_1 bag_union B_2) = (B_1 bag_union A).

theorem open_in_bag_1_eq ===
    B_1 = A
    ----------
    (B_1 in_bag B_2) = (A in_bag B_2).

theorem open_in_bag_2_eq ===
    B_2 = A
    ----------
    (B_1 in_bag B_2) = (B_1 in_bag A).

theorem open_the_bag_1_eq ===
    X = A
    ----------
    the_bag(X) = the_bag(A).

%%
%% Support for Window Inference - loads
%% bag specific information.
%%

require_tactic_file(wi_support,bag).


