%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: cprop.thy,v 1.20 2002/04/28 05:03:58 itp Exp $
%%
%% This is a theory of classical propositional logic.
%% It extends intuitionistic logic with the law of the excluded middle.


add_theory(cprop).
include_theory(stdops, share).
include_theory(iprop,  all).

axiom or_not ===
    A or not A.

close_theory.


%% Window inference opening rules can provide more context information
%% in this two-valued logic, so we prove some stronger opening rules.

% openrule(A => B,  [1], 
%          [not B], 
%          id, 
%          [=> ---> rev(=>),
%           =  ---> <=>],  
%          [A => B ---> A, 
%           not A  ---> A,
%           true    ---> false,
%           false   ---> true]).

theorem open_imp_1_fol ===
    hyp+++[not A] ---> C => B
    ------------------------------
    (C => A) <= (B => A).

theorem open_imp_1_imp ===
    hyp+++[not A] ---> C <= B
    ------------------------------
    (C => A) => (B => A).

theorem open_imp_1_iff ===
    hyp+++[not A] ---> C <=> B
    ------------------------------
    (C => A) <=> (B => A).


% openrule(A or B, [1],
%         [not B], 
%         id,
%         [=> ---> =>,
%          =  ---> <=>],
%         [A or B ---> A,
%          true   ---> true,
%          false  ---> false]).

theorem open_or_1_fol ===
    hyp+++[not A] ---> C <= B
    ------------------------------
    (C or A) <= (B or A).

theorem open_or_1_imp ===
    hyp+++[not A] ---> C => B
    ------------------------------
    (C or A) => (B or A).

theorem open_or_1_iff ===
    hyp+++[not A] ---> C <=> B
    ------------------------------
    (C or A) <=> (B or A).


% openrule(A or B, [2],
%         [not A],
%         id,
%         [=> ---> =>,
%          =  ---> <=>],
%         [A or B ---> B,
%          true   ---> true,
%          false  ---> false]).

theorem open_or_2_fol ===
    hyp+++[not B] ---> C <= A
    ------------------------------
    (B or C) <= (B or A).

theorem open_or_2_imp ===
    hyp+++[not B] ---> C => A
    ------------------------------
    (B or C) => (B or A).

theorem open_or_2_iff ===
    hyp+++[not B] ---> C <=> A
    ------------------------------
    (B or C) <=> (B or A).



%%
%% Other Classical-only Theorems
%% =============================
%%

theorem iff_true_or_false ===
    (A <=> true) or (A <=> false).

theorem not_not ===
    not (not P) <=> P.

theorem def_imp ===
    (P => Q) <=> not (P and not Q).

theorem split_imp ===
    not (A => B) <=> A and not B.

theorem deMorgan_and ===
    not (B and A) <=> not B or not A.

theorem not_or ===
    not A or A.

theorem imp_or_imp ===
    (A => B) or (B => A).

theorem contrapositive ===
    (A => B) <=> (not B => not A).

theorem contrapositive_c(CID:ctermid) ===
    hyp --- [CID::: not B] +++ [A] ---> B
    -------------------------------------
        ID::: not A
     provided context_search(hyp, ID, CID, not B).

theorem material_imp ===
    (A => B) <=> (not A or B).

theorem imp_elim_c ===
    hyp --- [CID::: A => B] +++ [not A] ---> C,
    hyp --- [CID::: A => B] +++ [B] ---> C
    --------------------------------------
    ID::: C
      provided context_search(hyp, ID, CID, A => B).

theorem deny_disj_l ===
    (not A => B) <=> A or B.

theorem deny_disj_r ===
    (not B => A) <=> A or B.

theorem deny_focus ===
   F <=> (not F => F).
 
theorem deny_focus_1 ===
   not F <=> (F => not F).

theorem or_cases ===
   C or B =>
   (A <=> (C=>A) and (B=>A)).
 
theorem imp_cases ===
   (C => B) =>
   (A <=> (not C=>A) and (B=>A)).

theorem not_or_not_not ===
  not A or not not A.    

%% Theorems for the prove_cprop tactic
%% ===================================
%% Are all of these meant to be in cprop?

theorem prove_cprop_1 ===
    (A <=> B) <=> (A or not B) and (B or not A).

theorem prove_cprop_2 ===
    A or (not A or B).

theorem prove_cprop_3 ===
    A and (not A and B) <=> false.

%theorem prove_cprop_4 ===
%    A and (B and C) <=> A and (C and B).
%
%theorem prove_cprop_5 ===
%    A and (B and C) <=> B and (A and C).

theorem prove_cprop_6 ===
    A or (B or C) <=> A or (C or B).

theorem prove_cprop_7 ===
    A or (B or C) <=> B or (A or C).


%% Theorems for the tableau tactic
%% ===============================

theorem iff_decomp ===
    (A <=> B) <=> (A and B) or (not A and not B).

theorem not_iff_decomp ===
    not (A <=> B) <=> (A and not B) or (not A and B).


%% Extra Unsorted Theorems
%% =======================

theorem case_or_not(B:thy(term)) ===
    hyp+++[B] ---> A,
    hyp+++[not B] ---> A
    --------------------
    A.

theorem not_l_imp_r__or ===
    not B => A
    ----------
    B or A.

theorem not_r_imp_l__or ===
    not A => B
    ----------
    B or A.

theorem hyp_not_imp_false_intro(A:thy(term)) ===
    hyp---[CA ::: not A => false]+++[A] ---> C
    ------------------------------------------
    ConcId ::: C
  provided
    context_search(hyp, ConcId, CA, not A => false).

theorem or_intro ===
    hyp +++ [not A] ---> B
   ----------------------- 
      A or B.

theorem not_not_iff ===
    not not A <=> A.

theorem false_iff_not_iff ===
    (false <=> not A) <=> A.   

theorem hyp_imp(ID:ctermid) ===
      hyp---[ID:::H] +++[not G] ---> (H => K),
      hyp+++[K] ---> G
      ----------------------
      C_0:::G
    provided context_search(hyp, C_0, ID, H).

theorem not_hyp_goal(ID:ctermid) ===
      hyp---[ID:::not H]+++[not G] ---> H
      -----------------------------------
      C_0:::G
    provided context_search(hyp, C_0, ID, not H).


