%% ##include tactic.copyright##


tactic cprop_safe ===
    (   begin_report(cprop_safe)
    ;   (   iprop_safe
        '|' rule_report( cprop_safe, or_not )
        '|' rule_report( cprop_safe, not_or )
        '|' rule_report( cprop_safe, id__not_not )
        '|' fail_report( cprop_safe )
        )
    ;   end_report(cprop_safe)
    ).

% tactic case_or_not__id(Pred,Conseq) ===
% %% Applied in the case where there is a hypothesis ( Pred => Conseq ).
% %% Creates two new proof nodes.
% %% The first has this hypothesis removed and the hypothesis ( Conseq ) added.
% %% The second has this hypothesis removed and the hypothesis ( not Pred ) added.
%     context(hyp,HypNo,'jprop.=>'(Pred,Conseq));
%     ( rule(case_or_not(Pred))
%     ;   [   rule( hyp_imp_intro(Pred,Conseq) )
%         ,   rule(hyp_delete(_, 'jprop.=>'(Pred,Conseq)))
%         ]
%     ).


% tactic case_or_not__not(Pred,Conseq) ===
% %% Applied in the case where there is a hypothesis ( not Pred => Conseq ).
% %% Creates two new proof nodes.
% %% The first has this hypothesis removed and the hypothesis ( Pred ) added.
% %% The second has this hypothesis removed and the hypothesis ( Conseq ) added.
%     (   context(hyp,HypNo,'jprop.=>'('jprop.not'(Pred),Conseq))
%     ;   (   rule(case_or_not(Pred))
%         ;   [   rule( hyp_delete(HypNo, 'jprop.=>'('jprop.not'(Pred),Conseq)))
%             ,   rule( hyp_imp_intro('jprop.not'(Pred),Conseq) )
%             ]
%         )
%     ).

%% TO DO: Rename case_or_not to something with hyp in the name.

tactic case_or_not(Pred,Conseq)
===
%% This tactic does a case analysis where a hypothesis of the form
%% ``Pred => Conseq'' or ``not Pred => Conseq'' appears.
%% Where a hypothesis is of the form ``not Pred => Conseq'' it will not
%% be treated as the former case.
%% Neither will either forms be treated if Conseq is false as this can
%% produce an endless loop ( not A to A => false to cases of A and not A ).
    (   context(hyp,HypNo,'jprop.=>'(Pred,Conseq))
    ;   fails(context(hyp,HypNo,'jprop.=>'(Pred,'jprop.false')))
    ;   !(  %% Handle the case not NotPred => Conseq.
            (   context(hyp,HypNo,'jprop.=>'('jprop.not'(NotPred),Conseq))
            ;   rule(case_or_not(NotPred))
            ;   [   rule( hyp_delete(HypNo, 'jprop.=>'('jprop.not'(NotPred),Conseq)))
                ,   rule( hyp_imp_intro('jprop.not'(NotPred),Conseq) )
                ]
            )
        '|' %% Handle the case Pred => Conseq where Pred \== not NotPred.
            (   rule(case_or_not(Pred))
            ;   [   rule( hyp_imp_intro(Pred,Conseq) )
                ,   rule(hyp_delete(_, 'jprop.=>'(Pred,Conseq)))
                ]
            )
        )
    )
.


tactic cprop_unsafe
===
    (   begin_report(cprop_unsafe)
    ;   (   rule_report( cprop_unsafe, not_l_imp_r__or )
        '|' rule_report( cprop_unsafe, not_r_imp_l__or )
        '|' iprop_unsafe
        '|' rule_report( cprop_unsafe, hyp_not_imp_false_intro )
        '|' tactic_report( cprop_unsafe, case_or_not(_,_) )
        '|' fail_report(cprop_unsafe)
        )
    ;   end_report(cprop_unsafe)
    )
.

