%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%%
%% CVS $Id: fixedpt.thy,v 1.2 2003/07/07 00:01:12 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% A theory about fixed points %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%
%% This theory is (mostly) a translation of the corresponding Isabelle theory.
%%

[zfc].
delete_theory_and_dependents(fixedpt).

add_theory(fixedpt).

include_theory(stdops, share).
include_theory(zfc,all).

%% Add some extra object variables.
declare_object_vars(w).
declare_object_vars(y).
declare_object_vars(z).

%% Fixed point defns.

define is_tfun_on(F,D) ===
     F:sets and D:sets and
     all x:sets x in D => ex_one y:sets (x |--> y) in F.

define bnd_mono(S,F) ===
     S:sets and is_tfun_on(F,power(S)) and (F at S) subset_of S and 
     (all x1:sets all x2:sets x1 subset_of x2 and x2 subset_of S =>
         (F at x1) subset_of (F at x2)).


define pre_lfp(S,F) ===
    set_of x x subset_of S and is_tfun_on(F, power(S)) and (F at x) subset_of x.

define lfp(S, F) ===  gen_inter(pre_lfp(S,F)).

define pre_gfp(S,F) ===
    set_of x x subset_of S and is_tfun_on(F, power(S)) and x subset_of (F at x).

define gfp(S,F) === gen_union(pre_gfp(S,F)).

% Theorems about bnd_mono

theorem bnd_mono_intro ===
      S:sets,
      is_tfun_on(F,power(S)),
      Y:sets,
      ordered_pair(S,Y) in F,
      Y subset_of S,
      all x1:sets all x2:sets x1 subset_of x2 and x2 subset_of S =>
                 (F at x1) subset_of (F at x2)
      ---------------------------------------------
        bnd_mono(S,F)
          provided x1 not_free_in x2, x1 not_free_in S, x1 not_free_in F,
                   x2 not_free_in S, x2 not_free_in F.


theorem bnd_mono_set_property ===
    bnd_mono(D,F)
 ----------------------
  (F at D) subset_of D.

theorem bnd_mono_mono ===
    bnd_mono(D,F), X:sets, Y:sets, X subset_of Y, Y subset_of D
  -----------------------------------------------
    (F at X) subset_of (F at Y).

theorem bnd_mono_subset ===
     bnd_mono(D,F), X:sets, X subset_of D
  -----------------------------------------------
    (F at X) subset_of D.

theorem is_tfun_on_type ===
    is_tfun_on(F,D) 
 ---------------------
      F:sets.

theorem bnd_mono_type ===
     bnd_mono(S,F)
  ----------------------
     S:sets.

theorem bnd_mono_tfun ===
   bnd_mono(S,F)
 --------------------
  is_tfun_on(F,power(S)).

theorem bnd_mono_union ===
    bnd_mono(D,F), X:sets, Y:sets, X subset_of D, Y subset_of D
  -----------------------------------------------
    ((F at X) union (F at Y)) subset_of (F at (X union Y)).

theorem bnd_mono_inter ===
    bnd_mono(D,F), X:sets, Y:sets, X subset_of D, Y subset_of D
  -----------------------------------------------
    (F at (X inter Y)) subset_of ((F at X) inter (F at Y)).

%% Theorems about least fixed points

theorem pre_lfp_exists ===
        S:sets
   -------------------------------
   ex_set x x subset_of S and is_tfun_on(F, power(S)) and (F at x) subset_of x
        provided x not_free_in S, x not_free_in F.

theorem pre_lfp_type ===
        S:sets
   -------------------------------
       pre_lfp(S,F):sets.

theorem pre_lfp_property ===
     X:sets, S:sets, is_tfun_on(F, power(S))
    -----------------------------------
     X in pre_lfp(S,F) 
     <=> 
     X subset_of S and (F at X) subset_of X.

theorem lfp_type ===
       S:sets
      --------------------------------
        lfp(S,F):sets.

theorem lfp_lowerbound ===
   A:sets, D:sets,  is_tfun_on(F, power(D)), 
   (F at A) subset_of A, A subset_of D
 --------------------------------------
     lfp(D, F) subset_of A.

theorem lfp_subset ===
   D:sets,  is_tfun_on(F, power(D))
   ----------------------------------
   lfp(D, F) subset_of D.

theorem lfp_greatest ===
   A:sets, D:sets, is_tfun_on(F, power(D)), (F at D) subset_of D,
   all x:sets x subset_of D and (F at x) subset_of x => A subset_of x
 ----------------------------------------------
   A subset_of lfp(D, F)
     provided x not_free_in D, x not_free_in F, x not_free_in A.

theorem lfp_lemma1 ===
    bnd_mono(D,F), A:sets, (F at A) subset_of A, A subset_of D
    ----------------------------------------
    (F at lfp(D,F)) subset_of A.

theorem lfp_lemma2 ===
     bnd_mono(D,F)
   ---------------------
    (F at lfp(D,F)) subset_of lfp(D,F).

theorem lfp_lemma3 ===
      bnd_mono(D,F)
   ---------------------
   lfp(D,F)  subset_of (F at lfp(D,F)).

theorem lfp_unfold ===
     bnd_mono(D,F)
    ---------------------
   lfp(D,F) = (F at lfp(D,F)).


theorem exists_lfp_set ===
    bnd_mono(D,F)
   ----------------------------
    ex_set x x in lfp(D,F) and P
     provided
       x not_free_in D, x not_free_in F.

theorem lfp_set_type ===
    bnd_mono(D,F)
    -------------------------
    (set_of x x in lfp(D,F) and P):sets
     provided
       x not_free_in D, x not_free_in F.

theorem lfp_set_property ===
    bnd_mono(D,F), A:sets
    -------------------------
    A in (set_of x x in lfp(D,F) and P) <=> A in lfp(D,F) and [A/x]P
     provided
       x not_free_in D, x not_free_in F, x not_free_in A.
      
theorem set_is_pre_fixedpt ===
    bnd_mono(D,F), 
    all x:sets x in F at (set_of y y in lfp(D,F) and [y/x]P) => P
  ------------------------------------
    (F at (set_of y y in lfp(D,F) and [y/x]P)) subset_of
       (set_of y y in lfp(D,F) and [y/x]P)
   provided
      x not_free_in D, x not_free_in F, 
      y not_free_in D, y not_free_in F, 
      y not_free_in x, y not_free_in P.


theorem induct(A:thy(term), x:obvar, P:thy(term)) ===
   bnd_mono(D,F), A:sets, A in lfp(D,F),
   all x:sets x in F at (set_of y y in lfp(D,F) and [y/x]P) => P
   ---------------------------------
   [A/x]P
  provided
      x not_free_in D, x not_free_in F, x not_free_in A, 
      y not_free_in D, y not_free_in F, y not_free_in A, 
      y not_free_in x, y not_free_in P.



%% Theorems about greatest fixed points

theorem pre_gfp_exists ===
        S:sets
   -------------------------------
   ex_set x x subset_of S and is_tfun_on(F, power(S)) and x subset_of (F at x)
        provided x not_free_in S, x not_free_in F.

theorem pre_gfp_type ===
        S:sets
   -------------------------------
       pre_gfp(S,F):sets.

theorem pre_gfp_property ===
     X:sets, S:sets, is_tfun_on(F, power(S))
    -----------------------------------
     X in pre_gfp(S,F) 
     <=> 
     X subset_of S and X subset_of (F at X).

theorem gfp_type ===
       S:sets
      --------------------------------
        gfp(S,F):sets.

theorem gfp_upperbound ===
   A:sets, D:sets,  is_tfun_on(F, power(D)), 
   A subset_of (F at A) , A subset_of D
 --------------------------------------
     A subset_of gfp(D, F).

theorem gfp_subset ===
   D:sets,  is_tfun_on(F, power(D))
   ----------------------------------
   gfp(D, F) subset_of D.

theorem gfp_least ===
   A:sets, bnd_mono(D, F),
   all x:sets x subset_of D and x subset_of (F at x)  => x subset_of A
 ----------------------------------------------
   gfp(D, F) subset_of A
     provided x not_free_in D, x not_free_in F, x not_free_in A.

theorem gfp_lemma1 ===
    bnd_mono(D,F), A:sets, A subset_of (F at A), A subset_of D
    ----------------------------------------
    A subset_of (F at gfp(D,F)).

theorem gfp_lemma2 ===
     bnd_mono(D,F)
   ---------------------
    gfp(D,F) subset_of (F at gfp(D,F)).

theorem gfp_lemma3 ===
      bnd_mono(D,F)
   ---------------------
   (F at gfp(D,F)) subset_of gfp(D,F).

theorem gfp_unfold ===
     bnd_mono(D,F)
    ---------------------
   gfp(D,F) = (F at gfp(D,F)).


theorem exists_gfp_set ===
    bnd_mono(D,F)
   ----------------------------
    ex_set x x in gfp(D,F) and P
     provided
       x not_free_in D, x not_free_in F.

theorem gfp_set_type ===
    bnd_mono(D,F)
    -------------------------
    (set_of x x in gfp(D,F) and P):sets
     provided
       x not_free_in D, x not_free_in F.

theorem gfp_set_property ===
    bnd_mono(D,F), A:sets
    -------------------------
    A in (set_of x x in gfp(D,F) and P) <=> A in gfp(D,F) and [A/x]P
     provided
       x not_free_in D, x not_free_in F, x not_free_in A.
      
theorem weak_coinduct ===
   bnd_mono(D,F), X:sets, A:sets, A in X, X subset_of D, X subset_of (F at X)
   -------------------------------
   A in gfp(D,F).

theorem coinduct_lemma ===
   bnd_mono(D,F), X:sets,  
   X subset_of D, X subset_of (F at (X union gfp(D,F)))
   ---------------------------------
   (X union gfp(D,F)) subset_of (F at (X union gfp(D,F))).

theorem coinduct ===
   bnd_mono(D,F), X:sets, A:sets, A in X, 
   X subset_of D, X subset_of (F at (X union gfp(D,F)))
   ---------------------------------
   A in gfp(D,F).

%% A theorem about bnd_mono lamda functions for use in constructing 
%% fixed points.

theorem lambda_bnd_mono ===
  D:sets, 
  all x:sets x in power(D) => 
         B:sets and B subset_of D and all y x = y => B = [y/x]B,
  all x:sets all x2:sets x subset_of x2 and x2 subset_of D =>
         B subset_of [x2/x]B
  -------------------------
  bnd_mono(D, lambda x:power(D) B)
    provided x not_free_in D, x not_free_in x2, x not_free_in y,
             y not_free_in B, y not_free_in D, y not_free_in x2,
             x2 not_free_in B, x2 not_free_in D.

save.

