%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%%
%% CVS $Id: func.thy,v 1.8 2003/05/28 03:14:50 itp Exp $

%%%%%%%%%%%%%%%%%%%
%% ZFC Functions %%
%%%%%%%%%%%%%%%%%%%

%% A theory of functions based on ZFC.


add_theory(func).
include_theory(stdops, share). 
include_theory(rel, all).

%% Add some extra object variables.
declare_object_vars(w).
declare_object_vars(y).
declare_object_vars(z).                                                         



%%%%%%%%%%%%%%%%%
%% Definitions %%
%%%%%%%%%%%%%%%%%

define partial_fun(D, R) === 
    set_of x x in rel(D, R) and 
         all y:sets y in domain(x) => ex_one z:sets (y |--> z) in x.


define total_fun(D,R) ===
  set_of x x in partial_fun(D, R) and domain(x) = D.



define injective_fun(D,R) ===
  set_of x x in partial_fun(D, R) and 
    all y:sets y in range(x) => ex_one z:sets (z |--> y) in x.


define total_injective_fun(D,R) ===
  set_of x x in injective_fun(D, R) and domain(x) = D.


define surjective_fun(D,R) ===
  set_of x x in partial_fun(D, R) and range(x) = R.


define total_surjective_fun(D,R) ===
  set_of x x in total_fun(D, R) and range(x) = R.


define bijection(D,R) ===
  set_of x x in total_injective_fun(D, R) and range(x) = R.


define const_fun(A, C) === A cross single(C).


%%%%%%%%%%%%%%
%% Theorems %%
%%%%%%%%%%%%%%

theorem ex_partial_fun ===
    D:sets, R:sets
   ------------------------
    ex_set x x in rel(D, R) and
         all y:sets y in domain(x) => ex_one z:sets (y |--> z) in x
     provided x not_free_in D, x not_free_in R, 
              x not_free_in y, x not_free_in z,
              y not_free_in z.

theorem partial_fun_type ===
    D:sets, R:sets
   ------------------------
    partial_fun(D, R):sets.

theorem partial_fun_property ===
    A:sets, D:sets, R:sets
  --------------------------
    A in partial_fun(D, R) 
    <=> 
    A in rel(D, R) and 
      all y:sets y in domain(A) => ex_one z:sets (y |--> z) in A
        provided y not_free_in A, z not_free_in A, y not_free_in z.
    
theorem ex_total_fun ===
   D:sets, R:sets
 ----------------------------
 ex_set x x in partial_fun(D, R) and domain(x) = D
   provided x not_free_in D, x not_free_in R.

theorem total_fun_type ===
    D:sets, R:sets
   ------------------------
    total_fun(D, R):sets.

theorem total_fun_property ===
    A:sets, D:sets, R:sets
  --------------------------
    A in total_fun(D, R)
    <=>
    A in partial_fun(D, R) and domain(A) = D.

theorem ex_injective_fun ===
   D:sets, R:sets
 ----------------------------
 ex_set x x in partial_fun(D, R) and
    all y:sets y in range(x) => ex_one z:sets (z |--> y) in x
  provided x not_free_in D, x not_free_in R,
           x not_free_in y, x not_free_in z,
           y not_free_in z.

theorem injective_fun_type ===
    D:sets, R:sets
   ------------------------
    injective_fun(D, R):sets.

theorem injective_fun_property ===
    A:sets, D:sets, R:sets
  --------------------------
    A in injective_fun(D, R)
    <=>
    A in partial_fun(D, R) and 
    all y:sets y in range(A) => ex_one z:sets (z |--> y) in A
  provided y not_free_in A, z not_free_in A,
           y not_free_in z.

theorem ex_total_injective_fun ===
   D:sets, R:sets
 ----------------------------
 ex_set x x in injective_fun(D, R) and domain(x) = D
   provided x not_free_in D, x not_free_in R.

theorem total_injective_fun_type ===
    D:sets, R:sets
   ------------------------
    total_injective_fun(D, R):sets.

theorem total_injective_fun_property ===
    A:sets, D:sets, R:sets
  --------------------------
    A in total_injective_fun(D, R)
    <=>
    A in injective_fun(D, R) and domain(A) = D.

theorem ex_surjective_fun ===
   D:sets, R:sets
 ----------------------------
 ex_set x x in partial_fun(D, R) and range(x) = R
   provided x not_free_in D, x not_free_in R.

theorem surjective_fun_type ===
    D:sets, R:sets
   ------------------------
    surjective_fun(D, R):sets.

theorem surjective_fun_property ===
    A:sets, D:sets, R:sets
  --------------------------
    A in surjective_fun(D, R)
    <=>
    A in partial_fun(D, R) and range(A) = R.

theorem ex_total_surjective_fun ===
   D:sets, R:sets
 ----------------------------
 ex_set x x in total_fun(D, R) and range(x) = R
   provided x not_free_in D, x not_free_in R.

theorem total_surjective_fun_type ===
    D:sets, R:sets
   ------------------------
    total_surjective_fun(D, R):sets.

theorem total_surjective_fun_property ===
    A:sets, D:sets, R:sets
  --------------------------
    A in total_surjective_fun(D, R)
    <=>
    A in total_fun(D, R) and range(A) = R.

theorem ex_bijection ===
   D:sets, R:sets
 ----------------------------
 ex_set x x in total_injective_fun(D,R) and range(x) = R
   provided x not_free_in D, x not_free_in R.

theorem bijection_type ===
    D:sets, R:sets
   ------------------------
    bijection(D,R):sets.

theorem bijection_property ===
    A:sets, D:sets, R:sets
  --------------------------
    A in bijection(D, R)
    <=>
    A in total_injective_fun(D, R) and range(A) = R.

theorem const_fun_type ===
    A:sets, C:sets
  --------------------
   const_fun(A,C):sets.

theorem const_fun_property ===
   X:sets, A:sets, C:sets
 ----------------------------
   X in const_fun(A,C) <=>  ex x:sets x in A and X = (x |--> C) 
      provided x not_free_in A, x not_free_in C, x not_free_in X.

theorem const_apply ===
  X:sets, A:sets, C:sets, X in A
 --------------------------------
  const_fun(A,C) at X = C.

theorem apply_fun ===
  X:sets, Y:sets, D:sets, R:sets, F:sets, 
  F in partial_fun(D, R), (X |--> Y) in F 
 -------------------------------------------------
  F at X  = Y.

%% Simplification of empty subterms

theorem empty_partial_fun_property ===
       X:sets, R:sets 
 ----------------------
  X in partial_fun(empty, R) <=> X = empty.

theorem partial_fun_empty_property ===
       X:sets, D:sets 
 -----------------------
 X in partial_fun(D, empty) <=> X = empty.

theorem empty_total_fun_property ===
       X:sets, R:sets 
  -------------------------
  X in total_fun(empty, R) <=> X = empty.

theorem empty_injective_fun_property ===
       X:sets, R:sets 
  -------------------------
  X in injective_fun(empty, R) <=> X = empty.

theorem injective_fun_empty_property ===
       X:sets, D:sets 
  -------------------------
  X in injective_fun(D, empty) <=> X = empty.

theorem empty_total_injective_fun_property ===
       X:sets, R:sets 
  --------------------------
  X in total_injective_fun(empty, R) <=> X = empty.


theorem surjective_fun_empty_property ===
       X:sets, D:sets 
  --------------------------
  X in surjective_fun(D, empty) <=> X = empty.


%% Inclusion properties

theorem const_fun_is_total_fun ===
 C:sets, A:sets, Y:sets, C in Y 
 ------------------------------
 const_fun(A,C) in total_fun(A, Y) .


theorem partial_fun_subset_rel ===
  D:sets, R:sets 
 ----------------------------
 partial_fun(D, R) subset_of rel(D, R).

theorem total_fun_subset_partial_fun ===
  D:sets, R:sets 
  --------------------------
  total_fun(D, R) subset_of partial_fun(D, R).

theorem injective_fun_subset_partial_fun ===
  D:sets, R:sets 
  ------------------------------
  injective_fun(D, R) subset_of partial_fun(D, R).

theorem total_injective_fun_subset_injective_fun ===
  D:sets, R:sets
 -------------------------------
 total_injective_fun(D, R) subset_of injective_fun(D, R).

theorem surjective_fun_subset_partial_fun ===
  D:sets, R:sets 
 --------------------------------
 surjective_fun(D, R) subset_of partial_fun(D, R).

theorem total_surjective_fun_subset_surjective_fun ===
  D:sets, R:sets 
  --------------------------------
  total_surjective_fun(D, R) subset_of surjective_fun(D, R).

theorem bijection_subset_total_surjective_fun ===
  D:sets, R:sets 
  ---------------------------------
  bijection(D, R) subset_of total_surjective_fun(D, R).

theorem bijection_subset_total_injective_fun ===
  D:sets, R:sets 
  ----------------------------------
  bijection(D, R) subset_of total_injective_fun(D, R).

 
%%%%%%%%%%%%%%%%%%%
%% Opening Rules %%
%%%%%%%%%%%%%%%%%%%


theorem open_partial_fun_1_eq ===
    D = A
    ----------
    ID:::partial_fun(D, B) = partial_fun(A, B)
      provided context_search(hyp, ID, _, D:sets),
               context_search(hyp, ID, _, B:sets).

theorem open_partial_fun_2_eq ===
    R = A
    ----------
    ID:::partial_fun(B, R) = partial_fun(B, A)
      provided context_search(hyp, ID, _, R:sets),
               context_search(hyp, ID, _, B:sets).

theorem open_total_fun_1_eq ===
    D = A
    ----------
    ID:::total_fun(D, B) = total_fun(A, B)
      provided context_search(hyp, ID, _, D:sets),
               context_search(hyp, ID, _, B:sets).

theorem open_total_fun_2_eq ===
    R = A
    ----------
    ID:::total_fun(B, R) = total_fun(B, A)
      provided context_search(hyp, ID, _, R:sets),
               context_search(hyp, ID, _, B:sets).

theorem open_injective_fun_1_eq ===
    D = A
    ----------
    ID:::injective_fun(D, B) = injective_fun(A, B)
      provided context_search(hyp, ID, _, D:sets),
               context_search(hyp, ID, _, B:sets).

theorem open_injective_fun_2_eq ===
    R = A
    ----------
    ID:::injective_fun(B, R) = injective_fun(B, A)
      provided context_search(hyp, ID, _, R:sets),
               context_search(hyp, ID, _, B:sets).


theorem open_total_injective_fun_1_eq ===
    D = A
    ----------
    ID:::total_injective_fun(D, B) = total_injective_fun(A, B)
      provided context_search(hyp, ID, _, D:sets),
               context_search(hyp, ID, _, B:sets).

theorem open_total_injective_fun_2_eq ===
    R = A
    ----------
    ID:::total_injective_fun(B, R) = total_injective_fun(B, A)
      provided context_search(hyp, ID, _, R:sets),
               context_search(hyp, ID, _, B:sets).

theorem open_surjective_fun_1_eq ===
    D = A
    ----------
    ID:::surjective_fun(D, B) = surjective_fun(A, B)
      provided context_search(hyp, ID, _, D:sets),
               context_search(hyp, ID, _, B:sets).

theorem open_surjective_fun_2_eq ===
    R = A
    ----------
    ID:::surjective_fun(B, R) = surjective_fun(B, A)
      provided context_search(hyp, ID, _, R:sets),
               context_search(hyp, ID, _, B:sets).

theorem open_total_surjective_fun_1_eq ===
    D = A
    ----------
    ID:::total_surjective_fun(D, B) = total_surjective_fun(A, B)
      provided context_search(hyp, ID, _, D:sets),
               context_search(hyp, ID, _, B:sets).

theorem open_total_surjective_fun_2_eq ===
    R = A
    ----------
    ID:::total_surjective_fun(B, R) = total_surjective_fun(B, A)
      provided context_search(hyp, ID, _, R:sets),
               context_search(hyp, ID, _, B:sets).

theorem open_bijection_1_eq ===
    D = A
    ----------
    ID:::bijection(D, B) = bijection(A, B)
      provided context_search(hyp, ID, _, D:sets),
               context_search(hyp, ID, _, B:sets).

theorem open_bijection_2_eq ===
    R = A
    ----------
    ID:::bijection(B, R) = bijection(B, A)
      provided context_search(hyp, ID, _, R:sets),
               context_search(hyp, ID, _, B:sets).

theorem open_const_fun_1_eq ===
    A = C
    ----------
    ID:::const_fun(A, B) = const_fun(C, B)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, B:sets).

theorem open_const_fun_2_eq ===
    B = C
    ----------
    ID:::const_fun(A, B) = const_fun(A, C)
      provided context_search(hyp, ID, _, A:sets),
               context_search(hyp, ID, _, B:sets).


%% Theorems about composition

theorem pfun_comp_pfun_is_pfun ===
    A:sets, B:sets, C:sets, F:sets, G:sets,
    F in partial_fun(A,B), G in partial_fun(B, C)
   -----------------------------------------------
    fwd_compose(F, G) in partial_fun(A, C).

theorem surj_comp_surj_is_surj ===
    A:sets, B:sets, C:sets, F:sets, G:sets,
    F in surjective_fun(A,B), G in surjective_fun(B, C)
   -----------------------------------------------
    fwd_compose(F, G) in surjective_fun(A, C).

%% Theorem about override

theorem rel_override_disjoint_pfuns ===
    A:sets, B:sets, C:sets, D:sets, F:sets, G:sets,
    F in partial_fun(A,B), G in partial_fun(C, D), A inter C = empty
   -----------------------------------------------
    rel_override(F, G) in partial_fun((A union C), (B union D)).

theorem rel_override_disjoint_surj ===
    A:sets, B:sets, C:sets, D:sets, F:sets, G:sets,
    F in surjective_fun(A,B), G in surjective_fun(C, D), A inter C = empty
   -----------------------------------------------
    rel_override(F, G) in surjective_fun((A union C), (B union D)).



   

%%
%% Support for Window Inference - loads
%% func specific information.
%%

require_global_tactic_file(wi_support).
require_tactic_file(wi_support,func).


