%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: iprop.thy,v 1.15 2002/04/28 05:03:59 itp Exp $
%%
%% This is an Ergo theory of Gentzen's intuitionistic propositional
%% natural deduction calculus.  Derived from his NJ calculus [1].
%% As shown by Gentzen [1], it is equivalent to Heyting's 
%% Intuitionistic Logic [2,3].
%%
%% It is built on top of the jprop theory (Johansson's Minimal Logic).
%% This theory extends that logic by adding the contradiction rule.
%% This results in a broader interpretation of implication:
%% (P => Q) can be proved from not(P).
%%
%%   [1] Gerhard Gentzen.  "Investigations into Logical Deduction."
%%       Chapter 3 in "The Collected Papers of Gerhard Gentzen",
%%       Edited by M. E. Szabo, North-Holland, 1969.  Pages 68-131.
%%       First published in German in 1934.
%%
%%   [2] A. Heyting. Die Formalen Regeln der Intuitionistischen Logik.
%%       Sitzungsberichte der Deutschen Akademie der Wissenschaften zu Berlin,
%%       Klasse fur Mathematik, Physik und Technik. pp 42-56, 1930. 
%%
%%   [3] A. Heyting. Intuitionism: An Introduction.  Studies in Logic and the
%%       Foundations of Mathematics. North-Holland. 1971.
%%  

add_theory(iprop).
include_theory(stdops, share).
include_theory(jprop,  all).

%% TO DO: Alias the old name of contradiction to false__id in theory iprop.

axiom false__id ===
    false
    -----
    P.

close_theory.

%% Distribution rules
%% ==================

theorem split_imp ===
    not (A => B) <=> not not A and not B.

theorem and_not_imp_not_imp ===
    B and not A => not (B => A).

theorem deMorgan_or ===
    not (B or A) <=> not B and not A.

theorem not_not_iff_imp ===
    not not (A => B) <=> (not not A => not not B).


%% Other rules
%% ===========

theorem or_elim_l ===
    not A and (A or B) => B.

theorem or_elim_r ===
    not B and (A or B) => A.

theorem not_imp ===
    not A => (A => B).

theorem contradiction2 ===
    A and not A => B.

theorem not_and_imp_not_imp ===
    not A and B => not (B => A).


%% Theorems about true and false
%% =============================

theorem not_true ===
    (not true) <=> false.

theorem not_false ===
    (not false) <=> true.

theorem false_and ===
    (false and A) <=> false.

theorem and_false ===
    (A and false) <=> false.

theorem false_or ===
    (false or A) <=> A.

theorem or_false ===
    (A or false) <=> A.

theorem imp_false ===
    (A => false) <=> not A.

theorem false_imp ===
    (false => A) <=> true.

theorem true_iff_false ===
    (true <=> false) <=> false.

theorem false_iff_true ===
    (false <=> true) <=> false.

theorem false_iff_imp_not ===
    (false <=> A) <=> not A.

theorem iff_false_imp_not ===
    (A <=> false) <=> not A.

theorem false_imp1 ===
    false => A.

theorem not_false1 ===
    not false.

theorem and_not ===
    (A and not A) <=> false.

theorem not_and ===
    (not A and A) <=> false.

theorem not_imp_iff_false ===
    not A => (A <=> false).

theorem imp_not_iff_false ===
    A => (not A <=> false).

%% Extra Unsorted Theorems
%% =======================

theorem from__false( CId:ctermid ) ===
    ConcId ::: A
  provided
    context_search( hyp, ConcId, CId, false ).

theorem contradiction(CId1:ctermid, CId2:ctermid) ===
    ConcId ::: A
  provided
    context_search( hyp, ConcId, CId1, B ),
    context_search( hyp, ConcId, CId2, not B ).
    
theorem not__iff_false ===
    not A
 --------------
  A <=> false.

theorem not_iff_false ===
      A
  -------------
  not A <=> false.

