%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: jpred.thy,v 1.27 2003/07/12 02:36:51 itp Exp $
%%
%% Johansson's Minimal Predicate Calculus (without equality)
%% ======================================
%%

add_theory(jpred).
include_theory(stdops, share).
include_theory(jprop, all).

%% Support for universal quantification
%% ====================================
%%
%%  Note: Gentzen's `all_intro' rule is (in Ergo syntax):
%%
%%                [x2/x]F
%%                -------
%%                all x F
%%
%%        with the two following restrictions:
%%
%%        1) the eigenvariable (x2) must not appear free in (all x F).
%%
%%        2) the eigenvariable (x2) must not appear free in any
%%           assumptions upon which the proof of [x2/x]F depends.
%%
%%        The second restriction is ensured by the context_nfi(x,ID1)
%%        constraint on the all_intro rule below.
%%
%%        In our formulation, the first restriction is trivially true,
%%        because we always choose the eigenvariable to be x, and 
%%        `x not_free_in (all x A)' is a tautology.
%%
%%        To see that this choice of eigenvariable is not a restriction
%%        consider applying this rule to a proof node containing `(all y B)'.
%%        The unification of `(all x A)' and `(all y B)' gives:
%%
%%            A = [x/y]B provided x not_free_in [$/y]B
%%
%%        and this proviso is equivalent to Gentzen's first restriction.
%%

declare all x _.

axiom all_intro(x:obvar) ===
    hyp ??? nfi(x) ---> A
    ---------------------
    all x A.

axiom all_elim(A:thy(term),x:obvar,T:thy(term)) ===
    all x A
    --------
    [T/x]A.

%%
%% Support for existential quantification
%% ======================================
%%
declare ex x _.

axiom ex_intro(x:obvar,W:thy(term)) ===
    [W/x]A
    -------
    ex x A.

axiom ex_elim(x:obvar,A:thy(term)) ===
    ID2::: ex x A,
    ID1::: hyp ??? nfi(x) +++ [A] ---> C
    ------------------------------------
    ID ::: C
  provided
    x not_free_in C.

close_theory.

theorem all_elim1(A:thy(term),x:obvar,T:thy(term)) ===
    ID2::: all x A,
    ID1::: hyp +++ [[T/x]A] ---> C
    ------------------------------------
    ID ::: C.

theorem all_elim_c(A:thy(term),x:obvar,T:thy(term)) ===
    hyp +++ [[T/x]A] ---> C
    ------------------------------------
    Node ::: C
  provided
    context_search(hyp, Node, CId2, all x A). 

theorem all_elim_c_once(A:thy(term),x:obvar,V:thy(term)) ===
    hyp --- [CId2 ::: all x A] +++ [[V/x]A] ---> C
    ------------------------------------
    Node ::: C
  provided
    context_search(hyp, Node, CId2, all x A).


theorem ex_elim_c(x:obvar,A:thy(term)) ===
    hyp --- [CId ::: ex x A] ??? nfi(x) +++ [A] ---> C
    ------------------------------------
    Node ::: C
  provided
    context_search(hyp, Node, CId, ex x A), 
    x not_free_in C.

theorem ex_imp_imp_ex ===
    ((ex x B) => A) => (ex x (B => A))
  provided
    x not_free_in A.

theorem imp_imp_all ===
    (all x (A => B)) => (A => (all x B))
  provided
    x not_free_in A.

theorem imp_imp_ex ===
    (all x0 (B => A)) => ((ex x0 B) => A)
  provided
    x0 not_free_in A.

theorem mk_inst ===
    (all x A) => [X/x]A.

theorem ex_intro2 ===
    [X/x]A => (ex x A).

theorem all_intro2 ===
    A <=> (all x A)
  provided
    x not_free_in A.


%% Include old jpred theorems and opening rules.

theorem all_all ===
  ((all x_0 all x_1 A) <=> (all x_1 all x_0 A)) provided
  x_0 not_free_in x_1, x_1 not_free_in x_0.  

theorem ex_ex ===
  ((ex x_0 ex x_1 A) <=> (ex x_1 ex x_0 A)) provided
  x_0 not_free_in x_1, x_1 not_free_in x_0.  

theorem all_and_iff_all_and_all ===
  (all x (A and B)) <=> (all x A) and (all x B).

theorem all_imp_ex ===
  (all x A) => ex x A.      

theorem all_imp_iff_imp_all ===
  (all x (A => B)) <=> (A => all x B) provided
  x not_free_in A.  

theorem all_imp_imp_all_imp_all ===
  (all x (A => B)) => ((all x A) => all x B).

theorem all_imp_trans ===
  (all x (A => B)) and (all x (B => C)) =>
    all x (A => C). 

theorem ex_all_imp_all_ex  ===
  (ex x all x1 A) => all x1 ex x A. 

theorem ex_and_imp_ex_and_ex ===
  (ex x (A and B)) => (ex x A) and (ex x B). 

theorem ex_and_nfi_iff_ex_and ===
  (ex x (A and B)) <=> (ex x A) and B 
    provided x not_free_in B. 

theorem ex_nfi_and_iff_and_ex ===
  (ex x (A and B)) <=> A and (ex x B) 
    provided x not_free_in A. 

theorem ex_elim2 ===
  ((ex x A) <=> A) provided
  x not_free_in A. 

theorem ex_imp_all_imp ===
  (ex x A) => (B <=> (all x (A => B))) provided
  x not_free_in B.      

theorem ex_imp_imp_imp_ex ===
  (ex x (A => B)) => (A => ex x B) provided
  x not_free_in A. 

theorem ex_not_imp_not_all ===
  (ex x (not A)) => not all x A. 

theorem mk_all ===
  (all x A) <=> [X/x]A and all x A.

theorem mk_ex ===
  (ex x A) <=> [X/x]A or ex x A.    

theorem not_ex_imp_all_not ===
  not (ex x A) => all x not A. 

theorem not_not_all_imp_all_not_not ===
  not not (all x A) => all x not not A.   

theorem or_all_imp_all_or ===
  A or (all x B) => all x (A or B) provided
  x not_free_in A.   

theorem all_true ===
  (all x true) <=> true. 

theorem ex_true ===
  (ex x true) <=> true. 

theorem all_false ===
  (all x false) <=> false.

theorem ex_false ===
  (ex x false) <=> false.


%% Opening Rules for all

%openrule(all x _, [b([x])],
%         [],
%         id,
%         [=> ---> =>,
%          =  ---> <=>],
%         [true    ---> true,
%          all x B ---> B]).

theorem open_all_1_fol( x0:obvar ) ===
  hyp ??? nfi(x0) ---> B <= A
  ------------------
  (all x0 B) <= (all x0 A).

theorem open_all_1_imp( x0:obvar ) ===
  hyp ??? nfi(x0) ---> B => A
  ------------------
  (all x0 B) => (all x0 A).

theorem open_all_1_iff( x0:obvar ) ===
  hyp ??? nfi(x0) ---> B <=> A
  -------------------
  (all x0 B) <=> (all x0 A).

%% Opening Rules for ex

theorem open_ex_1_fol( x0:obvar ) ===
  hyp ??? nfi(x0) ---> B <= A
  ----------------
  (ex x0 B) <= (ex x0 A).

theorem open_ex_1_imp( x0:obvar ) ===
  hyp ??? nfi(x0) ---> B => A
  ----------------
  (ex x0 B) => (ex x0 A).

theorem open_ex_1_iff( x0:obvar ) ===
  hyp ??? nfi(x0) ---> B <=> A
  -----------------
  (ex x0 B) <=> (ex x0 A).


%%
%% Support for Window Inference - loads
%% jpred specific information.
%%

require_tactic_file(wi_support,jpred).



