%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%%
%% Retrieves information about a horn clause from the database.
%%

horn_clause_get_seq(seq(_,[],Seq),node(Seq,_,_,_)).

horn_clause(Table, Formula, NewFormulae, Rule, Constraints) :-
    query_ancestors(Table(Formula, _, rule(Rule)),Thy),
    rule_query(Thy2, Rule, _, Formula, _, Seqs, Constraints),
    ancestor_of(Thy,Thy2),
    (Seqs = [] ->
        NewFormulae = []
     ;
        map(horn_clause_get_seq, Seqs, NewFormulae)
    ).


%%
%% Extract the hyps from the context
%%
horn_extract_hyps_support(c(_,Term),Term).

horn_extract_hyps(Context, Hyps) :-
    member('='(hyp,Hyps0),Context),
    map(horn_extract_hyps_support, [Hyps0, Hyps]).

%%
%% This code constructs a proof tree of the form
%% node(Goal, Rule, Children, Done)
%% where Goal is the goal of the (sub)tree,
%% Rule is the name of the rule used on Goal
%% Children is a list of node terms which are the 
%% premises of the rule.
%%
%% The tree is constructed in a breadth-first manner that works on
%% a list of open nodes. At each step each open node in this list is
%% examined to see if
%% 1. it can be inferred from the hyps (Rule = assump, Children = [])
%% 2. it can be inferred from other nodes in the tree (Rule = assump(Node),
%%    Children = [])
%% 3. a rule from the table can be applied.
%% A new list of open nodes is produced  for the next breadth-first step.
%%
%% Circular reasoning is avoided because such attempts are caught by the
%% occurs check in the unification Rule = assump(Node)
%%
%% A list of tactics is constructed from the leaves towards the root with the
%% exception being that if a node whose rule is assump(Node) then this
%% subtree is used to construct a tactic list before the tactic for the node.
%% This ensures all information is available when the tactic list is called.
%% The Done argument is used to avoid reprocessing subtrees.
%%
build_horn_proof(Table, Goal, Context, Tac, TacticList) :-
    horn_extract_hyps(Context, Hyps),
    RootNode = node(Goal, _Rule, _Children, _Done),
    build_horn_proof_aux([RootNode], [], Hyps, Table),
    build_horn_tactic([RootNode], Tac, TacticList, []),
    !.

build_horn_proof_aux([], _Nodes, _Hyps, _Table) :-
    !.
build_horn_proof_aux(OpenNodes, Nodes, Hyps, Table) :-
    horn_breadth_step(OpenNodes, Nodes, [], Hyps, 
                      NewNodes, NewOpenNodes, Table),
    build_horn_proof_aux(NewOpenNodes, NewNodes, Hyps, Table).

build_horn_tactic([], _Tac, T, T).
build_horn_tactic([node(Term, Rule, Children, Done)|Rest], Tac, 
                  TacticList, Link) :-
    ( atom(Done) ->
        build_horn_tactic(Rest, Tac, TacticList, Link)
      ;
      Rule = assump(Node) ->
        Done = done,
        build_horn_tactic([Node], Tac, TacticList, Link1),
        build_horn_tactic(Rest, Tac, Link1, Link)
      ;
      Rule = assump ->
        Done = done,
        build_horn_tactic(Rest, Tac, TacticList, Link)
      ;
        Done = done,
        build_horn_tactic(Children, Tac, TacticList, [Tac(Rule, Term)|Link1]),
        build_horn_tactic(Rest, Tac, Link1, Link)
    ).
        
    
horn_breadth_step([], Nodes, Open, _, Nodes, Open, _Table).
horn_breadth_step([Head|Open], Nodes, O, Hyps, NewNodes, NewOpen, Table):-
    Head = node(Term, Rule, Children, _Done),
    (     member(Term, Hyps),
          Rule = assump,
          Children = [],
          horn_breadth_step(Open, Nodes, O, Hyps, NewNodes, NewOpen, Table)
       ;
          member(N, Nodes), N = node(Term, _, _, _), Rule = assump(N),
          horn_breadth_step(Open, Nodes, O, Hyps, NewNodes, NewOpen, Table)
       ;
          horn_clause(Table, Term, Children, Rule, Constraints),
          append(O, Children, NO),
          call_predicate(Constraints),
          horn_breadth_step(Open, [Head|Nodes], NO, 
                            Hyps, NewNodes, NewOpen, Table)
    ).
	
