%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%%
%% CVS $Id: rel.thy,v 1.7 2003/05/28 03:14:50 itp Exp $

%%%%%%%%%%%%%%%%%%%
%% ZFC Relations %%
%%%%%%%%%%%%%%%%%%%

%% A theory of relations based on ZFC.




add_theory(rel).
include_theory(stdops, share).
include_theory(zfc, all).

%% Add some extra object variables.
declare_object_vars(w).
declare_object_vars(y).
declare_object_vars(z).                                                         


close_theory.

define rel(D,R) === power(D cross R).

%% Relation operations.

%% NOTE: Because this is an untyped set theory, sets can have mixtures
%% of pairs and non-pairs. The operations defined below apply to any
%% set but only do anything interesting to the subset of pairs -
%% i.e. the relational part of the set.
%% It is therefore conventient to be able to extract this subset.

define pairs_of(S) ===
  set_of z z in S and ex x:sets ex y:sets z = ordered_pair(x,y).

define domain(R) ===
  set_of x ex y:sets ((x |--> y) in pairs_of(R)).

define range(R) ===
  set_of x ex y:sets ((y |--> x) in pairs_of(R)).

define image(R,X) ===
  set_of y ex x:sets (x in X and ((x |--> y) in R)).

define ident(T) ===
  set_of x ex y:sets (y in T and x = (y |--> y)).

define inverse(R) ===
  set_of x ex y_1:sets ex y_2:sets
     ((y_1 |--> y_2) in R and x = (y_2 |--> y_1)).

define bkwd_compose(S,R) ===  
  set_of x ex y_1:sets ex y_2:sets ex y_3:sets (
     (y_1 |--> y_2) in R and (y_2 |--> y_3) in S and
      x = (y_1 |--> y_3)).


add_operator(525, nonassoc, bkwd_compose).

abbreviation fwd_compose(R,S) ===
  S bkwd_compose R.

add_operator(525, nonassoc, fwd_compose).

define range_restrict(R,A) ===
  (R fwd_compose ident(A)).

add_operator(525, nonassoc, range_restrict).

define domain_restrict(A,R) ===
  (ident(A) fwd_compose R).


add_operator(525, nonassoc, domain_restrict).

define range_subtract(R,A) ===
  (R range_restrict (range(R) diff A)).

add_operator(525, nonassoc, range_subtract).

define domain_subtract(A,R) ===
  ((domain(R) diff A) domain_restrict R).

add_operator(525, nonassoc, domain_subtract).

define rel_override(R,S) ===
  (((domain(R) diff domain(S)) domain_restrict R) union pairs_of(S)).

add_operator(525, nonassoc, rel_override).


%%%%%%%%%%%%%%
%% Theorems %%
%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Set existence, types and properties %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

theorem relation_type ===
   X:sets, Y:sets
 --------------------------
  rel(X, Y):sets.

theorem relation_property ===
   A:sets, X:sets, Y:sets
 ---------------------------
   A in rel(X, Y) <=> A subset_of (X cross Y).


theorem get_domain_is_function ===
  all x:sets ex_one y:sets (ex z1:sets ex z2:sets x = ordered_pair(z1,z2) and y
= z1) or not(ex z1:sets ex z2:sets x = ordered_pair(z1,z2)) and y = empty
   provided x not_free_in y,x not_free_in z1, x not_free_in z2,
            z1 not_free_in z2,  y not_free_in z1, y not_free_in z2.

theorem get_range_is_function ===
  all x:sets ex_one y:sets (ex z1:sets ex z2:sets x = ordered_pair(z1,z2) and y
= z2) or not(ex z1:sets ex z2:sets x = ordered_pair(z1,z2)) and y = empty
   provided x not_free_in y,x not_free_in z1, x not_free_in z2,
            z1 not_free_in z2,  y not_free_in z1, y not_free_in z2.





theorem ex_pairs_of ===
  A:sets =>
  ex_set z z in A and ex x:sets ex y:sets z = ordered_pair(x,y)
       provided x not_free_in y, x not_free_in A, y not_free_in A,
                x not_free_in z, y not_free_in z, z not_free_in A.

theorem pairs_of_type ===
   S:sets
 --------------------
  pairs_of(S):sets.

theorem pairs_of_property ===
   S:sets, A:sets
  ----------------------
   A in pairs_of(S) <=> A in S and ex x:sets ex y:sets A = ordered_pair(x,y)
       provided x not_free_in y, x not_free_in A, y not_free_in A.

theorem ex_domain ===
  R:sets => ex_set  x ex y:sets ((x |--> y) in pairs_of(R))
         provided x not_free_in y, x not_free_in R, y not_free_in R.

theorem domain_type ===
    R:sets
  -----------------
   domain(R):sets.

theorem domain_property ===
   Z:sets, R:sets
  ---------------------------------------
   Z in domain(R) <=> ex y:sets (Z |--> y) in R
        provided y not_free_in Z, y not_free_in R.

theorem ex_range ===
  R:sets  => ex_set y ex x:sets ((x |--> y) in pairs_of(R))
         provided x not_free_in y, x not_free_in R, y not_free_in R.

theorem range_type ===
         R:sets
  ----------------------
       range(R):sets.

theorem range_property ===
             Z:sets, R:sets
  ---------------------------------------
   Z in range(R) <=> ex x:sets (x |--> Z) in R
        provided x not_free_in Z, x not_free_in R.

theorem ex_image ===
  S:sets and R:sets =>
  ex_set y ex x:sets (x in S and ((x |--> y) in R))
         provided x not_free_in y, x not_free_in R, y not_free_in R,
                  x not_free_in S, y not_free_in S.

theorem image_type ===
  S:sets, R:sets
  ------------------
  image(R, S):sets.

theorem image_property ===
   Z:sets, S:sets, R:sets
   -----------------------------------
   Z in image(R, S) <=> ex x:sets x in S and (x |--> Z) in R
         provided x not_free_in Z, x not_free_in R, x not_free_in S.
 
theorem ex_ident  ===
  T:sets => ex_set x ex y:sets (y in T and x = (y |--> y))
        provided x not_free_in y, x not_free_in T, y not_free_in T.

theorem ident_type ===
      T:sets
  ----------------
   ident(T):sets.  

theorem ident_property ===
    X:sets, T:sets
 -----------------------------
  X in ident(T) <=> ex x:sets x in T and X = (x |--> x)
    provided x not_free_in X, x not_free_in T.


theorem ex_inverse ===
  R:sets => ex_set x ex y1:sets ex y2:sets
     ((y1 |--> y2) in R and x = (y2 |--> y1))
      provided x not_free_in R, y1 not_free_in R, y2 not_free_in R,
               x not_free_in y1, x not_free_in y2, y1 not_free_in y2.

theorem inverse_type ===
    R:sets
  ----------------
  inverse(R):sets.

theorem inverse_property ===
   Z:sets, R:sets
  --------------------
   Z in inverse(R) 
   <=> 
   ex y1:sets ex y2:sets ((y1 |--> y2) in R and Z = (y2 |--> y1))
      provided y1 not_free_in R, y2 not_free_in R,
               y1 not_free_in Z, y2 not_free_in Z,
               y1 not_free_in y2.


theorem ex_bkwd_compose ===
  S:sets and R:sets 
  =>  ex_set x ex y1:sets ex y2:sets ex y3:sets (
              (y1 |--> y2) in R and (y2 |--> y3) in S and
              x = (y1 |--> y3))
         provided x not_free_in R, x not_free_in S, x not_free_in y1,
                  x not_free_in y2, x not_free_in y3, y1 not_free_in R,
                  y1 not_free_in S, y1 not_free_in y2, y1 not_free_in y3,
                  y2 not_free_in R, y2 not_free_in S, y2 not_free_in y3,
                  y3 not_free_in R, y3 not_free_in S.

theorem bkwd_compose_type ===
  S:sets, R:sets
  -------------------------
  bkwd_compose(S,R):sets.

theorem bkwd_compose_property ===
  S:sets, R:sets, Z:sets
  ----------------------
  Z in bkwd_compose(S,R) 
  <=> 
  ex y1:sets ex y2:sets ex y3:sets (
     (y1 |--> y2) in R and (y2 |--> y3) in S and Z = (y1 |--> y3))
         provided y1 not_free_in R, y1 not_free_in S, y1 not_free_in Z,
                  y1 not_free_in y2, y1 not_free_in y3,
                  y2 not_free_in R, y2 not_free_in S, y2 not_free_in Z,
                  y2 not_free_in y3, 
                  y3 not_free_in R, y3 not_free_in S, y3 not_free_in Z.


theorem range_restrict_type ===
   A:sets, R:sets
 ------------------------
   range_restrict(R, A):sets.

theorem range_restrict_property ===
 X:sets, A:sets, R:sets
 ------------------------------
  X in range_restrict(R,A) 
  <=> ex x:sets ex y:sets y in A and X = ordered_pair(x,y) and X in R
     provided x not_free_in y, x not_free_in X, x not_free_in R, 
              y not_free_in X, y not_free_in R, x not_free_in A,
              y not_free_in A.

theorem domain_restrict_type ===
   A:sets, R:sets
 ------------------------
   domain_restrict(A, R):sets.

theorem domain_restrict_property ===
 X:sets, A:sets, R:sets
 ------------------------------
  X in domain_restrict(A,R) 
  <=> ex x:sets ex y:sets x in A and X = ordered_pair(x,y) and X in R
     provided x not_free_in y, x not_free_in X, x not_free_in R, 
              y not_free_in X, y not_free_in R, x not_free_in A,
              y not_free_in A.

theorem range_subtract_type ===
   A:sets, R:sets
 ------------------------
   range_subtract(R, A):sets.

theorem range_subtract_property ===
 X:sets, A:sets, R:sets
 ------------------------------
  X in range_subtract(R,A) 
  <=> ex x:sets ex y:sets y not_in A and X = ordered_pair(x,y) and X in R
     provided x not_free_in y, x not_free_in X, x not_free_in R, 
              y not_free_in X, y not_free_in R, x not_free_in A,
              y not_free_in A.

theorem domain_subtract_type ===
   A:sets, R:sets
 ------------------------
   domain_subtract(A, R):sets.

theorem domain_subtract_property ===
 X:sets, A:sets, R:sets
 ------------------------------
  X in domain_subtract(A,R) 
  <=> ex x:sets ex y:sets x not_in A and X = ordered_pair(x,y) and X in R
     provided x not_free_in y, x not_free_in X, x not_free_in R, 
              y not_free_in X, y not_free_in R, x not_free_in A,
              y not_free_in A.

theorem rel_override_type ===
   A:sets, R:sets
 ------------------------
   rel_override(A, R):sets.

theorem rel_override_property ===
 X:sets, R:sets, S:sets
 ------------------------------
  X in rel_override(R,S) 
  <=> ex x:sets ex y:sets X = ordered_pair(x,y) and (x not_in domain(S) and X in R or X in S)
     provided x not_free_in y, x not_free_in X, x not_free_in R, 
              y not_free_in X, y not_free_in R, x not_free_in S,
              y not_free_in S.



theorem domain_empty_is_empty ===
  domain(empty) = empty.

theorem range_empty_is_empty ===
  range(empty) = empty.

theorem ident_empty_is_empty ===
  ident(empty) = empty.

theorem inverse_empty_is_empty ===
  inverse(empty) = empty.

theorem empty_bkwd_compose_is_empty ===
  Z:sets
  -----------------------------------
  empty bkwd_compose Z = empty.

theorem bkwd_compose_empty_is_empty ===
  Z:sets
  -----------------------------
  Z bkwd_compose empty = empty.

theorem empty_range_restrict_is_empty ===
  Z:sets
  ------------------------------
  empty range_restrict Z = empty.

theorem range_restrict_empty_is_empty ===
  Z:sets
  ------------------------------
  Z range_restrict empty = empty.

theorem empty_domain_restrict_is_empty ===
  Z:sets
  --------------------------------
  empty domain_restrict Z = empty.

theorem domain_restrict_empty_is_empty ===
  Z:sets
  -------------------------------
  Z domain_restrict empty = empty.

theorem empty_range_subtract_is_empty ===
  Z:sets
  -------------------------------
  empty range_subtract Z = empty.

theorem domain_subtract_empty_is_empty ===
  Z:sets
  --------------------------------
  Z domain_subtract empty = empty.

theorem rel_override_empty_is_empty ===
  empty rel_override empty = empty.



%% Other theorems.

theorem domain_cross ===
  X:sets, Y:sets, not_empty(Y)
 -----------------------------
  domain(X cross Y) = X.

theorem in_domain ===
  X:sets, Y:sets, Z:sets
  -----------------------------------------------
  (ordered_pair(X,Y) in Z => X in domain(Z)).

theorem range_cross ===
  X:sets,Y:sets, not_empty(X)
 ----------------------------
  range(X cross Y) = Y.

theorem in_range ===
  X:sets, Y:sets, Z:sets
  -----------------------------------------------
  (ordered_pair(X,Y) in Z => Y in range(Z)).

theorem in_image ===
  X:sets, Y:sets, C:sets, Z:sets
  -----------------------------------------------
  (((ordered_pair(X,Y) in Z) and X in C) => Y in image(Z, C)).

theorem image_empty_1 ===
  X:sets
 ---------------------------
 image(empty,X) = empty.

theorem image_empty_2 ===
  Z:sets
  -----------------------------------------------
  image(Z,empty) = empty.

theorem in_domain_not_empty ===
    X:sets, Z:sets
  -----------------------------------------------
    X in domain(Z) => not_empty(Z).

theorem in_range_not_empty ===
    X:sets, Z:sets
  -----------------------------------------------
    X in range(Z) => not_empty(Z).

theorem domain_union ===
  R:sets, S:sets
  --------------------------------------------
  domain(R union S) = domain(R) union domain(S).

theorem range_union ===
  R:sets, S:sets
  ------------------------------------------
  range(R union S) = range(R) union range(S).

theorem domain_inter ===
  R:sets, S:sets
  ------------------------------------------
  domain(R inter S) subset_of domain(R) inter domain(S).

theorem range_inter ===
  R:sets, S:sets
  ------------------------------------------
  range(R inter S) subset_of range(R) inter range(S).

theorem bkwd_compose_associative ===
  R:sets, S:sets, T:sets
  ------------------------------------------
  R bkwd_compose (S bkwd_compose T) = (R bkwd_compose S) bkwd_compose T.

                       

%%%%%%%%%%%%%%%%%%%
%% Opening Rules %%
%%%%%%%%%%%%%%%%%%%

theorem open_pairs_of_1_eq ===
    X = Y
   --------------
   ID:::pairs_of(X) = pairs_of(Y)
     provided context_search(hyp, ID, _, X:sets).

theorem open_domain_1_eq ===
    X = Y
   --------------
   ID:::domain(X) = domain(Y)
     provided context_search(hyp, ID, _, X:sets).

theorem open_range_1_eq ===
    X = Y
   --------------
   ID:::range(X) = range(Y)
     provided context_search(hyp, ID, _, X:sets).

theorem open_image_1_eq ===
    X = Y
   --------------
   ID:::image(X,Z) = image(Y,Z)
     provided context_search(hyp, ID, _, X:sets),
              context_search(hyp, ID, _, Z:sets).

theorem open_image_2_eq ===
    X = Y
   --------------
   ID:::image(Z,X) = image(Z,Y)
     provided context_search(hyp, ID, _, X:sets),
              context_search(hyp, ID, _, Z:sets).

theorem open_rel_1_eq ===
    X = Y
   --------------
   ID:::rel(X, Z) = rel(Y, Z)
     provided context_search(hyp, ID, _, X:sets),
              context_search(hyp, ID, _, Z:sets).

theorem open_rel_2_eq ===
    X = Y
   --------------
   ID:::rel(Z, X) = rel(Z, Y)
     provided context_search(hyp, ID, _, X:sets),
              context_search(hyp, ID, _, Z:sets).

theorem open_ident_1_eq ===
    X = Y
   --------------
   ID:::ident(X) = ident(Y)
     provided context_search(hyp, ID, _, X:sets).

theorem open_inverse_1_eq ===
    X = Y
   --------------
   ID:::inverse(X) = inverse(Y)
     provided context_search(hyp, ID, _, X:sets).


theorem open_bkwd_compose_1_eq ===
    X = Y
   --------------
  ID::: bkwd_compose(X,Z) = bkwd_compose(Y,Z)
     provided context_search(hyp, ID, _, X:sets),
              context_search(hyp, ID, _, Z:sets).

theorem open_bkwd_compose_2_eq ===
    X = Y
   --------------
  ID::: bkwd_compose(Z,X) = bkwd_compose(Z,Y)
     provided context_search(hyp, ID, _, X:sets),
              context_search(hyp, ID, _, Z:sets).

theorem open_range_restrict_1_eq ===
    X = Y
   --------------
  ID::: range_restrict(X,Z) = range_restrict(Y,Z)
     provided context_search(hyp, ID, _, X:sets),
              context_search(hyp, ID, _, Z:sets).

theorem open_range_restrict_2_eq ===
    X = Y
   --------------
  ID::: range_restrict(Z,X) = range_restrict(Z,Y)
     provided context_search(hyp, ID, _, X:sets),
              context_search(hyp, ID, _, Z:sets).

theorem open_domain_restrict_1_eq ===
    X = Y
   --------------
  ID::: domain_restrict(X,Z) = domain_restrict(Y,Z)
     provided context_search(hyp, ID, _, X:sets),
              context_search(hyp, ID, _, Z:sets).

theorem open_domain_restrict_2_eq ===
    X = Y
   --------------
   ID:::domain_restrict(Z,X) = domain_restrict(Z,Y)
     provided context_search(hyp, ID, _, X:sets),
              context_search(hyp, ID, _, Z:sets).

theorem open_range_subtract_1_eq ===
    X = Y
   --------------
  ID::: range_subtract(X,Z) = range_subtract(Y,Z)
     provided context_search(hyp, ID, _, X:sets),
              context_search(hyp, ID, _, Z:sets).

theorem open_range_subtract_2_eq ===
    X = Y
   --------------
  ID::: range_subtract(Z,X) = range_subtract(Z,Y)
     provided context_search(hyp, ID, _, X:sets),
              context_search(hyp, ID, _, Z:sets).

theorem open_domain_subtract_1_eq ===
    X = Y
   --------------
  ID::: domain_subtract(X,Z) = domain_subtract(Y,Z)
     provided context_search(hyp, ID, _, X:sets),
              context_search(hyp, ID, _, Z:sets).

theorem open_domain_subtract_2_eq ===
    X = Y
   --------------
  ID::: domain_subtract(Z,X) = domain_subtract(Z,Y)
     provided context_search(hyp, ID, _, X:sets),
              context_search(hyp, ID, _, Z:sets).

theorem open_rel_override_1_eq ===
    X = Y
   --------------
  ID::: rel_override(X,Z) = rel_override(Y,Z)
     provided context_search(hyp, ID, _, X:sets),
              context_search(hyp, ID, _, Z:sets).

theorem open_rel_override_2_eq ===
    X = Y
   --------------
   ID:::rel_override(Z,X) = rel_override(Z,Y)
     provided context_search(hyp, ID, _, X:sets),
              context_search(hyp, ID, _, Z:sets).

theorem empty_rel_property ===
    X:sets, Y:sets
 -----------------------
 Y in rel(empty, X) <=> Y = empty.

theorem rel_empty_property ===
   X:sets, Y:sets
 -----------------------
  Y in rel(X, empty) <=> Y = empty.

theorem in_domain_in_domain_set ===
   A:sets, B:sets, R:sets, X:sets, R in rel(A,B), X in domain(R)
   -------------------------------------
    X in A.

theorem in_range_in_range_set ===
   A:sets, B:sets, R:sets, X:sets, R in rel(A,B), X in range(R)
   -------------------------------------
    X in B.

theorem in_rel_ex_pair ===
    F:sets, A:sets, B:sets, X:sets, F in rel(A,B), X in F
    ---------------------------------------------
      (ex x:sets ex y:sets X = (x |--> y))
           provided x not_free_in X, x not_free_in y, y not_free_in X.

theorem pairs_of_rel_eq ===
     A:sets, B:sets, R:sets, R in rel(A,B)
    --------------------------------------
        pairs_of(R) = R.

theorem domain_restrict_domain_eq ===
     A:sets, B:sets, R:sets, R in rel(A,B)
    --------------------------------------
       domain(R) domain_restrict R = R.
 
theorem range_restrict_range_eq ===
     A:sets, B:sets, R:sets, R in rel(A,B)
    --------------------------------------
       R  range_restrict range(R) = R.

theorem relation_union ===
     A:sets, B:sets, C:sets, D:sets, F:sets, G:sets,
     F in rel(A,B), G in rel(C,D)
    -------------------------------------
     (F union G) in rel(A union C, B union D). 
 

theorem domain_sets_disjoint_domains_disjoint ===
     A:sets, B:sets, C:sets, D:sets, F:sets, G:sets,
     F in rel(A,B), G in rel(C,D), A inter C = empty
    -------------------------------------
      domain(F) inter domain(G) = empty.

     
%%
%% Support for Window Inference - loads
%% rel specific information.
%%

require_global_tactic_file(wi_support).
require_tactic_file(wi_support,rel).

