%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%% 
%% CVS $Id: sm.thy,v 1.8 2002/07/28 05:03:14 itp Exp $

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Hilbert's Epsilon Operator %%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% This is a theory of Hilbert's Epsilon Operator.  It is a little stricter
%% Hilbert's operator in that `sm x B' where `B' is always false returns `?'
%% rather than an arbitrary element.


add_theory(sm).
include_theory(stdops, share).
include_theory(jpred_eq_t, all).
include_theory(apart, all).
include_theory(undef, all).


%% Add some extra object variables.
declare_object_vars(w).
declare_object_vars(y).
declare_object_vars(z).

%%%%%%%%%%%%%%%%%
%% Quantifiers %%
%%%%%%%%%%%%%%%%%

declare !!sm [x] Body.

add_operator(805, quant, sm).

define (!!sm [x:Type] Body) ===
  sm x (x:Type and Body).

define !!sm_one [x] Body ===
  (sm x Body) provided_that
  ex_one x Body.

add_operator(805, quant, sm_one).

define !!sm_one [x:Type] Body ===
  (sm x:Type Body) provided_that
  ex_one x:Type Body.

%%%%%%%%%%%%
%% Axioms %%
%%%%%%%%%%%%

axiom sm_true_def ===
  (ex x B) => ex x (B and x = sm x B).

axiom sm_false_def ===  %% This axiom makes sm stronger than normal.
  not (ex x B) => (sm x B) = (?).

axiom open_sm_1_eq(x:obvar) ===
  hyp ??? nfi(x) ---> B <=> A
  ---------------------------
  (sm x B) = (sm x A).  

close_theory.

%%%%%%%%%%%%%%%%%%%
%% Opening Rules %%
%%%%%%%%%%%%%%%%%%%

theorem open_sm_t_1_eq ===
  hyp ??? nfi(x) ---> T1 = T2
  --------------
  (sm x:T1 A) = (sm x:T2 A).

theorem open_smt_1_eq(x:obvar) ===
  hyp ??? nfi(x) +++ [x:T] ---> B <=> A
  ---------------------------
  (sm x:T B) = (sm x:T A)
     provided x not_free_in T.  

theorem open_sm_one_1_eq(x:obvar) ===
  hyp ??? nfi(x) ---> B <=> A
  ---------------------------
  (sm_one x B) = (sm_one x A).

theorem open_sm_one_t_1_eq ===
  hyp ??? nfi(x) ---> T1 = T2
  --------------
  (sm_one x:T1 A) = (sm_one x:T2 A).

theorem open_sm_onet_1_eq(x:obvar) ===
  hyp ??? nfi(x) +++ [x:T] ---> B <=> A
  ---------------------------
  (sm_one x:T B) = (sm_one x:T A)
     provided x not_free_in T.


%%%%%%%%%%%%%%
%% Theorems %%
%%%%%%%%%%%%%%

%% Simplification rules
theorem sm_false ===
  (sm x false) = (?).

theorem sm_false1 ===
  (sm x:T false) = (?).

theorem sm_one_false ===
  (sm_one x false) = (?).

theorem sm_one_false1 ===
  (sm_one x:T false) = (?).

%% Other rules
theorem equal_sm1 ===
  X = sm x (x = X) provided
  x not_free_in X.

theorem equal_sm2 ===
  X:T => X = sm x:T (x = X) provided
  x not_free_in X.

theorem equal_sm_one1 ===
  X = sm_one x (x = X) provided
  x not_free_in X.

theorem equal_sm_one2 ===
  X:T => X = sm_one x:T (x = X) provided
  x not_free_in X.


theorem ex_iff_ex_sm ===
  (ex x:T B) <=> ex x:T (B and x = sm x:T B) provided
  x not_free_in T.

theorem ex_one_iff_ex_one_sm_one ===
  (ex_one x:T B) <=> ex_one x:T (B and x = sm_one x:T B) provided
  x not_free_in T.

theorem ex_imp_sm_not_undef ===
    (ex x:T B) => not ((sm x:T B) = ? ).
 
theorem not_undef_imp_sm_type ===
     not ((sm x:T B) = ?) => (sm x:T B):T.

%%
%% Support for Window Inference - loads
%% sm specific information.
%%

require_tactic_file(wi_support,sm).



