%% ##Copyright##
%%  
%% Copyright (C) 1999 - 2002
%% Software Verification Research Centre
%% The University of Queensland
%% Australia 4072
%%
%% email: svrc@svrc.uq.edu.au
%%
%%
%% The Ergo System and Documentation
%%
%% COPYRIGHT NOTICE, LICENCE AND DISCLAIMER.
%%
%% Copyright 1999 - 2002 by 
%% The University of Queensland, Queensland 4072 Australia
%%
%% Permission to use, copy and distribute this software and associated
%% documentation for any non-commercial purpose and without fee is hereby
%% granted, subject to the following conditions:
%%
%% 1.     that the above copyright notice and this permission notice and
%%        warranty disclaimer appear in all copies and in supporting
%%        documentation;
%%
%% 2.     that the name of The University of Queensland not be used in
%%        advertising or publicity pertaining to distribution of the 
%%        software without specific, written prior permission;
%%
%% 3.     that users of this software should be responsible for determining
%%        the fitness of the software for the purposes for which the 
%%        software is employed by them;
%%
%% 4.     that no changes to the system or documentation are subsequently
%%        made available to third parties or redistributed without prior
%%        written consent from the SVRC; and
%%
%% 5.     that individuals, groups or institutions using this software
%%        register their use by completing the registration form and 
%%        sending the completed registration form to the SVRC.
%%
%% The University of Queensland disclaims all warranties with regard to this
%% software, including all implied warranties of merchantability and fitness
%% to the extent permitted by law. In no event shall The University of
%% Queensland be liable for any special, indirect or consequential damages 
%% or any damages whatsoever resulting from loss of use, data or profits, 
%% whether in an action of contract, negligence or other tortious action, 
%% arising out of or in connection with the use or performance of this 
%% software.
%%
%% THE UNIVERSITY OF QUEENSLAND MAKES NO REPRESENTATIONS ABOUT THE ACCURACY OR
%% SUITABILITY OF THIS MATERIAL FOR ANY PURPOSE.  IT IS PROVIDED "AS IS",
%% WITHOUT ANY EXPRESSED OR IMPLIED WARRANTIES.
%%
%%
%% For information on commercial use of this software contact the SVRC.
%%
%% ##Copyright##

%%
%% CVS $Id: zsets.thy,v 1.3 2003/04/09 04:58:12 itp Exp $

%%%%%%%%%%%%%
%% Z sets  %%
%%%%%%%%%%%%%

% A theory about the sets used within Z (typically as types).
%
% This theory has constructors that are like similar constructors
% in ZFC but are undefined if supplied with ill-typed arguments. 
[zfc].
delete_theory_and_dependents(zsets).
add_theory(zsets).

include_theory(stdops, share).
include_theory(iarith,all).
include_theory(undef,all).
include_theory(zfc,all).

%% Add some extra object variables.
declare_object_vars(w).
declare_object_vars(y).
declare_object_vars(z).

declare zsets.           %% The type Z sets

% All Z sets are sets
axiom zsets_are_sets ===
    X:zsets => X:sets.

% 'zin' is the same as 'in' but requires the arguments to be of
% appropriate types.

define zin(A,B) === (A in B) and A:sets and B:zsets. 

add_operator(700, nonassoc, zin).
       
% If a zset is used as a type then being of that type
% is the same as being in the set.
axiom zset_as_type ===
  T:zsets
 -----------------
  A:T <=> A zin T.

% Numbers are zsets
axiom nats_is_zset ===
    nats:zsets.

axiom ints_is_zset ===
    ints:zsets.

axiom power_zset_is_zset ===
    X:zsets
 --------------
  power(X):zsets.

axiom cross_zsets_is_zset ===
    X:zsets, Y:zsets
  ---------------------
   cross(X,Y):zsets.

% A subset of a zset is a zset
axiom subset_of_zset_is_zset ===
    A:sets, A subset_of B, B:zsets
  --------------------------------
      A:zsets.

close_theory.

% zpower is the same as power for zsets but undefined
% otherwise
define zpower(A) === power(A)
          provided_that A:zsets.

define zsubset_of(A,B) === subset_of(A,B) and A:sets and B:zsets. 

add_operator(525, nonassoc, zsubset_of).

define zp_subset_of(A,B) === zsubset_of(A,B) and not (A = B).

add_operator(525, nonassoc, zp_subset_of).

define zgen_union(S) === gen_union(S)
      provided_that ex x:zsets S zsubset_of zpower(x).

define zgen_inter(S) === gen_inter(S)
      provided_that not_empty(S) and ex x:zsets S zsubset_of zpower(x).

define zsingle(X) === single(X)
      provided_that ex x:zsets X zin x.

define zpair(A,B) === pair(A,B)
      provided_that ex x:zsets A zin x and B zin x.

define zordered_pair(X,Y) === zpair(zsingle(X), zpair(X,Y)).

define zunion(A,B) === union(A,B)
      provided_that ex x:zsets A zin zpower(x) and B zin zpower(x).

add_operator(525, leftassoc, zunion).

define zinter(A,B) === inter(A,B)
      provided_that ex x:zsets A zin zpower(x) and B zin zpower(x).

add_operator(520, leftassoc, zinter).

define zdiff(A,B) === diff(A,B)
      provided_that ex x:zsets A zin zpower(x) and B zin zpower(x).

add_operator(525,leftassoc, zdiff).

define zsym_diff(A,B) === sym_diff(A,B)
      provided_that ex x:zsets A zin zpower(x) and B zin zpower(x).

add_operator(525,leftassoc, zsym_diff).

define zpower_1(X) === zpower(X) zdiff zsingle(empty).

define zcross(X,Y) === cross(X,Y)
      provided_that X:zsets and Y:zsets.

add_operator(530, leftassoc, zcross).

%% Opening rules - note that these rules do not require
%% type constraints like the ZFC equivalents because if
%% the type constraints are not satisfied then both
%% sides are undefined and the rule is still valid.

theorem open_zin_1_iff ===
      X = Y
  ---------------
   X zin A <=> Y zin A.

theorem open_zin_2_iff ===
      X = Y
  -----------------
   A zin X <=> A zin Y.

theorem open_zsubset_of_1_iff ===
      X = Y
  ---------------
   X zsubset_of A <=> Y zsubset_of A.

theorem open_zsubset_of_2_iff ===
      X = Y
  -----------------
   A zsubset_of X <=> A zsubset_of Y.

theorem open_zp_subset_of_1_iff ===
      X = Y
  ---------------
   X zp_subset_of A <=> Y zp_subset_of A.

theorem open_zp_subset_of_2_iff ===
      X = Y
  -----------------
   A zp_subset_of X <=> A zp_subset_of Y.


theorem open_zpower_1_eq ===
      X = Y
 ------------------
  zpower(X) = zpower(Y).

theorem open_zpower_1_1_eq ===
      X = Y
 ------------------
  zpower_1(X) = zpower_1(Y).

theorem open_zgen_union_1_eq ===
      X = Y
 ------------------
  zgen_union(X) = zgen_union(Y).

theorem open_zgen_inter_1_eq ===
      X = Y
 ------------------
  zgen_inter(X) = zgen_inter(Y).

theorem open_zsingle_1_eq ===
      X = Y
 ------------------
  zsingle(X) = zsingle(Y).

theorem open_zunion_1_eq ===
      X = Y
 ------------------
  zunion(X,A) = zunion(Y,A).

theorem open_zunion_2_eq ===
      X = Y
 ------------------
  zunion(A,X) = zunion(A,Y).

theorem open_zinter_1_eq ===
      X = Y
 ------------------
  zinter(X,A) = zinter(Y,A).

theorem open_zinter_2_eq ===
      X = Y
 ------------------
  zinter(A,X) = zinter(A,Y).


theorem open_zpair_1_eq ===
      X = Y
 ------------------
  zpair(X,A) = zpair(Y,A).

theorem open_zpair_2_eq ===
      X = Y
 ------------------
  zpair(A,X) = zpair(A,Y).

theorem open_zordered_pair_1_eq ===
      X = Y
 ------------------
  zordered_pair(X,A) = zordered_pair(Y,A).

theorem open_zordered_pair_2_eq ===
      X = Y
 ------------------
  zordered_pair(A,X) = zordered_pair(A,Y).

theorem open_zdiff_1_eq ===
      X = Y
 ------------------
  zdiff(X,A) = zdiff(Y,A).

theorem open_zdiff_2_eq ===
      X = Y
 ------------------
  zdiff(A,X) = zdiff(A,Y).

theorem open_zsym_diff_1_eq ===
      X = Y
 ------------------
  zsym_diff(X,A) = zsym_diff(Y,A).

theorem open_zsym_diff_2_eq ===
      X = Y
 ------------------
  zsym_diff(A,X) = zsym_diff(A,Y).

theorem open_zcross_1_eq ===
      X = Y
 ------------------
  zcross(X,A) = zcross(Y,A).

theorem open_zcross_2_eq ===
      X = Y
 ------------------
  zcross(A,X) = zcross(A,Y).


%% Type and undef theorems

% A convenient rule for type inference.
theorem zset_type ===
    X:zsets
  -----------
    X:sets.

theorem zset_type_property ===
    X:zsets
 ---------------------
  A:X <=> A:sets and A in X.

theorem empty_is_zset === 
    empty:zsets.

% For each constructor we give rules about their types,
% about how they are related to the ZFC equivalent,
% about when a constructed term is undefined and
% how undefinedness 'moves through' the constructor.

theorem zpower_zset_is_zset ===
     X:zsets
   --------------
    zpower(X):zsets.

theorem zpower_of_zset_is_power ===
     X:zsets
 -------------------
  zpower(X) = power(X).

theorem undef_zpower ===
    not A:zsets <=> zpower(A) = (?).

theorem zpower_undef_undef ===
    zpower(?) = (?).

theorem zsubset_of_zset_is_subset ===
     Y:sets, X:zsets
 -------------------
  Y zsubset_of X <=> Y subset_of X.

theorem zin_of_zset_is_in ===
     Y:sets, X:zsets
  -------------------
   Y zin X <=> Y in X.

theorem not_undef_zin ===
    not  ? zin X.

theorem not_zin_undef ===
    not X zin (?).

theorem not_undef_zsubset_of ===
    not ? zsubset_of X.

theorem not_zsubset_of_undef ===
    not X zsubset_of (?).

theorem zsingle_zset_is_zset ===
    X:zsets, S:X
  -------------------
   zsingle(S):zsets.

theorem zsingle_zset_is_single ===
    X:zsets, S:X
 ---------------
  zsingle(S) = single(S).

theorem undef_zsingle ===
    not (ex x:zsets A:x) <=> zsingle(A) = (?)
       provided x not_free_in A.

theorem zsingle_undef_undef ===
    zsingle(?) = (?).

theorem zsingle_type ===
     A:zsets, X:A
   -----------------
     zsingle(X):zpower(A).

theorem zpair_zsets_is_zset ===
    X:zsets, A:X, B:X
  ---------------------
    zpair(A, B):zsets.

theorem zpair_zsets_is_pair ===
    X:zsets, A:X, B:X
  ---------------------
    zpair(A,B) = pair(A,B).

theorem undef_zpair ===
   not (ex x:zsets A:x and B:x) <=> zpair(A,B) = (?)
         provided x not_free_in A, x not_free_in B.

theorem undef_zpair_undef ===
    zpair(?, X) = (?).

theorem zpair_undef_undef ===
    zpair(X, ?) = (?).

theorem zpair_type ===
    A:zsets, X:A, Y:A
  ---------------------
   zpair(X,Y):zpower(A).

theorem zordered_pair_zsets_is_zset ===
    X:zsets, A:X, B:X
  ---------------------
    zordered_pair(A, B):zsets.

theorem zordered_pair_zsets_is_pair ===
    X:zsets, A:X, B:X
  ---------------------
    zordered_pair(A,B) = ordered_pair(A,B).

theorem undef_zordered_pair ===
   not (ex x:zsets A:x and B:x) <=> zordered_pair(A,B) = (?)
         provided x not_free_in A, x not_free_in B.

theorem undef_zordered_pair_undef ===
    zordered_pair(?, X) = (?).

theorem zordered_pair_undef_undef ===
    zordered_pair(X, ?) = (?).


theorem zgen_union_is_zset ===
     S:zpower(zpower(X))
  ------------------------------
     zgen_union(S):zsets.

theorem zgen_union_zsets_is_gen_union ===
     S:zpower(zpower(X))
  ------------------------------
   zgen_union(S) = gen_union(S).

theorem undef_zgen_union ===
    not (ex x:zsets S:zpower(zpower(x))) <=> zgen_union(S) = (?)
       provided x not_free_in S.

theorem zgen_union_undef_undef ===
    zgen_union(?) = (?).

theorem zgen_inter_is_zset ===
  not_empty(S), S:zpower(zpower(X))
  -----------------------------
     zgen_inter(S):zsets.

theorem zgen_inter_zsets_is_gen_inter ===
  not_empty(S), S:zpower(zpower(X))
  -----------------------------
    zgen_inter(S) = gen_inter(S).

theorem undef_zgen_inter ===
    (S = empty) or not (ex x:zsets S:zpower(zpower(x))) <=> zgen_inter(S) = (?)
       provided x not_free_in S.

theorem zgen_inter_undef_undef ===
    zgen_inter(?) = (?).

theorem zunion_zsets_is_zset ===
     A:zpower(X), B:zpower(X)
   --------------------------------
      zunion(A,B):zsets.

theorem zunion_zsets_is_union ===
     A:zpower(X), B:zpower(X)
   --------------------------------
     zunion(A,B) = union(A,B).

theorem undef_zunion ===
    not (ex x:zsets A:zpower(x) and B:zpower(x)) <=> zunion(A,B) = (?)
       provided x not_free_in A, x not_free_in B.

theorem undef_zunion_undef ===
     zunion(?, X) = (?).

theorem zunion_undef_undef ===
     zunion(X, ?) = (?).

theorem zinter_zsets_is_zset ===
    A:zpower(X), B:zpower(X)
   --------------------------------
      zinter(A,B):zsets.

theorem zinter_zsets_is_inter ===
     A:zpower(X), B:zpower(X)
   --------------------------------
     zinter(A,B) = inter(A,B).

theorem undef_zinter ===
    not (ex x:zsets A:zpower(x) and B:zpower(x)) <=> zinter(A,B) = (?)
       provided x not_free_in A, x not_free_in B.

theorem undef_zinter_undef ===
     zinter(?, X) = (?).

theorem zinter_undef_undef ===
     zinter(X, ?) = (?).

theorem zdiff_zsets_is_zset ===
    A:zpower(X), B:zpower(X)
   --------------------------------
      zdiff(A,B):zsets.
    
theorem zdiff_zsets_is_diff ===
     A:zpower(X), B:zpower(X)
   --------------------------------
     zdiff(A,B) = diff(A,B).

theorem undef_zdiff ===
    not (ex x:zsets A:zpower(x) and B:zpower(x)) <=> zdiff(A,B) = (?)
       provided x not_free_in A, x not_free_in B.

theorem undef_zdiff_undef ===
     zdiff(?, X) = (?).

theorem zdiff_undef_undef ===
     zdiff(X, ?) = (?).

theorem zsym_diff_zsets_is_zset ===
    A:zpower(X), B:zpower(X)
   --------------------------------
      zsym_diff(A,B):zsets.
    
theorem zsym_diff_zsets_is_sym_diff ===
     A:zpower(X), B:zpower(X)
   --------------------------------
     zsym_diff(A,B) = sym_diff(A,B).

theorem undef_zsym_diff ===
    not (ex x:zsets A:zpower(x) and B:zpower(x)) <=> zsym_diff(A,B) = (?)
       provided x not_free_in A, x not_free_in B.

theorem undef_zsym_diff_undef ===
     zsym_diff(?, X) = (?).

theorem zsym_diff_undef_undef ===
     zsym_diff(X, ?) = (?).

theorem zpower_1_zset_is_zset ===
    X:zsets
 -------------------
  zpower_1(X):zsets.

theorem zpower_1_zset_is_power_1 ===
    X:zsets
 -------------------
 zpower_1(X) = power_1(X).

theorem undef_zpower_1 ===
   not X:zsets <=> zpower_1(X) = (?).

theorem zpower_1_undef_undef ===
   zpower_1(?) = (?).

theorem zcross_zsets_is_zset ===
    X:zsets, Y:zsets
 ----------------------
  zcross(X,Y):zsets.

theorem zcross_zsets_is_cross ===
    X:zsets, Y:zsets
 ----------------------
  zcross(X,Y) = cross(X,Y).

theorem undef_zcross ===
   not (X:zsets and Y:zsets) <=> zcross(X,Y) = (?).

theorem undef_zcross_undef ===
   zcross(?, X) = (?).

theorem zcross_undef_undef ===
   zcross(X, ?) = (?).

%% Three theorems that are useful for other proofs.

theorem empty_type_zpower_zset ===
    A:zsets
 -------------
   empty:zpower(A).

theorem zsingle_empty_type_zpower_zpower_zset ===
    A:zsets
 ---------------
 zsingle(empty) : zpower(zpower(A)).


theorem zset_type_zpower ===
    A:zsets
 ------------------
  A:zpower(A).
   

%%------------------------------------------


theorem zempty_is_empty === all x:sets not(x zin empty).

theorem zempty_is_empty2 === 
    not (X zin empty).

theorem zempty_property ===
  X zin empty <=> false.

theorem zhas_elem_not_empty ===
               X:zsets 
  ----------------------------------
  not X = empty <=> ex x:sets x zin X
        provided x not_free_in X.

theorem zhas_elem_not_empty2 ===
   X zin Y
 ------------------
   not_empty(Y).

theorem zsubset_empty_empty === 
   X zsubset_of empty <=> X = empty.

theorem zgen_union_property ===
 Y zin zgen_union(X) <=> ex z:sets z zin X and Y zin z
        provided z not_free_in X, z not_free_in Y.

theorem zgen_inter_property ===
  Y zin zgen_inter(X) <=> not_empty(X) and all z:sets z zin X => Y zin z
        provided z not_free_in X, z not_free_in Y.
 
theorem zdiff_property ===
   Z zin zdiff(X,Y) <=> Z zin X and not (Z zin Y).

theorem zsym_diff_property ===
   Z zin zsym_diff(X,Y) <=> 
      Z zin X and not (Z zin Y) or Z zin Y and not (Z zin X).

theorem zpower_property ===
  X zin zpower(Y) <=> X zsubset_of Y.

theorem zpower_1_property ===
  Y zin zpower_1(X) <=> Y zsubset_of X and not_empty(Y).

theorem zsingle_property ===
    A:zsets, X:A
  ------------------------
  Y zin zsingle(X) <=> Y = X.

theorem zsingle_empty_property ===
  X zin zsingle(empty) <=> X = empty.

theorem zpower_zsingle_empty_property ===
   X zin zpower(zsingle(empty)) <=> (X = empty or X = zsingle(empty)).

theorem empty_not_eq_zsingle_empty ===
    not (empty = zsingle(empty)).


theorem zset_zsubset_of_zset ===
         X:zsets
    -------------------
     X zsubset_of X.

theorem empty_zsubset_of_zset ===
         X:zsets
    -------------------
     empty zsubset_of X.

theorem zsubset_zsubset_imp_eq ===
    X zsubset_of Y, Y zsubset_of X 
    ------------------------------
              X = Y.

theorem zsingle_not_empty ===
   not_empty(zsingle(X)).

theorem zpower_not_empty ===
    not_empty(zpower(X)).

theorem zgen_union_not_empty ===
  not_empty(X), not (X = zsingle(empty))
  -------------------------------
    not_empty(zgen_union(X)).

theorem not_empty_zunion_not_empty ===
  not_empty(X)
 -----------------------
  not_empty(X zunion Y).

theorem zunion_not_empty_not_empty ===
  not_empty(Y)
 ------------------------
  not_empty(X zunion Y).


theorem zpair_not_empty ===
  not_empty(pair(X,Y)).

theorem zpair_property ===
    A:zsets, X:A, Y:A
  ----------------------------------
  Z zin zpair(X,Y) <=> Z = X or Z = Y.

theorem zunion_property ===
  A:zsets, X:zpower(A), Y:zpower(A)
  --------------------------------
  Z zin X zunion Y <=> (Z zin X or Z zin Y).

theorem zinter_property ===
  A:zsets, X:zpower(A), Y:zpower(A)
  --------------------------------
  Z zin X zinter Y <=> (Z zin X and Z zin Y).


theorem eq_zordered_pairs ===
    A:zsets, X:A, Y:A, Z:A, W:A
  ---------------------------------------------
  zordered_pair(X,Y) = zordered_pair(Z,W) <=> X = Z and Y = W.

theorem zordered_pair_not_empty ===
 not_empty(zordered_pair(X,Y)).

theorem zordered_pair_zin_zpower_zpower ===
  Z:zsets, A:zpower(Z), B:zpower(Z)
  ---------------------------------------
  X zin A and Y zin B => zordered_pair(X,Y) zin zpower(zpower(A zunion B)).

theorem zcross_property ===
 Z zin zcross(X,Y) 
     <=> 
 ex x:sets ex y:sets x zin X and y zin Y and Z = zordered_pair(x,y)
        provided x not_free_in y, x not_free_in X, x not_free_in Y, 
                 x not_free_in Z, y not_free_in X, y not_free_in Y, 
                 y not_free_in Z.

theorem zcross_not_empty ===
      not_empty(X zcross Y).


theorem eq_zsingle ===
   A:zsets, X:A
 -----------------------------
 zsingle(X) = zsingle(Y) <=> X = Y.

theorem eq_pair ===
   Z:zsets, A:Z, B:Z
 ---------------------------------------------------
 zpair(A,B) = zpair(C,D) <=> A = C and B = D or A = D and B = C.

theorem zsingle_eq_pair ===
   Z:zsets, A:Z
 -----------------------------------
 zsingle(A) = zpair(B,C) <=> A = B and A = C.

theorem eq_zcross ===
   A:zsets, B:zsets,
   not_empty(A), not_empty(B), not_empty(C), not_empty(D)
  -------------------------------------------------------
     (A zcross B) = (C zcross D) <=> A = C and B = D.

theorem zcross_empty_eq_empty ===
   A:zsets 
 ----------------------
 A zcross empty = empty.

theorem empty_zcross_eq_empty ===
   A:zsets 
 ----------------------
 empty zcross A = empty.

theorem eq_zpower ===
   A:zsets
 ------------------------------
 zpower(A) = zpower(B) <=> A = B.




theorem empty_zdiff_eq_empty ===
    X:zsets 
 --------------------
 empty zdiff X = empty.

theorem zdiff_empty_eq ===
    X:zsets 
 ----------------
 X zdiff empty = X.

theorem zdiff_eq_empty ===
    X:zsets 
 ----------------
 X zdiff X = empty.

theorem zpower_empty_eq_zsingle_empty ===
    zpower(empty) = zsingle(empty).


theorem empty_zsubset ===
  X:zsets 
 -----------------
 empty zsubset_of X.

theorem zsubset_empty_eq_empty ===
 X zsubset_of empty <=> X = empty.

theorem empty_zunion_eq ===
  A:zsets
 -----------------
 empty zunion A = A.

theorem zunion_empty_eq ===
  A:zsets
 -----------------
 A zunion empty = A.

theorem empty_zinter_eq_empty ===
  A:zsets
 ---------------------
 empty zinter A = empty.

theorem zinter_empty_eq_empty ===
  A:zsets
 ---------------------
 A zinter empty = empty.

theorem zsym_diff_eq_empty ===
    X:zsets
 --------------------
 X zsym_diff X = empty.

theorem zgen_union_empty_eq_empty ===
    zgen_union(empty) = empty.



theorem zinter_zdiff_eq_empty ===
  A:zsets, X:zpower(A), Y:zpower(A)
 --------------------------
 X zinter (Y zdiff X) = empty.

theorem zin_zsubset ===
 (X zin A) and (A zsubset_of B) => (X zin B).

theorem zsubset_reflexive ===
  X:zsets 
 --------------
 X zsubset_of X.

theorem zsubset_transitive ===
 (X zsubset_of Y) and (Y zsubset_of Z) => (X zsubset_of Z).

theorem zp_subset_transitive ===
 (X zp_subset_of Y) and (Y zp_subset_of Z) => (X zp_subset_of Z).


theorem zsingle_is_zsubset ===
 (X zin Y) <=> zsingle(X) zsubset_of Y.

theorem zpair_commutes ===
 zpair(X,Y) = zpair(Y,X).

theorem zsubset_of_zpower === 
 A zsubset_of B => zpower(A) zsubset_of zpower(B).

theorem zin_zpower ===
  X:zsets 
 ---------------
 X zin zpower(X).

theorem zin_zsubset_zgen_union ===
  Z:zsets, Y:zpower(zpower(Z))
 ----------------------------
 X zin Y => X zsubset_of zgen_union(Y).

theorem zgen_union_zsingle ===
  X:zsets 
 ---------------
 X = zgen_union(zsingle(X)).

theorem zgen_inter_zin ===
 (X zin zgen_inter(A)) and (B zin A) => (X zin B).

theorem zin_zsubset_zgen_inter ===
  Z:zsets, Y:zpower(zpower(Z))
 -----------------------------
 X zin Y => zgen_inter(Y) zsubset_of X.

theorem zgen_inter_zsingle ===
  Z:zsets, X:zpower(Z)
 -----------------------
 X = zgen_inter(zsingle(X)).

theorem zin_zunion1 ===
  Z:zsets, A:zpower(Z), B:zpower(Z)
 ----------------------------
  X zin B => X zin (A zunion B).

theorem zin_zunion2 ===
  Z:zsets, A:zpower(Z), B:zpower(Z)
 ----------------------------
 X zin A => X zin (A zunion B).

theorem zunion_absorption ===
  Z:zsets, X:zpower(Z) 
 ----------------
 (X zunion X) = X.

theorem zsubset_zunion1 ===
 Z:zsets, X:zpower(Z), Y:zpower(Z) 
 ---------------------
 X zsubset_of (X zunion Y).

theorem zsubset_zunion2 ===
 Z:zsets, X:zpower(Z), Y:zpower(Z) 
 ---------------------
 Y zsubset_of (X zunion Y).

theorem zsubsets_zunion ===
 A:zsets, X:zpower(A), Z:zpower(A) 
 ---------------------------------------
 (X zsubset_of Y) and (Z zsubset_of Y) <=> (X zunion Z) zsubset_of Y.  

theorem zsubset_zunion ===
 A:zsets, Y:zpower(A), Z:zpower(A) 
 ---------------------------------------
 (X zsubset_of Y) and (X zsubset_of Z) => (X zsubset_of (Y zunion Z)).  

theorem zunion_zinter ===
 (X zsubset_of (Y zinter Z)) => (X zsubset_of (Y zunion Z)).  

theorem zunion_commutes ===
 (A zunion B) = (B zunion A).

theorem zunion_associates ===
 ((A zunion B) zunion C) = (A zunion (B zunion C)).

theorem zunion_zinter_dist ===
  ((A zunion B) zinter C) = (A zinter C) zunion (B zinter C).

theorem zgen_union_zunion_dist ===
 zgen_union(X zunion Y) = zgen_union(X) zunion zgen_union(Y).

theorem zgen_inter_zunion_dist ===
 zgen_inter(X zunion Y) = zgen_inter(X) zinter zgen_inter(Y).

theorem zinter_absorption ===
  Z:zsets, X:zpower(Z) 
 ------------------
 X zinter X = X.

theorem zinter_commutes ===
 (A zinter B) = (B zinter A).

theorem zinter_associates ===
 ((A zinter B) zinter C) = (A zinter (B zinter C)).

theorem zinter_zunion_dist ===
 ((A zinter B) zunion C) = (A zunion C) zinter (B zunion C).

theorem zinter_zsubset1 ===
  Z:zsets, X:zpower(Z), Y:zpower(Z) 
 -----------------------
 (X zinter Y) zsubset_of X.

theorem zinter_zsubset2 ===
  Z:zsets, X:zpower(Z), Y:zpower(Z) 
 -----------------------
 (X zinter Y) zsubset_of Y.

theorem zsubsets_zinter ===
 (X zsubset_of Y) and (Z zsubset_of Y) => ((X zinter Z) zsubset_of Y).  

theorem zsubset_zinter ===
  A:zsets, Y:zpower(A), Z:zpower(A) 
 -------------------------------------------------
 (X zsubset_of Y) and (X zsubset_of Z) <=> (X zsubset_of (Y zinter Z)).  

theorem zdiff_zsubset ===
  Z:zsets, X:zpower(Z), Y:zpower(Z) 
 -----------------------
 (X zdiff Y) zsubset_of X.

theorem zdiff_empty_zinter ===
  A:zsets, Y:zpower(A), Z:zpower(A) 
 -------------------------------------------------
  (X zsubset_of Y) and ((X zinter Z = empty)) <=> X zsubset_of (Y zdiff Z).

theorem zdiff_partition ===
  A:zsets, X:zpower(A), Y:zpower(A) 
 -----------------------
  X zsubset_of Y <=> X zunion (Y zdiff X) = Y.

theorem zdiff_zdiff ===
  A:zsets, X:zpower(A), Y:zpower(A), Z:zpower(A) 
 -------------------------------------------------
  X zsubset_of Y and Y zsubset_of Z => (Y zdiff (Z zdiff X)) = X.

theorem zdiff_zunion ===
  X zdiff (Y zunion Z) = (X zdiff Y) zinter (X zdiff Z).

theorem zdiff_zinter ===
  X zdiff (Y zinter Z) = (X zdiff Y) zunion (X zdiff Z).


theorem zordered_pair_zin_zcross ===
 zordered_pair(X,Y) zin A zcross B <=> X zin A and Y zin B.


theorem zsym_diff_commutes ===
 X zsym_diff Y = Y zsym_diff X.

theorem zsym_diff_associates ===
  X zsym_diff (Y zsym_diff Z) = (X zsym_diff Y) zsym_diff Z.

theorem zsym_diff_zsubset_zunion ===
  A:zsets, X:zpower(A), Y:zpower(A)
 -----------------------
 (X zsym_diff Y) zsubset_of (X zunion Y).

theorem zsym_diff_is_zunion ===
  A:zsets, X:zpower(A), Y:zpower(A)
 -----------------------
 (X zinter Y) = empty <=> (X zsym_diff Y) = (X zunion Y).


theorem zsubset_zpower_zgen_union_zsubset ===
    A zsubset_of zpower(B)
 ---------------------------------------
     zgen_union(A) zsubset_of B.

theorem zsubset_zpower_zgen_inter_zsubset ===
    not_empty(A), A zsubset_of zpower(B)
 ---------------------------------------
     zgen_inter(A) zsubset_of B.

theorem zpair_zin_zpower ===
    C:zsets, A:C, B:C
  ---------------------------------
    zpair(A,B):zpower(C).

%%------------------------------------------


%%
%% Support for Window Inference - loads
%% zsets specific information.
%%

require_global_tactic_file(wi_support).
require_tactic_file(wi_support,zsets).
